/*!
	\file easydiagnostics.h
	\brief Contains the EasyDiagnostics class
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-08-09
// Last time modified:	2008-03-01

#ifndef DEF_EASYDIAGNOSTICS_H
 #define DEF_EASYDIAGNOSTICS_H
 
 #include "diagnostic.h"
 #include "important_fwdecl.h"
 
 #include <boost/shared_ptr.hpp>

namespace HYPER_Compiler {
	
	class ErrorManager;
	
	/// EasyDiagnosticRef serves as a reference to a newly created diagnostic
	/// and makes it easy to add extra info to it. These references can be
	/// copied and when all instances of it are dead, the diagnostic is
	/// committed (added to the ErrorManager list).
	class EasyDiagnosticRef {
	public:
		static EasyDiagnosticRef createRef(Diagnostic const& d,
			ErrorManager& errMan);
		
		EasyDiagnosticRef const&
		addInfo(HYPER_Common::UTF8String msg, Location const* loc) const;
		
		EasyDiagnosticRef const&
		addInfo(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1) const
		{
			return addInfo(msg.compose(arg1), loc);
		}
		
		EasyDiagnosticRef const&
		addInfo(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1, HYPER_Common::UTF8String arg2) const
		{
			return addInfo(msg.compose(arg1, arg2), loc);
		}
		
		EasyDiagnosticRef const&
		addInfo(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1, HYPER_Common::UTF8String arg2,
			HYPER_Common::UTF8String arg3) const
		{
			return addInfo(msg.compose(arg1, arg2, arg3), loc);
		}
		
		EasyDiagnosticRef const&
		addInfo(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1, HYPER_Common::UTF8String arg2,
			HYPER_Common::UTF8String arg3, HYPER_Common::UTF8String arg4) const
		{
			HYPER_Common::UTF8String arr[] = { arg1, arg2, arg3, arg4 };
			return addInfo(msg.compose(4, arr), loc);
		}
		
	private:
		class DiagnosticCommitter {
		public:
			~DiagnosticCommitter();
			
		private:
			friend class EasyDiagnosticRef;
			
			DiagnosticCommitter(Diagnostic const& d, ErrorManager& errMan);
			
			Diagnostic fDiag;
			ErrorManager& fErrMan;
		};
		
		explicit EasyDiagnosticRef(DiagnosticCommitter* dc);
		
		boost::shared_ptr<DiagnosticCommitter> fDC;
	};
	
	/// Utility class for generating diagnostics
	class EasyDiagnostics {
	public:
		explicit EasyDiagnostics(ErrorManager& errMan);
		
		EasyDiagnosticRef add(Diagnostic const& d) const;
		
		EasyDiagnosticRef
		error(HYPER_Common::UTF8String msg, Location const* loc) const;
		
		EasyDiagnosticRef
		error(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1) const
		{
			return error(msg.compose(arg1), loc);
		}
		
		EasyDiagnosticRef
		error(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1, HYPER_Common::UTF8String arg2) const
		{
			return error(msg.compose(arg1, arg2), loc);
		}
		
		EasyDiagnosticRef
		error(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1, HYPER_Common::UTF8String arg2,
			HYPER_Common::UTF8String arg3) const
		{
			return error(msg.compose(arg1, arg2, arg3), loc);
		}
		
		EasyDiagnosticRef
		error(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1, HYPER_Common::UTF8String arg2,
			HYPER_Common::UTF8String arg3, HYPER_Common::UTF8String arg4) const
		{
			HYPER_Common::UTF8String arr[] = { arg1, arg2, arg3, arg4 };
			return error(msg.compose(4, arr), loc);
		}
		
		EasyDiagnosticRef
		error(Location const* loc, HYPER_Common::UTF8String msg,
			HYPER_Common::UTF8String arg1, HYPER_Common::UTF8String arg2,
			HYPER_Common::UTF8String arg3, HYPER_Common::UTF8String arg4,
			HYPER_Common::UTF8String arg5) const
		{
			HYPER_Common::UTF8String arr[] = { arg1, arg2, arg3, arg4, arg5 };
			return error(msg.compose(5, arr), loc);
		}
		
		
		EasyDiagnosticRef
		warning(HYPER_Common::UTF8String msg, Location const* loc) const;
		
		EasyDiagnosticRef
		note(HYPER_Common::UTF8String msg, Location const* loc) const;
		
	private:
		ErrorManager& fErrMan;
	};
}
#endif
