/*!
	\file errormanager.cxx
	\brief Implements the ErrorManager class
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-05-11
// Last time modified:	2008-06-29

#include "errormanager.h"

#include "asserts.h"

#include "common/filepath.h"
#include "common/utf8textflowstream.h"

namespace HYPER_Compiler {
	
	ErrorManager::ErrorManager(
		HYPER_Common::UTF8TextFlowStream& out, bool bufferErrors,
		FileMap const& filemap)
	 : fLastAdd(fList.end()), fTotalCount(0),
		fBufferErrors(bufferErrors), fOut(out), fFileMap(filemap)
	{
		//
	}
	
	HYPER_Common::FilePath const*
	ErrorManager::getFileName(FileID ID) const
	{
		FileMap::const_iterator it = fFileMap.find(ID);
		hyp_assert(it != fFileMap.end() && "File ID unknown");
		return it->second;
	}
	
	void
	ErrorManager::print(HYPER_Common::UTF8PushStream& out,
		Location const& loc) const
	{
		switch (loc.getType()) {
		case Location::kSource:
			{
				FileID file = loc.getFileID();
				out.print(
					UTF8STR("%1:%2:%2").compose(
						getFileName(file)->toString(), loc.getLine(),
						loc.getColumn()
					)
				);
				break;
			}
		case Location::kAutoGenerated:
			out.print(UTF8STR("<auto-generated>"));
			break;
		case Location::kBuiltIn:
			out.print(UTF8STR("<built-in>"));
			break;
		}
	}
	
	void ErrorManager::print(HYPER_Common::UTF8TextFlowStream& out,
		Diagnostic const& d) const
	{
		print(out, d.getLocation());
		out.print(UTF8STR(": "));
		out.print(d.getTypeString());
		out.print(UTF8STR(": "));
		{
			// indent when message is word wrapped
			HYPER_Common::Indent indent(out, 2);
			out.line(d.getData().message);
		}
		
		for (Diagnostic::InfoIterator it = d.getInfoBegin();
			it != d.getInfoEnd(); ++it)
		{
			DiagnosticPair const& info = *it;
			print(out, *info.location);
			out.print(UTF8STR(": "));
			
			// indent when message is word wrapped
			HYPER_Common::Indent indent(out, 2);
			out.line(info.message);
		}
		
		out.line();
	}
	
	void ErrorManager::print(Diagnostic const& d) const {
		print(fOut, d);
	}
	
	void ErrorManager::add(Diagnostic const& d) {
		//debug.line("ERROR added to ErrorManager");
		fCounts[d.getType()]++; // increase count for the diagnostic's type
		fTotalCount++;
		
		// print diagnostic directly if in `no-diagnostic-buffer' mode
		if (!fBufferErrors) {
			print(d);
		}
		
		// add diagnostic at the right place
		
		typedef std::list<Diagnostic>::iterator DIT;
		DIT end = fList.end();
		Location const& insertLoc = d.getLocation();
		
		// insertion at or after current position
		if (fLastAdd == end
			|| !strictlyOrdered(insertLoc, fLastAdd->getLocation()))
		{
			while (fLastAdd != end
				&& !strictlyOrdered(insertLoc, fLastAdd->getLocation()))
			{
				++fLastAdd;
			}
			
			// Either fLastAdd==end, which means that the new Diagnostic needs
			// to be put at the end of the list, or it is not the case, which means
			// that the Diagnostic needs to be put in front of the Diagnostic we're
			// pointing at.
			
			fLastAdd = fList.insert(fLastAdd, d);
			return;
		}
		
		// insertion needed before current position
		DIT begin = fList.begin();
		while (strictlyOrdered(insertLoc, fLastAdd->getLocation())) {
			if (fLastAdd == begin) {
				// already at first diagnostic in the list
				fLastAdd = fList.insert(fLastAdd, d);
				return;
			}
			--fLastAdd;
		}
		++fLastAdd;
		fLastAdd = fList.insert(fLastAdd, d);
		return;
	}
	
	bool
	ErrorManager::strictlyOrdered(Location const& loc1, Location const& loc2)
	{
		bool thisSrcLoc = (loc1.getType() == Location::kSource);
		bool otherSrcLoc = (loc2.getType() == Location::kSource);
		
		if (thisSrcLoc) {
			if (otherSrcLoc) {
				// comparing two locations of type 'source'
				if (loc1.getFileID() == loc2.getFileID()) { // compare sourcefiles
					// same source files, now check line number
					if (loc1.getLine() == loc2.getLine()) {
						// same line number, now look at column number
						return loc1.getColumn() < loc2.getColumn();
					}
					else {
						return loc1.getLine() < loc2.getLine();
					}
				}
				else {
					// different source files
					HYPER_Common::UTF8String name1 =
						getFileName(loc1.getFileID())->toString();
					HYPER_Common::UTF8String name2 =
						getFileName(loc2.getFileID())->toString();
					
					return name1 < name2;
				}
			}
			else { //   !otherSrcLoc
				return false; // other comes first
			}
		}
		else { //   !thisSrcLoc
			if (otherSrcLoc) {
				return true; // this comes first
			}
			else { //   !otherSrcLoc
				return loc1.getType() < loc2.getType();
			}
		}
	}
	
	unsigned ErrorManager::count() const {
		return fTotalCount;
	}
	
	unsigned ErrorManager::count(Diagnostic::Type t) const {
		if (fCounts.find(t) == fCounts.end()) {
			return 0; // not found, so count is zero
		}
		return fCounts.find(t)->second;
	}
	
	bool ErrorManager::errorsPresent() {
		return count(Diagnostic::kError) > 0;
	}
	
	ErrorManager::DiagnosticIterator ErrorManager::begin() const {
		return fList.begin();
	}
	
	ErrorManager::DiagnosticIterator ErrorManager::end() const {
		return fList.end();
	}
}
