/*!
	\file bufferedpullstream.cxx
	\brief Implements the BufferedPullStream class
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2008-02-20
// Last time modified: 	2008-04-27

#ifndef DEF_BUFFEREDPULLSTREAM_CXX
 #define DEF_BUFFEREDPULLSTREAM_CXX
 
 #include "bufferedpullstream.h"

namespace HYPER_Common {
	
	template <class T>
	BufferedPullStream<T>::BufferedPullStream(PullStream<T>& stream)
	 : fStream(stream), fPos(0)
	{
		fBuffer.reserve(1024); // initial capacity of 1kB
	}
	
	template <class T>
	bool BufferedPullStream<T>::streamOK() const {
		// stream is OK when in buffer
		if (fPos < fBuffer.size()) {
			return true; // because we use our 'memory', and that is always OK
		}
		
		// ... or when encapsulated stream is OK
		return fStream.streamOK();
	}
	
	template <class T>
	bool BufferedPullStream<T>::look(T& c) {
		if (fPos < fBuffer.size()) { // position is in buffer
			c = fBuffer[fPos];
			return true;
		}
		
		// now we're just past the end of the buffer
		
		if (fStream.look(c)) { // look ahead in encapsulated stream
			fBuffer.push_back(c); // and store in buffer
			return true;
		}
		else {
			return false;
		}
	}
	
	template <class T>
	bool BufferedPullStream<T>::pullNext() {
		unsigned long bufSize = fBuffer.size();
		typename PullStream<T>::PositionType streamPos =
			fStream.streamPosition();
		
		if (fPos < streamPos) { // just moving forward in buffer
			++fPos;
			return true;
		}
		
		if (fPos >= bufSize) { // don't have current char in buffer
			// Call "look" so that it stores the current character in the
			// buffer, before we move to the next one.
			T c;
			if (!this->look(c)) {
				return false;
			}
		}
		
		if (fStream.pullNext()) {
			++fPos;
			return true;
		}
		else {
			return false;
		}
	}
	
	template <class T>
	typename PullStream<T>::PositionType
	BufferedPullStream<T>::streamPosition() const {
		return fPos;
	}
	
	template <class T>
	bool
	BufferedPullStream<T>::goBackTo(typename PullStream<T>::PositionType pos)
	{
		if (pos > fBuffer.size()) { // check for out of range
			return false;
		}
		
		fPos = pos;
		return true;
	}
}
#endif
