/*!
	\file filepath.h
	\brief Contains the FilePath class
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2005-01-31
// Last time modified: 	2008-03-06

#ifndef DEF_FILEPATH_H
 #define DEF_FILEPATH_H
 
 #include "streams_fwdecl.h"
 #include "utf8string.h"
 
 #include <list>

/// Contains utility code
namespace HYPER_Common {
	
	/// Exception to be thrown when a string is not a valid (complete or
	/// partial, depending on the context) path name
	class InvalidPathStringException {
	public:
		InvalidPathStringException(UTF8String pathString)
		 : fStr(pathString)
		{
			// no action
		}
		
		/// Get the invalid path name
		UTF8String getString() const {
			return fStr;
		}
		
	private:
		UTF8String fStr;
	};
	
	/// Represents a path name for a file or directory. This class does not
	/// query the OS file system.
	class FilePath {
	public:
		/// Is the path absolute (not relative)?
		virtual bool isAbsolute() const = 0;
		
		/// Make this path absolute if it is relative. Use the specified
		/// path as a base directory. This base directory must be absolute
		/// already.
		virtual void makeAbsolute(FilePath const& baseDir) = 0;
		
		/// Gets a relative path out of this absolute one. The specified path
		/// must be a prefix of this path, and will be used as the part of the
		/// path that is no longer needed.
		virtual FilePath* getRelative(FilePath const& baseDir) const = 0;
		
		/// Gets the number of levels in the path. A file path without a
		/// specified parent directory would be level 1 for example.
		virtual unsigned countLevels() const = 0;
		
		/// Get the parent directory of the file or directory
		virtual FilePath* getParentDirectory() const = 0;
		
		/// Get a child of current path (only one level)
		virtual FilePath* getChildPath(UTF8String sub) const = 0;
		
		/// Mark the path as referring to a directory
		virtual void markAsDirectory() = 0;
		
		/// Is the path marked as a directory (like a trailing slash or so) ?
		virtual bool markedAsDirectory() const = 0;
		
		/// Make sure this path is considered as a file name.
		virtual void markAsFile() = 0;
		
		/// Get the largest common prefix of two pathnames. Both must be either
		/// absolute or relative. If there is no common prefix, 0 is returned.
		virtual FilePath* getLargestCommonPrefix(
			FilePath const& otherPath) const = 0;
		
		/// Get the largest common prefix of multiple pathnames.
		static FilePath* getLargestCommonPrefix(
			std::list<FilePath const*> const& paths);
		
		/// Does the path have an extension?
		virtual bool hasExtension() const {
			return this->getExtensionLength() != 0;
		}
		
		/// Does the path have one of some given extensions?
		bool hasExtension(std::list<UTF8String> const&) const;
		
		/// Get the length of the path extension. Returns zero for no extension.
		virtual unsigned int getExtensionLength() const = 0;
		
		/// Gets the path's extension.
		virtual UTF8String getExtension() const = 0;
		
		/// Replace the path's extension with another one.
		virtual void replaceExtension(UTF8String newExt) = 0;
		
		/// Append an (extra) extension to the path.
		virtual void addExtension(UTF8String ext) = 0;
		
		/// Gets the short path name as a string, without parent directories.
		virtual UTF8String getShortName() const = 0;
		
		/// Get the path name as a string.
		virtual UTF8String toString() const = 0;
		
		/// Create a copy of this path
		virtual FilePath* copy() const = 0;
		
		/// Compare two paths for equality. This implementation converts to
		/// strings and then compares those.
		virtual bool operator==(FilePath const& p) const {
			return this->toString() == p.toString();
		}
		
		/// Compare two paths for inequality.
		bool operator!=(FilePath const& p) const {
			return !operator==(p);
		}
		
		/// Compare two paths lexically. This implementation converts to strings
		/// and then compares those.
		virtual bool operator<(FilePath const& p) const {
			return this->toString() < p.toString();
		}
		
		/// Destructor.
		virtual ~FilePath() {
			// no action
		}
	};
}
#endif
