/*!
	\file filesystem.cxx
	\brief Implements the FileSystem class.
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-05-02
// Last time modified:	2008-04-28

#include "filesystem.h"

#include "unixfilepath.h" // Unix path names
#include "utf8stdinstream.h"
#include "utf8stdoutstream.h"
#include "utf8verifier.h"

#include <cassert>
#include <fstream>
#include <stdlib.h> // for getenv()

namespace HYPER_Common {
	
	FilePath* FileSystem::pathFromString(UTF8String path) {
		// implementation for Unix
		return UnixFilePath::createFromString(path);
		// TODO: implement for MS Windows
	}
	
	FilePath const* FileSystem::getWorkingDirectory(bool cached) {
		static FilePath const* workDirCache = 0;
		if (!cached || workDirCache == 0) {
			workDirCache = queryWorkingDirectory();
		}
		return workDirCache;
	}
	
	FilePath const* FileSystem::queryWorkingDirectory() {
		// implementation for Unix
		char const* path = getenv("PWD"); // get working directory
		assert(path != 0);
		UTF8String pathStr = UTF8String::fromCString(path);
		return UnixFilePath::createFromString(pathStr);
		// TODO: implement for MS Windows
	}
	
	PullStream<UTF8Byte>* FileSystem::openInputFile(FilePath const& f) {
		// FIXME: memory leaks
		
		boost::shared_array<const char> name = f.toString().toCString();
		std::ifstream* stdStream = new std::ifstream(name.get());
		UTF8StdInStream* inStream = new UTF8StdInStream(*stdStream);
		
		UTF8Verifier* v = new UTF8Verifier(true);
		return v->puller(*inStream);
	}
	
	UTF8PushStream* FileSystem::openOutputFile(FilePath const& f) {
		// FIXME: memory leaks
		
		boost::shared_array<const char> name = f.toString().toCString();
		std::ofstream* stdStream = new std::ofstream(name.get());
		
		return new UTF8StdOutStream(*stdStream);
	}
}
