/*!
	\file streamprocessor.cxx
	\brief Implements the StreamProcessor class
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2008-03-11
// Last time modified: 	2008-07-12

#ifndef DEF_STREAMPROCESSOR_CXX
 #define DEF_STREAMPROCESSOR_CXX
 
 #include "pullstream.h"
 #include "pushstream.h"
 #include "streamprocessor.h"
 
 //~ #include <iostream> // fallback for debugging this class
 
namespace HYPER_Common {
	
	template <class IN, class OUT>
	class StreamProcessor<IN, OUT>::PullWrapper : public PullStream<OUT> {
	public:
		PullWrapper(StreamProcessor<IN, OUT> const* p, PullStream<IN>& source)
		 : fSource(source), fProcessor(p), fPos(0), fOK(source.streamOK()),
			fState(p->getStartState())
		{
			//
		}
		
		typedef typename PullStream<OUT>::PositionType PositionType;
		
		bool streamOK() const {
			return fOK;
		}
		
		PositionType streamPosition() const {
			return fPos;
		}
		
		bool look(OUT& d) {
			//std::cout << "look called" << std::endl;
			if (fOut.empty()) {
				//std::cout << " no process output in store" << std::endl;
				processData();
				
				if (fOut.empty()) {
					fOK = false;
					return false; // error (EOF)
				}
			}
			
			d = fOut.back();
			return true;
		}
		
		bool pullNext() {
			//std::cout << "pullNext called" << std::endl;
			if (fOut.empty()) {
				//std::cout << " no process output in store" << std::endl;
				processData();
				
				if (fOut.empty()) {
					fOK = false;
					return false; // error (EOF)
				}
			}
			
			fOut.pop_back();
			++fPos;
			return true;
		}
		
	private:
		// pull some data through the processor
		void processData() {
			//std::cout << "processData called" << std::endl;
			
			ProcessStatus s;
			unsigned long triedInput = 0;
			do {
				bool inputOK = getEnoughInput();
				unsigned long curData = fIn.size();
				//std::cout << " input amount: " << curData << std::endl;
				
				// FIXME
				bool endOfInput = (fIn.size() < fState->fgInputNeeded);
				
				if (fIn.empty()) {
					return; // no input at all; bail out
				}
				
				// remember how much input we gave to the stream processor
				triedInput = curData;
				s = fProcessor->processData(fIn, endOfInput, fOut, *fState);
				
				curData = fIn.size();
				if (curData < triedInput) {
					// If the processor consumed some input, then pretend we
					// only gave the remaining input. This is in case the stream
					// processor has indicated that it needs more input.
					triedInput = curData;
				}
				
				switch (s) {
				case StreamProcessor<IN, OUT>::kOK:
					return; // OK, output produced  (put assert?)
				case StreamProcessor<IN, OUT>::kNeedMoreData:
					if (inputOK && triedInput < fState->fgInputNeeded) {
						// we'll try again, with more input
						continue;
					}
					else {
						// tried enough (or input fails), we accept our failure
						return;
					}
				}
			} while (true);
			
			
		}
		
		bool getEnoughInput() {
			//std::cout << "getEnoughInput called" << std::endl;
			IN i;
			while (fIn.size() < fState->fgInputNeeded) {
				if (fSource.streamOK() && fSource.look(i)) {
					fIn.push_back(i);
					fSource.pullNext();
				}
				else {
					return false;
				}
			}
			return true;
		}
		
		PullStream<IN>& fSource;
		StreamProcessor<IN, OUT> const* fProcessor;
		std::deque<IN> fIn;
		std::deque<OUT> fOut;
		PositionType fPos;
		bool fOK;
		boost::shared_ptr<State> fState;
	};
	
	template <class IN, class OUT>
	class StreamProcessor<IN, OUT>::PushWrapper : public PushStream<IN> {
	public:
		PushWrapper(StreamProcessor<IN, OUT> const* p, PushStream<OUT>& dest)
		 : fDest(dest), fProcessor(p), fPos(0), fState(p->getStartState())
		{
			//
		}
		
		typedef typename PullStream<OUT>::PositionType PositionType;
		
		bool streamOK() const {
			return fDest.streamOK();
		}
		
		PositionType streamPosition() const {
			return fPos;
		}
		
		bool push(IN d) {
			//~ std::cout << "PushWrapper::push called" << std::endl;
			if (!fDest.streamOK()) {
				return false;
			}
			
			fIn.push_back(d);
			++fPos;
			ProcessStatus s;
			while (fIn.size() >= fState->fgInputNeeded) {
				//~ std::cout << " enough input, giving it to processor" << std::endl;
				
				s = fProcessor->processData(fIn, false, fOut, *fState);
				if (s == StreamProcessor<IN, OUT>::kNeedMoreData) {
					// not enough input to get a result from the processor
					break;
				}
				
				while (!fOut.empty()) {
					OUT x = fOut.back();
					if (fDest.push(x)) {
						//~ std::cout << " consuming an output element" << std::endl;
						fOut.pop_back();
					}
					else {
						return false;
					}
				}
			}
			return true;
		}
		
	private:
		PushStream<OUT>& fDest;
		StreamProcessor<IN, OUT> const* fProcessor;
		std::deque<IN> fIn;
		std::deque<OUT> fOut;
		PositionType fPos;
		boost::shared_ptr<State> fState;
	};
	
	template<class IN, class OUT>
	PullStream<OUT>*
	StreamProcessor<IN, OUT>::puller(PullStream<IN>& source) const
	{
		return new PullWrapper(this, source);
	}
	
	template<class IN, class OUT>
	PushStream<IN>*
	StreamProcessor<IN, OUT>::pusher(PushStream<OUT>& dest) const
	{
		return new PushWrapper(this, dest);
	}
}
#endif
