/*!
	\file textflowstream.cxx
	\brief Implements the TextFlowStream and Indent classes
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-04-17
// Last time modified:	2008-02-13

#include "textflowstream.h"

#include <cassert>

namespace HYPER_Common {
	
	// ============
	// TextFlowStream class
	// ============
	
	TextFlowStream::TextFlowStream(PushStream<CodePoint>& output)
	 : fOut(output), fMargin(0), fAtIndentPos(true), fOpt(),
		fColumn(0)
	{
		// no action
	}
	
	TextFlowStream::TextFlowStream(PushStream<CodePoint>& output,
		TextFlowOptions opt)
	 : fOut(output), fMargin(0), fAtIndentPos(true), fOpt(opt),
		fColumn(0)
	{
		// no action
	}
	
	bool TextFlowStream::streamOK() const {
		return fOut.streamOK();
	}
	
	bool TextFlowStream::push(CodePoint c) {
		if (!fOpt.fEnableWrap) {
			// print immediately
			printChar(c);
			return streamOK();
		}
		
		// handle wrapping if needed
		
		fColumn++;
		do {
			if (isWrapChar(c)) {
				if (fColumn <= (fOpt.fWrapColumn + 1))
				{ // non-whites not yet passed the wrapping line
					// flush buffer
					while (!fBuf.empty()) {
						printChar(fBuf.front());
						fBuf.pop_front();
					}
					
					if (c.is(CodePoint::kLf)
							|| fColumn == (fOpt.fWrapColumn + 1))
					{ // white used as wrap char
						fColumn = 0;
						printNewline();
					}
					else {
						fBuf.push_back(c); // save white in buffer
					}
					
					break;
				}
				else { // we are already to the right of the wrapping line
					wrapNow(); // go to next line
					// proceed for what is left in the buffer
				}
			}
			else {
				fBuf.push_back(c);
				break;
			}
		} while (true);
		
		return streamOK();
	}
	
	unsigned long TextFlowStream::streamPosition() const {
		return fOut.streamPosition();
	}
	
	void TextFlowStream::increaseIndent(unsigned int amount)
	{
		if (amount == 0) {
			return; // no action
		}
		
		amount *= fOpt.fColIndent;
		
		assert(!fOpt.fEnableWrap || fOpt.fWrapColumn >= amount);
		fOpt.fWrapColumn -= amount;
		fMargin += amount;
	}
	
	void TextFlowStream::decreaseIndent(unsigned int amount)
	{
		if (amount == 0) {
			return; // no action
		}
		
		amount *= fOpt.fColIndent;
		
		assert(!fOpt.fEnableWrap || fMargin >= amount);
		fOpt.fWrapColumn += amount;
		fMargin -= amount;
	}
	
	inline bool TextFlowStream::isWrapChar(CodePoint c) {
		// maybe add some other chars later?
		return c.is(CodePoint::kSpace) || c.is(CodePoint::kLf);
	}
	
	void TextFlowStream::wrapNow() {
		if (isWrapChar(fBuf.front())) { // current line not empty
			// use buffer contents for next line
			
			// pop off previous wrapping chars
			while (isWrapChar(fBuf.front())) {
				fBuf.pop_front();
			}
		}
		else { // line is full without any wrap char in it
			// we need to wrap at a non-wrapping character
			
			// fill current line with first part of buffer
			unsigned int pop = fOpt.fWrapColumn;
			while (pop > 0) {
				pop --;
				printChar(fBuf.front());
				fBuf.pop_front();
			}
		}
		
		fColumn = fBuf.size() + 1;
		printNewline();
	}
	
	bool TextFlowStream::printChar(CodePoint c) {
		if (c.is(CodePoint::kLf)) {
			fOut.push(c);
			fAtIndentPos = true;
		}
		else if (fAtIndentPos) {
			printIndent();
			fAtIndentPos = false;
			fOut.push(c);
		}
		else {
			fOut.push(c);
		}
		
		return streamOK();
	}
	
	bool TextFlowStream::printNewline() {
		return printChar(CodePoint::linefeed());
	}
	
	bool TextFlowStream::printIndent() {
		for (unsigned int i = 0; i < (fMargin / fOpt.fColIndent); i++) {
			fOut.push(fOpt.fIndentChar);
		}
		return streamOK();
	}
	
	
	// ===============
	// Indent class
	// ===============
	
	Indent::Indent(TextFlowInterface& s, unsigned int amount)
	 : fAmount(amount), fOut(s)
	{
		fOut.increaseIndent(fAmount);
	}
	
	Indent::~Indent() {
		fOut.decreaseIndent(fAmount);
	}
}
