/*!
	\file textflowstream.h
	\brief Contains the TextFlowStream and Indent classes
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-04-17
// Last time modified:	2008-02-13

#ifndef DEF_TEXTFLOWSTREAM_H
 #define DEF_TEXTFLOWSTREAM_H
 
 #include "pushstream.h"
 #include "unicode.h"

 #include <deque>
 #include <cassert>

namespace HYPER_Common {
	
	class TextFlowInterface {
	public:
		virtual ~TextFlowInterface() { /*unused*/ }
		
		/** Increases indentation manually. Use with care. Using the Indent
		class is preferred. */
		virtual void increaseIndent(unsigned int amount = 1) = 0;
		
		/** Decreases indentation manually. Use with care. Using the Indent
		class is preferred. */
		virtual void decreaseIndent(unsigned int amount = 1) = 0;
	};
	
	/// Options for controlling text flow
	class TextFlowOptions {
	public:
		
		/**
			Default textflow options: wrapping disabled, and indentation
			uses space characters.
		*/
		TextFlowOptions()
		 : fIndentChar(' '), fColIndent(1), fEnableWrap(false),
			fWrapColumn(0)
		{
			// no action
		}
		
		/**
			Set indentation character.
			\param c  The character.
			\param columns  The 'width' of the character (e.g. 8 for a tab char)
		*/
		TextFlowOptions&
		setIndentChar(CodePoint c, unsigned columns)
		{
			assert(columns >= 1);
			fIndentChar = c;
			fColIndent = columns;
			return *this;
		}
		
		/**
			Enable text wrapping.
			\param column The column to wrap at
		*/
		TextFlowOptions&
		enableWrapping(unsigned column)
		{
			fEnableWrap = true;
			fWrapColumn = column;
			return *this;
		}
		
	private:
		friend class TextFlowStream;
		
		CodePoint fIndentChar;
		unsigned fColIndent;
		bool fEnableWrap;
		unsigned fWrapColumn;
	};
	
	/// Class to support text flow (indentation and wrapping)
	class TextFlowStream
		: public PushStream<CodePoint>, public TextFlowInterface
	{
	public:
		
		/**
			Creates a new TextFlowCharOutputStream for some output stream
			@param output The output stream to send output to
		*/
		TextFlowStream(PushStream<CodePoint>& output);
		
		/**
			Creates a new TextFlowCharOutputStream for some output stream
			@param output The output stream to send output to
			@param opt	Specify options other than the default
		*/
		TextFlowStream(PushStream<CodePoint>& output, TextFlowOptions opt);
		
		bool streamOK() const;
		bool push(CodePoint c);
		unsigned long streamPosition() const;
		
		void increaseIndent(unsigned int amount = 1);
		void decreaseIndent(unsigned int amount = 1);
		
	private:
		static bool isWrapChar(CodePoint c);
		void wrapNow();
		bool printNewline();
		bool printChar(CodePoint c);
		bool printIndent();
		bool isNewLine(CodePoint c);
		
		PushStream<CodePoint>& fOut;
		unsigned int fMargin;
		bool fAtIndentPos;
		TextFlowOptions fOpt;
		unsigned int fColumn;
		std::deque<CodePoint> fBuf;
	};
	
	/// Class to have a simple indentation and automatic undoing of the indentation
	class Indent {
	public:
		/// Create a new Indent object
		Indent(TextFlowInterface& s, unsigned int amount = 1);
		
		~Indent();
	private:
		unsigned int fAmount;
		TextFlowInterface& fOut;
	};
}
#endif
