/*!
	\file unicode.h
	\brief Contains general Unicode functionality
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2008-02-03
// Last time modified: 	2008-09-13

#ifndef DEF_UNICODE_H
 #define DEF_UNICODE_H

namespace HYPER_Common {
	
	/// Represents a Unicode code point
	class CodePoint {
	public:
		CodePoint() : fV(0) { /*empty*/ }
		explicit CodePoint(unsigned cp) : fV(cp) { /*empty*/ }
		explicit CodePoint(char c) : fV(c) { /* TODO : check for portability */ }
		
		unsigned toUnsigned() const { return fV; }
		
		/// Classification of Unicode characters
		enum Class {
			kAlphaNumeric, ///< ASCII letter or digit
			kAmpersand, ///<  ampersand
			kASCIILetter, ///<  A-Z or a-z
			kAsterisk, ///< asterisk
			kAt, ///< @ sign
			kBackslash, ///< backslash
			kColon, ///<  colon
			kComma, ///< comma
			kControl, ///<  control character
			kCr, ///<  carriage return
			kDigit, ///<  digit (0-9)
			kDollar, ///<  dollar sign
			kDot, ///<  dot (full stop)
			kEqual, ///<  equal sign
			kExclamationMark, ///<  exclamation mark
			kGreaterThan, ///<  greater-than sign
			kHashMark, ///<  hash mark (#)
			kHexDigit, ///<  hex digit (0-9 or A-F or a-f)
			kIndentWhite, ///< indentation whitespace (space or horizontal tab)
			kLeftCurlyBracket, ///<  left curly bracket
			kLeftParenthesis, ///<  left parenthesis
			kLeftSquareBracket, ///<  left square bracket ([)
			kLessThan, ///<  less-than sign
			kLf, ///<  linefeed
			kMinus, ///<  minus/hyphen sign
			kNull, ///<  null character
			kPercent, ///<  percent sign
			kPipe, ///< pipe / vertical bar
			kPlus, ///<  plus sign
			kQuestionMark, ///< question mark
			kRightCurlyBracket, ///<  right curly bracket
			kRightParenthesis, ///< right parenthesis
			kRightSquareBracket, ///< right square bracket (])
			kSemicolon, ///< semicolon
			kSlash, ///<  forward slash
			kSpace, ///<  space
			kTab, ///<  character tabulator
			kTilde, ///<  tilde
			kUnderscore, ///<  underscore
		};
		
		bool isNull() const {
			return fV == 0;
		}
		
		bool is(Class c) const {
			switch (c) {
			case kAlphaNumeric:
				return is(kASCIILetter) || is(kDigit);
			case kAmpersand:
				return fV == 0x26;
			case kASCIILetter:
				return (fV >= 0x40 && fV <= 0x5A)
					|| (fV >= 0x61 && fV <= 0x7A);
			case kAsterisk:
				return fV == 0x2A;
			case kAt:
				return fV == 0x40;
			case kBackslash:
				return fV == 0x5C;
			case kColon:
				return fV == 0x3A;
			case kComma:
				return fV == 0x2C;
			case kControl:
				// the C0 and C1 control codes
				return fV < 0x20 || (fV >= 0x7F && fV <= 0x9F);
			case kCr:
				return fV == 0x0D;
			case kDigit:
				return (fV >= 0x30 && fV <= 0x39);
			case kDollar:
				return fV == 0x24;
			case kDot:
				return fV == 0x2E;
			case kEqual:
				return fV == 0x3D;
			case kExclamationMark:
				return fV == 0x21;
			case kGreaterThan:
				return fV == 0x3E;
			case kHashMark:
				return fV == 0x23;
			case kHexDigit:
				return is(kDigit) || (fV >= 0x40 && fV <= 0x45)
					|| (fV >= 0x61 && fV <= 0x66);
			case kIndentWhite:
				return is(kSpace) || is(kTab);
			case kLeftCurlyBracket:
				return fV == 0x7B;
			case kLeftParenthesis:
				return fV == 0x28;
			case kLeftSquareBracket:
				return fV == 0x5B;
			case kLessThan:
				return fV == 0x3C;
			case kLf:
				return fV == 0x0A;
			case kMinus:
				return fV == 0x2D;
			case kNull:
				return fV == 0;
			case kPercent:
				return fV == 0x25;
			case kPipe:
				return fV == 0x7C;
			case kPlus:
				return fV == 0x2B;
			case kQuestionMark:
				return fV == 0x3F;
			case kRightCurlyBracket:
				return fV == 0x7D;
			case kRightParenthesis:
				return fV == 0x29;
			case kRightSquareBracket:
				return fV == 0x5D;
			case kSemicolon:
				return fV == 0x3B;
			case kSlash:
				return fV == 0x2F;
			case kSpace:
				return fV == 0x20;
			case kTab:
				return fV == 0x09;
			case kTilde:
				return fV == 0x7E;
			case kUnderscore:
				return fV == 0x5F;
			}
			
			return false; // invalid class
		}
		
		unsigned short getDigitValue() const {
			if (!is(kDigit)) {
				return 0;
			}
			
			return fV - 0x30;
		}
		
		CodePoint toLower() const {
			if (fV >= 0x40 && fV <= 0x5A) {
				return CodePoint(fV + 32);
			}
			return *this;
		}
		
		CodePoint toUpper() const {
			if (fV >= 0x61 && fV <= 0x7A) {
				return CodePoint(fV - 32);
			}
			return *this;
		}
		
		
		// === static character generator functions === //
		
		/// returns the copyright symbol
		static CodePoint copyright() {
			return CodePoint(0xA9u);
		}
		
		/// returns a white question mark in a black circle
		static CodePoint invertedQuestionMark() {
			return CodePoint(0xFFFDu);
		}
		
		static CodePoint linefeed() {
			return CodePoint(0x0Au);
		}
		
		static CodePoint tab() {
			return CodePoint('\t');
		}
		
		static CodePoint leftDoubleQuote() {
			return CodePoint(0x201Cu);
		}
		
		static CodePoint rightDoubleQuote() {
			return CodePoint(0x201Du);
		}
		
		static CodePoint leftSingleQuote() {
			return CodePoint(0x2018u);
		}
		
		static CodePoint rightSingleQuote() {
			return CodePoint(0x2019u);
		}
		
		
		// === comparison operators === //
		
		bool operator==(CodePoint c) const {
			return fV == c.fV;
		}
		
		bool operator!=(CodePoint c) const {
			return fV != c.fV;
		}
		
		bool operator<(CodePoint c) const {
			return fV < c.fV;
		}
		
		bool operator<=(CodePoint c) const {
			return fV <= c.fV;
		}
		
		bool operator>(CodePoint c) const {
			return fV > c.fV;
		}
		
		bool operator>=(CodePoint c) const {
			return fV >= c.fV;
		}
		
	private:
		unsigned fV;
	};
}
#endif
