/*!
	\file unixfilepath.cxx
	\brief Implements the FilePath class
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2005-09-12
// Last time modified: 	2008-03-16

#include "unixfilepath.h"

#include <cassert>

using std::deque;

namespace HYPER_Common {
	
	// *** UnixFilePath ***
	
	UnixFilePath::UnixFilePath()
	 : fHasRoot(false), fTrailingSlash(false)
	{
		// no action
	}
	
	UnixFilePath::UnixFilePath(UPDeque const& list, bool leadingSlash,
		bool trailingSlash)
	 : fHasRoot(leadingSlash), fTrailingSlash(trailingSlash), fLst(list)
	{
		// no action
	}
	
	bool UnixFilePath::simplify() {
		UPDeque lst;
		
		bool ignoredCur = false;
		unsigned int eatDirs = 0;
		
		for (UPDeque::const_reverse_iterator it = fLst.rbegin();
			it != fLst.rend(); ++it)
		{
			UTF8String cur = *it;
			
			if (cur == UTF8STR(".")) {
				ignoredCur = true; // NOTE: previous ignored '.' can be ignored safely
			}
			else if (cur == UTF8STR("..")) {
				ignoredCur = false;
				++eatDirs;
			}
			else {
				ignoredCur = false; // ignored '.' is now OK (no adaption afterwards)
				
				if (eatDirs > 0) {
					--eatDirs;
				}
				else {
					// transfer component name to new canonical name
					lst.push_front(cur);
				}
			}
		}
		
		if (eatDirs > 0) {
			if (isAbsolute()) { // path like "/../foobar"
				return false; // error
			}
			
			while (eatDirs > 0) {
				lst.push_front(UTF8STR(".."));
				--eatDirs;
			}
		}
		else if (ignoredCur && lst.empty()) {
			lst.push_front(UTF8STR("."));
		}
		
		fLst = lst;
		return true; // ok
	}
	
	bool UnixFilePath::checkValidComponent(UTF8String s) {
		UTF8String::ConstIterator it = s.begin();
		for (; it != s.end(); ++it) {
			if ((*it).is(CodePoint::kSlash)) { // slash in component...
				// .. not allowed
				return false;
			}
		}
		
		return true; // no violations found
	}
	
	UnixFilePath* UnixFilePath::createFromString(UTF8String path) {
		UnixFilePath p;
		//UTF8String pathBackup = path;
		
		if (path.startsWith(CodePoint('/'))) {
			p.fHasRoot = true;
			path = path.subString(++path.begin());
		}
		else {
			p.fHasRoot = false;
		}
		
		UTF8String::ConstIterator curBegin = path.begin();
		for (UTF8String::ConstIterator it = path.begin();
			it != path.end();
			++it)
		{
			if ((*it) == CodePoint('/')) {
				if (curBegin == it) {
					// found two consecutive slashes
					++curBegin;
				}
				else {
					p.fLst.push_back(path.subString(curBegin, it));
					curBegin = it.next();
				}
			}
		}
		
		if (path.endsWith(CodePoint('/'))) { // trailing slash ??
			p.fTrailingSlash = true;
		}
		else { // no trailing slash, add last part of path
			p.fLst.push_back(path.subString(curBegin));
		}
		
		bool ok = p.simplify(); // delete unneeded parts in the path
		
		if (ok) {
			return p.copy();
		}
		else {
			return 0;
		}
	}
	
	inline UnixFilePath const* UnixFilePath::toUnixPath(FilePath const& p)
	{
		return dynamic_cast<UnixFilePath const*>(&p);
	}
	
	inline bool UnixFilePath::isRootDir() const {
		return fHasRoot && fLst.empty();
	}
	
	unsigned UnixFilePath::countLevels() const {
		return fLst.size();
	}
	
	bool UnixFilePath::isAbsolute() const {
		return fHasRoot;
	}
	
	inline void UnixFilePath::markAsDirectory() {
		fTrailingSlash = true;
	}
	
	bool UnixFilePath::markedAsDirectory() const {
		return fTrailingSlash || (fLst.empty() && isAbsolute());
	}
	
	void UnixFilePath::markAsFile() {
		fTrailingSlash = false;
	}
	
	void UnixFilePath::makeAbsolute(FilePath const& baseDir) {
		assert(baseDir.isAbsolute());
		UnixFilePath const* base = dynamic_cast<UnixFilePath const*>(&baseDir);
		assert(base != 0);
		
		if (this->isAbsolute()) {
			return; // nothing to do here
		}
		
		for (UPDeque::const_reverse_iterator it = base->fLst.rbegin();
			it != base->fLst.rend();
			++it)
		{
			fLst.insert(fLst.begin(), *it);
		}
		
		fHasRoot = true;
		simplify();
	}
	
	UnixFilePath* UnixFilePath::getRelative(FilePath const& baseDir) const {
		assert(this->isAbsolute());
		assert(baseDir.isAbsolute());
		UnixFilePath const* base = toUnixPath(baseDir);
		assert(base != 0);
		
		// impossible prefix?
		if (base->fLst.size() > this->fLst.size()) {
			return 0;
		}
		
		for (unsigned int i = 0; i < fLst.size(); ++i) {
			if (i >= base->fLst.size() || base->fLst[i] != fLst[i]) {
				UnixFilePath* result = new UnixFilePath();
				
				// create part without prefix
				for (unsigned int j = i; j < fLst.size(); ++j) {
					result->fLst.push_back(fLst[j]);
				}
				
				return result;
			}
		}
		
		// This path is equal to the given prefix
		
		// return a "." path
		return createFromString(UTF8STR("."));
	}
	
	UnixFilePath* UnixFilePath::getParentDirectory() const {
		UnixFilePath* result = this->copy();
		
		if (result->fLst.size() > 1
			|| (result->fLst.size() == 1 && isAbsolute()))
		{
			// just pop off the last item
			result->fLst.pop_back();
			result->markAsDirectory();
			return result;
		}
		else if (isAbsolute()) {
			// absolute root path "/"
			delete result;
			return 0; // cannot create parent dir
		}
		else {
			// relative path with length 1
			delete result;
			return createFromString(UTF8STR("../"));
		}
	}
	
	UnixFilePath* UnixFilePath::getChildPath(UTF8String sub) const {
		if (sub == UTF8STR(".")) { // remain in the current directory
			// create a directory path
			UnixFilePath* result = this->copy();
			result->markAsDirectory();
			return result;
		}
		else if (sub == UTF8STR("..")) {
			// attempt to get parent directory
			return getParentDirectory();
		}
		
		if (!checkValidComponent(sub)) {
			throw InvalidPathStringException(sub);
		}
		
		UnixFilePath* result = this->copy();
		result->fLst.push_back(sub);
		return result;
	}
	
	UnixFilePath*
	UnixFilePath::getLargestCommonPrefix(FilePath const& otherPath) const
	{
		UnixFilePath const* other = toUnixPath(otherPath);
		assert(other != 0);
		
		// paths must both either be absolute or relative
		bool absoluteNess = this->isAbsolute();
		if (absoluteNess != other->isAbsolute()) {
			return 0;
		}
		
		if (this->fLst.empty() || other->fLst.empty()) {
			return 0; // no prefix
		}
		
		unsigned maxCommon =
			std::min(this->fLst.size() - 1, other->fLst.size() - 1);
		UPDeque prefix;
		
		for (unsigned int i = 0; i < maxCommon; ++i) {
			if (this->fLst[i] == other->fLst[i]) {
				prefix.push_back(this->fLst[i]);
			}
			else {
				break;
			}
		}
		
		// 1) absolute paths have the root "/" in common
		// 2) or there is a list of common parts
		if (this->isAbsolute() || !prefix.empty()) {
			return new UnixFilePath(
				prefix, this->isAbsolute(),
				prefix.size() < maxCommon
					|| maxCommon < this->fLst.size()
					|| maxCommon < other->fLst.size()
			);
		}
		else {
			return 0; // no common prefix
		}
	}
	
	unsigned int UnixFilePath::getExtensionLength() const {
		UTF8String back = fLst.back();
		
		if (back.size() < 3) {
			return 0; // no extension
		}
		
		unsigned len = 0;
		for (UTF8String::ConstIterator it = back.end().prev();
			it != back.begin();
			--it)
		{
			if (it->is(CodePoint::kDot)) {
				return len;
			}
			++len;
		}
		
		// no extension
		return 0;
	}
	
	UTF8String UnixFilePath::getExtension() const {
		UTF8String back = fLst.back();
		
		if (back.size() < 3) {
			return UTF8String(); // no extension
		}
		
		for (UTF8String::ConstIterator it = back.end().prev();
			it != back.begin();
			--it)
		{
			if (it->is(CodePoint::kDot)) {
				return back.subString(it.next());
			}
		}
		
		// no extension
		return UTF8String();
	}
	
	void UnixFilePath::replaceExtension(UTF8String newExt) {
		UTF8String back = fLst.back();
		
		for (UTF8String::ConstIterator it = back.end().prev();
			it != back.begin();
			--it)
		{
			if (it->is(CodePoint::kDot)) {
				fLst.back() =
					UTF8STR("%1.%2").compose(back.left(it), newExt);
			}
		}
		
		// no extension
		fLst.back() = UTF8STR("%1.%2").compose(back, newExt);
	}
	
	void UnixFilePath::addExtension(UTF8String ext) {
		fLst.back() = UTF8STR("%1.%2").compose(fLst.back(), ext);
	}
	
	UTF8String UnixFilePath::getShortName() const {
		return fLst.back();
	}
	
	UTF8String UnixFilePath::toString() const {
		UTF8StringBuilder sb;
		
		if (fHasRoot) {
			sb += UTF8STR("/");
		}
		
		for (unsigned int i = 0; i < fLst.size(); ++i) {
			sb += fLst[i];
			
			if (i < fLst.size() - 1) {
				sb += UTF8STR("/");
			}
		}
		
		if (fTrailingSlash) {
			sb += UTF8STR("/");
		}
		
		return sb.toString();
	}
	
	UnixFilePath* UnixFilePath::copy() const {
		return new UnixFilePath(*this);
	}
	
	bool UnixFilePath::operator==(FilePath const& p) const {
		UnixFilePath const* ufp = dynamic_cast<UnixFilePath const*>(&p);
		
		return ufp != 0 && fHasRoot == ufp->fHasRoot
			&& fTrailingSlash == ufp->fTrailingSlash && fLst == ufp->fLst;
	}
}
