/*!
	\file utf8.h
	\brief Contains UTF-8 functionality
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2008-02-02
// Last time modified: 	2008-09-06

#ifndef DEF_UTF8_H
 #define DEF_UTF8_H
 
 #include "unicode.h"

namespace HYPER_Common {
	
	typedef unsigned char Byte;
	
	class UTF8Byte;
	
	class UTF8Consumer {
	public:
		virtual ~UTF8Consumer() { /*empty*/ }
		virtual void consume(UTF8Byte b) = 0;
	};
	
	/// Represents a byte in a UTF-8 encoded stream
	class UTF8Byte {
	public:
		static const Byte kMaxSequenceLength = 4;
		
		UTF8Byte() { /*uninitialized*/ }
		explicit UTF8Byte(Byte b) : fB(b) { /*empty*/ }
		
		Byte toByte() const {
			return fB;
		}
		
		bool operator==(UTF8Byte b) const {
			return fB == b.fB;
		}
		
		bool operator!=(UTF8Byte b) const {
			return fB != b.fB;
		}
		
		/// Is this byte a single character on its own (ASCII range)?
		bool isSingleChar() const {
			return (fB & 128) == 0; // 0xxxxxxx
		}
		
		/// Is this a non-leading-byte of a UTF-8 encoded character?
		bool isFollowByte() const {
			return (fB & 192) == 128; // 10xxxxxx
		}
		
		/// Get the number of followers for this leading byte. Returns 0 for a
		/// non-leading or invalid byte.
		unsigned getFollowerCount() const {
			if ((fB & 224) == 192) { // 110xxxxx
				return 1;
			}
			
			if ((fB & 240) == 224) { // 1110xxxx
				return 2;
			}
			
			if ((fB & 248) == 240) { // 11110xxx
				return 3;
			}
			
			return 0; // invalid or non-leading byte
		}
		
		/// Is this byte the leading byte for a character encoded in UTF-8?
		bool isLeadingByte() const {
			if ((fB & 192) != 192) { // 11xxxxxx
				return false;
			}
			
			return getFollowerCount() > 0;
		}
		
		/// Is this an invalid byte value for UTF-8?
		bool isInvalid() const {
			switch (fB) {
			case 0xC0: case 0xC1:
			case 0xF5: case 0xF6: case 0xF7:
			case 0xF8: case 0xF9: case 0xFA: case 0xFB: case 0xFC: case 0xFD:
			case 0xFE: case 0xFF:
				return true; // invalid
			default:
				return false; // valid
			}
		}
		
		/// Decode a single char byte, assuming valid input
		CodePoint decodeSingleChar() const {
			// 0xxxxxxx
			// 00000000 00000000 0xxxxxxx
			return CodePoint(fB & 127u);
		}
		
		/// Decode a character of 2 UTF-8 bytes, assuming valid input
		static CodePoint decode2Byte(UTF8Byte a, UTF8Byte b) {
			// 110xxxxx 10yyyyyy
			// 00000000 00000xxx xxyyyyyy
			return CodePoint(((a.fB & 31u) << 6u) + (b.fB & 63u));
		}
		
		/// Decode a character of 3 UTF-8 bytes, assuming valid input
		static CodePoint decode3Byte(UTF8Byte a, UTF8Byte b, UTF8Byte c) {
			// 1110xxxx 10yyyyyy 10zzzzzz
			// 00000000 xxxxyyyy yyzzzzzz
			return CodePoint(((a.fB & 15u) << 12u) + ((b.fB & 63u) << 6u)
				+ (c.fB & 63u));
		}
		
		/// Decode a character of 4 UTF-8 bytes, assuming valid input
		static CodePoint
		decode4Byte(UTF8Byte a, UTF8Byte b, UTF8Byte c, UTF8Byte d)
		{
			// 11110www 10xxxxxx 10yyyyyy 10zzzzzz
			// 000wwwxx xxxxyyyy yyzzzzzz
			return CodePoint(((a.fB & 7u) << 18u) + ((b.fB & 63u) << 12u)
				+ ((c.fB & 63u) << 6u) + (d.fB & 63u));
		}
		
		/// Decode a character from a valid byte storage
		static CodePoint decodeFromValidStorage(UTF8Byte const* s) {
			if (s->isSingleChar()) {
				return s->decodeSingleChar();
			}
			else if (s->isLeadingByte()) {
				switch (s->getFollowerCount()) {
				case 1:
					return decode2Byte(s[0], s[1]);
				case 2:
					return decode3Byte(s[0], s[1], s[2]);
				case 3:
					return decode4Byte(s[0], s[1], s[2], s[3]);
				}
			}
			
			// should not come here
			return CodePoint::invertedQuestionMark();
		}
		
		static Byte bytesNeededFor(CodePoint cp) {
			unsigned u = cp.toUnsigned();
			
			if (u < 0x80) {
				return 1;
			}
			else if (u < 0x800) {
				return 2;
			}
			else if (u < 0x10000) {
				return 3;
			}
			else {
				return 4;
			}
		}
		
		/// Encode a Unicode character
		static void encode(CodePoint cp, UTF8Consumer& cons) {
			unsigned u = cp.toUnsigned();
			
			if (u < 0x80) {
				cons.consume(UTF8Byte(u & 127));
			}
			else if (u < 0x800) {
				cons.consume(UTF8Byte(((u >> 6) & 31) | 192));
				cons.consume(UTF8Byte((u & 63) | 128));
			}
			else if (u < 0x10000) {
				cons.consume(UTF8Byte(((u >> 12) & 15) | 224));
				cons.consume(UTF8Byte(((u >> 6) & 63) | 128));
				cons.consume(UTF8Byte((u & 63) | 128));
			}
			else {
				cons.consume(UTF8Byte(((u >> 18) & 7) | 240));
				cons.consume(UTF8Byte(((u >> 12) & 63) | 128));
				cons.consume(UTF8Byte(((u >> 6) & 63) | 128));
				cons.consume(UTF8Byte((u & 63) | 128));
			}
		}
		
		/// Encode an unsigned long int
		static void encode(unsigned long i, UTF8Consumer& cons) {
			encodeUnsignedIntegralNum(i, cons);
		}
		
		/// Encode a signed long int
		static void encode(signed long i, UTF8Consumer& cons) {
			encodeSignedIntegralNum(i, cons);
		}
		
		/// Encode an unsigned short int
		static void encode(unsigned short i, UTF8Consumer& cons) {
			encodeUnsignedIntegralNum(i, cons);
		}
		
		/// Encode a signed short int
		static void encode(signed short i, UTF8Consumer& cons) {
			encodeSignedIntegralNum(i, cons);
		}
		
		/// Encode an unsigned int
		static void encode(unsigned i, UTF8Consumer& cons) {
			encodeUnsignedIntegralNum(i, cons);
		}
		
		/// Encode a signed int
		static void encode(signed i, UTF8Consumer& cons) {
			encodeSignedIntegralNum(i, cons);
		}
		
		/// Encode an unsigned long long int
		static void encode(unsigned long long i, UTF8Consumer& cons) {
			encodeUnsignedIntegralNum(i, cons);
		}
		
		/// Encode a signed long long int
		static void encode(signed long long i, UTF8Consumer& cons) {
			encodeSignedIntegralNum(i, cons);
		}
		
		static void encode(double d, UTF8Consumer& cons) {
			encode(static_cast<long double>(d), cons);
		}
		
		static void encode(float f, UTF8Consumer& cons) {
			encode(static_cast<long double>(f), cons);
		}
		
		static void encode(long double d, UTF8Consumer& cons);
		
	private:
		template <class T>
		static void encodeUnsignedIntegralNum(T n, UTF8Consumer& c)
		{
			T div = 1;
			while (n / (div * 10) != 0) {
				div *= 10;
			}
			
			do {
				UTF8Byte b(n / div + 0x30);
				c.consume(b);
				n %= div;
				div /= 10;
			} while (div != 0);
		}
		
		template <class T>
		static void encodeSignedIntegralNum(T n, UTF8Consumer& c)
		{
			if (n < 0) {
				c.consume(UTF8Byte('-'));
				n = -n;
			}
			
			encodeUnsignedIntegralNum(n, c);
		}
		
		
		Byte fB;
	};
}
#endif
