/*!
	\file utf8string.cxx
	\brief Implements the UTF8String and UTF8StringBuilder classes
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2008-02-14
// Last time modified: 	2008-06-25

#include "utf8string.h"

namespace HYPER_Common {
	
	// *** UTF8String *** //
	
	UTF8String UTF8String::join(UTF8String const& s1, UTF8String const& s2) {
		if (s1.isEmpty()) {
			return s2;
		}
		
		if (s2.isEmpty()) {
			return s1;
		}
		
		Storage* st = new Storage(s1.fByteLen + s2.fByteLen);
		UTF8Byte* p = st->get();
		
		for (UTF8Byte const* p1 = s1.byteBegin(); p1 != s1.byteEnd(); ++p1)
		{
			*p = *p1;
			p++;
		}
		
		for (UTF8Byte const* p2 = s2.byteBegin(); p2 != s2.byteEnd(); ++p2)
		{
			*p = *p2;
			p++;
		}
		
		return UTF8String(
			boost::shared_ptr<Storage>(st), 0,
			s1.size() + s2.size(), st->getEnd() - st->get()
		);
	}
	
	UTF8String
	UTF8String::compose(unsigned componentCount,
		UTF8String const* components) const
	{
		ConstIterator it = begin();
		ConstIterator endIt = end();
		
		enum Mode {
			kNormal, kStartedComponent, kReadingComponentNumber
		};
		
		Mode mode = kNormal;
		unsigned number = 0;
		
		UTF8StringBuilder b;
		for (; it != endIt; ++it) {
			CodePoint c = *it;
			
			switch (mode) {
			case kNormal:
				if (c.is(CodePoint::kPercent)) {
					mode = kStartedComponent;
				}
				else {
					b.append(c);
				}
				break;
			case kStartedComponent:
				if (c.is(CodePoint::kDigit)) {
					mode = kReadingComponentNumber;
					number = c.getDigitValue();
				}
				else if (c.is(CodePoint::kPercent)) {
					// escaped percentage number
					mode = kNormal;
					b.append(c);
				}
				else {
					// other, invalid, character
					mode = kNormal;
					b.append(c);
				}
				break;
			case kReadingComponentNumber:
				if (c.is(CodePoint::kDigit)) {
					number *= 10;
					number += c.getDigitValue();
				}
				else {
					// end of component number reached
					
					// print component
					if (number >= 1 && number <= componentCount) {
						b.append(components[number - 1]);
					}
					number = 0;
					
					if (c.is(CodePoint::kPercent)) {
						mode = kStartedComponent;
					}
					else {
						mode = kNormal;
						b.append(c);
					}
				}
				break;
			}
		}
		
		// in case the last component number was at the end of the string
		if (number >= 1 && number <= componentCount) {
			b.append(components[number - 1]);
		}
		
		return b.toString();
	}
	
	
	// *** UTF8StringBuilder *** //
	
	//   all its functions are in the header file
	
	
	// *** UTF8StringLiteralPool *** //
	
	//   all its functions are in the header file
	
}	
