/*!
	\file utf8string.h
	\brief Contains the UTF8String, UTF8StringBuilder and UTF8StringLiteralPool classes
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2008-02-05
// Last time modified: 	2009-06-14

#ifndef DEF_UTF8STRING_H
 #define DEF_UTF8STRING_H
 
 #include "cxx_macros.h"
 #include "utf8.h"
 
 #include <boost/shared_array.hpp>
 #include <boost/shared_ptr.hpp>
 
 //#include <iostream> //debug
 #include <map>
 #include <string>
 #include <vector>

namespace HYPER_Common {
	
	/*
		<<< IMPORTANT >>>
		
		The following macro's depend on the C++ compiler's ability to optimize
		the use of identical string literals, for space efficiency reasons. See
		the class "UTF8StringLiteralPool" below.
	*/
	
	// Converts a string literal to a UTF8String object. Assumes valid UTF-8.
	#define UTF8STR(str) \
		HYPER_Common::UTF8StringLiteralPool::getInstance().literal(str)
	
	// Converts a string literal to a UTF8String object as above, except that
	// the string will be quoted by single (Unicode) quotes.
	#define QUOTED_UTF8STR(str) \
		HYPER_Common::UTF8StringLiteralPool::getInstance().quotedLiteral(str)
	
	// Converts a string literal to a UTF8String object as above, except that
	// the string will be quoted by double (Unicode) quotes.
	#define DOUBLEQUOTED_UTF8STR(str) \
		HYPER_Common::UTF8StringLiteralPool::getInstance().doubleQuotedLiteral(str)
	
	
	class UTF8StringBuilder;
	
	/// Represents a string that consists of UTF-8 bytes
	class UTF8String {
		// returns the maximum number of digits used by a number when printed
		template <class T>
		static unsigned digitCount()
		{
			/*
				This calulation is based on the formula:
				log_b(x) = log_k(x) / log_k(b)
				
				We need to find 1 + log_10(2 ^ (sizeof(type) * 8)). We use k = 2
				for easy calculation, and assume log_2(10) = 3 to round up the
				result.
			*/
			return 1 + sizeof(T) * 8 / 3;
		}
		
	public:
		/// Create empty string
		UTF8String()
		 : fStorage(getEmptyStorage()), fOffset(0), fCharLen(0), fByteLen(0)
		{
			//
		}
		
		/// Copy string
		UTF8String(UTF8String const& source)
		 : fStorage(source.fStorage), fOffset(source.fOffset),
			fCharLen(source.fCharLen), fByteLen(source.fByteLen)
		{
			//
		}
		
		/// Create string from Unicode char
		UTF8String(CodePoint c)
		 : fStorage(new Storage(UTF8Byte::bytesNeededFor(c))), fOffset(0),
			fCharLen(1), fByteLen(UTF8Byte::bytesNeededFor(c))
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(c, ca);
		}
		
		/// Create string from copied UTF-8 bytes (assumes valid stream!)
		UTF8String(UTF8Byte const* data, unsigned length)
		 : fStorage(new Storage(length)), fOffset(0), fByteLen(length)
		{
			unsigned chars = 0;
			for (unsigned i = 0; i < length; ++i) {
				fStorage->get()[i] = data[i];
				if (!data[i].isFollowByte()) {
					chars++;
				}
			}
			
			fCharLen = chars;
		}
		
		/// Create from traditional null-terminated C string
		/// (assumes valid UTF-8 bytes!)
		static UTF8String fromCString(char const* ptr) {
			unsigned length = 0;
			for (char const* ptr2 = ptr; *ptr2 != 0; ++ptr2) {
				length++;
			}
			
			return UTF8String(
				reinterpret_cast<UTF8Byte const*>(ptr), length
			);
		}
		
		UTF8String(signed long i)
		 : fStorage(new Storage(1 + digitCount<signed long>())), fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(unsigned long i)
		 : fStorage(new Storage(digitCount<unsigned long>() + 1)), fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(signed short i)
		 : fStorage(new Storage(1 + digitCount<signed short>())), fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(unsigned short i)
		 : fStorage(new Storage(digitCount<unsigned short>() + 1)), fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(signed i)
		 : fStorage(new Storage(1 + digitCount<signed>())), fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(unsigned i)
		 : fStorage(new Storage(digitCount<unsigned>() + 1)), fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(signed long long i)
		 : fStorage(new Storage(1 + digitCount<signed long long>())), fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(unsigned long long i)
		 : fStorage(new Storage(digitCount<unsigned long long>() + 1)),
			fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(i, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(float f)
		 : fStorage(new Storage(50)), // FIXME: storage amount
			fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(f, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(double d)
		 : fStorage(new Storage(100)), // FIXME: storage amount
			fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(d, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		UTF8String(long double d)
		 : fStorage(new Storage(150)), // FIXME: storage amount
			fOffset(0)
		{
			ConsumerAdapter ca(fStorage->get());
			UTF8Byte::encode(d, ca);
			fCharLen = ca.countConsumed();
			fByteLen = fCharLen; // each digit is one byte
		}
		
		~UTF8String() {
			//
		}
		
		UTF8String& operator=(UTF8String const& s) {
			if (this != &s) { // avoid self-assignment
				fStorage = s.fStorage;
				fOffset = s.fOffset;
				fCharLen = s.fCharLen;
				fByteLen = s.fByteLen;
			}
			return *this;
		}
		
		bool isEmpty() const {
			return fCharLen == 0;
		}
		
		/// Returns the size of the string in number of characters
		unsigned length() const {
			return fCharLen;
		}
		
		/// Returns the size of the string in number of characters
		unsigned size() const {
			return fCharLen;
		}
		
		// Check for equality
		bool operator==(UTF8String const& s) const {
			if (&s == this) { // comparison with itself
				return true;
			}
			
			if (fByteLen != s.fByteLen || fCharLen != s.fCharLen) {
				return false;
			}
			
			UTF8Byte const* myP = &fStorage->get()[fOffset];
			UTF8Byte const* hisP = &s.fStorage->get()[s.fOffset];
			
			for (unsigned i = 0; i < fByteLen; ++i) {
				if (*myP != *hisP) {
					return false;
				}
				myP++;
				hisP++;
			}
			
			return true;
		}
		
		// Check for inequality
		bool operator!=(UTF8String const& s) const {
			return !operator==(s);
		}
		
		class ConstIterator {
			UTF8String const* fOwner;
			UTF8Byte const* fPtr;
			
			friend class UTF8String;
			
			ConstIterator(UTF8String const* s, UTF8Byte const* p)
			 : fOwner(s), fPtr(p)
			{
				//
			}
			
			void goForward() {
				fPtr += 1 + fPtr->getFollowerCount();
			}
			
			void goBackward() {
				do {
					fPtr--;
				} while (fPtr->isFollowByte());
			}
			
		public:
			//~ UTF8String const& getOwner() {
				//~ return *fOwner;
			//~ }
			
			bool operator==(ConstIterator const& it) const {
				return fOwner == it.fOwner && fPtr == it.fPtr;
			}
			
			bool operator!=(ConstIterator const& it) const {
				return !operator==(it);
			}
			
			ConstIterator next() const {
				ConstIterator it(*this); // copy
				it.goForward();
				return it;
			}
			
			ConstIterator prev() const {
				ConstIterator it(*this); // copy
				it.goBackward();
				return it;
			}
			
			// pre-increment
			ConstIterator& operator++() {
				this->goForward();
				return *this;
			}
			
			// post-increment
			ConstIterator operator++(int) {
				ConstIterator it(*this); // copy
				this->goForward();
				return it;
			}
			
			// pre-decrement
			ConstIterator& operator--() {
				this->goBackward();
				return *this;
			}
			
			// post-decrement
			ConstIterator operator--(int) {
				ConstIterator it(*this); // copy
				this->goBackward();
				return it;
			}
			
			// print the bytes of this character to a UTF8Consumer
			void bytes(UTF8Consumer& cons) const {
				for (unsigned i = 0; i <= fPtr->getFollowerCount(); ++i) {
					cons.consume(fPtr[i]);
				}
			}
			
			CodePoint operator*() const {
				return UTF8Byte::decodeFromValidStorage(fPtr);
			}
			
			boost::shared_ptr<CodePoint> operator->() const {
				CodePoint c(UTF8Byte::decodeFromValidStorage(fPtr));
				return boost::shared_ptr<CodePoint>(new CodePoint(c));
			}
		};
		
		ConstIterator begin() const {
			return ConstIterator(this, byteBegin());
		}
		
		ConstIterator end() const {
			return ConstIterator(this, byteEnd());
		}
		
		void print(UTF8Consumer& cons) const {
			UTF8Byte const* it = byteBegin();
			UTF8Byte const* endIt = byteEnd();
			
			while (it != endIt) {
				cons.consume(*it);
				++it;
			}
		}
		
		bool startsWith(CodePoint c) const {
			if (isEmpty()) {
				return false;
			}
			
			return *begin() == c;
		}
		
		bool endsWith(CodePoint c) const {
			if (isEmpty()) {
				return false;
			}
			
			return *--end() == c;
		}
		
		boost::shared_array<char const> toCString() const {
			char* mem = new char[fByteLen + 1];
			const UTF8Byte* b = byteBegin();
			for (unsigned long i = 0; i < fByteLen; ++b, ++i) {
				mem[i] = b->toByte();
			}
			mem[fByteLen] = 0;  // null pointer to terminate C-string
			
			return boost::shared_array<char const>(mem);
		}
		
		UTF8String subString(ConstIterator begin, ConstIterator end) const {
			unsigned offset = begin.fPtr - fStorage->get();
			
			// count characters
			unsigned charCount = 0;
			while (begin != end && begin.fPtr != fStorage->getEnd()) {
				++begin;
				++charCount;
			}
			
			unsigned endOffset = begin.fPtr - fStorage->get();
			unsigned byteLen = endOffset - offset;
			
			return UTF8String(fStorage, offset, charCount, byteLen);
		}
		
		UTF8String subString(ConstIterator begin, unsigned chars) const {
			unsigned offset = begin.fPtr - fStorage->get();
			
			unsigned charCount = 0;
			while (charCount < chars && begin.fPtr != fStorage->getEnd()) {
				++begin;
				++charCount;
			}
			
			unsigned endOffset = begin.fPtr - fStorage->get();
			unsigned byteLen = endOffset - offset;
			
			return UTF8String(fStorage, offset, charCount, byteLen);
		}
		
		UTF8String subString(ConstIterator begin) const {
			return subString(begin, this->end());
		}
		
		UTF8String left(unsigned chars) const {
			return subString(begin(), chars);
		}
		
		UTF8String left(ConstIterator end) const {
			return subString(begin(), end);
		}
		
		static UTF8String join(UTF8String const& s1, UTF8String const& s2);
		
		UTF8String compose(unsigned componentCount,
			UTF8String const* components) const;
		
		UTF8String compose(UTF8String component1) const {
			return compose(1, &component1);
		}
		
		UTF8String compose(UTF8String component1, UTF8String component2) const {
			UTF8String arr[] = {component1, component2};
			return compose(2, arr);
		}
		
		UTF8String
		compose(UTF8String component1, UTF8String component2,
			UTF8String component3) const
		{
			UTF8String arr[] = {component1, component2, component3};
			return compose(3, arr);
		}
		
		
	private:
		class Storage {
			UTF8Byte* const fPtr;
			UTF8Byte* const fEndPtr;
		public:
			explicit Storage(unsigned length)
			 : fPtr(new UTF8Byte[length]), fEndPtr(&fPtr[length])
			{
				//
			}
			
			~Storage() {
				delete[] fPtr;
			}
			
			/// Get pointer to first storage byte
			UTF8Byte* get() const {
				return fPtr;
			}
			
			/// Get pointer to one past the last storage byte
			UTF8Byte* getEnd() const {
				return fEndPtr;
			}
		};
		
		UTF8String(boost::shared_ptr<Storage> s, unsigned o, unsigned chars,
			unsigned bytes)
		 : fStorage(s), fOffset(o), fCharLen(chars), fByteLen(bytes)
		{
			//
		}
		
		static boost::shared_ptr<Storage>
		getEmptyStorage() {
			static boost::shared_ptr<Storage> st(new Storage(1));
			return st;
		}
		
		UTF8Byte* byteBegin() const {
			return &fStorage->get()[fOffset];
		}
		
		UTF8Byte* byteEnd() const {
			return &byteBegin()[fByteLen];
		}
		
		class ConsumerAdapter : public UTF8Consumer {
			UTF8Byte* fPtr;
			unsigned fCount;
		public:
			ConsumerAdapter(UTF8Byte* p)
			 : fPtr(p), fCount(0)
			{
				//
			}
			
			void consume(UTF8Byte b) {
				*fPtr = b;
				fPtr++;
				fCount++;
			}
			
			unsigned countConsumed() const {
				return fCount;
			}
		};
		
		boost::shared_ptr<Storage> fStorage;
		unsigned fOffset;
		unsigned fCharLen; // length in no. of characters, not bytes
		unsigned fByteLen; // length in UTF-8 bytes
	};
	
	inline UTF8String
	operator+(UTF8String const& s1, UTF8String const& s2) {
		return UTF8String::join(s1, s2);
	}
	
	inline bool operator<(UTF8String const& s1, UTF8String const& s2) {
		UTF8String::ConstIterator i1 = s1.begin();
		UTF8String::ConstIterator i2 = s2.begin();
		UTF8String::ConstIterator e1 = s1.end();
		UTF8String::ConstIterator e2 = s2.end();
		
		do {
			if (i1 == e1 && i2 != e2) {
				return true;
			}
			else if (i2 == e2) {
				return false;
			}
			
			CodePoint c1 = *i1;
			CodePoint c2 = *i2;
			
			if (c1 != c2) {
				return c1 < c2;
			}
			
			// still equal if we get here
			++i1;
			++i2;
		} while (true);
	}
	
	inline bool operator<=(UTF8String const& s1, UTF8String const& s2) {
		return s1 < s2 || s1 == s2;
	}
	
	inline bool operator>(UTF8String const& s1, UTF8String const& s2) {
		return !(s2 <= s1);
	}
	
	inline bool operator>=(UTF8String const& s1, UTF8String const& s2) {
		return !(s2 < s1);
	}
	
	/// Utility to build UTF8String instances from several pieces.
	class UTF8StringBuilder {
	public:
		UTF8StringBuilder()
		 : fLatestValid(true)
		{
			//
		}
		
		void append(CodePoint c) {
			fLatestValid = false;
			ConsumerAdapter adapt(fBuf);
			UTF8Byte::encode(c, adapt);
		}
		
		void append(UTF8String s) {
			if (s.isEmpty()) {
				return; // no real change
			}
			
			fLatestValid = false;
			ConsumerAdapter adapt(fBuf);
			s.print(adapt);
		}
		
		void operator+=(CodePoint c) {
			append(c);
		}
		
		void operator+=(UTF8String s) {
			append(s);
		}
		
		void newLine() {
			append(CodePoint::linefeed());
		}
		
		UTF8String toString() {
			if (fLatestValid) {
				return fLatest;
			}
			else {
				fLatest = UTF8String(&fBuf[0], fBuf.size());
				fLatestValid = true;
				return fLatest;
			}
		}
		
	private:
		class ConsumerAdapter : public UTF8Consumer {
			std::vector<UTF8Byte>& fPtr;
			unsigned fCount;
		public:
			ConsumerAdapter(std::vector<UTF8Byte>& p)
			 : fPtr(p), fCount(0)
			{
				//
			}
			
			void consume(UTF8Byte b) {
				fPtr.push_back(b);
				fCount++;
			}
			
			unsigned countConsumed() const {
				return fCount;
			}
		};
		
		std::vector<UTF8Byte> fBuf;
		UTF8String fLatest;
		bool fLatestValid;
	};
	
	/// Keeps a pool of UTF8String instances for char pointer literals.
	/*
		Depends on a compiler treating string literals as read-only, which
		will cause the char-pointers of identical string literals to be equal.
	*/
	class UTF8StringLiteralPool {
		typedef std::map<char const*, UTF8String> PoolMap;
		PoolMap fPool;
		PoolMap fQuotedPool;
		PoolMap fDoubleQuotedPool;
	public:
		static UTF8StringLiteralPool& getInstance() {
			static UTF8StringLiteralPool pool;
			return pool;
		}
		
		UTF8String literal(char const* lit) {
			PoolMap::iterator it = fPool.find(lit);
			if (it != fPool.end()) {
				// found
				return it->second;
			}
			
			// not found
			UTF8String str = UTF8String::fromCString(lit);
			
			// store it
			fPool[lit] = str;
			return str;
		}
		
		UTF8String quotedLiteral(char const* lit) {
			PoolMap::iterator it = fQuotedPool.find(lit);
			if (it != fQuotedPool.end()) {
				// found
				return it->second;
			}
			
			// not found; construct it first
			UTF8String str =
				literal("%1%2%3").compose(
					CodePoint::leftSingleQuote(),
					literal(lit),
					CodePoint::rightSingleQuote()
				);
			
			// and store it
			fQuotedPool[lit] = str;
			return str;
		}
		
		UTF8String doubleQuotedLiteral(char const* lit) {
			PoolMap::iterator it = fDoubleQuotedPool.find(lit);
			if (it != fDoubleQuotedPool.end()) {
				// found
				return it->second;
			}
			
			// not found; construct it first
			UTF8String str =
				literal("%1%2%3").compose(
					CodePoint::leftDoubleQuote(),
					literal(lit),
					CodePoint::rightDoubleQuote()
				);
			
			// and store it
			fDoubleQuotedPool[lit] = str;
			return str;
		}
	};
}

inline HYPER_Common::UTF8String quote(HYPER_Common::UTF8String s) {
	return UTF8STR("%1%2%3").compose(
		HYPER_Common::CodePoint::leftSingleQuote(),
		s,
		HYPER_Common::CodePoint::rightSingleQuote()
	);
}

inline HYPER_Common::UTF8String doubleQuote(HYPER_Common::UTF8String s) {
	return UTF8STR("%1%2%3").compose(
		HYPER_Common::CodePoint::leftDoubleQuote(),
		s,
		HYPER_Common::CodePoint::rightDoubleQuote()
	);
}
#endif
