/*!
	\file utf8verifier.cxx
	\brief Implements the UTF8Verifier class
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2008-03-16
// Last time modified: 	2008-03-24

#include "utf8verifier.h"

//#include <iostream>

namespace HYPER_Common {
	
	class UTF8Verifier::UTF8Dequer : public UTF8Consumer {
	public:
		UTF8Dequer(std::deque<UTF8Byte>& d) : fDeque(d)
		{
			//
		}
		
		void consume(UTF8Byte b) {
			fDeque.push_front(b);
		}
		
	private:
		std::deque<UTF8Byte>& fDeque;
	};
	
	UTF8Verifier::ProcessStatus
	UTF8Verifier::processData(std::deque<UTF8Byte>& input,
		bool endOfInputReached, std::deque<UTF8Byte>& output,
		State& processState) const
	{
		// get state object
		VerifyState& state = dynamic_cast<VerifyState&>(processState);
		
		//std::cout << "verifier running" << std::endl;
		
		// skip BOM if present and skipping is allowed
		if (state.fgAtStartOfInput && state.fgMaySkipBOM
				&& input[0] == UTF8Byte(0xEF))
		{
			if (input.size() < 3) {
				// cannot check if it's really a BOM
				state.fgInputNeeded = 3;
				return kNeedMoreData;
			}
			
			// check for BOM presence
			if (input[1] == UTF8Byte(0xBB) && input[2] == UTF8Byte(0xBF)) {
				// BOM present, get rid of it
				input.pop_front();
				input.pop_front();
				input.pop_front();
				state.fgAtStartOfInput = false;
				
				if (input.empty()) {
					state.fgInputNeeded = 1;
					return kNeedMoreData;
				}
			}
		}
		
		UTF8Byte firstB = input.front();
		
		//std::cout << " first byte: " << (int)firstB.toByte() << std::endl;
		
		if (firstB.isInvalid()) {
			input.pop_front();
			UTF8Dequer d(output);
			UTF8Byte::encode(CodePoint::invertedQuestionMark(), d);
			return kOK;
		}
		
		if (firstB.isFollowByte()) { // follow byte without lead
			// we'll skip up to three follow bytes
			if (input.size() < 3) {
				state.fgInputNeeded = 3;
				return kNeedMoreData;
			}
			
			UTF8Dequer d(output);
			UTF8Byte::encode(CodePoint::invertedQuestionMark(), d);
			input.pop_front(); // pop first follow byte
			if (input[0].isFollowByte()) {
				input.pop_front(); // pop second follow byte
				if (input[0].isFollowByte()) {
					input.pop_front(); // pop third follow byte
				}
			}
			
			state.fgInputNeeded = 1;
			return kOK;
		}
		
		// At this point we either have a lead byte or a single byte.
		// First check that we have enough input bytes.
		
		unsigned bytesNeeded = 1 + firstB.getFollowerCount();
		//std::cout << " bytes needed: " << bytesNeeded << std::endl;
		if (bytesNeeded > input.size()) {
			state.fgInputNeeded = bytesNeeded;
			return kNeedMoreData;
		}
		
		//std::cout << " going to compose character" << std::endl;
		
		// put bytes in small array buffer
		UTF8Byte buf[4];
		buf[0] = firstB;
		input.pop_front(); // pop first byte
		bool followersOK = true;
		for (unsigned i = 1; i < bytesNeeded; ++i) {
			buf[i] = input.front();
			if (buf[i].isFollowByte()) {
				input.pop_front();
			}
			else {
				// not enough (valid) follow bytes
				followersOK = false;
				break;
			}
		}
		
		if (followersOK) {
			//std::cout << " char OK" << std::endl;
			for (unsigned i = 0; i < bytesNeeded; ++i) {
				//std::cout << "  pushing byte " << (int)buf[i].toByte() << std::endl;
				output.push_front(buf[i]);
			}
		}
		else {
			//std::cout << " followers NOT OK" << std::endl;
			// one or more follow bytes not OK; invalid codepoint
			UTF8Dequer d(output);
			UTF8Byte::encode(CodePoint::invertedQuestionMark(), d);
		}
		
		state.fgInputNeeded = 1; // next byte could be single char
		return kOK;
	}
}
