/*!
	\file accesschecker.cxx
	\brief Contains the AccessChecker class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-11-07
// Last time modified:	2008-06-26

#include "accesschecker.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "class.h"
#include "classcontext.h"
#include "compileinfo.h"
#include "declareditem.h"
#include "field.h"
#include "overloadedconstructor.h"
#include "overloadedprocedure.h"
#include "type.h"

namespace HYPER_Compiler {
	
	bool
	AccessChecker::check(ClassContext const& from,
		DeclaredItem const& item, Location const* loc, CompileInfo& info)
	{
		//debug << "DIRECT ACCESS CHECK; to item " + item.getDescription() << '\n';
		
		Class const* destClass = 0;
		ClassMemberAccess memberAccess =
			ClassMemberAccess::createPublic();
		
		// Check for parent class of 'item' and if 'item' is a public member
		getInfo(item, memberAccess, destClass);
		
		if (destClass == 0) { // class unknown
			return true; // can't check / check not needed
		}
		
		// the class we start the access attempt from
		Class const* startClass = from.getClass();
		hyp_assert(startClass != 0);
		
		if (startClass == destClass) {
			// OK
			return true;
		}
		
		if (isInside(startClass, destClass)) {
			// OK
			return true;
		}
		
		while (startClass != 0 && !isInside(destClass, startClass)) {
			startClass = startClass->getParentClass();
		}
		
		if (!memberAccess.isPublic()) {
			//debug << "accessing non-public item from outside" <<'\n';
			
			info.diag().error(
				loc,
				UTF8STR("cannot access %1 %2 in %3 from this context"),
				memberAccess,
				item.getDescription(),
				destClass->getDescription()
			).addInfo(
				item.getLocation(),
				UTF8STR("declaration of %1 is here"),
				item.getDescription()
			);
			
			return false;
		}
		
		//checkAccessInto(startClass, destClass, loc, info);
		return true;
	}
	
	bool
	AccessChecker::checkMember(ClassContext const& from,
		DeclaredItem const& item, HYPER_Common::UTF8String memberName,
		Location const* loc, CompileInfo& info)
	{
		//debug << "MEMBER ACCESS CHECK; to item " + item.getDescription() <<'\n';
		
		Class const* itemOwnerClass = 0;
		ClassMemberAccess memberAccess = ClassMemberAccess::createPublic();
		
		// Check for parent class of 'item' and if 'item' is a public member
		getInfo(item, memberAccess, itemOwnerClass);
		
		if (itemOwnerClass == 0) { // class unknown
			return true; // can't check / check not needed
		}
		
		if (memberAccess.isPublic()) { // 'item' is public
			// access granted
			return true;
		}
		
		// the class we start the access attempt from
		Class const* startClass = from.getClass();
		hyp_assert(startClass != 0);
		
		if (startClass == itemOwnerClass
			|| isInside(startClass, itemOwnerClass))
		{
			return true; // OK
		}
		else {
			//debug << "accessing non-public member from outside" <<'\n';
			
			info.diag().error(
				loc,
				UTF8STR("cannot access %1 member %2 in %3 from this context"),
				memberAccess,
				quote(memberName),
				itemOwnerClass->getDescription()
			).addInfo(
				item.getLocation(),
				UTF8STR("declaration of %1 is here"),
				item.getDescription()
			);
			
			return false;
		}
	}
	
	void
	AccessChecker::getInfo(DeclaredItem const& item,
		ClassMemberAccess& memberAccess, Class const*& memberOwner)
	{
		memberOwner = item.getOwnerClass();
		if (memberOwner != 0) {
			memberAccess = item.getAccessInOwner();
		}
	}
	
	bool AccessChecker::isInside(Class const* a, Class const* b) {
		hyp_assert(a != b);
		Class const* curClass = a;
		
		while (true) {
			if (curClass == b) {
				// a is inside b
				
				//debug << a->getTypeName() << " is inside "
				//	<< b->getTypeName() <<'\n';
				return true;
			}
			else if (curClass == 0) {
				// a is not inside b
				
				//debug << a->getTypeName() << " is NOT inside "
				//	<< b->getTypeName() <<'\n';
				return false;
			}
			else {
				// go one class higher
				curClass = curClass->getParentClass();
			}
		}
	}
	
	// The code below is not needed yet, and has gotten out of date by the way.
	
	/*void AccessChecker::checkAccessInto(Class const* a, Class const* b,
		Location const* loc, CompileInfo& info)
	{
		// b should be inside a
		// a can be 0, this means access from outside the root parent of b
		
		//debug << "checkAccessInto : from "
		//	<< (a == 0 ? "<root>" : a->getTypeName()) << " to "
		//	<< b->getTypeName() <<'\n';
		
		Class const * curC = b;
		
		do {
			ClassMemberAccess acc = curC->getAccess();
			Class const* nextC = curC->getParentClass();
			
			if (!acc.isPublic() && nextC != a) { // no public access ?
				//debug << "non-public: " << curC->getTypeName() <<'\n';
				
				Error* e =
					new Error(
						"access error", loc
					);
				info.getErrorManager().add(e);
			}
			
			curC = nextC;
		} while (curC != 0 && curC != a);
	}*/
}
