/*!
	\file arrayclass.cxx
	\author Kevin André
	\brief Implements the Array class
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-08-29
// Last time modified:	2009-08-30

#include "arrayclass.h"

#include "base/easydiagnostics.h"
#include "base/location.h"

#include "arrayoperator.h"
#include "builtinconstructor.h"
#include "builtinprocedure.h"
#include "builtintypes.h"
#include "compileinfo.h"
#include "declareditem.h"
#include "expression.h"
#include "initializationhandler.h"
#include "parameter.h"
#include "type.h"

//~ #include "base/debug.h"

namespace HYPER_Compiler {
	
	// *** Util classes ***
	
	class Array::ArrayConstructor : public BuiltInConstructor {
	public:
		ArrayConstructor(Array& ac);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParam;
	};
	
	class Array::IndexOperator : public ArrayOperator {
	public:
		IndexOperator(std::vector<Parameter const*> parameters,
			Type const* base);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* operand,
			std::vector<ConstantValue const*> params);
		
		// TODO : change return type
		virtual void generateCode();
	
	private:
		std::vector<Parameter const*> fParam;
		Type const* fBase;
	};
	
	/// Serves as the procedures "size()", "size0(), "size1()" etc...
	class Array::ProcedureSize : public BuiltInProcedure {
	public:
		ProcedureSize(Array& arr, unsigned dimNo, bool multiDimension);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		Array& fArr;
		std::vector<Parameter const*> fParams;
	};
	
	/// Initialization handler for array types.
	class Array::ArrayInitHandler : public InitializationHandler {
	public:
		TriBool
		checkInitByDefaultValue(ClassContext const& ctxt,
			Location const* useLoc, CompileInfo& info);
		
		TriBool
		checkInitByExpression(Expression* e, ClassContext const& ctxt,
			Location const* loc, CompileInfo& info);
		
		//void checkAccess(ClassContext const& ctxt,
		//	Location const* loc, CompileInfo& info);
		
		ArrayInitHandler(Array const* arr);
		
	private:
		Array const* fArray;
		InitializationHandler* fBaseInit;
	};
	
	
	// *** Array class ***
	
	Array::Array(CompileInfo& info, const Type* baseType,
		unsigned int dimensions)
	 : BuiltInClass(info, kSealed), fDims(dimensions), fBaseType(baseType)
	{
		// no action
	}
	
	Array::~Array() {
		// no action
	}
	
	HYPER_Common::UTF8String Array::getTypeName() const {
		//~ debug.line(UTF8STR("Array::getTypeName() called"));
		HYPER_Common::UTF8StringBuilder sb;
		
		for (unsigned i = 0; i < fDims; ++i) {
			sb += UTF8STR("[ ]");
		}
		
		sb += UTF8STR(" ");
		sb += fBaseType->toString();
		return sb.toString();
	}
	
	HYPER_Common::UTF8String Array::getDeclarationName() const {
		return getTypeName();
	}
	
	Type const* Array::getBaseType() const {
		return fBaseType;
	}
	
	unsigned int Array::getDimensions() const {
		return fDims;
	}
	
	void Array::buildSymbolTable() {
		createConstructors();
		createMembers();
		createOperators();
	}
	
	void Array::createConstructors() {
		registerConstructor(new ArrayConstructor(*this), false);
	}
	
	void Array::createMembers() {
		// create member procedures
		bool multiDim = fDims > 1;
		for (unsigned int i = 0; i < fDims; ++i) {
			registerProcedure(new ProcedureSize(*this, i, multiDim));
		}
	}
	
	void Array::createOperators() {
		// parameter list for array operator
		std::vector<Parameter const*> args;
		for (unsigned int i = 0; i < fDims; ++i) {
			args.push_back(
				createParameter(
					UTF8STR("i") + HYPER_Common::UTF8String(i + 1),
					BuiltInTypes::kNat
				)
			);
		}
		
		// create array operator
		registerOperator(new IndexOperator(args, fBaseType));
	}
	
	InitializationHandler* Array::getInitHandler() const {
		return new ArrayInitHandler(this);
	}
	
	
	/* ArrayConstructor class */
	
	Array::ArrayConstructor::ArrayConstructor(Array& ac)
	 : BuiltInConstructor(&ac, true)
	{
		// no action
	}
	
	const std::vector<Parameter const*>*
	Array::ArrayConstructor::getParameterList() const {
		return &fParam;
	}
	
	
	/* IndexOperator class */
	
	Array::IndexOperator::IndexOperator(std::vector<Parameter const*> params,
		Type const* base)
	 : ArrayOperator(
			params, Type::createExtraPointer(base), Location::getBuiltIn()
		), fParam(params), fBase(base)
	{
		//
	}
	
	ConstantValue const*
	Array::IndexOperator::evaluateCompileTime(
		ConstantValue const* operand,
		std::vector<ConstantValue const*> params)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Array::IndexOperator::generateCode() {
		// TODO
	}
	
	
	/* ProcedureSize class */
	
	Array::ProcedureSize::ProcedureSize(Array& arr, unsigned dimNo,
		bool multiDimension)
	 : BuiltInProcedure(
			&arr,
			UTF8STR("size")
				+ (multiDimension ? HYPER_Common::UTF8String(dimNo) : UTF8STR("")),
			kSealed, ClassMemberAccess::createPublic(),
			true, BuiltInTypes::createNat(arr.getCompileInfo())
		),
		fArr(arr)
	{
		// no parameters
	}
	
	const std::vector<Parameter const*>*
	Array::ProcedureSize::getParameterList() const {
		return &fParams;
	}
	
	
	// *** ArrayInitHandler ***
	
	TriBool
	Array::ArrayInitHandler::checkInitByDefaultValue(
		ClassContext const& ctxt,
		Location const* useLoc, CompileInfo& info)
	{
		if (fArray->getBaseType()->isNonNullPointer()) {
			// array of pointer type -> OK
			// an array like this is initialized with null pointer values
			return true;
		}
		
		// get initialization handler for base type
		fBaseInit = fArray->getBaseType()->getInitHandler();
		
		// check initialization of array element
		return fBaseInit->checkInitByDefaultValue(
			ctxt, useLoc, info
		);
	}
	
	TriBool
	Array::ArrayInitHandler::checkInitByExpression(Expression* e,
		ClassContext const& ctxt, Location const* loc, CompileInfo& info)
	{
		// initializing an array with an expression is *not* possible!
		info.diag().error(
			UTF8STR("an array cannot be initialized with an expression"),
			loc
		).addInfo(
			UTF8STR("expression: ") + e->toString(),
			e->getLocation()
		);
		return false;
	}
	
	//void
	//ArrayInitHandler::checkAccess(ClassContext const& ctxt,
	//	Location const* loc, CompileInfo& info)
	//{
		// TODO
	//}
	
	Array::ArrayInitHandler::ArrayInitHandler(Array const* arr)
	 : fArray(arr), fBaseInit(0)
	{
		//
	}
}
