/*!
	\file arrayoperatorexpression.cxx
	\brief Implements the ArrayOperatorExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-10
// Last time modified:	2009-08-23

#include "arrayoperatorexpression.h"

#include "arrayoperator.h"
#include "arraytype.h"
#include "class.h"
#include "compileinfo.h"
#include "constantvalue.h"
#include "overloadedarrayoperator.h"
#include "overloadresolution.h"
#include "type.h"

#include <algorithm>

namespace HYPER_Compiler {
	
	ArrayOperatorExpression::ArrayOperatorExpression(
		Expression* arrayExpr, Expression* arraySelector)
	 : Expression(arrayExpr->getCompileInfo(), arrayExpr->getLocation()),
		fArrayExpr(arrayExpr),
		fIsClassLValue(TriBool::uncertain),
		fIsPointerLValue(TriBool::uncertain)
	{
		fSelectors.push_back(arraySelector);
	}
	
	ArrayOperatorExpression::~ArrayOperatorExpression() {
		// nothing here
	}
	
	void ArrayOperatorExpression::addSelector(Expression* s) {
		fSelectors.push_back(s);
	}
	
	HYPER_Common::UTF8String ArrayOperatorExpression::toString() const {
		HYPER_Common::UTF8StringBuilder sb;
		sb += fArrayExpr->toString();
		sb += UTF8STR("[");
		
		for (unsigned int i = 0; i < fSelectors.size(); ++i) {
			if (i > 0) {
				sb += UTF8STR(", ");
			}
			
			sb += fSelectors[i]->toString();
		}
		
		sb += UTF8STR("]");
		
		return sb.toString();
	}
	
	void ArrayOperatorExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		fArrayExpr->putInSymbolTable(loc);
		for (unsigned int i = 0; i < fSelectors.size(); ++i) {
			fSelectors[i]->putInSymbolTable(loc);
		}
	}
	
	bool ArrayOperatorExpression::doFullResolve(Trace const& trace) {
		bool resolveOK = fArrayExpr->resolveFull(trace);
		resolveOK &= fArrayExpr->checkReturnsValue();
		
		bool selectorsOK = true;
		for (unsigned int i = 0; i < fSelectors.size(); ++i) {
			Expression* selector = fSelectors[i];
			selectorsOK &= selector->resolveFull(trace);
			selector->checkReturnsValue();
			selectorsOK &= selector->getResult().isResolved();
		}
		resolveOK &= selectorsOK;
		
		ExpressionResult const& arrRes = fArrayExpr->getResult();
		if (arrRes.isResolved() && selectorsOK) {
			// copy selector list (need a vector of pointers to const Expressions)
			std::vector<Expression const*> selectors;
			selectors.resize(fSelectors.size());
			std::copy(fSelectors.begin(), fSelectors.end(), selectors.begin());
			
			// check array selectors
			resolveOK &=
				checkSelectors(
					selectors, 0, arrRes.getType()->getDereferenced()
				);
		}
		
		return resolveOK;
	}
	
	TriBool ArrayOperatorExpression::isClassLValue() const {
		return fIsClassLValue;
	}
	
	TriBool ArrayOperatorExpression::isPointerLValue() const {
		return fIsPointerLValue;
	}
	
	TriBool ArrayOperatorExpression::isCall() const {
		return false;
	}
	
	TriBool ArrayOperatorExpression::hasResult() const {
		return true;
	}
	
	TriBool ArrayOperatorExpression::resultAtCompileTime() const {
		return false; // FIXME
	}
	
	HYPER_Common::UTF8String
	ArrayOperatorExpression::printSelectors(unsigned begin, unsigned end) const
	{
		HYPER_Common::UTF8StringBuilder sb;
		
		for (unsigned i = begin; i < end; ++i)
		{
			sb += UTF8STR("[");
			sb += fSelectors[i]->toString();
			sb += UTF8STR("]");
		}
		
		return sb.toString();
	}
	
	HYPER_Common::UTF8String
	ArrayOperatorExpression::printSelectorTypes(unsigned begin,
		unsigned end) const
	{
		HYPER_Common::UTF8StringBuilder sb;
		
		for (unsigned i = begin; i < end; ++i)
		{
			sb += UTF8STR("[");
			sb += fSelectors[i]->getResult().getType()->toString();
			sb += UTF8STR("]");
		}
		
		return sb.toString();
	}
	
	HYPER_Common::UTF8String
	ArrayOperatorExpression::printSelectionBase(unsigned beginExclude) const {
		if (beginExclude > 0) {
			HYPER_Common::UTF8StringBuilder sb;
			
			// put between brackets
			sb += UTF8STR("(");
			sb += fArrayExpr->toString();
			sb += UTF8STR(")");
			
			for (unsigned i = 0; i < beginExclude; ++i) {
				sb += UTF8STR("[");
				sb += fSelectors[i]->toString();
				sb += UTF8STR("]");
			}
			
			return sb.toString();
		}
		else {
			return fArrayExpr->toString();
		}
	}
	
	bool
	ArrayOperatorExpression::checkSelectors(
		const std::vector<Expression const*>& selectors, unsigned firstSelector,
		Type const* curType)
	{
		unsigned int maxSelectors = fSelectors.size() - firstSelector;
		
		// get array class
		const Class* cls = curType->getDereferenced()->getClass();
		
		OverloadedArrayOperator const* ovArrOp = cls->getArrayOperators();
		if (ovArrOp == 0) { // no array operators in class
			// create error message
			HYPER_Common::UTF8String ops =
				printSelectors(firstSelector, fSelectors.size());
			
			HYPER_Common::UTF8String operand =
				printSelectionBase(firstSelector);
			
			getCompileInfo().diag().error(
				UTF8STR(
					"array selection operator(s) used on class %1 which does "
						"not support the array operator"
				).compose(
					quote(curType->getDereferenced()->toString())
				),
				fSelectors[firstSelector]->getLocation()
			).addInfo(
				UTF8STR("operand: ") + operand,
				fArrayExpr->getLocation()
			).addInfo(
				UTF8STR("operator(s) used: ") + ops,
				fSelectors[firstSelector]->getLocation()
			);
			
			return false;
		}
		
		unsigned classArrDims = ovArrOp->getDimensionCount();
		
		// too few selectors for the array operators?
		if (maxSelectors < classArrDims) {
			// create error message
			HYPER_Common::UTF8StringBuilder sb;
			
			sb += UTF8STR("array selection ");
			sb += quote(printSelectors(firstSelector, fSelectors.size()));
			sb += UTF8STR(" provides only ");
			
			if (maxSelectors == 1) {
				sb += UTF8STR("1 dimension");
			}
			else {
				sb += maxSelectors;
				sb += UTF8STR(" dimensions");
			}
			
			sb += UTF8STR(", while it is used on operand ");
			sb += quote(printSelectionBase(firstSelector));
			sb += UTF8STR(" that has type ");
			sb += quote(curType->toString());
			sb += UTF8STR(" and requires ");
			sb += classArrDims;
			sb += UTF8STR(" dimensions");
			
			getCompileInfo().diag().error(
				sb.toString(), selectors[firstSelector]->getLocation()
			);
			
			return false;
		}
		
		typedef OverloadResolution<OverloadedArrayOperator> ORA;
		ORA::Result call;
		
		// can we use the entire list of selectors?
		if (firstSelector == 0 && classArrDims == fSelectors.size()) { // yes
			// then do so
			call = ORA::resolveCall(ovArrOp, selectors, getCompileInfo());
		}
		else { // no
			// FIXME: find a way to get rid of this copying
			
			// create a new vector for the selectors we need
			std::vector<Expression const*> args;
			for (unsigned i = 0; i < classArrDims; ++i) {
				args.push_back(selectors[firstSelector + i]);
			}
			
			// now do overload resolution
			call = ORA::resolveCall(ovArrOp, args, getCompileInfo());
		}
		
		switch (call.getResult()) {
		case ORA::Result::kNotResolved:
			return false;
			
		case ORA::Result::kNoCandidates: {
			HYPER_Common::UTF8StringBuilder sb;
			sb += UTF8STR("no match for array selection ");
			sb +=
				quote(printSelectorTypes(
					firstSelector, firstSelector + classArrDims
				));
			sb += UTF8STR(", used on type ");
			sb += quote(curType->toString());
			sb += UTF8STR(" and needed for expression ");
			sb += quote(printSelectionBase(firstSelector));
			
			getCompileInfo().diag().error(sb.toString(), getLocation());
			
			return false; }
			
		case ORA::Result::kMultipleCandidates: {
			HYPER_Common::UTF8StringBuilder sb;
			sb += UTF8STR("multiple matches for array selection ");
			sb +=
				quote(printSelectorTypes(
					firstSelector, firstSelector + classArrDims
				));
			sb += UTF8STR(", used on type ");
			sb += quote(curType->toString());
			sb += UTF8STR(" and needed for expression ");
			sb += quote(printSelectionBase(firstSelector));
			
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(sb.toString(), getLocation());
			
			call.listCandidates(e);
			
			return false; }
		
		case ORA::Result::kSingleCandidate:
			break; // handled below the switch statement
		}
		
		bool resolveOK = true;
		
		// on a built-in array type, check selectors for out-of-range errors
		if (curType->isArray()) {
			const std::vector<Dimension>& dims =
				curType->getArrayDimensions();
			
			for (unsigned i = 0; i < classArrDims; ++i) {
				Dimension const& dim = dims[i];
				if (dim.isOpen() || dim.isUnknown()) continue;
				
				ConstantValue::nat_t dimSize = dim.getSize();
				
				Expression const* indexExpr = fSelectors[firstSelector + i];
				ExpressionResult const& indexRes = indexExpr->getResult();
				if (!indexRes.knownValue()) continue;
				if (indexRes.getValue()->getConstantType()
						!= ConstantValue::kUnsignedInt)
				{
					continue;
				}
				
				ConstantValue::nat_t indexValue =
					indexRes.getValue()->getUnsigned();
				
				if (indexValue >= dimSize) {
					// index out of range
					
					getCompileInfo().diag().error(
						UTF8STR(
							"array index out of range; value %1 should be "
							"in the range 0 .. %2"
						).compose(
							indexValue, dimSize
						),
						indexExpr->getLocation()
					);
					
					resolveOK = false;
				}
			}
		}
		
		if (maxSelectors == classArrDims) { // all selectors used now...
			// now we know the result of this entire array expression
			
			Type const* resultT =
				call.getCandidates()[0]->getEntity()->getResultType();
			fIsClassLValue = resultT->isNonNullPointer();
			fIsPointerLValue = resultT->isDoublePointer();
			
			setResolved();
			setResultType(resultT);
		}
		else { // we still have selectors left...
			// check recursively
			resolveOK &= 
				checkSelectors(
					selectors, firstSelector + classArrDims,
					call.getCandidates()[0]->getEntity()->getResultType()
				);
		}
		
		return resolveOK;
	}
}
