/*!
	\file arraytype.cxx
	\brief Implements the ArrayType class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-12-16
// Last time modified:	2008-10-25

#include "arraytype.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "arrayclass.h"
#include "classrepository.h"
#include "compileinfo.h"
#include "constantvalue.h"
#include "conversions.h"
#include "expression.h"
#include "type.h"

//~ #include "base/debug.h"

namespace HYPER_Compiler {
	
	ArrayType::ArrayType(std::vector<Expression*> dimensions,
		DeclaredType* base, const Location* loc)
	: DeclaredType(base->getCompileInfo(), loc), fBaseType(base),
		fArrayDims(dimensions)
	{
		// the base type must be allocatable directly (as part of the array)
		fBaseType->setRestrictions(true, mustHaveCompileTimeSize());
	}
	
	ArrayType::ArrayType(ArrayType const& at)
	: DeclaredType(at), fBaseType(at.fBaseType), fArrayDims(at.fArrayDims)
	{
		// copy constructor should not be used
		hyp_assert_unreachable();
	}
	
	bool ArrayType::isArray() const {
		return true;
	}
	
	bool ArrayType::isPointer() const {
		return false;
	}
	
	bool ArrayType::isDoublePointer() const {
		return false;
	}
	
	bool ArrayType::isClass() const {
		return true;
	}
	
	bool ArrayType::isConst() const {
		// an array is const iff its base type is const
		return fBaseType->isConst();
	}
	
	void ArrayType::makeConst() {
		// an array is const iff its base type is const
		fBaseType->makeConst();
	}
	
	HYPER_Common::UTF8String ArrayType::toString() const {
		//~ debug.line(UTF8STR("ArrayType::toString called"));
		HYPER_Common::UTF8StringBuilder sb;
		
		for (unsigned int i = 0; i < fArrayDims.size(); ++i) {
			Expression* dim = fArrayDims[i];
			sb += UTF8STR("[");
			sb += (dim != 0 ? dim->toString() : UTF8STR(""));
			sb += UTF8STR("]");
		}
		
		// (array is never `const')
		sb += UTF8STR(" ");
		sb += fBaseType->toString();
		return sb.toString();
	}
	
	void ArrayType::putInSymbolTable(LookupLocation const* loc) {
		DeclaredType::putInSymbolTable(loc);
		fBaseType->putInSymbolTable(loc);
		for (unsigned int i = 0; i < fArrayDims.size(); ++i) {
			Expression* dim = fArrayDims[i];
			if (dim) {
				dim->putInSymbolTable(loc);
			}
		}
	}
	
	bool ArrayType::doInterfaceResolve(Trace const& trace) {
		//debug.line("ArrayType::doInterfaceResolve called");
		
		// resolve array base type
		bool resolveOK = fBaseType->resolveInterface(trace);
		
		CompileInfo& info = getCompileInfo();
		
		// check array dimensions
		bool openArrayErrorGiven = false;
		std::vector<Dimension> tDims;
		for (unsigned int i = 0; i < fArrayDims.size(); ++i) {
			Expression* dim = fArrayDims[i];
			
			if (dim == 0) { // dimension size omitted...
				tDims.push_back(Dimension::createOpen());
				
				if (mustBeDirectAllocatable() && !openArrayErrorGiven) {
					info.diag().error(
						getLocation(),
						UTF8STR(
							"open array %1 not allowed in "
							"this context; all dimension sizes that are "
							"not behind a pointer need to be specified "
							"in order to allocate space for the array"
						),
						quote(toString())
					);
					openArrayErrorGiven = true;
				}
			}
			else { // dimension size present...
				resolveOK &= dim->resolveFull(trace);
				resolveOK &= resolveDim(dim);
				
				if (dim->resultAtCompileTime().isTrue()) {
					// dimension has size known at compile-time
					ConstantValue::nat_t dimSize =
						dim->getResult().getValue()->getUnsigned();
					
					if (dimSize == 0) {
						// Correct zero size; a diagnostic was issued in the
						// method "resolveDim".
						dimSize = 1;
					}
					
					tDims.push_back(Dimension::create(dimSize));
				}
				else {
					// dimension size not known at compile time
					tDims.push_back(Dimension::createUnknown());
					
					if (mustHaveCompileTimeSize()) {
						info.diag().error(
							dim->getLocation(),
							UTF8STR(
								"array size %1 in type %2 should "
								"be known at compile time but is not"
							),
							quote(dim->toString()),
							quote(toString())
						);
						resolveOK = false;
					}
				}
			}
		}
		resolveOK &= !openArrayErrorGiven;
		
		hyp_assert(tDims.size() == fArrayDims.size());
		
		if (resolveOK) {
			Type const* resolvedBase = fBaseType->getResolvedType();
			hyp_assert(resolvedBase != 0); // because resolving succeeded
			
			setResolvedType(
				Type::createArrayType(tDims, fBaseType->getResolvedType())
			);
		}
		
		return resolveOK;
	}
	
	bool ArrayType::resolveDim(Expression* dim) {
		if (!dim->checkReturnsValue()) {
			return false;
		}
		
		// check size type
		
		Type const* t = dim->getResult().getType();
		ValuePassing vp =
			BuiltInTypes::getNatConversion(
				getCompileInfo(), t->getDereferenced()
			);
		
		if (!vp.isPossible()) { // no conversion...
			getCompileInfo().diag().error(
				dim->getLocation(),
				UTF8STR(
					"array index expression %1 has type %2, which is "
					"not automatically convertable to %3"
				),
				quote(dim->toString()),
				quote(t->toString()),
				QUOTED_UTF8STR("nat")
			);
			return false;
		}
		
		// check compile-time value
		ConstantValue const* dimSize = dim->getResult().getValue();
		if (dimSize != 0
			&& dimSize->getConstantType() == ConstantValue::kUnsignedInt)
		{
			if (dimSize->getUnsigned() == 0) {
				getCompileInfo().diag().error(
					dim->getLocation(),
					UTF8STR(
						"a zero-sized array dimension is not allowed; "
						"the offending dimension expression is %1"
					),
					quote(dim->toString())
				);
				return false;
			}
		}
		
		// TODO : check access
		//vp.checkAccess(*fContext, test->getLocation(), info);
		
		return true;
	}
}
