/*!
	\file assignmentstatement.cxx
	\brief Implementation of the AssignmentStatement class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-12
// Last time modified:	2009-01-05

#include "assignmentstatement.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

#include "assignmentoperator.h"
#include "compileinfo.h"
#include "expression.h"
#include "operatorcalls.h"
#include "simpleparameter.h"
#include "statementcontext.h"
#include "type.h"

#include "base/debug.h"

namespace HYPER_Compiler {
	
	AssignmentStatement::AssignmentStatement(Expression* lValue,
		Expression* rValue, Operators::AssignmentOperator op,
		StatementContext const* c, const Location* pos, CompileInfo& info)
	 : Statement(info, pos), fLValue(lValue), fRValue(rValue), fOpSym(op),
		fOp(0), fContext(c)
	{
		hyp_assert(lValue != 0);
		hyp_assert(rValue != 0);
	}
	
	AssignmentStatement::~AssignmentStatement() {
		// nothing here
	}
	
	HYPER_Common::UTF8String AssignmentStatement::toString() const {
		return UTF8STR("%1 %2 %3").compose(
			fLValue->toString(), Operators::toString(fOpSym),
			fRValue->toString()
		);
	}
	
	HYPER_Common::UTF8String AssignmentStatement::getSyntaxName() const {
		return UTF8STR("Assignment");
	}
	
	void
	AssignmentStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const
	{
		ind.line(getSyntaxName());
		
		HYPER_Common::Indent out(ind);
		
		ind.line(UTF8STR("LValue: ") + fLValue->toString());
		ind.line(UTF8STR("Operator: ") + Operators::toString(fOpSym));
		ind.line(UTF8STR("RValue: ") + fRValue->toString());
	}
	
	Expression* AssignmentStatement::getLValue() const {
		return fLValue;
	}
	
	Expression* AssignmentStatement::getRValue() const {
		return fRValue;
	}
	
	HYPER_Common::UTF8String
	AssignmentStatement::getResolveDescription() const
	{
		return UTF8STR("assignment statement");
	}
	
	void AssignmentStatement::putInSymbolTable(ScopeLocation const& loc) {
		Statement::putInSymbolTable(loc);
		LookupLocation const* ll = new LookupLocation(loc);
		fLValue->putInSymbolTable(ll);
		fRValue->putInSymbolTable(ll);
	}
	
	bool AssignmentStatement::doFullResolve(Trace const& trace) {
		bool resolveOK = true;
		
		resolveOK &= fLValue->resolveFull(trace);
		resolveOK &= fRValue->resolveFull(trace);
		bool lValOK = fLValue->checkReturnsValue();
		bool rValOK = fRValue->checkReturnsValue();
		
		if (!lValOK || !rValOK) {
			resolveOK = false;
		}
		else if (fOpSym == Operators::kPointerAssign) { // pointer assignment
			resolveOK &= resolvePointerAssignment(getCompileInfo());
		}
		else { // object assignment
			resolveOK &= resolveOverloading(getCompileInfo());
		}
		
		return resolveOK;
	}
	
	bool AssignmentStatement::resolvePointerAssignment(CompileInfo& info)
	{
		TriBool lvTest = fLValue->isPointerLValue();
		if (lvTest.isFalse()) {
			info.diag().error(
				UTF8STR(
					"left-side expression of pointer assignment is not a "
						"valid pointer L-value"
				),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("expression: ") + fLValue->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("expression type: ")
					+ fLValue->getResult().getType()->toString(),
				fLValue->getLocation()
			);
			return false;
		}
		
		const Type* leftT = fLValue->getResult().getType();
		
		if (!leftT->isNonNullPointer()) { // L-value is not a pointer...
			info.diag().error(
				UTF8STR(
					"pointer assignment tries to assign to a non-pointer value"
				),
				getLocation()
			).addInfo(
				UTF8STR("L-value: ") + fLValue->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("L-value type: ") + leftT->toString(),
				fLValue->getLocation()
			);
			return false;
		}
		
		const Type* leftSinglePointerT =
			leftT->isDoublePointer() ? leftT->getBaseType() : leftT;
		
		if (leftSinglePointerT->isConst()) { // pointer is const...
			info.diag().error(
				UTF8STR("pointer assignment tries to write to a const pointer"),
				getLocation()
			).addInfo(
				UTF8STR("L-value: ") + fLValue->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("L-value type: ") + leftT->toString(),
				fLValue->getLocation()
			);
			return false;
		}
		
		SimpleParameter param(leftSinglePointerT, SimpleParameter::kCopyValue);
		const Type* rightT = fRValue->getResult().getType();
		ValuePassing vp = ValuePassing::check(rightT, param);
		
		if (vp.isPossible()) {
			return vp.checkAccess(*fContext, getLocation(), info);
		}
		else {
			info.diag().error(
				UTF8STR("pointer assignment has incompatible R-value"),
				fRValue->getLocation()
			).addInfo(
				UTF8STR("R-value: ") + fRValue->toString(),
				fRValue->getLocation()
			).addInfo(
				UTF8STR("R-value type: ") + rightT->toString(),
				fRValue->getLocation()
			).addInfo(
				UTF8STR("expected type: ") + leftSinglePointerT->toString(),
				fLValue->getLocation()
			);
			return false;
		}
	}
	
	bool AssignmentStatement::resolveOverloading(CompileInfo& info) {
		if (fRValue->getResult().getType() == 0) {
			debug.line(UTF8STR("Assignment R-Value type not resolved"));
			return false; // no overload resolution possible
		}
		
		OperatorCalls::AssignmentCallResult result=
			OperatorCalls::resolveAssignmentOperator(
				*fLValue, fRValue->getResult().getType(), fOpSym, info
			);
		
		if (result.fgErrorLValue) {
			info.diag().error(
				UTF8STR(
					"left-side expression of assignment statement is not "
						"a valid class L-value"
				),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("expression: ") + fLValue->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("expression type: ")
					+ fLValue->getResult().getType()->toString(),
				fLValue->getLocation()
			);
			return false;
		}
		
		bool resolveOK = true;
		
		if (result.fgErrorConst) {
			info.diag().error(
				UTF8STR("cannot assign to const object"),
				getLocation()
			).addInfo(
				UTF8STR("L-value: ") + fLValue->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("L-value type: ")
					+ fLValue->getResult().getType()->toString(),
				fLValue->getLocation()
			);
			resolveOK = false;
		}
		
		if (result.fgErrorNoOperator) {
			info.diag().error(
				UTF8STR(
					"this type of assignment operator is not supported by the "
						"class of the L-value used here"
				),
				getLocation()
			).addInfo(
				UTF8STR("L-value: ") + fLValue->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("L-value class: ")
					+ fLValue->getResult().getType()->getDereferenced()->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("assignment operator: ") + Operators::toString(fOpSym),
				getLocation()
			);
			resolveOK = false;
		}
		else if (result.fgErrorNoCandidates) {
			info.diag().error(
				UTF8STR("no suitable match found for assignment"),
				getLocation()
			).addInfo(
				UTF8STR("L-value: ") + fLValue->toString(),
				fLValue->getLocation()
			).addInfo(
				UTF8STR("L-value class: ")
					+ fLValue->getResult().getType()->getDereferenced()->toString(),
				fLValue->getLocation()
			).addInfo(
				getLocation(),
				UTF8STR("operator requested: %1(%2)"),
				Operators::toString(fOpSym),
				fRValue->getResult().getType()->toString()
			);
			resolveOK = false;
		}
		else if (result.fgErrorMultipleCandidates) {
			EasyDiagnosticRef e =
				info.diag().error(
					UTF8STR("multiple matches found for assignment"),
					getLocation()
				).addInfo(
					UTF8STR("L-value: ") + fLValue->toString(),
					fLValue->getLocation()
				).addInfo(
					UTF8STR("L-value class: ")
						+ fLValue->getResult().getType()->getDereferenced()->toString(),
					fLValue->getLocation()
				).addInfo(
					getLocation(),
					UTF8STR("operator requested: %1(%2)"),
					Operators::toString(fOpSym),
					fRValue->getResult().getType()->toString()
				);
			
			OperatorCalls::listCandidates(
				fLValue->getResult().getType()->getDereferenced(),
				result.fgOperators, e
			);
			
			resolveOK = false;
		}
		
		if (result.fgOperators.size() == 1) { // 1 candidate
			// OK
			fOp = result.fgOperators[0]; // save operator called
			
			ValuePassing argVP = fOp->getUsedSubstitution();
			resolveOK &= argVP.checkAccess(*fContext, getLocation(), info);
			
			// TODO : check accessibility of operator
		}
		
		return resolveOK;
	}
	
	SimpleProbability AssignmentStatement::getReturnProbability() const {
		return SimpleProbability::never;
	}
}
