/*!
	\file binaryoperatorexpression.cxx
	\brief Implements the BinaryOperatorExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-04
// Last time modified:	2009-08-16

#include "binaryoperatorexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "binaryoperator.h"
#include "builtintypes.h"
#include "compileinfo.h"
#include "conversions.h"
#include "operatorcalls.h"
#include "type.h"

#include "base/debug.h"

namespace HYPER_Compiler {
	
	// *** class BinaryOperatorExpression ***
	
	BinaryOperatorExpression::BinaryOperatorExpression(
		Expression* leftExpr, Expression* rightExpr,
		Operators::BinaryOperator op)
	 : Expression(leftExpr->getCompileInfo(), leftExpr->getLocation()),
		fLeft(leftExpr), fRight(rightExpr), fOpSym(op)
	{
		// nothing here
	}
	
	BinaryOperatorExpression::~BinaryOperatorExpression() {
		// nothing here
	}
	
	HYPER_Common::UTF8String BinaryOperatorExpression::toString() const {
		return UTF8STR("(%1 %2 %3)").compose(
			fLeft->toString(), Operators::toString(fOpSym), fRight->toString()
		);
	}
	
	void BinaryOperatorExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		fLeft->putInSymbolTable(loc);
		fRight->putInSymbolTable(loc);
	}
	
	bool BinaryOperatorExpression::doFullResolve(Trace const& trace) {
		bool resolveOK = true;
		resolveOK &= fLeft->resolveFull(trace);
		resolveOK &= fRight->resolveFull(trace);
		resolveOK &= fLeft->checkReturnsValue();
		resolveOK &= fRight->checkReturnsValue();
		
		if (fLeft->isResolvedAndReturnsResult()
			&& fRight->isResolvedAndReturnsResult())
		{
			if (fOpSym == Operators::kPointerEquals
				|| fOpSym == Operators::kPointerDiffers)
			{
				resolveOK &= resolve3PointerOp();
			}
			else {
				resolveOK &= resolve3NonPointerOp();
			}
		}
		
		return resolveOK;
	}
	
	TriBool BinaryOperatorExpression::isClassLValue() const {
		return false;
	}
	
	TriBool BinaryOperatorExpression::isPointerLValue() const {
		return false;
	}
	
	TriBool BinaryOperatorExpression::isCall() const {
		return false;
	}
	
	TriBool BinaryOperatorExpression::hasResult() const {
		return true;
	}
	
	TriBool BinaryOperatorExpression::resultAtCompileTime() const {
		return fLeft->resultAtCompileTime()
			&& fRight->resultAtCompileTime();
	}
	
	bool BinaryOperatorExpression::resolve3PointerOp() {
		Type const* leftT = fLeft->getResult().getType();
		Type const* rightT = fRight->getResult().getType();
		
		if (!leftT->isPointerOrNull() && !rightT->isPointerOrNull()) {
			getCompileInfo().diag().error(
				this->getLocation(),
				UTF8STR(
					"both operands of pointer comparison %1 do not have "
					"pointer types"
				),
				quote(toString())
			).addInfo(
				UTF8STR("left operand type: ") + leftT->toString(),
				fLeft->getLocation()
			).addInfo(
				UTF8STR("right operand type: ") + rightT->toString(),
				fRight->getLocation()
			);
			return false;
		}
		
		bool leftNull = leftT->isNull();
		bool rightNull = rightT->isNull();
		
		if (leftNull && rightNull) {
			getCompileInfo().diag().error(
				this->getLocation(),
				UTF8STR(
					"both operands of pointer comparison %1 are %2"
				),
				quote(toString()), QUOTED_UTF8STR("null")
			);
			
			return false;
		}
		
		Type const* leftPointerT = leftT;
		if (!leftPointerT->isPointerOrNull()) {
			leftPointerT =
				Type::createSinglePointer(
					false, leftPointerT->getAllConstType()
				);
		}
		else if (leftPointerT->isDoublePointer()) {
			leftPointerT = leftPointerT->getBaseType()->getAllConstType();
		}
		else {
			leftPointerT = leftPointerT->getAllConstType();
		}
		
		Type const* rightPointerT = rightT;
		if (!rightPointerT->isPointerOrNull()) {
			rightPointerT =
				Type::createSinglePointer(
					false, rightPointerT->getAllConstType()
				);
		}
		else if (rightPointerT->isDoublePointer()) {
			rightPointerT = rightPointerT->getBaseType()->getAllConstType();
		}
		else {
			rightPointerT = rightPointerT->getAllConstType();
		}
		
		if (!rightNull) {
			// check conversion of left pointer to type of right pointer
			//~ debug.line(UTF8STR("right pointer not null type"));
			ValuePassing leftToRightVP =
				ValuePassing::check(
					leftT,
					SimpleParameter(rightPointerT, SimpleParameter::kCopyValue)
				);
			fPointerLeftVP.reset(new ValuePassing(leftToRightVP));
		}
		
		if (!leftNull) {
			// check conversion of right pointer to type of left pointer
			//~ debug.line(UTF8STR("left pointer not null type"));
			ValuePassing rightToLeftVP =
				ValuePassing::check(
					rightT,
					SimpleParameter(leftPointerT, SimpleParameter::kCopyValue)
				);
			fPointerRightVP.reset(new ValuePassing(rightToLeftVP));
		}
		
		if (fPointerLeftVP.get() != 0 && fPointerLeftVP->isPossible()) {
			// OK (left to right conversion)
		}
		else if (fPointerRightVP.get() != 0 && fPointerRightVP->isPossible()) {
			// OK (right to left conversion)
		}
		else {
			getCompileInfo().diag().error(
				this->getLocation(),
				UTF8STR(
					"the operands of pointer comparison %1 do not have "
					"compatible types"
				),
				quote(toString())
			).addInfo(
				UTF8STR("left operand type: ") + leftT->toString(),
				fLeft->getLocation()
			).addInfo(
				UTF8STR("right operand type: ") + rightT->toString(),
				fRight->getLocation()
			);
			return false;
		}
		
		setResolved();
		setResultType(BuiltInTypes::createBool(getCompileInfo()));
		return true;
	}
	
	bool BinaryOperatorExpression::resolve3NonPointerOp() {
		//debug << "RESOLVING BinaryOperatorExpression 3 non-pointer" << newl;
		//DirectIndentDebug did;
		
		bool nullErr = false;
		if (fLeft->isNullLiteral()) {
			getCompileInfo().diag().error(
				fLeft->getLocation(),
				UTF8STR(
					"%1 cannot be used as an operand for binary operator %2"
				),
				QUOTED_UTF8STR("null"),
				Operators::toString(fOpSym)
			);
			nullErr = true;
		}
		if (fRight->isNullLiteral()) {
			getCompileInfo().diag().error(
				fRight->getLocation(),
				UTF8STR(
					"%1 cannot be used as an operand for binary operator %2"
				),
				QUOTED_UTF8STR("null"),
				Operators::toString(fOpSym)
			);
			nullErr = true;
		}
		if (nullErr) {
			return false;
		}
		
		Type const* leftT = fLeft->getResult().getType();
		Type const* rightT = fRight->getResult().getType();
		
		std::vector<BinaryOperatorCandidate const*> candidates =
			OperatorCalls::resolveBinaryOperatorExpression(
				leftT, rightT, fOpSym, getCompileInfo()
			);
		
		if (candidates.empty()) { // no candidates
			// error message
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no match for %1, which is needed for expression %2"
				),
				BinaryOperator::getPrintPrototype(fOpSym, leftT, rightT),
				quote(toString())
			);
			return false;
		}
		else if (candidates.size() > 1) { // multiple candidates
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR(
						"multiple matches found for %1, which is needed for "
						"expression %2"
					),
					BinaryOperator::getPrintPrototype(fOpSym, leftT, rightT),
					quote(toString())
				);
			
			for (unsigned int i = 0; i < candidates.size(); ++i) {
				//debug << "print candidate " << i << " : " << (unsigned int)(candidates[i]) << newl;
				BinaryOperator const * binOp = candidates[i]->getOperator();
				e.addInfo(
					binOp->getLocation(),
					UTF8STR("candidate %1 : %2"),
					i + 1,
					binOp->toString()
				);
			}
			
			return false;
		}
		else {
			// operator call is OK
			BinaryOperatorCandidate const* binOp = candidates[0];
			fOp = binOp;
			setResolved();
			setResultType(candidates[0]->getOperator()->getResultType());
			
			ConstantValue const* leftV = fLeft->getResult().getValue();
			ConstantValue const* rightV = fRight->getResult().getValue();
			if (leftV || rightV) {
				ConstantValue const* val;
				
				if (binOp->getUse() == BinaryOperatorCandidate::kOnlyReversedUse)
				{
					// reversed use
					val =
						binOp->getOperator()->evaluateCompileTime(
							rightV, leftV
						);
				}
				else {
					// normal use
					val =
						binOp->getOperator()->evaluateCompileTime(
							leftV, rightV
						);
				}
				
				if (val) {
					setResultValue(val);
				}
			}
			
			return true;
		}
	}
}
