/*!
	\file boolclass.cxx
	\author Kevin André
	\brief Implements the Bool class
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-09-25
// Last time modified:	2008-12-26

#include "boolclass.h"

#include "base/location.h"

#include "assignmentoperator.h"
#include "binaryoperator.h"
#include "builtintypes.h"
#include "constantvalue.h"
#include "parameter.h"
#include "pointertype.h"

namespace HYPER_Compiler {
	
	class Bool::BoolConstructor : public BuiltInConstructor {
	public:
		BoolConstructor(Bool& bc, bool defaultConstructor);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		bool fDefaultConstr;
		std::vector<Parameter const*> fParam;
	};
	
	class Bool::ProcedureNegate : public BuiltInProcedure {
	public:
		ProcedureNegate(Bool& b);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
		Type* fReturnType;
	};
	
	class Bool::NotOperator : public UnaryOperator {
	public:
		NotOperator(Bool& b);
		
		ConstantValue const*
			evaluateCompileTime(ConstantValue const* operand);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::AndOperator : public BinaryOperator {
	public:
		AndOperator(Bool& b);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::OrOperator : public BinaryOperator {
	public:
		OrOperator(Bool& b);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::ConsequenceOperator : public BinaryOperator {
	public:
		ConsequenceOperator(Bool& b);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::EqualsOperator : public BinaryOperator {
	public:
		EqualsOperator(Bool& b);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::DiffersOperator : public BinaryOperator {
	public:
		DiffersOperator(Bool& b);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::BoolAssignOp : public AssignmentOperator {
	public:
		BoolAssignOp(Bool& b);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::AndAssignOp : public AssignmentOperator {
	public:
		AndAssignOp(Bool& b);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Bool::OrAssignOp : public AssignmentOperator {
	public:
		OrAssignOp(Bool& b);
		
		// TODO : change return type
		void generateCode();
	};
	
	// *** Bool ***
	
	Bool::Bool(CompileInfo& info)
	 : BuiltInClass(info, kSealed)
	{
		// no action
	}
	
	Bool::~Bool() {
		// no action
	}
	
	HYPER_Common::UTF8String Bool::getTypeName() const {
		return UTF8STR("bool");
	}
	
	HYPER_Common::UTF8String Bool::getDeclarationName() const {
		return UTF8STR("bool");
	}
	
	void Bool::buildSymbolTable() {
		createConstructors();
		createMembers();
		createOperators();
	}
	
	void Bool::createConstructors() {
		registerConstructor(new BoolConstructor(*this, true), false);
		registerConstructor(new BoolConstructor(*this, false), true);
	}
	
	void Bool::createMembers() {
		// create member procedures
		registerProcedure(new ProcedureNegate(*this));
	}
	
	void Bool::createOperators() {
		registerOperator(new NotOperator(*this));
		
		registerOperator(new OrOperator(*this));
		registerOperator(new AndOperator(*this));
		registerOperator(new ConsequenceOperator(*this));
		registerOperator(new EqualsOperator(*this));
		registerOperator(new DiffersOperator(*this));
		
		registerOperator(new BoolAssignOp(*this));
		registerOperator(new AndAssignOp(*this));
		registerOperator(new OrAssignOp(*this));
	}
	
	
	// *** BoolConstructor ***
	
	Bool::BoolConstructor::BoolConstructor(Bool& bc, bool defaultConstructor)
	 : BuiltInConstructor(&bc, ClassMemberAccess::createPublic(), true),
		fDefaultConstr(defaultConstructor)
	{
		if (!fDefaultConstr) {
			fParam.push_back(
				bc.createParameter(UTF8STR("b"), BuiltInTypes::kBool)
			);
		}
	}
	
	/// Gets the constructor's parameter list
	const std::vector<Parameter const*>*
	Bool::BoolConstructor::getParameterList() const {
		return &fParam;
	}
	
	
	// *** ProcedureNegate ***
	
	Bool::ProcedureNegate::ProcedureNegate(Bool& bc)
	 : BuiltInProcedure(
			&bc, UTF8STR("negate"), kSealed, ClassMemberAccess::createPublic(),
			false, 0
		)
	{
		// no action
	}
	
	const std::vector<Parameter const*>*
	Bool::ProcedureNegate::getParameterList() const {
		return &fParams;
	}
	
	
	// *** NotOperator ***
	
	Bool::NotOperator::NotOperator(Bool& bc)
	 : UnaryOperator(Operators::kNot,
		BuiltInTypes::createBool(bc.getCompileInfo()),
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Bool::NotOperator::evaluateCompileTime(ConstantValue const* operand) {
		return new ConstantValue(!operand->getBool());
	}
	
	// TODO : change return type
	void Bool::NotOperator::generateCode() {
		// TODO
	}
	
	
	// *** AndOperator ***
	Bool::AndOperator::AndOperator(Bool& bc)
	 : BinaryOperator(Operators::kAnd,
		BuiltInTypes::createBool(bc.getCompileInfo()),
		BuiltInTypes::createBool(bc.getCompileInfo()), true,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Bool::AndOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && leftOp->getBool() == false) {
			return leftOp; // return false
		}
		else if (rightOp && rightOp->getBool() == false) {
			return rightOp; // return false
		}
		else if (leftOp && rightOp) {
			return new ConstantValue(true); // return true
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void Bool::AndOperator::generateCode() {
		// TODO
	}
	
	
	// *** OrOperator ***
	Bool::OrOperator::OrOperator(Bool& bc)
	 : BinaryOperator(Operators::kOr,
		BuiltInTypes::createBool(bc.getCompileInfo()),
		BuiltInTypes::createBool(bc.getCompileInfo()), true,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Bool::OrOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && leftOp->getBool() == true) {
			return leftOp; // return true
		}
		else if (rightOp && rightOp->getBool() == true) {
			return rightOp; // return true
		}
		else if (leftOp && rightOp) {
			return new ConstantValue(false); // return false
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void Bool::OrOperator::generateCode() {
		// TODO
	}
	
	
	// *** ConsequenceOperator ***
	Bool::ConsequenceOperator::ConsequenceOperator(Bool& bc)
	 : BinaryOperator(Operators::kConsequence,
		BuiltInTypes::createBool(bc.getCompileInfo()),
		BuiltInTypes::createBool(bc.getCompileInfo()), false,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Bool::ConsequenceOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && leftOp->getBool() == false) {
			// false -> (...) == true
			return new ConstantValue(true);
		}
		else if (leftOp && rightOp) {
			// leftOp is true, so consequence implies rightOp is the result
			return rightOp;
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void Bool::ConsequenceOperator::generateCode() {
		// TODO
	}
	
	
	// *** EqualsOperator ***
	Bool::EqualsOperator::EqualsOperator(Bool& bc)
	 : BinaryOperator(Operators::kEquals,
		BuiltInTypes::createBool(bc.getCompileInfo()),
		BuiltInTypes::createBool(bc.getCompileInfo()), true,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Bool::EqualsOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(leftOp->getBool() == rightOp->getBool());
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void Bool::EqualsOperator::generateCode() {
		// TODO
	}
	
	
	// *** DiffersOperator ***
	Bool::DiffersOperator::DiffersOperator(Bool& bc)
	 : BinaryOperator(Operators::kDiffers,
		BuiltInTypes::createBool(bc.getCompileInfo()),
		BuiltInTypes::createBool(bc.getCompileInfo()), true,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Bool::DiffersOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(leftOp->getBool() != rightOp->getBool());
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void Bool::DiffersOperator::generateCode() {
		// TODO
	}
	
	
	// *** BoolAssignOp ***
	Bool::BoolAssignOp::BoolAssignOp(Bool& bc)
	 : AssignmentOperator(Operators::kSimpleAssign,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Bool::BoolAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** AndAssignOp ***
	Bool::AndAssignOp::AndAssignOp(Bool& bc)
	 : AssignmentOperator(Operators::kAndAssign,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Bool::AndAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** OrAssignOp ***
	Bool::OrAssignOp::OrAssignOp(Bool& bc)
	 : AssignmentOperator(Operators::kOrAssign,
		BuiltInTypes::createBool(bc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Bool::OrAssignOp::generateCode() {
		// TODO
	}
}
