/*!
	\file boolconditionexpression.cxx
	\brief Implements the BoolConditionExpression class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-04-19
// Last time modified:	2008-10-15

#include "boolconditionexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "builtintypes.h"
#include "compileinfo.h"
#include "conversions.h"
#include "type.h"

namespace HYPER_Compiler {
	
	BoolConditionExpression::BoolConditionExpression(Expression* expr,
		ClassContext const* context)
	 : Expression(expr->getCompileInfo(), expr->getLocation()),
		fExpr(expr), fCtxt(context)
	{
		// no action
	}
	
	BoolConditionExpression::~BoolConditionExpression() {
		// no action
	}
	
	HYPER_Common::UTF8String BoolConditionExpression::toString() const {
		return fExpr->toString();
	}
	
	void BoolConditionExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		fExpr->putInSymbolTable(loc);
	}
	
	bool BoolConditionExpression::doFullResolve(Trace const& trace) {
		//debug.line("BoolConditionExpression::doFullResolve called");
		bool resolveOK = fExpr->resolveFull(trace);
		
		if (fExpr->checkHasResult() && fExpr->getResult().isResolved())
		{
			Type const* t = fExpr->getResult().getType();
			hyp_assert(t != 0);
			
			// check if expr type is convertable to 'bool'
			ValuePassing vp =
				BuiltInTypes::getBoolConversion(getCompileInfo(), t);
			
			if (vp.isPossible()) {
				resolveOK &=
					vp.checkAccess(
						*fCtxt, fExpr->getLocation(), getCompileInfo()
					);
			}
			else { // no conversion...
				getCompileInfo().diag().error(
					fExpr->getLocation(),
					UTF8STR(
						"condition expression %1 has type %2, which is "
						"not automatically convertable to type %3"
					),
					quote(fExpr->toString()),
					quote(t->toString()),
					QUOTED_UTF8STR("bool")
				);
				resolveOK = false;
			}
		}
		
		setResolved();
		setResultType(BuiltInTypes::createBool(getCompileInfo()));
		return resolveOK;
	}
	
	TriBool BoolConditionExpression::isClassLValue() const {
		return false;
	}
	
	TriBool BoolConditionExpression::isPointerLValue() const {
		return false;
	}
	
	TriBool BoolConditionExpression::isCall() const {
		return fExpr->isCall();
	}
	
	TriBool BoolConditionExpression::hasResult() const {
		return true;
	}
	
	TriBool BoolConditionExpression::resultAtCompileTime() const {
		return fExpr->resultAtCompileTime();
	}
}
