/*!
	\file builtinclass.cxx
	\author Kevin André
	\brief Implements the BuiltInClass class
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2005-09-23
// Date modified:	2008-10-15

#include "builtinclass.h"

#include "base/location.h"

#include "arrayoperator.h"
#include "assignmentoperator.h"
#include "binaryoperator.h"
#include "crementoperator.h"
#include "declareditem.h"
#include "overloadedarrayoperator.h"
#include "overloadedassignmentoperator.h"
#include "overloadedbinaryoperator.h"
#include "overloadedconstructor.h"
#include "overloadedprocedure.h"
#include "parameter.h"
#include "unaryoperator.h"

namespace HYPER_Compiler {
	
	BuiltInClass::BuiltInClass(CompileInfo& info, AbstractSealedStatic ass,
		Class* parent)
	 :	Class(info), fParent(parent),
		fBuiltInLoc(Location::getBuiltIn()), fASS(ass),
		fPublicAccess(ClassMemberAccess::createPublic()), fArrayOps(0)
	{
		// no action
	}
	
	BuiltInClass::~BuiltInClass() {
		//
	}
	
	AbstractSealedStatic BuiltInClass::getASS() const {
		return fASS;
	}
	
	ClassMemberAccess BuiltInClass::getAccess() const {
		return fPublicAccess;
	}
	
	const Location* BuiltInClass::getLocation() const {
		return fBuiltInLoc;
	}
	
	void BuiltInClass::registerOperator(ArrayOperator* op) {
		if (fArrayOps == 0) {
			fArrayOps = new OverloadedArrayOperator(op);
		}
		else {
			fArrayOps->addOverload(op);
		}
	}
	
	void BuiltInClass::registerOperator(AssignmentOperator* op) {
		OverloadedAssignmentOperator* & oao = fAssignOps[op->getOperator()];
		if (oao == 0) {
			oao = new OverloadedAssignmentOperator(op);
		}
		else {
			oao->addOverload(op);
		}
	}
	
	void BuiltInClass::registerOperator(BinaryOperator* op) {
		OverloadedBinaryOperator* & obp = fBinOps[op->getOperator()];
		if (obp == 0) {
			obp = new OverloadedBinaryOperator(op);
		}
		else {
			obp->addOverload(op);
		}
	}
	
	void BuiltInClass::registerOperator(CrementOperator* op) {
		fCremOps[op->getOperator()] = op;
	}
	
	void BuiltInClass::registerOperator(UnaryOperator* op) {
		fUnOps[op->getOperator()] = op;
	}
	
	Class::BinOpsMap const& BuiltInClass::getBinaryOperators() const {
		return fBinOps;
	}
	
	Class::CremOpsMap const& BuiltInClass::getCrementOperators() const {
		return fCremOps;
	}
	
	Class::UnOpsMap const& BuiltInClass::getUnaryOperators() const {
		return fUnOps;
	}
	
	Class::AssignOpsMap const& BuiltInClass::getAssignmentOperators() const {
		return fAssignOps;
	}
	
	OverloadedArrayOperator const* BuiltInClass::getArrayOperators() const {
		return fArrayOps;
	}
	
	Class const* BuiltInClass::getParentClass() const {
		return fParent;
	}
	
	Parameter*
	BuiltInClass::createParameter(HYPER_Common::UTF8String name,
		BuiltInTypes::Primitive type, SimpleParameter::Passing p) const
	{
		return new Parameter(
			BuiltInTypes::create(getCompileInfo(), type, false), p, name,
			Location::getBuiltIn()
		);
	}
	
	Parameter*
	BuiltInClass::createParameter(HYPER_Common::UTF8String name,
		Type const* type, SimpleParameter::Passing p) const
	{
		return new Parameter(type, p, name, Location::getBuiltIn());
	}
}
