/*!
	\file builtinintegralclass.cxx
	\author Kevin André
	\brief Implements the BuiltInIntegralClass class
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2007-06-23
// Last modified:	2009-08-16

#include "builtinintegralclass.h"

#include "base/asserts.h"
#include "base/compileoptions.h"
#include "base/location.h"

#include "assignmentoperator.h"
#include "binaryoperator.h"
#include "builtinconstructor.h"
#include "builtinprocedure.h"
#include "builtintypes.h" // for BuiltInTypes::createBool
#include "compileinfo.h"
#include "crementoperator.h"
#include "parameter.h"
#include "type.h"
#include "unaryoperator.h"

namespace HYPER_Compiler {
	
	ConstantValue::nat_t
	BuiltInIntegralClass::getMaxUnsignedValue(IntegerSize s)
	{
		switch (s) {
		case k8Bit:
			return 255u;
		case k16Bit:
			return 65535u;
		case k32Bit:
			return 4294967295uL;
		case k64Bit:
			return 18446744073709551615uLL;
		default:
			hyp_assert_unreachable();
		}
	}
	
	ConstantValue::int_t
	BuiltInIntegralClass::getMaxSignedValue(IntegerSize s)
	{
		switch (s) {
		case k8Bit:
			return 127;
		case k16Bit:
			return 32767;
		case k32Bit:
			return 2147483647L;
		case k64Bit:
			return 9223372036854775807LL;
		default:
			hyp_assert_unreachable();
		}
	}
	
	ConstantValue::int_t
	BuiltInIntegralClass::getMinSignedValue(IntegerSize s)
	{
		switch (s) {
		case k8Bit:
			return -128;
		case k16Bit:
			return -32768;
		case k32Bit:
			return -2147483647L - 1;
		case k64Bit:
			return -9223372036854775807LL - 1;
		default:
			hyp_assert_unreachable();
		}
	}
	
	BuiltInIntegralClass::BuiltInIntegralClass(CompileInfo& info,
		HYPER_Common::UTF8String simpleName, IntegerSize s, bool isSigned)
	 : BuiltInClass(info, kSealed), fType(0), fName(simpleName), fSize(s),
		fSigned(isSigned)
	{
		// no action
	}
	
	BuiltInIntegralClass::~BuiltInIntegralClass() {
		// no action
	}
	
	HYPER_Common::UTF8String BuiltInIntegralClass::getTypeName() const {
		return fName;
	}
	
	HYPER_Common::UTF8String BuiltInIntegralClass::getDeclarationName() const {
		return fName;
	}
	
	IntegerSize BuiltInIntegralClass::getSize() const {
		return fSize;
	}
	
	bool BuiltInIntegralClass::isSigned() const {
		return fSigned;
	}
	
	void BuiltInIntegralClass::buildSymbolTable() {
		fType = getClassType();
		createConstructors();
		createMembers();
		createOperators();
	}
	
	Type const* BuiltInIntegralClass::getType() const {
		return fType;
	}
	
	ConstantValue::nat_t
	BuiltInIntegralClass::getMaxUnsignedValue() const {
		return getMaxUnsignedValue(fSize);
	}
	
	ConstantValue::int_t
	BuiltInIntegralClass::getMaxSignedValue() const {
		return getMaxSignedValue(fSize);
	}
	
	ConstantValue::int_t
	BuiltInIntegralClass::getMinSignedValue() const {
		return getMinSignedValue(fSize);
	}
	
	bool BuiltInIntegralClass::canBeInvisiblyPassedByValue() const {
		//~ Platform::WordSize s =
			//~ getCompileInfo().getOptions().fgTargetPlatform->getWordSize();
		//~ return getSize() <= toIntegralSize(s);
		
		// for now we'll assume that all integer sizes can be passed by value:
		return true;
	}
	
	
	// *** Begin class declarations ***
	
	class BuiltInIntegralClass::BIICConstructor : public BuiltInConstructor {
	public:
		BIICConstructor(BuiltInIntegralClass* c, bool defaultConstructor)
		 : BuiltInConstructor(c, true), fClass(c)
		{
			if (!defaultConstructor) {
				fParam.push_back(
					c->createParameter(UTF8STR("copySource"), c->getType())
				);
			}
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParam;
		}
		
	private:
		BuiltInIntegralClass const* fClass;
		std::vector<Parameter const*> fParam;
	};
	
	class BuiltInIntegralClass::BIICConvConstructor : public BuiltInConstructor {
	public:
		BIICConvConstructor(BuiltInIntegralClass* c, Type const* paramT,
			bool paramIsSigned)
		 : BuiltInConstructor(c, true), fClass(c),
			fConvertFromSigned(paramIsSigned)
		{
			fParam.push_back(
				c->createParameter(UTF8STR("copySource"), paramT)
			);
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParam;
		}
		
	private:
		std::vector<Parameter const*> fParam;
		BuiltInIntegralClass const* fClass;
		bool fConvertFromSigned;
	};
	
	class BuiltInIntegralClass::BIICProcedureTurnSign : public BuiltInProcedure {
	public:
		BIICProcedureTurnSign(BuiltInIntegralClass* c)
		 : BuiltInProcedure(
				c, UTF8STR("turnSign"), kSealed,
				ClassMemberAccess::createPublic(), false, 0
			)
		{
			// no action
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParams;
		}
		
	private:
		std::vector<Parameter const*> fParams;
	};
	
	class BuiltInIntegralClass::BIICProcedureAbs : public BuiltInProcedure {
	public:
		BIICProcedureAbs(BuiltInIntegralClass* c)
		 : BuiltInProcedure(
				c, UTF8STR("abs"), kSealed, ClassMemberAccess::createPublic(),
				true, c->getAbsResultType()
			), fClass(c)
		{
			// no action
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParams;
		}
		
	private:
		std::vector<Parameter const*> fParams;
		BuiltInIntegralClass const* fClass;
	};
	
	class BuiltInIntegralClass::BIICProcedureTruncateToSigned : public BuiltInProcedure {
	public:
		BIICProcedureTruncateToSigned(BuiltInIntegralClass* c)
		 : BuiltInProcedure(
				c, UTF8STR("truncateToSigned"), kSealed,
				ClassMemberAccess::createPublic(), true,
				c->getTruncateToSignedResultType()
			), fClass(c)
		{
			// no action
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParams;
		}
		
	private:
		std::vector<Parameter const*> fParams;
		BuiltInIntegralClass const* fClass;
	};
	
	class BuiltInIntegralClass::BIICProcedureTruncateToByte : public BuiltInProcedure {
	public:
		BIICProcedureTruncateToByte(BuiltInIntegralClass* c)
		 : BuiltInProcedure(
				c, UTF8STR("truncateToByte"), kSealed,
				ClassMemberAccess::createPublic(), true,
				BuiltInTypes::createByte(c->getCompileInfo())
			)
		{
			// no action
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParams;
		}
		
	private:
		std::vector<Parameter const*> fParams;
	};
	
	class BuiltInIntegralClass::BIICProcedureMin : public BuiltInProcedure {
	public:
		BIICProcedureMin(BuiltInIntegralClass* c)
		 : BuiltInProcedure(
				c, UTF8STR("min"), kStatic,
				ClassMemberAccess::createPublic(), false,
				c->getType()
			), fClass(c)
		{
			// no action
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParams;
		}
		
	private:
		std::vector<Parameter const*> fParams;
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICProcedureMax : public BuiltInProcedure {
	public:
		BIICProcedureMax(BuiltInIntegralClass* c)
		 : BuiltInProcedure(
				c, UTF8STR("max"), kStatic,
				ClassMemberAccess::createPublic(), false,
				c->getType()
			), fClass(c)
		{
			// no action
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParams;
		}
		
	private:
		std::vector<Parameter const*> fParams;
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICPositiveOperator : public UnaryOperator {
	public:
		BIICPositiveOperator(BuiltInIntegralClass* c)
		 : UnaryOperator(Operators::kPositive, c->getType(),
			c->getType(), Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* operand) {
			return operand;
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICNegativeOperator : public UnaryOperator {
	public:
		BIICNegativeOperator(BuiltInIntegralClass* c)
		 : UnaryOperator(Operators::kNegative, c->getType(),
			c->getNegateResultType(), Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* operand) {
			return new ConstantValue(-operand->getUnsigned());
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICSuccessorOperator : public UnaryOperator {
	public:
		BIICSuccessorOperator(BuiltInIntegralClass* c)
		 : UnaryOperator(Operators::kSuccessor, c->getType(),
			c->getType(), Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* operand)
		{
			if (fClass->isSigned()) {
				ConstantValue::int_t op = operand->getSigned();
				if (op < fClass->getMaxSignedValue()) {
					return new ConstantValue(op + 1);
				}
				else {
					// TODO
					return 0;
				}
			}
			else {
				ConstantValue::nat_t op = operand->getUnsigned();
				if (op < fClass->getMaxUnsignedValue()) {
					return new ConstantValue(op + 1);
				}
				else {
					// TODO
					return 0;
				}
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICPredecessorOperator : public UnaryOperator {
	public:
		BIICPredecessorOperator(BuiltInIntegralClass* c)
		 : UnaryOperator(Operators::kPredecessor, c->getType(),
			c->getType(), Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* operand) {
			if (fClass->isSigned()) {
				ConstantValue::int_t op = operand->getSigned();
				if (op > fClass->getMinSignedValue()) {
					return new ConstantValue(op - 1);
				}
				else {
					// TODO
					return 0;
				}
			}
			else {
				ConstantValue::nat_t op = operand->getUnsigned();
				if (op > 0) {
					return new ConstantValue(op - 1);
				}
				else {
					// TODO
					return 0;
				}
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICEqualsOperator : public BinaryOperator {
	public:
		BIICEqualsOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kEquals, c->getType(), c->getType(),
			true, BuiltInTypes::createBool(c->getCompileInfo()),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					return new ConstantValue(
						leftOp->getSigned() == rightOp->getSigned()
					);
				}
				else {
					return new ConstantValue(
						leftOp->getUnsigned() == rightOp->getUnsigned()
					);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICDiffersOperator : public BinaryOperator {
	public:
		BIICDiffersOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kDiffers, c->getType(), c->getType(),
			true, BuiltInTypes::createBool(c->getCompileInfo()),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					return new ConstantValue(
						leftOp->getSigned() != rightOp->getSigned()
					);
				}
				else {
					return new ConstantValue(
						leftOp->getUnsigned() != rightOp->getUnsigned()
					);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICLessOperator : public BinaryOperator {
	public:
		BIICLessOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kLess, c->getType(), c->getType(),
			false, BuiltInTypes::createBool(c->getCompileInfo()),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					return new ConstantValue(
						leftOp->getSigned() < rightOp->getSigned()
					);
				}
				else {
					return new ConstantValue(
						leftOp->getUnsigned() < rightOp->getUnsigned()
					);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICLessEqOperator : public BinaryOperator {
	public:
		BIICLessEqOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kLessOrEqual, c->getType(), c->getType(),
			false, BuiltInTypes::createBool(c->getCompileInfo()),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					return new ConstantValue(
						leftOp->getSigned() <= rightOp->getSigned()
					);
				}
				else {
					return new ConstantValue(
						leftOp->getUnsigned() <= rightOp->getUnsigned()
					);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICGreaterOperator : public BinaryOperator {
	public:
		BIICGreaterOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kGreater, c->getType(), c->getType(),
			false, BuiltInTypes::createBool(c->getCompileInfo()),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					return new ConstantValue(
						leftOp->getSigned() > rightOp->getSigned()
					);
				}
				else {
					return new ConstantValue(
						leftOp->getUnsigned() > rightOp->getUnsigned()
					);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICGreaterEqOperator : public BinaryOperator {
	public:
		BIICGreaterEqOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kGreaterOrEqual, c->getType(), c->getType(),
			false, BuiltInTypes::createBool(c->getCompileInfo()),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					return new ConstantValue(
						leftOp->getSigned() >= rightOp->getSigned()
					);
				}
				else {
					return new ConstantValue(
						leftOp->getUnsigned() >= rightOp->getUnsigned()
					);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICAddOperator : public BinaryOperator {
	public:
		BIICAddOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kAdd,
			c->getType(), c->getType(), true, c->getType(),
			Location::getBuiltIn()),
			fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					ConstantValue::int_t val =
						leftOp->getSigned() + rightOp->getSigned();
					
					if (val <= fClass->getMaxSignedValue()
						&& val >= fClass->getMinSignedValue())
					{
						return new ConstantValue(val);
					}
					else {
						// TODO : overflow
						return 0;
					}
				}
				else {
					ConstantValue::nat_t val =
						leftOp->getUnsigned() + rightOp->getUnsigned();
					if (val <= fClass->getMaxUnsignedValue()) {
						return new ConstantValue(val);
					}
					else {
						// TODO : overflow
						return 0;
					}
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICSubtractOperator : public BinaryOperator {
	public:
		BIICSubtractOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kSubtract,
			c->getType(), c->getType(), true, c->getType(),
			Location::getBuiltIn()),
			fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					ConstantValue::int_t val =
						leftOp->getSigned() - rightOp->getSigned();
					
					if (val <= fClass->getMaxSignedValue()
						&& val >= fClass->getMinSignedValue())
					{
						return new ConstantValue(val);
					}
					else {
						// TODO : overflow
						return 0;
					}
				}
				else {
					ConstantValue::nat_t val =
						leftOp->getUnsigned() - rightOp->getUnsigned();
					
					if (val <= fClass->getMaxUnsignedValue()) {
						return new ConstantValue(val);
					}
					else {
						// TODO : overflow
						return 0;
					}
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICMultiplyOperator : public BinaryOperator {
	public:
		BIICMultiplyOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kMultiply,
			c->getType(), c->getType(), true, c->getType(),
			Location::getBuiltIn()),
			fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					ConstantValue::int_t val =
						leftOp->getSigned() * rightOp->getSigned();
					
					if (val <= fClass->getMaxSignedValue()
						&& val >= fClass->getMinSignedValue())
					{
						return new ConstantValue(val);
					}
					else {
						// TODO : overflow
						return 0;
					}
				}
				else {
					ConstantValue::nat_t val =
						leftOp->getUnsigned() * rightOp->getUnsigned();
					
					if (val <= fClass->getMaxUnsignedValue()) {
						return new ConstantValue(val);
					}
					else {
						// TODO : overflow
						return 0;
					}
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICDivideOperator : public BinaryOperator {
	public:
		BIICDivideOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kDivide,
			c->getType(), c->getType(), true, c->getType(),
			Location::getBuiltIn()),
			fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					ConstantValue::int_t val =
						leftOp->getSigned() / rightOp->getSigned();
					
					return new ConstantValue(val);
					
				}
				else {
					ConstantValue::nat_t val =
						leftOp->getUnsigned() / rightOp->getUnsigned();
					
					return new ConstantValue(val);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICModuloOperator : public BinaryOperator {
	public:
		BIICModuloOperator(BuiltInIntegralClass* c)
		 : BinaryOperator(Operators::kModulo,
			c->getType(), c->getType(), true, c->getType(),
			Location::getBuiltIn()),
			fClass(c)
		{
			// no action
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			if (leftOp && rightOp) {
				if (fClass->isSigned()) {
					ConstantValue::int_t val =
						leftOp->getSigned() % rightOp->getSigned();
					
					return new ConstantValue(val);
					
				}
				else {
					ConstantValue::nat_t val =
						leftOp->getUnsigned() % rightOp->getUnsigned();
					
					return new ConstantValue(val);
				}
			}
			else {
				return 0; // don't know result
			}
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICIncrementOperator : public CrementOperator {
	public:
		BIICIncrementOperator(BuiltInIntegralClass* c)
		 : CrementOperator(Operators::kIncrement, Location::getBuiltIn()),
			fClass(c)
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICDecrementOperator : public CrementOperator {
	public:
		BIICDecrementOperator(BuiltInIntegralClass* c)
		 : CrementOperator(Operators::kDecrement, Location::getBuiltIn()),
			fClass(c)
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICAssignOp : public AssignmentOperator {
	public:
		BIICAssignOp(BuiltInIntegralClass* c)
		 : AssignmentOperator(Operators::kSimpleAssign, c->getType(),
			Location::getBuiltIn())
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
	};
	
	class BuiltInIntegralClass::BIICAddAssignOp : public AssignmentOperator {
	public:
		BIICAddAssignOp(BuiltInIntegralClass* c)
		 : AssignmentOperator(Operators::kAddAssign, c->getType(),
			Location::getBuiltIn())
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
	};
	
	class BuiltInIntegralClass::BIICSubtractAssignOp : public AssignmentOperator {
	public:
		BIICSubtractAssignOp(BuiltInIntegralClass* c)
		 : AssignmentOperator(Operators::kSubtractAssign, c->getType(),
			Location::getBuiltIn())
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
	};
	
	class BuiltInIntegralClass::BIICMultiplyAssignOp : public AssignmentOperator {
	public:
		BIICMultiplyAssignOp(BuiltInIntegralClass* c)
		 : AssignmentOperator(Operators::kMultiplyAssign, c->getType(),
			Location::getBuiltIn())
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
	};
	
	class BuiltInIntegralClass::BIICDivideAssignOp : public AssignmentOperator {
	public:
		BIICDivideAssignOp(BuiltInIntegralClass* c)
		 : AssignmentOperator(Operators::kDivideAssign, c->getType(),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	class BuiltInIntegralClass::BIICModuloAssignOp : public AssignmentOperator {
	public:
		BIICModuloAssignOp(BuiltInIntegralClass* c)
		 : AssignmentOperator(Operators::kModuloAssign, c->getType(),
			Location::getBuiltIn()), fClass(c)
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {}
		
	private:
		BuiltInIntegralClass* fClass;
	};
	
	// *** End class declarations ***
	
	
	void BuiltInIntegralClass::createConstructors() {
		// default & copy constructor
		registerConstructor(new BIICConstructor(this, true), false);
		registerConstructor(new BIICConstructor(this, false), true);
	}
	
	void
	BuiltInIntegralClass::addConversionConstructor(Type const* paramT,
		bool paramSignedT)
	{
		registerConstructor(
			new BIICConvConstructor(this, paramT, paramSignedT), false
		);
	}
	
	void BuiltInIntegralClass::createMembers() {
		if (fSigned) { // signed type
			registerProcedure(new BIICProcedureTurnSign(this));
			registerProcedure(new BIICProcedureAbs(this));
		}
		else { // unsigned type
			if (getTruncateToSignedResultType() != 0) { // if supported
				registerProcedure(new BIICProcedureTruncateToSigned(this));
			}
			
			if (getSize() > k8Bit) {
				registerProcedure(new BIICProcedureTruncateToByte(this));
			}
		}
		
		registerProcedure(new BIICProcedureMin(this));
		registerProcedure(new BIICProcedureMax(this));
	}
	
	void BuiltInIntegralClass::createOperators() {
		registerOperator(new BIICPositiveOperator(this));
		registerOperator(new BIICSuccessorOperator(this));
		registerOperator(new BIICPredecessorOperator(this));
		if (getNegateResultType() != 0) { // if supported
			registerOperator(new BIICNegativeOperator(this));
		}
		
		registerOperator(new BIICEqualsOperator(this));
		registerOperator(new BIICDiffersOperator(this));
		registerOperator(new BIICLessOperator(this));
		registerOperator(new BIICLessEqOperator(this));
		registerOperator(new BIICGreaterOperator(this));
		registerOperator(new BIICGreaterEqOperator(this));
		
		registerOperator(new BIICAddOperator(this));
		registerOperator(new BIICSubtractOperator(this));
		registerOperator(new BIICMultiplyOperator(this));
		registerOperator(new BIICDivideOperator(this));
		registerOperator(new BIICModuloOperator(this));
		
		registerOperator(new BIICIncrementOperator(this));
		registerOperator(new BIICDecrementOperator(this));
		
		registerOperator(new BIICAssignOp(this));
		registerOperator(new BIICAddAssignOp(this));
		registerOperator(new BIICSubtractAssignOp(this));
		registerOperator(new BIICMultiplyAssignOp(this));
		registerOperator(new BIICDivideAssignOp(this));
		registerOperator(new BIICModuloAssignOp(this));
	}
}
