/*!
	\file builtintypes.cxx
	\brief Implements the BuiltInTypes class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-12-16
// Last time modified:	2009-01-05

#include "builtintypes.h"

#include "base/asserts.h"
#include "base/location.h"

//~ #include "class.h"  // added for debug
#include "classrepository.h"
#include "compileinfo.h"
#include "conversions.h"
#include "doublepointertype.h"
#include "simpleparameter.h"
#include "type.h"

//~ #include "base/debug.h"

namespace HYPER_Compiler {
	
	BuiltInTypes::Primitive
	BuiltInTypes::fromKeyword(Keywords::Keyword k)
	{
		switch (k) {
		case Keywords::kBool:
			return BuiltInTypes::kBool;
		case Keywords::kByte:
			return BuiltInTypes::kByte;
		case Keywords::kChar:
			return BuiltInTypes::kChar;
		case Keywords::kDouble:
			return BuiltInTypes::kDouble;
		case Keywords::kInt:
			return BuiltInTypes::kInt;
		case Keywords::kInt16:
			return BuiltInTypes::kInt16;
		case Keywords::kInt32:
			return BuiltInTypes::kInt32;
		case Keywords::kInt64:
			return BuiltInTypes::kInt64;
		case Keywords::kNat:
			return BuiltInTypes::kNat;
		case Keywords::kNat16:
			return BuiltInTypes::kNat16;
		case Keywords::kNat32:
			return BuiltInTypes::kNat32;
		case Keywords::kNat64:
			return BuiltInTypes::kNat64;
		case Keywords::kObject:
			return BuiltInTypes::kObject;
		case Keywords::kReal:
			return BuiltInTypes::kReal;
		case Keywords::kSingle:
			return BuiltInTypes::kSingle;
		case Keywords::kString:
			return BuiltInTypes::kString;
		default:
			return BuiltInTypes::kNone;
		}
	}
	
	Type const*
	BuiltInTypes::create(CompileInfo& info, Primitive p, bool isConst,
		const Location* loc)
	{
		if (loc == 0) {
			loc = Location::getBuiltIn();
		}
		
		// get class for the given primitive type
		Class* c = info.getClassRepository().getClassOfBuiltin(p);
		hyp_assert(c != 0);
		
		//~ debug.line(
			//~ UTF8STR("Creating type for built-in class ") + c->getTypeName()
		//~ );
		
		return Type::createClassType(isConst, c);
	}
	
	Type const*
	BuiltInTypes::createBool(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kBool, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createByte(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kByte, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createChar(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kChar, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createDouble(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kDouble, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createInt(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kInt, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createInt16(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kInt16, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createInt32(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kInt32, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createInt64(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kInt64, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createNat(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kNat, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createNat16(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kNat16, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createNat32(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kNat32, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createNat64(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kNat64, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createObject(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kObject, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createReal(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kReal, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createSingle(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kSingle, isConst, loc);
	}
	
	Type const*
	BuiltInTypes::createString(CompileInfo& info, bool isConst,
		const Location* loc)
	{
		return create(info, kString, isConst, loc);
	}
	
	ValuePassing
	BuiltInTypes::getBoolConversion(CompileInfo& info, Type const* t)
	{
		static Type const* boolT = BuiltInTypes::createBool(info);
		SimpleParameter param(boolT);
		return ValuePassing::check(t, param);
	}
	
	ValuePassing
	BuiltInTypes::getNatConversion(CompileInfo& info, Type const* t)
	{
		static Type const* natT = BuiltInTypes::createNat(info);
		SimpleParameter param(natT);
		return ValuePassing::check(t, param);
	}
}
