/*!
	\file callstatement.cxx
	\brief Implements the CallStatement class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-12
// Last time modified:	2008-06-26

#include "callstatement.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

#include "compileinfo.h"
#include "expression.h"

namespace HYPER_Compiler {
	
	CallStatement::CallStatement(Expression* callExpression,
		const Location* pos, CompileInfo& info)
	 : Statement(info, pos), fCall(callExpression)
	{
		hyp_assert(callExpression != 0);
	}
	
	CallStatement::~CallStatement() {
		// nothing here
	}
	
	HYPER_Common::UTF8String CallStatement::getSyntaxName() const {
		return UTF8STR("Call");
	}
	
	void CallStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.line(getSyntaxName() + UTF8STR(" ") + fCall->toString());
	}
	
	Expression* CallStatement::getCall() const {
		return fCall;
	}
	
	HYPER_Common::UTF8String CallStatement::getResolveDescription() const {
		return UTF8STR("procedure call statement");
	}
	
	void CallStatement::putInSymbolTable(ScopeLocation const& loc) {
		Statement::putInSymbolTable(loc);
		fCall->putInSymbolTable(new LookupLocation(loc));
	}
	
	bool CallStatement::doFullResolve(Trace const& trace) {
		bool resolveOK = fCall->resolveFull(trace);
		
		if (fCall->isCall().isFalse()) {
			getCompileInfo().diag().error(
				fCall->getLocation(),
				UTF8STR(
					"expression `%1' cannot be used as a statement; "
					"it is not a procedure call"
				),
				fCall->toString()
			);
			return false;
		}
		return resolveOK;
	}
	
	SimpleProbability CallStatement::getReturnProbability() const {
		return SimpleProbability::never;
	}
}
