/*!
	\file chainedcompareexpression.cxx
	\brief Implements the ChainedCompareExpression class
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-01-15
// Last time modified:	2008-10-15

#include "chainedcompareexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "binaryoperator.h"
#include "builtintypes.h"
#include "compileinfo.h"
#include "conversions.h"
#include "operatorcalls.h"
#include "type.h"

namespace HYPER_Compiler {
	
	ChainedCompareExpression::ChainedCompareExpression(
		std::vector<Expression*> exprs,
		std::vector<Operators::BinaryOperator> ops,
		ClassContext const* context)
	: Expression(exprs[0]->getCompileInfo(), exprs[0]->getLocation()),
		fExprs(exprs), fOpSymbols(ops),
		fContext(context)
	{
		// check arguments
		hyp_assert(exprs.size() - 1 == ops.size());
		for (unsigned int i = 0; i < ops.size(); ++i) {
			hyp_assert(Operators::isRegularCompare(ops[i]));
		}
	}
	
	ChainedCompareExpression::~ChainedCompareExpression() {
		// no action
	}
	
	HYPER_Common::UTF8String ChainedCompareExpression::toString() const {
		HYPER_Common::UTF8StringBuilder sb;
		sb += UTF8STR("(");
		for (unsigned int i = 0; i < fExprs.size(); ++i) {
			if (i > 0) {
				sb += UTF8STR(" ");
				sb += Operators::toString(fOpSymbols[i - 1]);
				sb += UTF8STR(" ");
			}
			sb += fExprs[i]->toString();
		}
		sb += UTF8STR(")");
		return sb.toString();
	}
	
	void ChainedCompareExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		for (unsigned int i = 0; i < fExprs.size(); ++i) {
			fExprs[i]->putInSymbolTable(loc);
		}
	}
	
	bool ChainedCompareExpression::doFullResolve(Trace const& trace) {
		bool resolveOK = true;
		for (unsigned int i = 0; i < fExprs.size(); ++i) {
			resolveOK &= fExprs[i]->resolveFull(trace);
			resolveOK &= fExprs[i]->checkReturnsValue();
			
			if (fExprs[i]->isNullLiteral()) {
				getCompileInfo().diag().error(
					UTF8STR(
						"`null' is not allowed as an operand in this expression"
					),
					fExprs[i]->getLocation()
				);
				resolveOK = false;
			}
			
			if (i > 0 && fExprs[i - 1]->isResolvedAndNotNull()
				&& fExprs[i]->isResolvedAndNotNull())
			{
				BinaryOperatorCandidate const* cand =
					findAndCheckOpCandidate(i - 1);
				fOps.push_back(cand);
			}
		}
		
		// This expression is always successfully resolved, because its result
		// is guaranteed to have type 'bool'.
		setResolved();
		setResultType(BuiltInTypes::createBool(getCompileInfo(), true, 0));
		return resolveOK;
	}
	
	BinaryOperatorCandidate const*
	ChainedCompareExpression::findAndCheckOpCandidate(
		unsigned int opIndex) const
	{
		Operators::BinaryOperator op = fOpSymbols[opIndex];
		
		Expression const* prev = fExprs[opIndex];
		Expression const* curr = fExprs[opIndex + 1];
		ExpressionResult const& prevR = prev->getResult();
		ExpressionResult const& currR = curr->getResult();
		
		Type const* prevT =
			prevR.getType()->getDereferenced();
		Type const* currT =
			currR.getType()->getDereferenced();
		
		std::vector<BinaryOperatorCandidate const*> candidates =
			OperatorCalls::resolveBinaryOperatorExpression(
				prevT, currT, op, getCompileInfo()
			);
		
		if (candidates.empty()) { // no candidates
			// error message
			getCompileInfo().diag().error(
				prev->getLocation(),
				UTF8STR(
					"no match for operator%1(%2, %3), which is needed "
					"for expression `%4 %1 %5'"
				),
				Operators::toString(op), prevT->toString(), currT->toString(),
				prev->toString(), curr->toString()
			);
			
			return 0;
		}
		else if (candidates.size() > 1) { // multiple candidates
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR(
					"multiple matches for operator%1(%2, %3), which is needed "
					"for expression `%4 %1 %5'"
					),
					Operators::toString(op), prevT->toString(),
					currT->toString(), prev->toString(), curr->toString()
				);
			
			//debug << "candidates: " << candidates.size() << newl;
			for (unsigned int i = 0; i < candidates.size(); ++i) {
				//debug << "print candidate " << i << " : " << (unsigned int)(candidates[i]) << newl;
				BinaryOperator const * binOp = candidates[i]->getOperator();
				e.addInfo(
					binOp->getLocation(),
					UTF8STR("candidate %1 : %2"),
					i + 1, binOp->toString()
				);
			}
			
			return 0;
		}
		
		// operator call OK; now check its result type
		Type const* result =
			candidates[0]->getOperator()->getResultType();
		
		ValuePassing vp =
			BuiltInTypes::getBoolConversion(getCompileInfo(), result);
		
		if (vp.isPossible()) {
			// FIXME: use boolean result from access check
			vp.checkAccess(*fContext, prev->getLocation(), getCompileInfo());
			return candidates[0]; // return the one candidate
		}
		else {
			getCompileInfo().diag().error(
				prev->getLocation(),
				UTF8STR(
					"the result of `%1 %2 %3' has type `%4' and cannot be "
					"implicitly converted to `bool'"
				),
				prev->toString(), Operators::toString(op), curr->toString(),
				result->toString()
			);
			
			return 0;
		}
	}
	
	TriBool ChainedCompareExpression::isClassLValue() const {
		return false;
	}
	
	TriBool ChainedCompareExpression::isPointerLValue() const {
		return false;
	}
	
	TriBool ChainedCompareExpression::isCall() const {
		return false;
	}
	
	TriBool ChainedCompareExpression::hasResult() const {
		return true;
	}
	
	TriBool ChainedCompareExpression::resultAtCompileTime() const {
		return false; // FIXME: implement constant folding
	}
}
