/*!
	\file charclass.cxx
	\author Kevin André
	\brief Implements the Char class
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-10-08
// Last time modified:	2008-12-26

#include "charclass.h"

#include "base/location.h"

#include "assignmentoperator.h"
#include "binaryoperator.h"
#include "constantvalue.h"
#include "parameter.h"

namespace HYPER_Compiler {
	
	class Char::CharConstructor : public BuiltInConstructor {
	public:
		CharConstructor(Char & cc);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParam;
	};
	
	class Char::EqualsOperator : public BinaryOperator {
	public:
		EqualsOperator(Char & cc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Char::DiffersOperator : public BinaryOperator {
	public:
		DiffersOperator(Char & cc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Char::LessOperator : public BinaryOperator {
	public:
		LessOperator(Char & cc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Char::LessEqOperator : public BinaryOperator {
	public:
		LessEqOperator(Char & cc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Char::GreaterOperator : public BinaryOperator {
	public:
		GreaterOperator(Char & cc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Char::GreaterEqOperator : public BinaryOperator {
	public:
		GreaterEqOperator(Char & cc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Char::AssignOp : public AssignmentOperator {
	public:
		AssignOp(Char & cc);
		
		// TODO : change return type
		void generateCode();
	};
	
	// *** Char ***
	
	Char::Char(CompileInfo& info)
	 : BuiltInClass(info, kSealed)
	{
		// no action
	}
	
	Char::~Char() {
		// no action
	}
	
	HYPER_Common::UTF8String Char::getTypeName() const {
		return UTF8STR("char");
	}
	
	HYPER_Common::UTF8String Char::getDeclarationName() const {
		return UTF8STR("char");
	}
	
	void Char::buildSymbolTable() {
		createConstructors();
		createMembers();
		createOperators();
	}
	
	void Char::createConstructors() {
		// only copy constructor at this time
		registerConstructor(new CharConstructor(*this), true);
	}
	
	void Char::createMembers() {
		// none yet
	}
	
	void Char::createOperators() {
		registerOperator(new EqualsOperator(*this));
		registerOperator(new DiffersOperator(*this));
		registerOperator(new LessOperator(*this));
		registerOperator(new LessEqOperator(*this));
		registerOperator(new GreaterOperator(*this));
		registerOperator(new GreaterEqOperator(*this));
		
		registerOperator(new AssignOp(*this));
	}
	
	
	// *** CharConstructor ***
	
	Char::CharConstructor::CharConstructor(Char & cc)
	 : BuiltInConstructor(&cc, ClassMemberAccess::createPublic(), true)
	{
		fParam.push_back(
			cc.createParameter(UTF8STR("c"), BuiltInTypes::kChar)
		);
	}
	
	const std::vector<Parameter const*>*
	Char::CharConstructor::getParameterList() const {
		return &fParam;
	}
	
	
	// *** EqualsOperator ***
	Char::EqualsOperator::EqualsOperator(Char & cc)
	 : BinaryOperator(Operators::kEquals,
		BuiltInTypes::createChar(cc.getCompileInfo()),
		BuiltInTypes::createChar(cc.getCompileInfo()), true,
		BuiltInTypes::createBool(cc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Char::EqualsOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		return 0; // TODO
	}
	
	// TODO : change return type
	void Char::EqualsOperator::generateCode() {
		// TODO
	}
	
	
	// *** DiffersOperator ***
	Char::DiffersOperator::DiffersOperator(Char & cc)
	 : BinaryOperator(Operators::kDiffers,
		BuiltInTypes::createChar(cc.getCompileInfo()),
		BuiltInTypes::createChar(cc.getCompileInfo()), true,
		BuiltInTypes::createBool(cc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Char::DiffersOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		return 0; // TODO
	}
	
	// TODO : change return type
	void Char::DiffersOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessOperator ***
	Char::LessOperator::LessOperator(Char & cc)
	 : BinaryOperator(Operators::kLess,
		BuiltInTypes::createChar(cc.getCompileInfo()),
		BuiltInTypes::createChar(cc.getCompileInfo()), false,
		BuiltInTypes::createBool(cc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Char::LessOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		return 0; // TODO
	}
	
	// TODO : change return type
	void Char::LessOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessEqOperator ***
	Char::LessEqOperator::LessEqOperator(Char & cc)
	 : BinaryOperator(Operators::kLessOrEqual,
		BuiltInTypes::createChar(cc.getCompileInfo()),
		BuiltInTypes::createChar(cc.getCompileInfo()), false,
		BuiltInTypes::createBool(cc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Char::LessEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		return 0; // TODO
	}
	
	// TODO : change return type
	void Char::LessEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterOperator ***
	Char::GreaterOperator::GreaterOperator(Char & cc)
	 : BinaryOperator(Operators::kGreater,
		BuiltInTypes::createChar(cc.getCompileInfo()),
		BuiltInTypes::createChar(cc.getCompileInfo()), false,
		BuiltInTypes::createBool(cc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Char::GreaterOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		return 0; // TODO
	}
	
	// TODO : change return type
	void Char::GreaterOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterEqOperator ***
	Char::GreaterEqOperator::GreaterEqOperator(Char & cc)
	 : BinaryOperator(Operators::kGreaterOrEqual,
		BuiltInTypes::createChar(cc.getCompileInfo()),
		BuiltInTypes::createChar(cc.getCompileInfo()), false,
		BuiltInTypes::createBool(cc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Char::GreaterEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		return 0; // TODO
	}
	
	// TODO : change return type
	void Char::GreaterEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** AssignOp ***
	Char::AssignOp::AssignOp(Char & cc)
	 : AssignmentOperator(Operators::kSimpleAssign,
		BuiltInTypes::createChar(cc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Char::AssignOp::generateCode() {
		// TODO
	}
}
