/*!
	\file class.cxx
	\author Kevin André
	\brief Implements the Class class
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-05-22
// Last time modified:	2009-08-23

#include "class.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "field.h"
#include "initializationhandler.h"
#include "overloadedconstructor.h"
#include "overloadedprocedure.h"
#include "overloadresolution.h"

namespace HYPER_Compiler {
	
	/// Default initialization handler for class types.
	class Class::ClassInitHandler : public InitializationHandler {
	public:
		TriBool
		checkInitByDefaultValue(ClassContext const& ctxt,
			Location const* useLoc, CompileInfo& info);
		
		TriBool
		checkInitByExpression(Expression* e, ClassContext const& ctxt,
			Location const* loc, CompileInfo& info);
		
		//void checkAccess(ClassContext const& ctxt,
		//	Location const* loc, CompileInfo& info);
		
		ClassInitHandler(Class const* c);
		
	private:
		TriBool
		checkInitByConstructor(ClassContext const& ctxt,
			const Location* loc, std::vector<Expression const*> args,
			CompileInfo& info);
		
		Class const* fClass;
		Expression* fExpr;
		Constructor const* fConstr;
	};
	
	
	// *** Class ***
	
	Class::Class(CompileInfo& info)
	 : fCompileInfo(info), fConstrList(0), fCopyConstr(0)
	{
		// no action
	}
	
	Class::~Class() {
		// no action
	}
	
	CompileInfo& Class::getCompileInfo() const {
		return fCompileInfo;
	}
	
	NamedDeclaredItem* Class::findMember(HYPER_Common::UTF8String name) const {
		return findLocalSymbol(name);
	}
	
	OverloadedConstructor const* Class::getConstructors() const {
		hyp_assert(getASS() != kStatic);
		return fConstrList;
	}
	
	HYPER_Common::UTF8String Class::getDescription() const {
		return UTF8STR("class %1").compose(quote(getTypeName()));
	}
	
	Class const* Class::getOwnerClass() const {
		return getParentClass();
	}
	
	ClassMemberAccess Class::getAccessInOwner() const {
		return getAccess();
	}
	
	InitializationHandler* Class::getInitHandler() const {
		return new ClassInitHandler(this);
	}
	
	void Class::putInSymbolTable(Scope& parent) {
		//debug.line("Class::putInSymbolTable called for class " + getDeclarationName());
		registerInParent(parent);
	}
	
	void Class::registerClass(Class* c) {
		//debug.line("Class::registerClass called");
		registerSymbolWithDiagnostic(c, fCompileInfo);
		c->putInSymbolTable(*this);
	}
	
	void Class::registerConstructor(Constructor* c, bool isCopyConstructor) {
		//debug.line("Class::registerConstructor called");
		if (fConstrList == 0) {
			fConstrList = new OverloadedConstructor(c);
		}
		else {
			fConstrList->addOverload(c);
		}
		c->putInSymbolTable(*this);
		
		if (isCopyConstructor) {
			registerCopyConstructor(c);
		}
	}
	
	void Class::registerCopyConstructor(Constructor const* c) {
		hyp_assert(fCopyConstr == 0);
		fCopyConstr = c;
	}
	
	Constructor const* Class::getCopyConstructor() const {
		return fCopyConstr;
	}
	
	void Class::registerProcedure(Procedure* p) {
		HYPER_Common::UTF8String caption = p->getCaption();
		//debug.line("Class::registerProcedure called for `" + caption + "'");
		
		p->putInSymbolTable(*this);
		
		NamedDeclaredItem* item = findLocalSymbol(caption);
		
		if (item != 0) {
			if (OverloadedProcedure* overload =
					dynamic_cast<OverloadedProcedure*>(item))
			{
				overload->addOverload(p);
				//debug.line(" procedure added to overload");
			}
			else {
				// name clash with something that is not a procedure
				fCompileInfo.diag().error(
					p->getLocation(),
					UTF8STR("declaration of %1 conflicts with..."),
					quote(p->getDescription())
				).addInfo(
					UTF8STR("...this declaration: ") + item->getDescription(),
					item->getLocation()
				);
				return;
			}
		}
		else {
			OverloadedProcedure* ovp = new OverloadedProcedure(p);
			registerSymbol(ovp); // register name in the scope of this class
			fProcList.push_back(ovp);
			//debug.line(" procedure put in new overload");
		}
	}
	
	void Class::registerField(Field* f) {
		//debug.line("Class::registerField called");
		f->putInSymbolTable(*this);
		
		std::vector<FieldDecl*> const& decl = f->getDeclarations();
		for (unsigned int i = 0; i < decl.size(); ++i) {
			registerSymbolWithDiagnostic(decl[i], getCompileInfo());
		}
	}
	
	bool Class::checkOverloads() {
		//debug.line("Class::checkOverloads called for `" + getTypeName() + "'");
		bool ok = true;
		
		if (getASS() != kStatic) {
			ok &= fConstrList->checkOverloads(fCompileInfo.diag());
		}
		
		for (unsigned i = 0; i < fProcList.size(); ++i) {
			ok &= fProcList[i]->checkOverloads(fCompileInfo.diag());
		}
		
		return ok;
	}
	
	HYPER_Common::UTF8String Class::getResolveDescription() const {
		return UTF8STR("class %1").compose(quote(getTypeName()));
	}
	
	const Location* Class::getResolveLocation() const {
		return getLocation();
	}
	
	
	// *** ClassInitHandler ***
	
	Class::ClassInitHandler::ClassInitHandler(Class const* c)
	 : fClass(c), fExpr(0)
	{
		//
	}
	
	TriBool
	Class::ClassInitHandler::checkInitByDefaultValue(ClassContext const& ctxt,
		Location const* useLoc, CompileInfo& info)
	{
		std::vector<Expression const*> args; // empty arguments list
		return checkInitByConstructor(ctxt, useLoc, args, info);
	}
	
	TriBool
	Class::ClassInitHandler::checkInitByExpression(Expression* e,
		ClassContext const& ctxt, Location const* useLoc, CompileInfo& info)
	{
		fExpr = e;
		std::vector<Expression const*> args;
		args.push_back(fExpr); // one argument
		return checkInitByConstructor(ctxt, useLoc, args, info);
	}
	
	//void
	//ClassInitHandler::checkAccess(ClassContext const& ctxt,
	//	Location const* loc, CompileInfo& info)
	//{
		// TODO
	//}
	
	TriBool
	Class::ClassInitHandler::checkInitByConstructor(ClassContext const& ctxt,
		const Location* useLoc, std::vector<Expression const*> args,
		CompileInfo& info)
	{
		AbstractSealedStatic ass = fClass->getASS();
		if (ass == kStatic || ass == kAbstract) {
			// Cannot have storage that is made of an abstract or static class.
			// Don't give an error, it has already been generated.
			return false;
		}
		
		OverloadedConstructor const* ovc = fClass->getConstructors();
		//debug << "class has " << ovc->getOverloadList().size() << " constructors available" << newl;
		
		typedef OverloadResolution<OverloadedConstructor> ORC;
		
		// do overload resolution for constructor call
		ORC::Result call = ORC::resolveCall(ovc, args, info);
		
		switch (call.getResult()) {
		case ORC::Result::kNotResolved:
			// cannot check
			return false;
			
		case ORC::Result::kNoCandidates:
			info.diag().error(
				UTF8STR(
					"no match found for the constructor call that is needed "
						"for the initialization of this variable"
				),
				useLoc
			).addInfo(
				useLoc,
				UTF8STR("call needed: %1.%2"),
				fClass->getTypeName(),
				ORC::getCallPrototype(QualifiedName(UTF8STR("new")), args)
			);
			return false;
			
		case ORC::Result::kMultipleCandidates: {
			EasyDiagnosticRef e =
				info.diag().error(
					UTF8STR(
						"multiple matches for the constructor call that is "
							"needed for the initialization of this variable"
					),
					useLoc
				).addInfo(
					useLoc,
					UTF8STR("call needed: %1.%2"),
					fClass->getTypeName(),
					ORC::getCallPrototype(QualifiedName(UTF8STR("new")), args)
				);
			
			call.listCandidates(e);
			
			return false; }
		
		case ORC::Result::kSingleCandidate:
			break; // handled below the switch statement
		}
		
		bool ok = true;
		
		const Constructor* constr = call.getCandidates()[0]->getEntity();
		
		// check access
		ok &= ORC::checkAccess(call.getCandidates()[0], ctxt, useLoc, info);
		ok &= AccessChecker::check(ctxt, *constr, useLoc, info);
		
		// check for implicit use of non-implicit constructor
		if (args.size() == 1 && !constr->isImplicit()) {
			info.diag().error(
				UTF8STR(
					"implicit use of non-implicit constructor in the "
						"initialization of this variable"
				),
				useLoc
			).addInfo(
				UTF8STR("constructor: ") + constr->headerToString(),
				constr->getLocation()
			);
			
			ok = false;
		}
		
		// FIXME
		//cnds[0]->getConversions()[0]->checkAccess(
		//	ctxt, args[0]->getLocation(), info
		//);
		
		fConstr = constr;
		
		return ok;
	}
}
