/*!
	\file class.h
	\author Kevin André
	\brief Contains the Class class
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-05-22
// Last time modified:	2008-12-26

#ifndef DEF_CLASS_H
 #define DEF_CLASS_H
 
 #include "important_fwdecl.h"
 #include "membertypes.h"
 #include "nameddeclareditem.h"
 #include "operators.h"
 #include "qualifiedname.h"
 #include "resolveable.h"
 #include "scopes.h"
 #include "tribool.h"
 
 #include <map>
 #include <vector>

namespace HYPER_Compiler {
	
	class Constructor;
	class CrementOperator;
	class Field;
	class InitializationHandler;
	class OverloadedArrayOperator;
	class OverloadedAssignmentOperator;
	class OverloadedBinaryOperator;
	class OverloadedConstructor;
	class OverloadedProcedure;
	class Procedure;
	class UnaryOperator;
	
	/// Represents the general concept of a class
	class Class : public NamedDeclaredItem, public Scope, public Resolveable {
	public:
		
		typedef std::map<Operators::BinaryOperator, OverloadedBinaryOperator*> BinOpsMap;
		typedef std::map<Operators::CrementOperator, CrementOperator*> CremOpsMap;
		typedef std::map<Operators::UnaryOperator, UnaryOperator*> UnOpsMap;
		typedef std::map<Operators::AssignmentOperator, OverloadedAssignmentOperator*> AssignOpsMap;
		
		/// Destructor
		virtual ~Class();
		
		/// Get the CompileInfo
		CompileInfo& getCompileInfo() const;
		
		/// Returns the type 'name' of the class
		virtual HYPER_Common::UTF8String getTypeName() const = 0;
		
		/// Returns the class member access level of this class inside its parent
		virtual ClassMemberAccess getAccess() const = 0;
		
		/// Returns the abstract/sealed/static/none qualifier of this class
		virtual AbstractSealedStatic getASS() const = 0;
		
		/// Returns the binary operators of the class
		virtual BinOpsMap const& getBinaryOperators() const = 0;
		
		/// Returns the crement operators of the class
		virtual CremOpsMap const& getCrementOperators() const = 0;
		
		/// Returns the unary operators of the class
		virtual UnOpsMap const& getUnaryOperators() const = 0;
		
		/// Returns the assignment operators of the class
		virtual AssignOpsMap const& getAssignmentOperators() const = 0;
		
		/// Returns the array operator(s) of the class
		virtual OverloadedArrayOperator const* getArrayOperators() const = 0;
		
		/// Returns the location of the class in the source code
		virtual const Location* getLocation() const = 0;
		
		/// Get the parent of this class. Returns 0 for no parent.
		virtual Class const* getParentClass() const = 0;
		
		/// Searches the class scope for a named member
		NamedDeclaredItem* findMember(HYPER_Common::UTF8String name) const;
		
		/// Get the class' copy constructor
		Constructor const* getCopyConstructor() const;
		
		/// Get the class' constructors (not for static classes!!)
		OverloadedConstructor const* getConstructors() const;
		
		HYPER_Common::UTF8String getDescription() const;
		Class const* getOwnerClass() const;
		
		ClassMemberAccess getAccessInOwner() const;
		
		virtual InitializationHandler* getInitHandler() const;
		
		void putInSymbolTable(Scope& parent);
		
		/// Returns true if an (input) parameter of this type can be passed in
		/// a register without effect on user code.
		virtual bool canBeInvisiblyPassedByValue() const {
			return false; // default is NO
		}
		
	protected:
		/// Constructor
		Class(CompileInfo& info);
		
		HYPER_Common::UTF8String getResolveDescription() const;
		const Location* getResolveLocation() const;
		
		virtual bool doInterfaceResolve(Trace const& trace) { return true; }
		virtual bool doFullResolve(Trace const& trace) { return true; }
		
		void registerClass(Class* c);
		void registerConstructor(Constructor* c, bool isCopyConstructor);
		void registerCopyConstructor(Constructor const* c);
		void registerProcedure(Procedure* p);
		void registerField(Field* f);
		bool checkOverloads();
		
	private:
		class ClassInitHandler;
		
		CompileInfo& fCompileInfo;
		OverloadedConstructor* fConstrList;
		Constructor const* fCopyConstr;
		std::vector<OverloadedProcedure*> fProcList;
	};
}
#endif
