/*!
	\file classrepository.cxx
	\brief Implements the ClassRepository class
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-09-29
// Last time modified:	2009-02-08

#include "classrepository.h"

#include "base/asserts.h"

#include "class.h"
#include "compileinfo.h"
#include "declareditem.h"
#include "namespace.h"
#include "overloadedbinaryoperator.h"
#include "type.h"

#include "arrayclass.h"
#include "boolclass.h"
#include "byteclass.h"
#include "charclass.h"
#include "doubleclass.h"
#include "int16class.h"
#include "int32class.h"
#include "int64class.h"
#include "intclass.h"
#include "nat16class.h"
#include "nat32class.h"
#include "nat64class.h"
#include "natclass.h"
#include "objectclass.h"
#include "realclass.h"
#include "singleclass.h"
#include "stringclass.h"

//~ #include "base/debug.h"

using std::set;
using std::map;

namespace HYPER_Compiler {
	
	// TODO: get rid of these   (use static vars in the get*Class() functions)
	Bool* ClassRepository::fBoolClass;
	Byte* ClassRepository::fByteClass;
	Char* ClassRepository::fCharClass;
	Double* ClassRepository::fDoubleClass;
	Nat16* ClassRepository::fNat16Class;
	Nat32* ClassRepository::fNat32Class;
	Nat64* ClassRepository::fNat64Class;
	Nat* ClassRepository::fNatClass;
	Object* ClassRepository::fObjectClass;
	Int16* ClassRepository::fInt16Class;
	Int32* ClassRepository::fInt32Class;
	Int64* ClassRepository::fInt64Class;
	Int* ClassRepository::fIntClass;
	Real* ClassRepository::fRealClass;
	Single* ClassRepository::fSingleClass;
	String* ClassRepository::fStringClass;
	
	ClassRepository::ClassRepository(CompileInfo& info)
	 : fCompileInfo(info)
	{
		fCompileInfo.setClassRepository(this);
		registerBuiltIns();
	}
	
	const ClassRepository::BinOpsMap&
	ClassRepository::getBinaryOperators() const
	{
		return fAllBinOps;
	}
	
	Class* ClassRepository::getClassOfBuiltin(BuiltInTypes::Primitive p) {
		switch (p) {
		case BuiltInTypes::kBool:
			return getBoolClass();
		case BuiltInTypes::kByte:
			return getByteClass();
		case BuiltInTypes::kChar:
			return getCharClass();
		case BuiltInTypes::kDouble:
			return getDoubleClass();
		case BuiltInTypes::kInt:
			return getIntClass();
		case BuiltInTypes::kInt16:
			return getInt16Class();
		case BuiltInTypes::kInt32:
			return getInt32Class();
		case BuiltInTypes::kInt64:
			return getInt64Class();
		case BuiltInTypes::kNat:
			return getNatClass();
		case BuiltInTypes::kNat16:
			return getNat16Class();
		case BuiltInTypes::kNat32:
			return getNat32Class();
		case BuiltInTypes::kNat64:
			return getNat64Class();
		case BuiltInTypes::kObject:
			return getObjectClass();
		case BuiltInTypes::kReal:
			return getRealClass();
		case BuiltInTypes::kSingle:
			return getSingleClass();
		case BuiltInTypes::kString:
			return getStringClass();
		case BuiltInTypes::kNone:
		default:
			hyp_assert_unreachable();
		}
	}
	
	Object* ClassRepository::getObjectClass() {
		if (fObjectClass == 0) {
			return fObjectClass = new Object(fCompileInfo);
		}
		else {
			return fObjectClass;
		}
	}
	
 	Bool* ClassRepository::getBoolClass() {
		if (fBoolClass == 0) {
			return fBoolClass = new Bool(fCompileInfo);
		}
		else {
			return fBoolClass;
		}
	}
	
	Byte* ClassRepository::getByteClass() {
		if (fByteClass == 0) {
			return fByteClass = new Byte(fCompileInfo);
		}
		else {
			return fByteClass;
		}
	}
	
	Char* ClassRepository::getCharClass() {
		if (fCharClass == 0) {
			return fCharClass = new Char(fCompileInfo);
		}
		else {
			return fCharClass;
		}
	}
	
	Double* ClassRepository::getDoubleClass() {
		if (fDoubleClass == 0) {
			return fDoubleClass = new Double(fCompileInfo);
		}
		else {
			return fDoubleClass;
		}
	}
	
	Int16* ClassRepository::getInt16Class() {
		if (fInt16Class == 0) {
			return fInt16Class = new Int16(fCompileInfo);
		}
		else {
			return fInt16Class;
		}
	}
	
	Int32* ClassRepository::getInt32Class() {
		if (fInt32Class == 0) {
			return fInt32Class = new Int32(fCompileInfo);
		}
		else {
			return fInt32Class;
		}
	}
	
	Int64* ClassRepository::getInt64Class() {
		if (fInt64Class == 0) {
			return fInt64Class = new Int64(fCompileInfo);
		}
		else {
			return fInt64Class;
		}
	}
	
	Int* ClassRepository::getIntClass() {
		if (fIntClass == 0) {
			Platform::WordSize ws =
				fCompileInfo.getTargetPlatform().getWordSize();
			
			fIntClass = new Int(fCompileInfo, toIntegerSize(ws));
		}
		
		return fIntClass;
	}
	
	Nat16* ClassRepository::getNat16Class() {
		if (fNat16Class == 0) {
			return fNat16Class = new Nat16(fCompileInfo);
		}
		else {
			return fNat16Class;
		}
	}
	
	Nat32* ClassRepository::getNat32Class() {
		if (fNat32Class == 0) {
			return fNat32Class = new Nat32(fCompileInfo);
		}
		else {
			return fNat32Class;
		}
	}
	
	Nat64* ClassRepository::getNat64Class() {
		if (fNat64Class == 0) {
			return fNat64Class = new Nat64(fCompileInfo);
		}
		else {
			return fNat64Class;
		}
	}
	
	Nat* ClassRepository::getNatClass() {
		if (fNatClass == 0) {
			Platform::WordSize ws =
				fCompileInfo.getTargetPlatform().getWordSize();
			
			fNatClass = new Nat(fCompileInfo, toIntegerSize(ws));
		}
		
		return fNatClass;
	}
	
	Real* ClassRepository::getRealClass() {
		if (fRealClass == 0) {
			return fRealClass = new Real(fCompileInfo);
		}
		else {
			return fRealClass;
		}
	}
	
	Single* ClassRepository::getSingleClass() {
		if (fSingleClass == 0) {
			return fSingleClass = new Single(fCompileInfo);
		}
		else {
			return fSingleClass;
		}
	}
	
	String* ClassRepository::getStringClass() {
		if (fStringClass == 0) {
			return fStringClass = new String(fCompileInfo);
		}
		else {
			return fStringClass;
		}
	}
	
	Class*
	ClassRepository::getArrayClass(Type const* elemType,
		unsigned dimensionCount)
	{
		hyp_assert(elemType != 0);
		
		// get hash string  (concatenate dim count and element type)
		HYPER_Common::UTF8String tStr = dimensionCount + elemType->toString();
		
		//~ debug.line(
			//~ UTF8STR("Get ARRAY TYPE for element type: %1").compose(
				//~ elemType->toString()
			//~ )
		//~ );
		
		// search for requested array type in lookup table
		std::map<HYPER_Common::UTF8String, Array*>::iterator it =
			fArrayClasses.find(tStr);
		
		if (it == fArrayClasses.end()) { // not found in lookup table
			// create new array class
			Array* cls = new Array(fCompileInfo, elemType, dimensionCount);
			
			// register the class (for building symbol table)
			registerBuiltInClass(cls, 0);
			
			// put class in lookup table
			fArrayClasses[tStr] = cls;
			
			return cls;
		}
		else { // found in lookup table
			// return the class that we found in the lookup table
			return it->second;
		}
	}
	
	void ClassRepository::registerClass(Class* c) {
		hyp_assert(c != 0);
		fClasses.insert(c);
	}
	
	void
	ClassRepository::registerBuiltInClass(BuiltInClass* c, Namespace* parent)
	{
		//debug.line(
		//	"ClassRepository::registerBuiltInClass : for class "
		//		+ c->getTypeName()
		//);
		registerClass(c);
		putInSymbolTable(c, parent);
	}
	
	void
	ClassRepository::putInSymbolTable(BuiltInClass* c, Namespace* parent)
	{
		if (parent != 0) {
			parent->registerClass(c);
			// c->putInSymbolTable is called by the namespace
		}
		else {
			c->putInSymbolTable(fCompileInfo.getRootScope());
		}
		c->buildSymbolTable();
	}
	
	void ClassRepository::registerBuiltIns() {
		//~ debug.line(UTF8STR("registering built-in primitive types"));
		registerBuiltInClass(getBoolClass());
		registerBuiltInClass(getByteClass());
		registerBuiltInClass(getCharClass());
		registerBuiltInClass(getDoubleClass());
		registerBuiltInClass(getNat16Class());
		registerBuiltInClass(getNat32Class());
		registerBuiltInClass(getNat64Class());
		registerBuiltInClass(getNatClass());
		registerBuiltInClass(getObjectClass());
		registerBuiltInClass(getInt16Class());
		registerBuiltInClass(getInt32Class());
		registerBuiltInClass(getInt64Class());
		registerBuiltInClass(getIntClass());
		registerBuiltInClass(getRealClass());
		registerBuiltInClass(getSingleClass());
		registerBuiltInClass(getStringClass());
	}
	
	void ClassRepository::buildOperatorsLists() {
		typedef set<Class*>::iterator Iterator;
		
		for (Iterator it = fClasses.begin(); it != fClasses.end(); ++it) {
			typedef BinOpsMap::const_iterator Iterator2;
			
			for (Iterator2 it2 = (*it)->getBinaryOperators().begin();
					it2 != (*it)->getBinaryOperators().end(); ++it2)
			{
				OverloadedBinaryOperator* ov = it2->second;
				
				OverloadedBinaryOperator* target =
					fAllBinOps[ov->getOperator()];
				
				if (target == 0) {
					// Take a copy of the class' overloadedbinaryoperator
					// as a start.
					fAllBinOps[ov->getOperator()] =
						new OverloadedBinaryOperator(*ov);
				}
				else {
					target->addOverloadList(ov);
				}
			}
		}
	}
}
