/*!
	\file classrepository.h
	\brief Contains the ClassRepository class
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-09-29
// Last time modified:	2008-10-19

#ifndef DEF_CLASSREPOSITORY_H
 #define DEF_CLASSREPOSITORY_H
 
 #include "operators.h"
 #include "builtintypes.h"
 #include "important_fwdecl.h"
 
 #include <set>
 #include <map>

namespace HYPER_Compiler {
	
	class BuiltInClass;
	class OverloadedBinaryOperator;
	
	class Array;
	class Bool;
	class Byte;
	class Char;
	class Double;
	class Nat16;
	class Nat32;
	class Nat64;
	class Nat;
	class Object;
	class Int16;
	class Int32;
	class Int64;
	class Int;
	class Real;
	class Single;
	class String;
	
	class ClassRepository {
	public:
		ClassRepository(CompileInfo& info);
		
		void buildOperatorsLists();
		
		typedef std::map<Operators::BinaryOperator,
			OverloadedBinaryOperator*> BinOpsMap;
		
		const BinOpsMap& getBinaryOperators() const;
		
		Class* getClassOfBuiltin(BuiltInTypes::Primitive p);
		Class* getArrayClass(Type const* elemType, unsigned dimensionCount);
		
		void registerClass(Class* c);
		void registerBuiltInClass(BuiltInClass* c, Namespace* parent = 0);
		
	private:
		void putInSymbolTable(BuiltInClass* c, Namespace* parent);
		void registerBuiltIns();
		
		Bool* getBoolClass();
		Byte* getByteClass();
		Char* getCharClass();
		Double* getDoubleClass();
		Nat16* getNat16Class();
		Nat32* getNat32Class();
		Nat64* getNat64Class();
		Nat* getNatClass();
		Object* getObjectClass();
		Int16* getInt16Class();
		Int32* getInt32Class();
		Int64* getInt64Class();
		Int* getIntClass();
		Real* getRealClass();
		Single* getSingleClass();
		String* getStringClass();
		
		static Bool* fBoolClass;
		static Byte* fByteClass;
		static Char* fCharClass;
		static Double* fDoubleClass;
		static Nat16* fNat16Class;
		static Nat32* fNat32Class;
		static Nat64* fNat64Class;
		static Nat* fNatClass;
		static Object* fObjectClass;
		static Int16* fInt16Class;
		static Int32* fInt32Class;
		static Int64* fInt64Class;
		static Int* fIntClass;
		static Real* fRealClass;
		static Single* fSingleClass;
		static String* fStringClass;
		
		std::set<Class*> fClasses;
		std::map<HYPER_Common::UTF8String, Array*> fArrayClasses;
		BinOpsMap fAllBinOps;
		CompileInfo& fCompileInfo;
	};
}
#endif
