/*!
	\file classtype.cxx
	\brief Implements the ClassType class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-12-16
// Last time modified:	2009-01-05

#include "classtype.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "accesschecker.h"
#include "class.h"
#include "compileinfo.h"
#include "declareditem.h"
#include "namespace.h"
#include "type.h"

//~ #include "base/debug.h"

namespace HYPER_Compiler {
	
	ClassType::ClassType(Class* c, bool isConst, const Location* loc)
	 : DeclaredType(c->getCompileInfo(), loc), fTypeName(c->getTypeName()),
		fConst(isConst), fClassRef(c)
	{
		setResolvedType(Type::createClassType(isConst, c));
	}
	
	ClassType::ClassType(CompileInfo& info, QualifiedName const& t,
		bool isConst, const Location* loc)
	 : DeclaredType(info, loc), fTypeName(t), fConst(isConst), fClassRef(0)
	{
		//
	}
	
	ClassType::ClassType(ClassType const& c)
	 : DeclaredType(c.getCompileInfo(), c.getLocation()),
		fTypeName(c.fTypeName), fConst(c.fConst), fClassRef(c.fClassRef)
	{
		// copy constructor should not be used
		hyp_assert_unreachable();
	}
	
	bool ClassType::isArray() const {
		return false;
	}
	
	bool ClassType::isPointer() const {
		return false;
	}
	
	bool ClassType::isDoublePointer() const {
		return false;
	}
	
	bool ClassType::isClass() const {
		return true;
	}
	
	bool ClassType::isConst() const {
		return fConst;
	}
	
	void ClassType::makeConst() {
		fConst = true;
	}
	
	HYPER_Common::UTF8String ClassType::toString() const {
		HYPER_Common::UTF8String constant =
			fConst ? UTF8STR("const ") : UTF8STR("");
		
		if (fClassRef) {
			return constant + fClassRef->getTypeName();
		}
		else {
			return constant + fTypeName.toString();
		}
	}
	
	bool ClassType::doInterfaceResolve(Trace const& trace) {
		//debug.line("ClassType::doInterfaceResolve called");
		if (!resolveClassName()) { // class not resolved ?
			return false; // cannot continue
		}
		
		// check for static
		if (fClassRef->getASS() == kStatic) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR("static class used in type %1"),
				quote(toString())
			);
			
			return false;
		}
		
		// Check direct allocation
		if (mustBeDirectAllocatable()) {
			//~ debug.line(UTF8STR("ClassType:  mustBeDirectAllocatable() == true"));
			if (fClassRef->getASS() == kAbstract) {
				// user wants to allocate an abstract class ...
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR(
						"abstract class used without a pointer to it in type %1"
					),
					quote(toString())
				);
				
				return false;
			}
			else {
				// interface of class must be completely resolved
				return fClassRef->resolveInterface(trace);
			}
		}
		
		return true;
	}
	
	bool ClassType::resolveClassName() {
		if (fClassRef != 0) { // already know what class ?
			// no type resolving to do here
			return true;
		}
		
		//debug.line(
		//	UTF8STR("ClassType: looking up typename `%1'").compose(
		//		fTypeName.toString()
		//	)
		//);
		
		// look up root name
		HYPER_Common::UTF8String root = fTypeName.getRootName();
		NamedDeclaredItem* item = getLookupLocation()->findSymbol(root);
		if (item == 0) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR("%1 undeclared"),
				quote(root)
			);
			return false;
		}
		
		Namespace const* ns = 0;
		Class* cl = 0;
		Scope const* curScope = 0;
		unsigned int nameIndex = 0;
		bool lastName = (nameIndex == fTypeName.countNames() - 1);
		QualifiedName curFullName(root);
		HYPER_Common::UTF8String curName = root;
		
		do {
			cl = dynamic_cast<Class*>(item);
			ns = dynamic_cast<Namespace const*>(item);
			
			if (lastName && cl == 0) {
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR("%1 is not a class"),
					quote(curFullName.toString())
				);
				return false;
			}
			else if (cl == 0 && ns == 0) {
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR("%1 is not a namespace or a class"),
					quote(curFullName.toString())
				);
				return false;
			}
			else {
				curScope = dynamic_cast<Scope const*>(item);
				hyp_assert(curScope != 0 && "not a Scope ???");
			}
			
			if (lastName) {
				break; // we're done now
			}
			else {
				++nameIndex;
				curName = fTypeName[nameIndex];
				
				item = curScope->findLocalSymbol(curName);
				if (item == 0) {
					getCompileInfo().diag().error(
						getLocation(),
						UTF8STR("%1 does not have a member called %2"),
						quote(curFullName.toString()),
						quote(curName)
					);
					return false;
				}
				
				curFullName.addIdentifier(curName);
				lastName = (nameIndex == fTypeName.countNames() - 1);
			}
		} while (true);
		
		hyp_assert(cl != 0);
		fClassRef = cl;
		//debug.line(UTF8STR(" typename was found"));
		
		setResolvedType(Type::createClassType(fConst, cl));
		return true;
	}
}
