/*!
	\file compiler.cxx
	\brief Implements the Compiler class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-04-03
// Last time modified:	2008-11-16

/* compiler version milestones are now in the file version.h */

#include "compiler.h"

#include "common/console.h"
#include "common/filesystem.h"
#include "common/utf8textflowstream.h"

#include "base/asserts.h"
#include "base/compilationfailederror.h"
#include "base/debug.h"
#include "base/easydiagnostics.h"
#include "base/hostplatform.h"

#include "lexer/fileexceptions.h"
#include "lexer/lexer.h"
#include "lexer/reader.h"

#include "classrepository.h"
#include "parser.h"
#include "resolver.h"
#include "sourcefile.h"
#include "version.h"

#include <boost/shared_ptr.hpp>

using HYPER_Common::CodePoint;
using HYPER_Common::FileSystem;
using HYPER_Common::PullStream;
using HYPER_Common::UTF8Byte;
using HYPER_Common::UTF8Decoder;
using HYPER_Common::UTF8PushStream;
using HYPER_Common::UTF8TextFlowStream;

namespace HYPER_Compiler {
	
	Compiler::Compiler(CompileOptions o)
	 : fMainSourcefile(0)
	{
		fInfo.setOptions(o);
	}
	
	Compiler::~Compiler() {
		// nothing to delete
	}
	
	HYPER_Common::UTF8String Compiler::getVersion() {
 		return HYPER_Common::UTF8String::fromCString(Version::get());
	}
	
	void Compiler::dumpSpecs() {
		unsigned wordsize = Platform::toNumeric(gHostPlatform.getWordSize());
		char const* endianness =
			Platform::toString(gHostPlatform.getEndianness());
		
		UTF8PushStream& out = HYPER_Common::Console::getStdOut();
		
		out.line(UTF8STR("compiler_version: ") + getVersion());
		out.line(UTF8STR("host_wordsize: ") + wordsize);
		out.line(UTF8STR("host_endianness: ") + UTF8STR(endianness));
	}
	
	void Compiler::Compile(HYPER_Common::FilePath const& filename) {
		//hyp_assert(filename.isAbsolute());
		
		// enable/disable debug output
		Debug::enableDebugOutput(fInfo.getOptions().fgDebug);
		
		// print compiler identifier if debug output enabled
		debug.line();
		debug.line(
			UTF8STR("[ Hyper Compiler %1 ]").compose(Compiler::getVersion())
		);
		debug.line();
		
		// prepare shared data
		ClassRepository* repo = new ClassRepository(fInfo);
		
		// read & parse the file
		
		// FIXME: when implementing multiple sourcefiles, remove fixed file ID
		FileID fileID(1);
		fInfo.registerFile(fileID, filename);
		PullStream<UTF8Byte>* stream = FileSystem::openInputFile(filename);
		
		if (stream->streamOK()) {
			UTF8Decoder decoder;
			PullStream<CodePoint>* decodeStr = decoder.puller(*stream);
			Reader reader(*decodeStr, false);
			
			try {
				Lexer lexer(reader, fileID, &fInfo);
				Parser parser(ParserInfo(lexer, fInfo));
				
				fMainSourcefile =
					parser.doTranslationUnit(
						repo, filename, fileID, !fInfo.getOptions().fgDontLink
					);
			}
			catch (FatalParseError const& fpe) {
				// Add note saying in simple words that the parser stopped
				// before finishing and that the resolver was not run.
				fInfo.diag().note(
					UTF8STR(
						"error checking was halted because of a confusing "
							"syntax error"
					),
					fpe.getLocation()
				);
				throw CompilationFailedError(true);
			}
		}
		else {
			// cannot read file
			throw FileNotReadableException();
		}
		
		// print syntax tree after parsing, if this was requested
		if (fInfo.getOptions().fgPrint_AST) {
			HYPER_Common::UTF8TextFlowStream& out =
				HYPER_Common::Console::getTextFlowStdOut();
			
			out.line();
			out.line(UTF8STR("/# *** begin AST ***"));
			out.line();
			
			fMainSourcefile->myPrint(out);
			
			out.line();
			out.line(UTF8STR("\\# *** end AST ***"));
			out.line();
		}
		
		if (fInfo.getOptions().fgSyntaxOnly) {
			if (fInfo.errorsPresent()) {
				throw CompilationFailedError(false);
			}
			else {
				return;
			}
		}
		
		// semantic analysis
		Resolver resolver(*fMainSourcefile, fInfo);
		resolver.run();
		
		// bail out if errors found
		if (fInfo.errorsPresent()) {
			throw CompilationFailedError(false);
		}
		
		// Generate code & optimize
		if (fInfo.getOptions().fgNoOutput == false) {
			// TODO: generate code
		}
		
		// Clean up
		delete fMainSourcefile;
	}
	
	ErrorManager const& Compiler::getErrorManager() const {
		return fInfo.getErrorManager();
	}
}
