/*!
	\file conditionalexpression.cxx
	\brief Implements the ConditionalExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-04
// Last time modified:	2009-03-13

#include "conditionalexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "boolconditionexpression.h"
#include "compileinfo.h"
#include "conversions.h"
#include "simpleparameter.h"
#include "type.h"

namespace HYPER_Compiler {
	
	ConditionalExpression::ConditionalExpression(
		BoolConditionExpression* decision, Expression* trueExpr,
		Expression* falseExpr, ClassContext const* cCtxt)
	 : Expression(decision->getCompileInfo(), decision->getLocation()),
		fDecision(decision), fTrue(trueExpr), fFalse(falseExpr),
		fContext(cCtxt)
	{
		//
	}
	
	ConditionalExpression::~ConditionalExpression() {
		//
	}
	
	HYPER_Common::UTF8String ConditionalExpression::toString() const {
		return UTF8STR("(%1 ? %2 : %3)").compose(
			fDecision->toString(), fTrue->toString(), fFalse->toString()
		);
	}
	
	void ConditionalExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		fDecision->putInSymbolTable(loc);
		fTrue->putInSymbolTable(loc);
		fFalse->putInSymbolTable(loc);
	}
	
	bool ConditionalExpression::doFullResolve(Trace const& trace) {
		bool resolveOK = true;
		resolveOK &= fDecision->resolveFull(trace);
		resolveOK &= fTrue->resolveFull(trace);
		resolveOK &= fFalse->resolveFull(trace);
		fDecision->checkReturnsValue();
		fTrue->checkReturnsValue();
		fFalse->checkReturnsValue();
		
		// Checking if the condition expression is a bool or converts to a bool
		// is not needed as it is done by the expression itself (The condition
		// expression is a BoolConditionExpression).
		
		if (fTrue->isResolvedAndReturnsResult()
			&& fFalse->isResolvedAndReturnsResult())
		{
			resolveOK &= determineResultType();
		}
		
		return resolveOK;
	}
	
	bool ConditionalExpression::determineResultType() {
		Type const* trueT =
			fTrue->getResult().getType();
		Type const* falseT =
			fFalse->getResult().getType();
		
		// check if both have the exact same type
		if (trueT->constEquals(falseT)) {
			setResolved();
			setResultType(trueT);
			return true;
		}
		
		SimpleParameter trueParam(trueT, SimpleParameter::kCopyValue);
		SimpleParameter falseParam(falseT, SimpleParameter::kCopyValue);
		ValuePassing trueToFalse = ValuePassing::check(trueT, falseParam);
		ValuePassing falseToTrue = ValuePassing::check(falseT, trueParam);
		bool trueToFalseConv = trueToFalse.isPossible();
		bool falseToTrueConv = falseToTrue.isPossible();
		
		if (trueToFalseConv && falseToTrueConv) {
			// ambiguous conversions
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"the true and false parts of conditional "
					"expression %1 have "
					"different types, and implicit conversion "
					"can be done in both directions. Resolve "
					"this by explicitly converting one of the "
					"operands"
				),
				quote(toString())
			).addInfo(
				fTrue->getLocation(),
				UTF8STR("true part has type %1"),
				quote(fTrue->getResult().getType()->toString())
			).addInfo(
				fFalse->getLocation(),
				UTF8STR("false part has type %1"),
				quote(fFalse->getResult().getType()->toString())
			);
			return false;
		}
		else if (!trueToFalseConv && !falseToTrueConv) {
			// totally incompatible types
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"the true and false parts of conditional "
					"expression %1 have "
					"different types, and implicit conversion "
					"cannot be done"
				),
				quote(toString())
			).addInfo(
				fTrue->getLocation(),
				UTF8STR("true part has type %1"),
				quote(fTrue->getResult().getType()->toString())
			).addInfo(
				fFalse->getLocation(),
				UTF8STR("false part has type %1"),
				quote(fFalse->getResult().getType()->toString())
			);
			return false;
		}
		else if (trueToFalseConv) {
			setResolved();
			setResultType(falseT);
			return trueToFalse.checkAccess(
				*fContext, fTrue->getLocation(), getCompileInfo()
			);
		}
		else {
			setResolved();
			setResultType(trueT);
			return falseToTrue.checkAccess(
				*fContext, fFalse->getLocation(), getCompileInfo()
			);
		}
	}
	
	TriBool ConditionalExpression::isClassLValue() const {
		if (!this->isResolvedAndReturnsResult()) {
			return TriBool::uncertain;
		}
		
		return getResult().getType()->isNonNullPointer();
	}
	
	TriBool ConditionalExpression::isPointerLValue() const {
		if (!this->isResolvedAndReturnsResult()) {
			return TriBool::uncertain;
		}
		
		return getResult().getType()->isDoublePointer();
	}
	
	TriBool ConditionalExpression::isCall() const {
		return false;
	}
	
	TriBool ConditionalExpression::hasResult() const {
		return true;
	}
	
	TriBool ConditionalExpression::resultAtCompileTime() const {
		return false; // FIXME
		//return fDecision->resultAtCompileTime()
		//	&& fTrue->resultAtCompileTime() && fFalse->resultAtCompileTime();
	}
}
