/*!
	\file constructor.h
	\brief Contains the Constructor class.
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-08-10
// Last time modified:	2008-12-30

#ifndef DEF_CONSTRUCTOR_H
 #define DEF_CONSTRUCTOR_H
 
 #include "declareditem.h"
 #include "membertypes.h"
 #include "overloadable.h"
 
 #include <vector>

namespace HYPER_Compiler {
	
	class Parameter;
	
	/// Represents a constructor
	class Constructor : public Overloadable, public DeclaredItem {
	public:
		/// Constructor of the constructor :-)
		Constructor(CompileInfo& info);
		
		/// Destructor of the constructor :-)
		virtual ~Constructor();
		
		/// Get the CompileInfo
		CompileInfo& getCompileInfo() const;
		
		/// Get the class the constructor is in
		virtual Class* getClass() const = 0;
		
		/// Gets the header of the constructor as would appear in sourcecode
		virtual HYPER_Common::UTF8String headerToString() const;
		
		HYPER_Common::UTF8String getDescription() const;
		
		Class const* getOwnerClass() const;
		ClassMemberAccess getAccessInOwner() const;
		
		/// Gets the class member access level of this constructor inside its parent
		virtual ClassMemberAccess getAccess() const = 0;
		
		/// Checks if the constructor is an 'implicit' one
		virtual bool isImplicit() const = 0;
		
		/// Makes this constructor implicit
		virtual void makeImplicit() = 0;
		
		/// Get the number of parameters
		virtual unsigned getParameterCount() const = 0;
		
		/// Constructor has no parameters?
		bool hasNoParameters() const {
			return getParameterCount() == 0;
		}
		
		/// Gets the constructor's parameter list
		virtual const std::vector<Parameter const*>* getParameterList() const = 0;
		
		/// Get the location of the constructor in the source code
		virtual const Location* getLocation() const = 0;
		
		HYPER_Common::UTF8String getOverloadHash() const;
		
		/// Mark this constructor as a duplicate of another one
		virtual void markAsDuplicate() = 0;
		
		/// Check if this constructor is marked as a duplicate
		virtual bool isDuplicate() const = 0;
		
		virtual void putInSymbolTable(Scope& parent) {}
		
	private:
		CompileInfo& fCompileInfo;
	};
}
#endif
