/*!
	\file conversions.cxx
	\brief Implements the Conversions class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-05-28
// Last time modified:	2009-08-23

#include "conversions.h"

#include "base/asserts.h"
#include "base/debug.h" // used by ValuePassing::dump()

#include "accesschecker.h"
#include "arrayclass.h"
#include "arraytype.h"
#include "class.h"
#include "classrepository.h"
#include "constructor.h"
#include "declareditem.h"
#include "overloadedconstructor.h"
#include "parameter.h"
#include "type.h"

#include <boost/assign/list_of.hpp>

namespace HYPER_Compiler {
	
	HYPER_Common::UTF8String Conversions::nameOf(Possibility p) {
		switch (p) {
		case kImpossible:
			return UTF8STR("kImpossible");
		case kNotNeeded:
			return UTF8STR("kNotNeeded");
		case kPossibleByLevel1:
			return UTF8STR("kPossibleByLevel1");
		case kPossibleByLevel2:
			return UTF8STR("kPossibleByLevel2");
		default:
			hyp_assert_unreachable();
		}
	}
	
	HYPER_Common::UTF8String Conversions::toString(Kind c) {
		switch (c) {
		case kNoConversion:
			return UTF8STR("kNoConversion");
		case kDerivedToBaseConversion:
			return UTF8STR("kDerivedToBaseConversion");
		case kUserConversion:
			return UTF8STR("kUserConversion");
		default:
			hyp_assert_unreachable();
		}
	}
	
	Conversions::Kind Conversions::toConversion(Possibility p) {
		switch (p) {
		case kImpossible:
			// cannot get a conversion type for an impossible conversion
			hyp_assert_unreachable();
		case kNotNeeded:
			return kNoConversion;
		case kPossibleByLevel1:
			return kDerivedToBaseConversion;
		case kPossibleByLevel2:
			return kUserConversion;
		default:
			hyp_assert_unreachable();
		}
	}
	
	int Conversions::compare(Kind c1, Kind c2) {
		if (c1 < c2) {
			return -1;  // c1 is better
		}
		else if (c1 > c2) {
			return 1;  // c2 is better
		}
		else {
			return 0;  // equal
		}
	}
	
	bool Conversions::conversionOk(Kind c, AllowedLevel a) {
		switch (c) {
		case kNoConversion:
			return true;
		case kDerivedToBaseConversion:
			switch (a) {
			case kNoneAllowed:
				return false;
			case kLevel1Allowed:
				return true;
			case kLevel2Allowed:
				return true;
			default:
				hyp_assert_unreachable();
			}
		case kUserConversion:
			switch (a) {
			case kNoneAllowed:
				return false;
			case kLevel1Allowed:
				return false;
			case kLevel2Allowed:
				return true;
			default:
				hyp_assert_unreachable();
			}
		default:
			hyp_assert_unreachable();
		}
	}
	
	HYPER_Common::UTF8String Conversions::nameOf(Referencing r) {
		switch (r) {
		case kNoReferencing:
			return UTF8STR("kNoReferencing");
		case kReference:
			return UTF8STR("kReference");
		case kDereference:
			return UTF8STR("kDereference");
		case kDoubleDereference:
			return UTF8STR("kDoubleDereference");
		}
		return UTF8STR("(invalid/corrupted value)");
	}
	
	
	// *** ValuePassing ***
	
	ValuePassing::ValuePassing(Type const* from, SimpleParameter const& param,
		Conversions::Possibility possible,
		Conversions::Referencing preRef, Constructor const* constr,
		Conversions::Referencing postRef)
	 : fFrom(from), fTo(param.getType()), fPassing(param.getPassing()),
		fPossible(possible), fPreRef(preRef), fConstr(constr), fPostRef(postRef)
	{
		//
	}
	
	void ValuePassing::dump() const {
		debug.line(UTF8STR("ValuePassing:"));
		IndentDebug ind;
		debug.print(UTF8STR("from type: "));
		fFrom->dump();
		debug.print(UTF8STR("to type  : "));
		fTo->dump();
		debug.line(UTF8STR("passing  : ") + SimpleParameter::nameOf(fPassing));
		debug.line(UTF8STR("pre-ref  : ") + Conversions::nameOf(fPreRef));
		if (fConstr != 0) {
			debug.line(UTF8STR("constr   : ") + fConstr->headerToString());
		}
		else {
			debug.line(UTF8STR("constr   : (none)"));
		}
		debug.line(UTF8STR("post-ref : ") + Conversions::nameOf(fPostRef));
		debug.line(UTF8STR("possible : ") + Conversions::nameOf(fPossible));
		debug.line();
	}
	
	Conversions::Kind ValuePassing::conversionLevelNeeded() const {
		hyp_assert(isPossible());
		return Conversions::toConversion(fPossible);
	}
	
	int ValuePassing::compareTo(ValuePassing const& vp) const {
		if (vp.fPossible == fPossible) {
			return 0;
		}
		else if (vp.fPossible == Conversions::kImpossible) {
			return -1;
		}
		else if (fPossible == Conversions::kImpossible) {
			return 1;
		}
		else {
			return Conversions::compare(
				conversionLevelNeeded(), vp.conversionLevelNeeded()
			);
		}
	}
	
	bool
	ValuePassing::checkAccess(ClassContext const& context, Location const* loc,
		CompileInfo& info) const
	{
		if (fConstr != 0) {
			// Check access to constructor
			return AccessChecker::check(context, *fConstr, loc, info);
		}
		else {
			return true; // all OK
		}
	}
	
	ValuePassing
	ValuePassing::createUnsuitable(Type const* from,
		SimpleParameter const& param)
	{
		return ValuePassing(
			from, param, Conversions::kImpossible,
			Conversions::kNoReferencing, 0, Conversions::kNoReferencing
		);
	}
	
	ValuePassing
	ValuePassing::createRefOp(Type const* from, SimpleParameter const& param,
		Conversions::Referencing ref)
	{
		return ValuePassing(
			from, param, Conversions::kNotNeeded, ref, 0,
			Conversions::kNoReferencing
		);
	}
	
	ValuePassing
	ValuePassing::createLevel1(Type const* from, SimpleParameter const& param,
		Conversions::Referencing ref)
	{
		return ValuePassing(
			from, param, Conversions::kPossibleByLevel1, ref, 0,
			Conversions::kNoReferencing
		);
	}
	
	ValuePassing
	ValuePassing::check(Type const* value, SimpleParameter const& param,
		Conversions::AllowedLevel level)
	{
		SimpleParameter::Passing passing = param.getPassing();
		
		if (value->isNull() && passing == SimpleParameter::kInOut) {
			// "null" cannot be passed to an "inout" parameter
			return createUnsuitable(value, param);
		}
		
		Type const* to = param.getType();
		bool toPointer = to->isNonNullPointer(); // to any non-pointer type (not null)?
		bool toDoublePointer = to->isDoublePointer();
		bool toSinglePointer = toPointer && !toDoublePointer;
		
		//~ debug.line(UTF8STR("ValuePassing::check called"));
		//~ IndentDebug id;
		//~ debug.line(UTF8STR("from:    ") + value->toString());
		//~ debug.line(UTF8STR("to:      ") + to->toString());
		//~ debug.line(UTF8STR("passing: ") + SimpleParameter::nameOf(passing));
		
		
		// -- (1) -- Handle the cases where null is passed
		
		if (value->isNull()) {
			if (toSinglePointer) {
				// no conversion takes place
				return createRefOp(value, param, Conversions::kNoReferencing);
			}
			else if (toPointer) { // double pointer (or 'reference' in the future) ...
				// "null" is not allowed
				//~ debug.line(UTF8STR("incompatible"));
				return createUnsuitable(value, param);
			}
			else {
				// cannot pass null to non-pointer type
				//~ debug.line(UTF8STR("incompatible"));
				return createUnsuitable(value, param);
			}
		}
		
		// -- (2) -- Check for no conversion or trivial conversion
		
		ValuePassing trivialVP = checkTrivialOrNoConversion(value, param);
		if (trivialVP.isPossible()) {
			return trivialVP;
		}
		
		// -- (3) -- We need at least a level 1 conversion
		
		if (level == Conversions::kNoneAllowed
				|| passing == SimpleParameter::kInOut)
		{
			// bail out if no conversions are allowed or if it's an inout parameter
			//~ debug.line(UTF8STR("incompatible"));
			return createUnsuitable(value, param);
		}
		
		Type const* fromClassT = value->getDereferenced();
		Type const* toClassT = to->getDereferenced();
		Class const* toClass = toClassT->getClass();
		
		// conversion to unresolved class type or to static class?
		if (toClass == 0 || toClass->getASS() == kStatic) {
			//~ debug.line(UTF8STR("incompatible"));
			return createUnsuitable(value, param);
		}
		
		bool fromArray = fromClassT->isArray();
		bool toArray = toClassT->isArray();
		
		if (fromArray && toArray) {
			// TODO: when cast from derived to base class is supported, then
			// this check can be done for all types and not just arrays
			ValuePassing vp = checkLevel1Conversion(value, param);
			if (vp.isPossible()) {
				return vp; // level 1 conversion possible, so use that
			}
		}
		
		// -- (4) -- We need a level 2 conversion
		
		if (level == Conversions::kLevel1Allowed) {
			// bail out if no level 2 conversions are allowed
			//~ debug.line(UTF8STR("incompatible"));
			return createUnsuitable(value, param);
		}
		
		if (to->isNonNullPointer()) {
			// No conversion available, because implicit constructors are not
			// called when a pointer is needed.
			
			//~ debug.line(UTF8STR("incompatible"));
			return createUnsuitable(value, param);
		}
		
		return checkLevel2Conversion(value, param);
	}
	
	ValuePassing
	ValuePassing::checkTrivialOrNoConversion(Type const* value,
		SimpleParameter const& param)
	{
		// 1) (const) X      to     (const) X
		// 2) (const) * (const) X     to     (const) X
		// 3) X     to     (const) * (const) X
		// 4) (const) * X     to     (const) * (const) X
		// 5) const X     to     (const) * const X
		// 6) * (const) X    to    ** const X
		// 7) *(*) X    to    ** X
		//
		// Possible type changes here are adding const, removing const when
		// the const value will be copied anyway, and referencing and
		// dereferencing.
		//
		// TODO: check const correctness in the rules above
		//
		
		Type const* fromDerefT = value->getDereferenced();
		Type const* toDerefT = param.getType()->getDereferenced();
		
		if (!fromDerefT->nonConstEquals(toDerefT)) {
			return createUnsuitable(value, param);
		}
		
		// If we get here, we only need a simple pointer 'conversion' or even
		// nothing at all.
		
		bool fromPointer = value->isNonNullPointer();
		bool fromDoublePointer = value->isDoublePointer();
		Type const* to = param.getType();
		bool toPointer = to->isNonNullPointer(); // to any non-pointer type (not null)?
		bool toDoublePointer = to->isDoublePointer();
		
		bool ok;
		bool noLoseConst = !fromDerefT->isConst() || toDerefT->isConst();
		Conversions::Referencing ref;
		
		SimpleParameter::Passing passing = param.getPassing();
		
		if (toDoublePointer) {
			// to double pointer (not for "inout" parameters)
			
			ok =
				fromPointer // 'from' must be single or double pointer
				&& (value->isDoublePointer() || !value->isConst()) // 'from' cannot be const
				&& noLoseConst; // value stays const if it is const
			
			//~ debug.line(UTF8STR("ok = ") + ok);
			//~ debug.line(UTF8STR("fromPointer = ") + fromPointer);
			//~ debug.line(UTF8STR("!from->isConst() = ") + !from->isConst());
			//~ debug.line(UTF8STR("noLoseConst = ") + noLoseConst);
			
			if (fromDoublePointer) {
				ref = Conversions::kNoReferencing;
			}
			else { // single pointer
				ref = Conversions::kReference;
			}
		}
		else if (passing == SimpleParameter::kInOut) {
			// "inout" parameter
			
			ok =
				(fromPointer || !toPointer) // cannot reference twice (inout)
				&& noLoseConst; // value stays const if it is const
			
			if (toPointer) { // to single pointer (double is not allowed)
				// this requires:  fromPointer == true
				if (fromDoublePointer) {
					ref = Conversions::kNoReferencing;
				}
				else {
					ref = Conversions::kReference;
				}
			}
			else { // to a class type
				bool passByValueClass =
					to->getDereferenced()->canBeInvisiblyPassedByValue();
				
				if (fromDoublePointer) {
					ref = Conversions::kDereference;
				}
				else if (!fromPointer && passByValueClass) {
					ref = Conversions::kReference;
				}
				else {
					ref = Conversions::kNoReferencing;
				}
			}
		}
		else if (!to->canBeInvisiblyPassedByValue()
			&& passing == SimpleParameter::kCopyValue)
		{
			// need to copy a class instance, check copy constructor
			
			//~ debug.line(UTF8STR("copy class instance"));
			
			Constructor const* copyConstr =
				to->getClass()->getCopyConstructor();
			
			if (copyConstr == 0) { // no copy constructor
				return createUnsuitable(value, param);
			}
			
			// use the copy constructor as some sort of 'conversion'
			// (see level 2)
			
			// access checking for the constructor is done afterwards (it is not
			// a part of value pass checking, and it shouldn't be)
			return ValuePassing(
				value, param, Conversions::kNotNeeded,
				Conversions::kNoReferencing, copyConstr,
				Conversions::kNoReferencing
			);
		}
		else { // to single pointer or class type
			ok =
				!toPointer // value is just copied (const doesn't matter)
				|| noLoseConst; // value stays const if it is const
			
			//~ debug.line(UTF8STR("toPointer = ") + toPointer);
			
			bool passByValueClass =
				to->getDereferenced()->canBeInvisiblyPassedByValue();
			
			// assume we need a raw single pointer (pointer to class struct)
			if (fromDoublePointer) {
				ref = Conversions::kDereference;
			}
			else if (!fromPointer && passByValueClass) {
				ref = Conversions::kReference;
			}
			else {
				ref = Conversions::kNoReferencing;
			}
			
			// adjust if we convert to a pass-by-value class (like "int")
			if (!toPointer && passByValueClass) {
				// - kReference becomes kNoReferencing
				// - kNoReferencing becomes kDereference
				// - kDereference becomes kDoubleDereference
				switch (ref) {
				case Conversions::kDoubleDereference:
					hyp_assert_unreachable();
				case Conversions::kDereference:
					ref = Conversions::kDoubleDereference;
					break;
				case Conversions::kNoReferencing:
					ref = Conversions::kDereference;
					break;
				case Conversions::kReference:
					ref = Conversions::kNoReferencing;
					break;
				}
			}
		}
		
		if (ok) {
			return createRefOp(value, param, ref);
		}
		else {
			return createUnsuitable(value, param);
		}
	}
	
	ValuePassing
	ValuePassing::checkLevel1Conversion(Type const* from,
		SimpleParameter const& param)
	{
		if (param.getType()->isDoublePointer()) {
			// level 1 conversion not available for creating double pointers
			return createUnsuitable(from, param);
		}
		
		Type const* fromT = from->getDereferenced();
		Type const* toT = param.getType()->getDereferenced();
		
		while (true) {
			if (fromT->isArray() && toT->isArray()) {
				//debug << " getting at arrays" << newl;
				
				const std::vector<Dimension>& fromDims =
					fromT->getArrayDimensions();
				const std::vector<Dimension>& toDims =
					toT->getArrayDimensions();
				
				if (fromDims.size() != toDims.size()) {
					return createUnsuitable(from, param);
				}
				
				for (unsigned int i = 0; i < fromDims.size(); ++i) {
					if (toDims[i].isOpen()) { // to open dimension?
						continue; //OK; skip check
					}
					
					if (fromDims[i].isOpen() || fromDims[i].isUnknown()) {
						// e.g. :
						//   (1) "[] int" to "[5] int"
						//   (2) "[x] int" to "[5] int"
						// cannot go from non-constant to non-open dimension
						return createUnsuitable(from, param);
					}
					
					// now we need to check that dimension sizes are equal
					if (fromDims[i].getSize() == toDims[i].getSize()) {
						// OK, sizes match
					}
					else {
						return createUnsuitable(from, param);
					}
				}
				
				fromT = fromT->getBaseType();
				toT = toT->getBaseType();
			}
			else if (fromT->isNonNullPointer() && toT->isNonNullPointer()) {
				fromT = fromT->getBaseType();
				toT = toT->getBaseType();
			}
			else if (fromT->isClass() && toT->isClass()) {
				if (fromT->getClass() != toT->getClass()) {
					return createUnsuitable(from, param);
				}
				
				if (fromT->isConst() && !toT->isConst()) {
					return createUnsuitable(from, param);
				}
				else {
					// OK, end of type, exit loop
					break;
				}
			}
			else {
				return createUnsuitable(from, param);
			}
		} // end while
		
		// if we get here, then we can do a level 1 conversion (pointer cast)
		
		// Only a dereference is needed sometimes (in case we get a double
		// pointer), otherwise we always already have a pointer (because an
		// expression of class type is referred to by its "this" pointer).
		Conversions::Referencing ref;
		if (from->isDoublePointer()) {
			ref = Conversions::kDereference;
		}
		else {
			ref = Conversions::kNoReferencing;
		}
		
		return createLevel1(from, param, ref);
	}
	
	ValuePassing
	ValuePassing::checkLevel2Conversion(Type const* from,
		SimpleParameter const& param)
	{
		// "to" is a class type, get its class
		Class const* toClass = param.getType()->getClass();
		
		// get constructors from toClass
		OverloadedConstructor const* ovc = toClass->getConstructors();
		const std::vector<Constructor*>& ovl = ovc->getOverloadList();
		
		const Constructor* constr = 0;
		
		// Don't use overload resolution here, just look for an exact match.
		// Otherwise we would run in circles.
		
		for (unsigned int i = 0; i < ovl.size(); ++i) {
			Constructor* c = ovl[i];
			
			if (c->getParameterList() == 0) {
				continue; // parameters of constructor not resolved
			}
			
			// only look at constructors with one parameter
			if (c->getParameterList()->size() != 1) {
				continue;
			}
			
			SimpleParameter const* constrParam = (*c->getParameterList())[0];
			
			// Do ValuePassing check, but don't allow any conversions.
			ValuePassing vp =
				check(from, *constrParam, Conversions::kNoneAllowed);
			
			if (vp.isPossible()) {
				// We found our conversion constructor; only still need to check
				// if it can be called implicitly.
				constr = c;
				break;
			}
		}
		
		// found a suitable constructor?
		if (constr != 0 && constr->isImplicit()) {
			/*
				PreRef is not needed, because it is handled by the nested
				ValuePassing for calling the constructor. PostRef is only needed
				when "to" is a class type that is always passed by value.
			*/
			
			Conversions::Referencing postRef;
			if (toClass->canBeInvisiblyPassedByValue()) {
				postRef = Conversions::kDereference;
			}
			else {
				postRef = Conversions::kNoReferencing;
			}
			
			return ValuePassing(
				from, param, Conversions::kPossibleByLevel2,
				Conversions::kNoReferencing, constr, postRef
			);
		}
		
		// if we get here, then there is no conversion available
		return createUnsuitable(from, param);
	}
	
}
