/*!
	\file conversions.h
	\brief Contains the Conversions class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-05-28
// Last time modified:	2009-01-04

#ifndef DEF_CONVERSIONS_H
 #define DEF_CONVERSIONS_H
 
 #include "important_fwdecl.h"
 
 #include "common/cxx_macros.h"
 #include "common/utf8string.h"
 
 #include "simpleparameter.h" //  for enum SimpleParameter::Passing
 
namespace HYPER_Compiler {
	
	class Constructor;
	
	/// Collection of functions for doing advanced things with types
	class Conversions {
	public:
		/// Specifies a kind of conversion that needs to be done.
		enum Kind {
			kNoConversion = 0,
			kDerivedToBaseConversion,	// in level 1 & 2
			kUserConversion		// only in level 2
		};
		
		static HYPER_Common::UTF8String toString(Kind c);
		static int compare(Kind c1, Kind c2);
		
		/// Specifies the maximum level of conversion that can be done.
		enum AllowedLevel {
			kNoneAllowed = 0,
			kLevel1Allowed,
			kLevel2Allowed
		};
		
		static bool conversionOk(Kind c, AllowedLevel a);
		
		/// Specifies what (de)referencing needs to be done when passing an
		/// expression to a parameter of some type.
		enum Referencing {
			kDoubleDereference,
			kDereference,
			kNoReferencing,
			kReference
		};
		
		// for debug purposes
		static HYPER_Common::UTF8String nameOf(Referencing r);
		
	private:
		friend class HYPER_Compiler::ValuePassing;
		
		Conversions() {
			// inaccessible constructor  (it's a static class)
		}
		
		enum Possibility {
			kImpossible = 0,
			kNotNeeded, ///< conversion not needed
			kPossibleByLevel1,
			kPossibleByLevel2
		};
		
		static HYPER_Common::UTF8String nameOf(Possibility p);
		static Kind toConversion(Possibility p);
	};
	
	/// Contains the data needed for passing a value to some expected type
	class ValuePassing {
	public:
		
		/// Check if it is possible to pass a value to some parameter, and
		/// figure out how to do it if possible. The amount of conversion
		/// allowed can be specified; otherwise all implicit conversions are
		/// allowed to happen. The checks assumes that pointers are passed
		/// directly, and that class types are passed as the value of their
		/// "this" pointer.
		static ValuePassing
		check(Type const* value, SimpleParameter const& param,
			Conversions::AllowedLevel level = Conversions::kLevel2Allowed);
		
		bool isPossible() const {
			return fPossible != Conversions::kImpossible;
		}
		
		Conversions::Kind conversionLevelNeeded() const;
		
		int compareTo(ValuePassing const& vp) const;
		
		bool ATTR_WARN_UNUSED_RESULT
		checkAccess(ClassContext const& context, Location const* loc,
			CompileInfo& info) const;
		
		/// Print the ValuePassing to debug output
		void dump() const;
		
	private:
		ValuePassing(Type const* from, SimpleParameter const& param,
			Conversions::Possibility possible,
			Conversions::Referencing preRef, Constructor const* constr,
			Conversions::Referencing postRef);
		
		static ValuePassing createUnsuitable(Type const* from,
			SimpleParameter const& param);
		
		static ValuePassing createRefOp(Type const* from,
			SimpleParameter const& param, Conversions::Referencing ref);
		
		static ValuePassing createLevel1(Type const* from,
			SimpleParameter const& param, Conversions::Referencing ref);
		
		static ValuePassing checkTrivialOrNoConversion(Type const* value,
			SimpleParameter const& param);
		
		static ValuePassing checkLevel1Conversion(Type const* from,
			SimpleParameter const& param);
		
		static ValuePassing checkLevel2Conversion(Type const* from,
			SimpleParameter const& param);
		
		Type const* fFrom;
		Type const* fTo;
		SimpleParameter::Passing fPassing;
		Conversions::Possibility fPossible;
		Conversions::Referencing fPreRef;
		Constructor const* fConstr;
		Conversions::Referencing fPostRef;
	};
}
#endif
