/*!
	\file crementstatement.cxx
	\brief Implementation of the CrementStatement class.
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-11-02
// Last time modified:	2009-01-25

#include "crementstatement.h"

#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

#include "class.h"
#include "compileinfo.h"
#include "expression.h"
#include "operatorcalls.h"
#include "type.h"

namespace HYPER_Compiler {
	
	CrementStatement::CrementStatement(Expression* lValue,
		Operators::CrementOperator op, const Location* loc, CompileInfo& info)
	 : Statement(info, loc), fLValue(lValue), fOpSym(op)
	{
		// nothing here
	}
	
	CrementStatement::~CrementStatement() {
		// nothing here
	}
	
	HYPER_Common::UTF8String CrementStatement::toString() const {
		return fLValue->toString() + UTF8STR(" ") + Operators::toString(fOpSym);
	}
	
	HYPER_Common::UTF8String CrementStatement::getSyntaxName() const {
		return UTF8STR("CrementStatement");
	}
	
	void CrementStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const
	{
		ind.line(getSyntaxName());
		
		HYPER_Common::Indent out(ind);
		
		ind.line(UTF8STR("LValue: ") + fLValue->toString());
		ind.line(UTF8STR("Operator: ") + Operators::toString(fOpSym));
	}
	
	Expression* CrementStatement::getLValue() const {
		return fLValue;
	}
	
	HYPER_Common::UTF8String CrementStatement::getResolveDescription() const {
		return UTF8STR("%1 statement").compose(
			(fOpSym == Operators::kIncrement)
				? UTF8STR("increment")
				: UTF8STR("decrement")
		);
	}
	
	void CrementStatement::putInSymbolTable(ScopeLocation const& loc) {
		Statement::putInSymbolTable(loc);
		fLValue->putInSymbolTable(new LookupLocation(loc));
	}
	
	bool CrementStatement::doFullResolve(Trace const& trace) {
		bool resolveOK = fLValue->resolveFull(trace);
		
		if (!fLValue->getResult().isResolved()) {
			// cannot continue
			return false;
		}
		
		OperatorCalls::CrementCallResult result =
			OperatorCalls::resolveCrementOperator(*fLValue, fOpSym);
		
		if (result.fgErrorLValue) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"invalid L-value `%1' in %2 statement `%3'"
				),
				fLValue->toString(),
				(fOpSym == Operators::kIncrement)
					? UTF8STR("increment")
					: UTF8STR("decrement"),
				toString()
			);
			resolveOK = false;
		}
		
		if (result.fgErrorConst) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"cannot %1 the (dereferenced) L-value `%2' because it "
					"has const type `%3'"
				),
				(fOpSym == Operators::kIncrement)
					? UTF8STR("increment")
					: UTF8STR("decrement"),
				fLValue->toString(),
				fLValue->getResult().getType()->toString()
			);
			resolveOK = false;
		}
		
		if (result.fgErrorNoOperator) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no %1 operator in class `%2', which is needed for "
					"statement `%3'"
				),
				(fOpSym == Operators::kIncrement)
					? UTF8STR("increment")
					: UTF8STR("decrement"),
				fLValue->getResult().getType()->getDereferenced()->toString(),
				toString()
			);
			resolveOK = false;
		}
		
		const CrementOperator* op = result.fgOperator;
		if (op != 0) {
			// OK
			
		}
		
		return resolveOK;
	}
	
	SimpleProbability CrementStatement::getReturnProbability() const {
		return SimpleProbability::never;
	}
}
