/*!
	\file declaredparameter.cxx
	\brief Implements the DeclaredParameter class.
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-06-01
// Last time modified:	2008-11-09

#include "declaredparameter.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "compileinfo.h"
#include "declaredtype.h"
#include "type.h"
#include "scopes.h" // for LookupLocation

namespace HYPER_Compiler {
	
	DeclaredParameter::DeclaredParameter(CompileInfo& info,
		HYPER_Common::UTF8String name, DeclaredType* type, const Location* loc,
		SimpleParameter::Passing passing)
	 : fCompileInfo(info), fName(name), fType(type), fPassing(passing),
		fLoc(loc), fResolvedParam(0)
	{
		// disallow runtime sizes
		fType->setRestrictions(false, true);
	}
	
	HYPER_Common::UTF8String DeclaredParameter::getName() const {
		return fName;
	}
	
	const Location* DeclaredParameter::getLocation() const {
		return fLoc;
	}
	
	Parameter const* DeclaredParameter::getResolvedParameter() const {
		return fResolvedParam;
	}
	
	HYPER_Common::UTF8String DeclaredParameter::toString() const {
		HYPER_Common::UTF8StringBuilder sb;
		
		// add passing prefix if needed
		if (fPassing == SimpleParameter::kInOut) {
			sb += UTF8STR("inout ");
		}
		
		sb += fName;
		sb += UTF8STR(" : ");
		sb += fType->toString();
		
		return sb.toString();
	}
	
	HYPER_Common::UTF8String DeclaredParameter::getDescription() const {
		return UTF8STR("parameter %1").compose(quote(getName()));
	}
	
	Class const* DeclaredParameter::getOwnerClass() const {
		return 0;
	}
	
	ClassMemberAccess DeclaredParameter::getAccessInOwner() const {
		hyp_assert_unreachable();
	}
	
	HYPER_Common::UTF8String DeclaredParameter::getDeclarationName() const {
		return getName();
	}
	
	void DeclaredParameter::putInSymbolTable(ScopeLocation const& loc) {
		fType->putInSymbolTable(new LookupLocation(loc));
	}
	
	CompileInfo& DeclaredParameter::getCompileInfo() const {
		return fCompileInfo;
	}
	
	HYPER_Common::UTF8String DeclaredParameter::getResolveDescription() const {
		return UTF8STR("parameter %1").compose(quote(fName));
	}
	
	const Location* DeclaredParameter::getResolveLocation() const {
		return fLoc;
	}
	
	bool DeclaredParameter::doInterfaceResolve(Trace const& trace) {
		if (fPassing == SimpleParameter::kInOut && fType->isConst()) {
			getCompileInfo().diag().error(
				fType->getLocation(),
				UTF8STR("inout parameter cannot have a %1 type"),
				QUOTED_UTF8STR("const")
			);
		}
		
		if  (!fType->resolveInterface(trace)) {
			// type could not be resolved
			return false;
		}
		
		Type const* t = fType->getResolvedType();
		fResolvedParam = new Parameter(t, fPassing, fName, fLoc);
		
		return true;
	}
	
	bool DeclaredParameter::doFullResolve(Trace const& trace) {
		// TODO: get rid of "fullResolve" for DeclaredType instances (if possible)
		return fType->resolveFull(trace);
	}
}
