/*!
	\file declaredtype.h
	\brief Contains the DeclaredType class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-01
// Last time modified:	2008-10-19

#ifndef DEF_DECLAREDTYPE_H
 #define DEF_DECLAREDTYPE_H
 
 #include "important_fwdecl.h"
 #include "resolveable.h"
 #include "tribool.h"
 
 #include <vector>

namespace HYPER_Compiler {
	
	class InitializationHandler;
	
	/// Represents a type, as declared in a source file
	class DeclaredType : public Resolveable {
	public:
		
		/// Check if the type is a double pointer
		virtual bool isDoublePointer() const = 0;
		
		/// Check if the type is a pointer or double pointer
		virtual bool isPointer() const = 0;
		
		/// Check if the type is an array
		virtual bool isArray() const = 0;
		
		/// Checks if the type is a class type, either built-in or user defined.
		/// This includes array types.
		virtual bool isClass() const = 0;
		
		/// Is this part of the type const?
		virtual bool isConst() const = 0;
		
		/// Make the type const
		virtual void makeConst() = 0;
		
		/// Get type syntax as a string
		virtual HYPER_Common::UTF8String toString() const = 0;
		
		/// Print the type to debug output
		void dump() const;
		
		virtual ~DeclaredType();
		
		/// Let the type know where it is in the symbol table, i.e. what
		/// its parent scope is.
		virtual void putInSymbolTable(LookupLocation const* loc);
		
		CompileInfo& getCompileInfo() const { return fCompileInfo; }
		
		void setRestrictions(bool directAllocate, bool compileTimeSize);
		
		/// Get the location of the type (where it was referred to)
		const Location* getLocation() const;
		
		/// Get the resolved type (returns 0 if not successfully resolved)
		Type const* getResolvedType() const;
		
	protected:
		void setResolvedType(Type const* t);
		
		explicit DeclaredType(CompileInfo& info, Location const* loc);
		DeclaredType(DeclaredType const& t); // copy constructor
		
		HYPER_Common::UTF8String getResolveDescription() const {
			return UTF8STR("type %1").compose(quote(toString()));
		}
		
		const Location* getResolveLocation() const {
			return getLocation();
		}
		
		bool doFullResolve(Trace const& trace) { return true; }
		
		LookupLocation const* getLookupLocation() const;
		bool mustBeDirectAllocatable() const;
		bool mustHaveCompileTimeSize() const;
		
		// ban the assignment operator
		void operator=(const DeclaredType& t);
		
	private:
		CompileInfo& fCompileInfo;
		const Location* fLoc;
		LookupLocation const* fLookupLoc;
		bool fDirectAllocate;
		bool fCompileTimeSize;
		Type const* fResolvedType;
	};
}
#endif
