/*!
	\file doubleclass.cxx
	\author Kevin André
	\brief Implements the Double class
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-04-19
// Last time modified:	2008-12-26

#include "doubleclass.h"

#include "base/location.h"

#include "parameter.h"

namespace HYPER_Compiler {
	
	class Double::DoubleConstructor : public BuiltInConstructor {
	public:
		// default & copy constructor
		DoubleConstructor(Double& dc, bool defaultConstructor);
		
		// conversion constructors
		DoubleConstructor(Double& dc, BuiltInTypes::Primitive type);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParam;
	};
	
	class Double::PositiveOperator : public UnaryOperator {
	public:
		PositiveOperator(Double& dc);
		
		ConstantValue const*
			evaluateCompileTime(ConstantValue const* operand);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::NegativeOperator : public UnaryOperator {
	public:
		NegativeOperator(Double& dc);
		
		ConstantValue const*
			evaluateCompileTime(ConstantValue const* operand);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::EqualsOperator : public BinaryOperator {
	public:
		EqualsOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::DiffersOperator : public BinaryOperator {
	public:
		DiffersOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::LessOperator : public BinaryOperator {
	public:
		LessOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::LessEqOperator : public BinaryOperator {
	public:
		LessEqOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::GreaterOperator : public BinaryOperator {
	public:
		GreaterOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::GreaterEqOperator : public BinaryOperator {
	public:
		GreaterEqOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::AddOperator : public BinaryOperator {
	public:
		AddOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::SubtractOperator : public BinaryOperator {
	public:
		SubtractOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::MultiplyOperator : public BinaryOperator {
	public:
		MultiplyOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::DivideOperator : public BinaryOperator {
	public:
		DivideOperator(Double& dc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::AssignOp : public AssignmentOperator {
	public:
		AssignOp(Double& dc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::AddAssignOp : public AssignmentOperator {
	public:
		AddAssignOp(Double& dc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::SubtractAssignOp : public AssignmentOperator {
	public:
		SubtractAssignOp(Double& dc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::MultiplyAssignOp : public AssignmentOperator {
	public:
		MultiplyAssignOp(Double& dc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Double::DivideAssignOp : public AssignmentOperator {
	public:
		DivideAssignOp(Double& dc);
		
		// TODO : change return type
		void generateCode();
	};
	
	
	// *** Double ***
	
	Double::Double(CompileInfo& info)
	 : BuiltInClass(info, kSealed)
	{
		// no action
	}
	
	Double::~Double() {
		// no action
	}
	
	HYPER_Common::UTF8String Double::getTypeName() const {
		return UTF8STR("double");
	}
	
	HYPER_Common::UTF8String Double::getDeclarationName() const {
		return UTF8STR("double");
	}
	
	void Double::buildSymbolTable() {
		createConstructors();
		createMembers();
		createOperators();
	}
	
	void Double::createConstructors() {
		registerConstructor(new DoubleConstructor(*this, true), false);
		registerConstructor(new DoubleConstructor(*this, false), true);
		registerConstructor(
			new DoubleConstructor(*this, BuiltInTypes::kByte), false
		);
		registerConstructor(
			new DoubleConstructor(*this, BuiltInTypes::kInt16), false
		);
		registerConstructor(
			new DoubleConstructor(*this, BuiltInTypes::kInt32), false
		);
		registerConstructor(
			new DoubleConstructor(*this, BuiltInTypes::kNat16), false
		);
		registerConstructor(
			new DoubleConstructor(*this, BuiltInTypes::kSingle), false
		);
	}
	
	void Double::createMembers() {
		// none yet
	}
	
	void Double::createOperators() {
		registerOperator(new PositiveOperator(*this));
		registerOperator(new NegativeOperator(*this));
		
		registerOperator(new EqualsOperator(*this));
		registerOperator(new DiffersOperator(*this));
		registerOperator(new LessOperator(*this));
		registerOperator(new LessEqOperator(*this));
		registerOperator(new GreaterOperator(*this));
		registerOperator(new GreaterEqOperator(*this));
		registerOperator(new AddOperator(*this));
		registerOperator(new SubtractOperator(*this));
		registerOperator(new MultiplyOperator(*this));
		registerOperator(new DivideOperator(*this));
		
		registerOperator(new AssignOp(*this));
		registerOperator(new AddAssignOp(*this));
		registerOperator(new SubtractAssignOp(*this));
		registerOperator(new MultiplyAssignOp(*this));
		registerOperator(new DivideAssignOp(*this));
	}
	
	
	// *** DoubleConstructor ***
	
	Double::DoubleConstructor::DoubleConstructor(Double& dc,
		bool defaultConstructor)
	 : BuiltInConstructor(&dc, true)
	{
		if (!defaultConstructor) {
			fParam.push_back(
				dc.createParameter(UTF8STR("b"), BuiltInTypes::kDouble)
			);
		}
	}
	
	Double::DoubleConstructor::DoubleConstructor(Double& dc,
		BuiltInTypes::Primitive type)
	 : BuiltInConstructor(&dc, true)
	{
		fParam.push_back(
			dc.createParameter(UTF8STR("b"), type)
		);
	}
	
	const std::vector<Parameter const*>*
	Double::DoubleConstructor::getParameterList() const {
		return &fParam;
	}
	
	
	// *** PositiveOperator ***
	
	Double::PositiveOperator::PositiveOperator(Double& dc)
	 : UnaryOperator(Operators::kPositive, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::PositiveOperator::evaluateCompileTime(ConstantValue const* operand)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::PositiveOperator::generateCode() {
		// TODO
	}
	
	
	// *** NegativeOperator ***
	
	Double::NegativeOperator::NegativeOperator(Double& dc)
	 : UnaryOperator(Operators::kNegative, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::NegativeOperator::evaluateCompileTime(ConstantValue const* operand)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::NegativeOperator::generateCode() {
		// TODO
	}
	
	
	// *** EqualsOperator ***
	Double::EqualsOperator::EqualsOperator(Double& dc)
	 : BinaryOperator(Operators::kEquals, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), true, BuiltInTypes::createBool(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::EqualsOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::EqualsOperator::generateCode() {
		// TODO
	}
	
	
	// *** DiffersOperator ***
	Double::DiffersOperator::DiffersOperator(Double& dc)
	 : BinaryOperator(Operators::kDiffers, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), true, BuiltInTypes::createBool(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::DiffersOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::DiffersOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessOperator ***
	Double::LessOperator::LessOperator(Double& dc)
	 : BinaryOperator(Operators::kLess, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), false, BuiltInTypes::createBool(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::LessOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::LessOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessEqOperator ***
	Double::LessEqOperator::LessEqOperator(Double& dc)
	 : BinaryOperator(Operators::kLessOrEqual, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), false, BuiltInTypes::createBool(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::LessEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::LessEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterOperator ***
	Double::GreaterOperator::GreaterOperator(Double& dc)
	 : BinaryOperator(Operators::kGreater, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), false, BuiltInTypes::createBool(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::GreaterOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::GreaterOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterEqOperator ***
	Double::GreaterEqOperator::GreaterEqOperator(Double& dc)
	 : BinaryOperator(Operators::kGreaterOrEqual, BuiltInTypes::createDouble(dc.getCompileInfo()),
		BuiltInTypes::createDouble(dc.getCompileInfo()), false, BuiltInTypes::createBool(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::GreaterEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::GreaterEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** AddOperator ***
	Double::AddOperator::AddOperator(Double& dc)
	 : BinaryOperator(Operators::kAdd,
		BuiltInTypes::createDouble(dc.getCompileInfo()), BuiltInTypes::createDouble(dc.getCompileInfo()), true,
		BuiltInTypes::createDouble(dc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::AddOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::AddOperator::generateCode() {
		// TODO
	}
	
	
	// *** SubtractOperator ***
	Double::SubtractOperator::SubtractOperator(Double& dc)
	 : BinaryOperator(Operators::kSubtract,
		BuiltInTypes::createDouble(dc.getCompileInfo()), BuiltInTypes::createDouble(dc.getCompileInfo()), false,
		BuiltInTypes::createDouble(dc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::SubtractOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::SubtractOperator::generateCode() {
		// TODO
	}
	
	
	// *** MultiplyOperator ***
	Double::MultiplyOperator::MultiplyOperator(Double& dc)
	 : BinaryOperator(Operators::kMultiply,
		BuiltInTypes::createDouble(dc.getCompileInfo()), BuiltInTypes::createDouble(dc.getCompileInfo()), true,
		BuiltInTypes::createDouble(dc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::MultiplyOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::MultiplyOperator::generateCode() {
		// TODO
	}
	
	
	// *** DivideOperator ***
	Double::DivideOperator::DivideOperator(Double& dc)
	 : BinaryOperator(Operators::kDivide,
		BuiltInTypes::createDouble(dc.getCompileInfo()), BuiltInTypes::createDouble(dc.getCompileInfo()), false,
		BuiltInTypes::createDouble(dc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Double::DivideOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Double::DivideOperator::generateCode() {
		// TODO
	}
	
	
	// *** AssignOp ***
	Double::AssignOp::AssignOp(Double& dc)
	 : AssignmentOperator(Operators::kSimpleAssign, BuiltInTypes::createDouble(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Double::AssignOp::generateCode() {
		// TODO
	}
	
	
	// *** AddAssignOp ***
	Double::AddAssignOp::AddAssignOp(Double& dc)
	 : AssignmentOperator(Operators::kAddAssign, BuiltInTypes::createDouble(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Double::AddAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** SubtractAssignOp ***
	Double::SubtractAssignOp::SubtractAssignOp(Double& dc)
	 : AssignmentOperator(Operators::kSubtractAssign, BuiltInTypes::createDouble(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Double::SubtractAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** MultiplyAssignOp ***
	Double::MultiplyAssignOp::MultiplyAssignOp(Double& dc)
	 : AssignmentOperator(Operators::kMultiplyAssign, BuiltInTypes::createDouble(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Double::MultiplyAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** DivideAssignOp ***
	Double::DivideAssignOp::DivideAssignOp(Double& dc)
	 : AssignmentOperator(Operators::kDivideAssign, BuiltInTypes::createDouble(dc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Double::DivideAssignOp::generateCode() {
		// TODO
	}
}
