/*!
	\file enum.cxx
	\author Kevin André
	\brief Implements the Enum class
	
	Copyright (C) 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2009-09-05
// Last time modified:	2009-09-06

#include "enum.h"

#include "base/location.h"

#include "compiler/assignmentoperator.h"
#include "compiler/binaryoperator.h"
#include "compiler/builtinconstructor.h"
#include "compiler/builtintypes.h"
#include "compiler/overloadedassignmentoperator.h"
#include "compiler/overloadedbinaryoperator.h"
#include "compiler/parameter.h"
#include "compiler/type.h"

namespace HYPER_Compiler {
	
	class Enum::EnumConstructor : public BuiltInConstructor {
	public:
		
		/// Create default constructor
		EnumConstructor(Enum& e, EnumValue const* defaultValue)
		 : BuiltInConstructor(&e, true), fDefaultValue(defaultValue)
		{
			//
		}
		
		// Create copy constructor
		EnumConstructor(Enum& e)
		 : BuiltInConstructor(&e, true), fDefaultValue(0)
		{
			fParam.push_back(
				new Parameter(
					Type::createClassType(false, &e),
					SimpleParameter::kInByReference,
					UTF8STR("e"),
					Location::getBuiltIn()
				)
			);
		}
		
		const std::vector<Parameter const*>* getParameterList() const {
			return &fParam; // no parameters
		}
		
	private:
		std::vector<Parameter const*> fParam;
		EnumValue const* fDefaultValue;
	};
	
	class Enum::EnumAssignOp : public AssignmentOperator {
	public:
		EnumAssignOp(Enum& e)
		 : AssignmentOperator(
				Operators::kSimpleAssign, Type::createClassType(false, &e),
				Location::getBuiltIn()
			)
		{
			//
		}
		
	};
	
	class Enum::EnumEqualsOp : public BinaryOperator {
	public:
		EnumEqualsOp(Enum& e)
		 : BinaryOperator(
				Operators::kEquals, Type::createClassType(false, &e),
				Type::createClassType(false, &e), true,
				BuiltInTypes::createBool(e.getCompileInfo()),
				Location::getBuiltIn()
			)
		{
			//
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			// TODO
			return 0;
		}
		
	};
	
	class Enum::EnumDiffersOp : public BinaryOperator {
	public:
		EnumDiffersOp(Enum& e)
		 : BinaryOperator(
				Operators::kDiffers, Type::createClassType(false, &e),
				Type::createClassType(false, &e), true,
				BuiltInTypes::createBool(e.getCompileInfo()),
				Location::getBuiltIn()
			)
		{
			//
		}
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp)
		{
			// TODO
			return 0;
		}
		
	};
	
	
	// *** Enum ***
	
	Enum::Enum(CompileInfo& info, ClassMemberAccess access, Class* parent)
	 : Class(info), fParent(parent), fAccess(access)
	{
		//
	}
	
	Enum::~Enum() {
		//
	}
	
	HYPER_Common::UTF8String Enum::getDescription() const {
		return UTF8STR("enum %1").compose(quote(getTypeName()));
	}
	
	ClassMemberAccess Enum::getAccess() const {
		return fAccess;
	}
	
	void Enum::createDefaultConstructor(EnumValue const* defaultValue) {
		registerConstructor(new EnumConstructor(*this, defaultValue), false);
	}
	
	void Enum::createCopyConstructor() {
		registerConstructor(new EnumConstructor(*this), true);
	}
	
	void Enum::createAssignmentOperator() {
		EnumAssignOp* op = new EnumAssignOp(*this);
		
		OverloadedAssignmentOperator* & oao =
			fAssignOps[Operators::kSimpleAssign];
		
		if (oao == 0) {
			oao = new OverloadedAssignmentOperator(op);
		}
		else {
			oao->addOverload(op);
		}
	}
	
	void Enum::createEqualityOperator() {
		EnumEqualsOp* op = new EnumEqualsOp(*this);
		
		OverloadedBinaryOperator* & obp = fBinOps[op->getOperator()];
		if (obp == 0) {
			obp = new OverloadedBinaryOperator(op);
		}
		else {
			obp->addOverload(op);
		}
	}
	
	void Enum::createDiffersOperator() {
		EnumDiffersOp* op = new EnumDiffersOp(*this);
		
		OverloadedBinaryOperator* & obp = fBinOps[op->getOperator()];
		if (obp == 0) {
			obp = new OverloadedBinaryOperator(op);
		}
		else {
			obp->addOverload(op);
		}
	}
	
	Class::BinOpsMap const& Enum::getBinaryOperators() const {
		return fBinOps;
	}
	
	Class::CremOpsMap const& Enum::getCrementOperators() const {
		return fCremOps;
	}
	
	Class::UnOpsMap const& Enum::getUnaryOperators() const {
		return fUnOps;
	}
	
	Class::AssignOpsMap const& Enum::getAssignmentOperators() const {
		return fAssignOps;
	}
	
	OverloadedArrayOperator const* Enum::getArrayOperators() const {
		return fArrayOps;
	}
	
}
