/*!
	\file enum.h
	\author Kevin André
	\brief Contains the Enum class
	
	Copyright (C) 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2009-09-05
// Last time modified:	2009-09-06

#ifndef DEF_ENUM_H
 #define DEF_ENUM_H
 
 #include "class.h"
 
namespace HYPER_Compiler {
	
	class EnumValue;
	
	/// Represents a class implemented by the user
	class Enum : public Class {
	public:
		
		virtual ~Enum();
		
		HYPER_Common::UTF8String getDescription() const;
		
		ClassMemberAccess getAccess() const;
		AbstractSealedStatic getASS() const { return kSealed; }
		
		Class const* getParentClass() const { return fParent; }
		
		BinOpsMap const& getBinaryOperators() const;
		CremOpsMap const& getCrementOperators() const;
		UnOpsMap const& getUnaryOperators() const;
		AssignOpsMap const& getAssignmentOperators() const;
		OverloadedArrayOperator const* getArrayOperators() const;
		
	protected:
		/// Constructor
		Enum(CompileInfo& info, ClassMemberAccess access, Class* parent = 0);
		
		void createDefaultConstructor(EnumValue const* defaultValue);
		void createCopyConstructor();
		void createAssignmentOperator();
		void createEqualityOperator();
		void createDiffersOperator();
		
	private:
		class EnumConstructor;
		class EnumAssignOp;
		class EnumEqualsOp;
		class EnumDiffersOp;
		
		Class* fParent;
		ClassMemberAccess fAccess;
		
		UnOpsMap fUnOps;
		BinOpsMap fBinOps;
		CremOpsMap fCremOps;
		AssignOpsMap fAssignOps;
		OverloadedArrayOperator* fArrayOps;
	};
	
	class EnumValue : public NamedDeclaredItem {
	public:
		EnumValue(Enum* parent, HYPER_Common::UTF8String name,
			Location const* loc)
		 : fName(name), fParent(parent), fLoc(loc)
		{
			//
		}
		
		HYPER_Common::UTF8String getDeclarationName() const {
			return fName;
		}
		
		HYPER_Common::UTF8String getDescription() const {
			return UTF8STR("enum value %1").compose(quote(fName));
		}
		
		Enum const* getOwnerClass() const {
			return fParent;
		}
		
		ClassMemberAccess getAccessInOwner() const {
			return ClassMemberAccess::createPublic();
		}
		
		const Location* getLocation() const {
			return fLoc;
		}
		
	private:
		HYPER_Common::UTF8String fName;
		Enum* fParent;
		Location const* fLoc;
	};
	
}
#endif
