/*!
	\file expression.cxx
	\brief Implements the Expression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2004-07-04
// Date modified:	2008-10-15

#include "expression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "compileinfo.h"
#include "constantvalue.h"
#include "declareditem.h"
#include "instanceprovider.h"
#include "scopes.h"

namespace HYPER_Compiler {
	
	// *** ExpressionResult ***
	
	bool ExpressionResult::isResolved() const {
		return fResolved;
	}
	
	Type const* ExpressionResult::getType() const {
		hyp_assert(isResolved());
		return fType;
	}
	
	DeclaredItem* ExpressionResult::getDecl() const {
		hyp_assert(isResolved());
		return fDecl;
	}
	
	bool ExpressionResult::knownValue() const {
		return fValue != 0;
	}
	
	ConstantValue const* ExpressionResult::getValue() const {
		return fValue;
	}
	
	TriBool ExpressionResult::getBoolValue() const {
		// TODO: assert that the expression type is bool
		if (knownValue()) {
			return fValue->getBool();
		}
		else {
			return TriBool::uncertain;
		}
	}
	
	ExpressionResult::ExpressionResult(Expression& expr)
	 : fExpr(expr), fResolved(false), fValue(0), fType(0), fDecl(0)
	{
		// no action
	}
	
	// *** Expression ***
	
	Expression::Expression(CompileInfo& info, const Location* loc)
	 : fCompileInfo(info), fResult(*this), fOwner(0), fLoc(loc), fLookupLoc(0),
		fIsCalled(false)
	{
		// no action
	}
	
	Expression::~Expression() {
		// no action
	}
	
	CompileInfo& Expression::getCompileInfo() const {
		return fCompileInfo;
	}
	
	ExpressionResult const& Expression::getResult() const {
		hyp_assert(!hasResult().isFalse());
		return fResult;
	}
	
	const Location* Expression::getLocation() const {
		return fLoc;
	}
	
	void Expression::putInSymbolTable(LookupLocation const* loc) {
		//debug.line("Expression::putInSymbolTable called for " + toString());
		fLookupLoc = loc;
	}
	
	HYPER_Common::UTF8String Expression::getResolveDescription() const {
		return UTF8STR("expression %1").compose(quote(toString()));
	}
	
	void Expression::setResolved(bool isResolved) {
		fResult.fResolved = isResolved;
	}
	
	void Expression::setResultType(Type const* t) {
		hyp_assert(t != 0);
		fResult.fType = t;
	}
	
	void Expression::setOwnerInstance(InstanceProvider const& i) {
		fOwner = new InstanceProvider(i);
	}
	
	InstanceProvider const* Expression::getOwnerInstance() const {
		return fOwner;
	}
	
	void Expression::setResultDecl(DeclaredItem* d) {
		hyp_assert(d != 0);
		fResult.fDecl = d;
	}
	
	void Expression::setResultValue(ConstantValue const* cv) {
		hyp_assert(cv != 0);
		hyp_assert(fResult.fResolved);
		hyp_assert(fResult.fType != 0);
		fResult.fValue = cv;
	}
	
	LookupLocation const* Expression::getLookupLocation() const {
		return fLookupLoc;
	}
	
	bool Expression::checkHasResult() const {
		if (hasResult().isFalse()) {
			fCompileInfo.diag().error(
				getLocation(),
				UTF8STR(
					"expression %1 is used in a context "
					"where it should return a result, but it does not"
				),
				quote(toString())
			);
			return false; // error
		}
		else {
			return true; // OK
		}
	}
	
	TriBool Expression::returnsValue() const {
		return hasResult() && fResult.fResolved && fResult.fType != 0;
	}
	
	bool Expression::checkReturnsValue() const {
		if (hasResult().isFalse() || (fResult.fResolved && fResult.fType == 0))
		{
			HYPER_Common::UTF8StringBuilder sb;
			sb +=
				UTF8STR(
					"expression %1 is used in a context "
					"where it should return a result, but it does not"
				).compose(quote(toString()));
			
			if (fResult.fDecl != 0) {
				sb += UTF8STR(". It refers to ");
				sb += fResult.fDecl->getDescription();
			}
			
			fCompileInfo.diag().error(sb.toString(), getLocation());
			return false; // error
		}
		else {
			return fResult.fResolved;
		}
	}
	
	void Expression::markAsCalled() {
		hyp_assert(resolveNotYetStarted());
		fIsCalled = true;
	}
	
	bool Expression::isMarkedAsCalled() {
		return fIsCalled;
	}
}
