/*!
	\file expression.h
	\brief Contains the Expression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2004-07-04
// Date modified:	2008-07-09

#ifndef DEF_EXPRESSION_H
 #define DEF_EXPRESSION_H
 
 #include "important_fwdecl.h"
 #include "resolveable.h"
 #include "tribool.h"

/// Contains all code of the compiler
namespace HYPER_Compiler {
	
	class InstanceProvider;
	
	/// Describes the result of an expression
	class ExpressionResult {
	public:
		/// Was the expression successfully resolved?
		bool isResolved() const;
		
		/// Get the resulting type if the expression returned a value
		Type const* getType() const;
		
		/// Get the declaration the expression refers to
		DeclaredItem* getDecl() const;
		
		/// Is the resulting value of the expression known at compile time?
		bool knownValue() const;
		
		/// Get the compile-time value of the expression.
		ConstantValue const* getValue() const;
		
		/// Get bool result value of the expression (must have bool type!)
		TriBool getBoolValue() const;
		
	private:
		ExpressionResult(Expression& expr);
		friend class Expression;
		
		Expression& fExpr;
		bool fResolved;
		ConstantValue const* fValue;
		Type const* fType;
		DeclaredItem* fDecl;
	};
	
	/// Base class for all sorts of expressions
	class Expression : public Resolveable {
	public:
		/// Destructor
		virtual ~Expression();
		
		CompileInfo& getCompileInfo() const;
		
		/// Returns a string representation of the expression
		virtual HYPER_Common::UTF8String toString() const = 0;
		
		/// Gets information about the expression's result
		ExpressionResult const& getResult() const;
		
		InstanceProvider const* getOwnerInstance() const;
		
		/// Get the location of the expression
		const Location* getLocation() const;
		
		virtual void putInSymbolTable(LookupLocation const* loc);
		
		virtual HYPER_Common::UTF8String getResolveDescription() const;
		
		const Location* getResolveLocation() const { return getLocation(); }
		
		virtual bool doInterfaceResolve(Trace const& trace) { return true; }
		
		/// Ask if the dereferenced expression can be used as an L-value
		virtual TriBool isClassLValue() const = 0;
		
		/// Ask if the pointer expression can be used as an L-value
		virtual TriBool isPointerLValue() const = 0;
		
		/// Ask if the expression is a call
		virtual TriBool isCall() const = 0;
		
		/// Ask if the expression if it returns something (a value or a
		/// declareditem)
		virtual TriBool hasResult() const = 0;
		
		/// Ask the expression if its result is known at compile time
		virtual TriBool resultAtCompileTime() const = 0;
		
		/// Check if the expression returns something
		bool checkHasResult() const;
		
		/**
			\brief Asks the expression if it returns a result value, not just a declareditem
			\return True if it returns a result
		*/
		TriBool returnsValue() const;
		
		/**
			\brief Check if the expression returns a result, and if not generates an error message
			\return True if no error was generated, false otherwise.
		*/
		bool checkReturnsValue() const;
		
		/// Check if the expression is 'null'
		virtual bool isNullLiteral() const {
			return false; // most expressions are not
		}
		
		/// Check if the expression is resolved and returns a result
		bool isResolvedAndReturnsResult() const {
			return this->returnsValue().isTrue()
				&& this->getResult().isResolved();
		}
		
		/// Check if the expression is resolved, returns a result and is not
		/// 'null'
		bool isResolvedAndNotNull() const {
			return this->returnsValue().isTrue()
				&& this->getResult().isResolved()
				&& !this->isNullLiteral();
		}
		
		void markAsCalled();
		bool isMarkedAsCalled();
		
	protected:
		Expression(CompileInfo& info, const Location* loc);
		
		void setResolved(bool isResolved = true);
		void setResultType(Type const* t);
		void setOwnerInstance(InstanceProvider const& i);
		void setResultDecl(DeclaredItem* d);
		void setResultValue(ConstantValue const* cv);
		
		LookupLocation const* getLookupLocation() const;
		
	private:
		CompileInfo& fCompileInfo;
		ExpressionResult fResult;
		InstanceProvider const* fOwner;
		const Location* fLoc;
		LookupLocation const* fLookupLoc;
		bool fIsCalled;
	};
}
#endif
