/*!
	\file expressionparser.cxx
	\brief Implements the ExpressionParser class.
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-09-21
// Last time modified:	2008-10-15

#include "expressionparser.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "arrayoperatorexpression.h"
#include "binaryoperatorexpression.h"
#include "boolconditionexpression.h"
#include "chainedcompareexpression.h"
#include "conditionalexpression.h"
#include "functioncallexpression.h"
#include "identifierexpression.h"
#include "literalexpression.h"
#include "memberselectionexpression.h"
#include "newoperatorexpression.h"
#include "operators.h"
#include "typeparser.h"
#include "unaryoperatorexpression.h"

#include <sstream>

using HYPER_Common::UTF8String;
using HYPER_Common::UTF8StringBuilder;

namespace HYPER_Compiler {
	
	ExpressionParser::ExpressionParser(ParserInfo const& pi,
		ClassContext const* cCtxt)
	: ParserBase(pi), fCtxt(cCtxt)
	{
		//
	}
	
	bool ExpressionParser::canBeStartOfExpression(Token const& t) {
		if (t.isLiteral() || t.isIdentifier() || isClassNameKeyword(t)) {
			return true;
		}
		else if (t.isKeyword()) {
			// (class name keywords like 'int' are also allowed, see previous 'if')
			switch (t.getKeyword()) {
			case Keywords::kFalse:
			case Keywords::kNew:
			case Keywords::kNull:
			case Keywords::kSystem:
			case Keywords::kThis:
			case Keywords::kTrue:
				return true;
			default:
				return false;
			}
		}
		else if (t.isSymbol()) {
			switch (t.getSymbol()) {
			case Token::kMinus:
			case Token::kMinusMinus:
			case Token::kNot:
			case Token::kOpenParen:
			case Token::kPlus:
			case Token::kPlusPlus:
				return true;
			default:
				return false;
			}
		}
		else {
			// fallback (EOF, newline, etc...)
			return false;
		}
	}
	
	Expression* ExpressionParser::doExpression() {
		//debug.line("trying to parse expression");
		Expression* e = doConditionalExpression();
		return e;
	}
	
	BoolConditionExpression* ExpressionParser::doBoolResultExpression() {
		return new BoolConditionExpression(doExpression() ,fCtxt);
	}
	
	Expression*
	ExpressionParser::doConditionalExpression()
	{
		Expression* e = doOrExpression();
		
		if (lookToken().isSymbol(Token::kQuestionMark)) {
			nextToken(); // eat '?' symbol
			Expression* te = doExpression();
			doSymbol(Token::kColon);
			Expression* fe = doExpression();
			e =
				new ConditionalExpression(
					new BoolConditionExpression(e, fCtxt), te, fe, fCtxt
				);
		}
		
		return e;
	}
	
	Expression* ExpressionParser::doOrExpression() {
		Expression* e = doAndExpression();
		
		while (true) {
			if (lookToken().isSymbol(Token::kOr)) {
				// `|' operator
			}
			else {
				break;
			}
			
			nextToken(); // eat "|" symbol
			e =
				new BinaryOperatorExpression(
					e, doAndExpression(), Operators::kOr
				);
		};
		
		return e;
	}
	
	Expression* ExpressionParser::doAndExpression() {
		Expression* e = doConsequenceExpression();
		
		while (true) {
			if (lookToken().isSymbol(Token::kAnd)) {
				// `&' operator
			}
			else {
				break;
			}
			
			nextToken(); // eat "&" symbol
			e =
				new BinaryOperatorExpression(
					e, doConsequenceExpression(), Operators::kAnd
				);
		};
		
		return e;
	}
	
	Expression*
	ExpressionParser::doConsequenceExpression()
	{
		Expression* e = doCompareExpression();
		
		while (true) {
			if (lookToken().isSymbol(Token::kArrow)) {
				// `->' operator
			}
			else {
				break;
			}
			
			nextToken(); // eat "->" symbol
			e =
				new BinaryOperatorExpression(
					e, doCompareExpression(), Operators::kConsequence
				);
		};
		
		return e;
	}
	
	Expression*
	ExpressionParser::doCompareExpression()
	{
		std::vector<Expression*> exprs;
		std::vector<Operators::BinaryOperator> ops;
		
		exprs.push_back(doAddExpression());
		
		while (true) {
			Operators::BinaryOperator op;
			
			if (lookToken().isSymbol(Token::kEqual)) {
				op = Operators::kEquals;
			}
			else if (lookToken().isSymbol(Token::kNotEqual)) {
				op = Operators::kDiffers;
			}
			else if (lookToken().isSymbol(Token::kLessThan)) {
				op = Operators::kLess;
			}
			else if (lookToken().isSymbol(Token::kLessOrEqual)) {
				op = Operators::kLessOrEqual;
			}
			else if (lookToken().isSymbol(Token::kGreaterThan)) {
				op = Operators::kGreater;
			}
			else if (lookToken().isSymbol(Token::kGreaterOrEqual)) {
				op = Operators::kGreaterOrEqual;
			}
			else if (lookToken().isSymbol(Token::kPointerEqual)) {
				op = Operators::kPointerEquals;
			}
			else if (lookToken().isSymbol(Token::kPointerNotEqual)) {
				op = Operators::kPointerDiffers;
			}
			else {
				break;
			}
			
			nextToken(); // eat operator symbol
			ops.push_back(op);
			
			exprs.push_back(doAddExpression());
		};
		
		unsigned int eCount = exprs.size();
		if (eCount == 1) { // no compares at all..
			return exprs[0];
		}
		else if (eCount == 2) { // single compare...
			return new BinaryOperatorExpression(exprs[0], exprs[1], ops[0]);
		}
		else { // chain of compares...
			return new ChainedCompareExpression(exprs, ops, fCtxt);
		}
	}
	
	Expression* ExpressionParser::doAddExpression() {
		Expression* e = doMulExpression();
		
		while (true) {
			Operators::BinaryOperator op;
			
			if (lookToken().isSymbol(Token::kPlus)) {
				op = Operators::kAdd;
			}
			else if	(lookToken().isSymbol(Token::kMinus)) {
				op = Operators::kSubtract;
			}
			else {
				break;
			}
			
			nextToken(); // eat operator symbol
			e = new BinaryOperatorExpression(e, doMulExpression(), op);
		}
		
		return e;
	}
	
	Expression* ExpressionParser::doMulExpression() {
		Expression* e = doUnaryExpression();
		
		while (true) {
			Operators::BinaryOperator op;
			
			if (lookToken().isSymbol(Token::kAsterisk)) {
				op = Operators::kMultiply;
			}
			else if (lookToken().isSymbol(Token::kSlash)) {
				op = Operators::kDivide;
			}
			else if (lookToken().isSymbol(Token::kPercentage)) {
				op = Operators::kModulo;
			}
			else {
				break;
			}
			
			nextToken(); // eat operator symbol
			e = new BinaryOperatorExpression(e, doUnaryExpression(), op);
		}
		
		return e;
	}
	
	Expression* ExpressionParser::doUnaryExpression() {
		//debug.line("trying to parse unary expression");
		Token tok = lookToken();
		UTF8String desc = tok.getDescription();
		const Location* p = lookPos();
		Operators::UnaryOperator op;
		
		if (tok.isSymbol(Token::kMinus)) {
			op = Operators::kNegative;
		}
		else if (tok.isSymbol(Token::kPlus)) {
			op = Operators::kPositive;
		}
		else if (tok.isSymbol(Token::kNot)) {
			op = Operators::kNot;
		}
		else if (tok.isSymbol(Token::kPlusPlus)) {
			op = Operators::kSuccessor;
		}
		else if (tok.isSymbol(Token::kMinusMinus)) {
			op = Operators::kPredecessor;
		}
		else {
			return doSingleExpression();
		}
		
		nextToken(); // eat operator symbol
		
		return new UnaryOperatorExpression(doUnaryExpression(), op, p);
	}
	
	Expression* ExpressionParser::doSingleExpression() {
		//debug.line("trying to parse single expression");
		Token tok = lookToken();
		
		if (tok.isKeyword(Keywords::kNull)) {
			// FIXME: remove this special case for `null', and move it with other literals
			
			Expression* e =
				LiteralExpression::createNull(lookPos(), fCtxt);
			nextToken(); // eat 'null' keyword
			
			return e;
		}
		else {
			return doPostfixExpression();
		}
	}
	
	Expression* ExpressionParser::doPostfixExpression() {
		//debug.line("trying to parse postfix expression");
		Expression* e;
		const Location* p = lookPos();
		Token const tok = lookToken();
		bool firstTokenParseError = false;
		
		if (tok.isSymbol(Token::kOpenParen)) {
			// expression between brackets
			nextToken(); // eat '(' symbol
			e = doExpression();
			doSymbol(Token::kCloseParen);
		}
		else if (isClassNameKeyword(tok)) {
			// handle class name keywords before keywords in general
			e = new IdentifierExpression(tok.getKeyword(), p, fCtxt);
			nextToken();
		}
		else if (tok.isKeyword()) {
			switch (tok.getKeyword()) {
			case Keywords::kFalse:
				{
					e =
						LiteralExpression::createFromBool(
							false, lookPos(), fCtxt
						);
					nextToken();
				}
				break;
			case Keywords::kNew:
				{
					nextToken(); // eat 'new' keyword
					TypeParser tp(this->parserInfo(), fCtxt);
					DeclaredType* type = tp.doType(false);
					
					std::vector<Expression*> args;
					if (lookToken().isSymbol(Token::kOpenParen)) {
						noWhitesBeforeCurrentToken();
						args = doArgumentsExpression();
					}
					
					e =
						new NewOperatorExpression(
							type, args, p, getCompileInfo()
						);
				}
				break;
			case Keywords::kSystem:
				{
					nextToken(); // eat 'system' keyword
					
					e = new IdentifierExpression(UTF8STR("system"), p, fCtxt);
					
					if (lookToken().isSymbol(Token::kDot)) {
						noWhitesBeforeCurrentToken();
						nextToken(); // eat '.' symbol
						p = lookPos();
					}
					else {
						fatal(error(
							UTF8STR("using the \"system\" namespace by itself "
								"in this context is not allowed; you can only "
								"use its members"),
							lookPos()
						));
					}
					
					e = new MemberSelectionExpression(e, doIdentifier(), p, fCtxt);
				}
				break;
			case Keywords::kThis:
				{
					e = LiteralExpression::createThis(lookPos(), fCtxt);
					nextToken();
				}
				break;
			case Keywords::kTrue:
				{
					e =
						LiteralExpression::createFromBool(
							true, lookPos(), fCtxt
						);
					nextToken();
				}
				break;
			default:
				firstTokenParseError = true;
			}
		}
		else if (tok.isNumericLiteral()) {
			e = doNumberLiteral();
		}
		else if (tok.isChar()) {
			e = doCharLiteral();
		}
		else if (tok.isString()) {
			e = doStringLiteral();
		}
		else if (tok.isIdentifier()) {
			e = new IdentifierExpression(doIdentifier(), p, fCtxt);
		}
		else {
			firstTokenParseError = true;
		}
		
		if (firstTokenParseError) { // parse error in first token
			if (canBeStartOfExpression(tok)) {
				fatal(error(
					UTF8STR(
						"%1 cannot be used as the start of a postfix expression"
					).compose(
						tok.getDescriptiveString()
					),
					tok.getPosition()
				));
			}
			else {
				fatal(errorExpected(
					UTF8STR("expression"), tok.getPosition(), tok
				));
			}
			e = 0; // to avoid 'unitialized use' warnings
		}
		
		bool stop = false;
		//debug.line("now checking for postfixes");
		
		do {
			if (lookToken().isSymbol(Token::kDot)) { // handle member selection
				//debug.line("parsing member selection");
				noWhitesBeforeCurrentToken();
				nextToken(); // skip the . (dot)
				noWhitesBeforeCurrentToken();
				p = lookPos();
				e = new MemberSelectionExpression(e, doIdentifier(), p, fCtxt);
			}
			else if (lookToken().isSymbol(Token::kOpenParen)) { // handle ( ) operator
				noWhitesBeforeCurrentToken();
				e =
					new FunctionCallExpression(
						e, doArgumentsExpression(), fCtxt
					);
			}
			else if (lookToken().isSymbol(Token::kOpenSquareParen)) { // handle [] operator
				noWhitesBeforeCurrentToken();
				e = doArraySelectionExpressions(e);
			}
			else {
				//debug.line("no (more) postfixes");
				stop = true;
			}
		} while (!stop);
		
		return e;
	}
	
	std::vector<Expression*> ExpressionParser::doArgumentsExpression() {
		doSymbol(Token::kOpenParen);
		
		bool stop = false;
		
		if (lookToken().isSymbol(Token::kCloseParen)) {
			stop = true; // don't go into the loop
		}
		
		std::vector<Expression*> args;
		
		while (!stop) {
			args.push_back(doExpression());
			
			if (lookToken().isSymbol(Token::kCloseParen)) {
				stop = true;
			}
			else if (lookToken().isSymbol(Token::kComma)) {
				nextToken(); // eat ',' symbol
			}
			else {
				fatal(errorExpected(
					UTF8STR("',' or ')' symbol"), lookPos(), lookToken()
				));
			}
		}
		
		doSymbol(Token::kCloseParen);
		
		return args;
	}
	
	Expression*
	ExpressionParser::doArraySelectionExpressions(
		Expression* arrayExpression)
	{
		// see if the given expression is already an array selection
		ArrayOperatorExpression* aoe =
			dynamic_cast<ArrayOperatorExpression*>(arrayExpression);
		
		while (lookToken().isSymbol(Token::kOpenSquareParen)) {
			nextToken(); // skip the "["
			
			if (aoe == 0) { // if we have no array selection expression object yet ...
				// create one, of course :)
				aoe =
					new ArrayOperatorExpression(
						arrayExpression, doExpression()
					);
			}
			else {
				// use the object we already have
				aoe->addSelector(doExpression());
			}
			
			while (lookToken().isSymbol(Token::kComma)) {
				nextToken(); // eat ',' symbol
				aoe->addSelector(doExpression());
			}
			
			doSymbol(Token::kCloseSquareParen);
		}
		
		return aoe;
	}
	
	Expression* ExpressionParser::doNumberLiteral() {
		//debug.line("parsing numeric literal");
		const Location* pos = lookPos();
		UTF8String numStr;
		bool fpnum = false;
		
		if (lookToken().isWholeNumber()) {
			numStr = lookToken().getDescription();
			nextToken(); // eat number
			// ok
		}
		else if (lookToken().isFloatingPoint()) {
			fpnum = true;
			numStr = lookToken().getDescription();
			nextToken(); // eat number
			// ok
		}
		else {
			fatal(errorExpected(UTF8STR("number"), lookPos(), lookToken()));
		}
		
		// FIXME: get rid of istringstream etc.
		boost::shared_array<char const> numCStr = numStr.toCString();
		std::istringstream iss(std::string(numCStr.get()));
		if (fpnum) {
			long double f;
			iss >> f;
			return LiteralExpression::createFromFloatNumber(f, pos, fCtxt);
		}
		else {
			unsigned long long i;
			iss >> i;
			return LiteralExpression::createFromUnsignedIntNumber(
				i, pos, fCtxt
			);
		}
	}
	
	Expression* ExpressionParser::doStringLiteral() {
		const Location* p = lookPos();
		UTF8String s = doString();
		return LiteralExpression::createFromString(s, p, fCtxt);
	}
	
	Expression* ExpressionParser::doCharLiteral() {
		const Location* p = lookPos();
		UTF8String s = doChar();
		
		/*if (s.size() > 1) {
			nonfatal(error(
				UTF8STR("only simple char literals are currently supported"), p
			));
		}*/
		hyp_assert(s.size() == 1);
		
		return LiteralExpression::createFromChar(
			HYPER_Common::CodePoint(*s.begin()), p, fCtxt
		);
	}
}
