/*!
	\file field.cxx
	\brief Implementation of the Field class.
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-01-31
// Last time modified:	2008-10-16

#include "field.h"

#include "common/utf8textflowstream.h"

#include "accesschecker.h"
#include "class.h"
#include "declaredtype.h"
#include "fieldcontext.h"
#include "initializer.h"
#include "overloadresolution.h"
#include "overloadedconstructor.h"

#include <set>

namespace HYPER_Compiler {
	
	// *** Field ***
	
	Field::Field(
		Class const* c, bool isStatic, ClassMemberAccess access,
		std::vector<HYPER_Common::UTF8String> const& names,
		std::vector<Location const*> const& locs,
		DeclaredType* t, Initializer* init, const Location* loc,
		FieldContext const* ctxt)
	 :  fClass(c), fType(t), fInit(init), fStatic(isStatic), fAccess(access),
		fBeginLoc(loc), fContext(ctxt),
		fCompileInfo(c->getCompileInfo())
	{
		hyp_assert(names.size() == locs.size());
		
		for (unsigned int i = 0; i < names.size(); ++i) {
			fDecls.push_back(new FieldDecl(this, names[i], locs[i]));
		}
	}
	
	Field::~Field() {
		// nothing here
	}
	
	HYPER_Common::UTF8String Field::getSyntaxName() const {
		return UTF8STR("Field");
	}
	
	void Field::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.print(getSyntaxName() + UTF8STR(" [") + fAccess + UTF8STR("]"));
		if (fStatic) {
			ind.print(UTF8STR(" [static]"));
		}
		ind.line();
		
		HYPER_Common::Indent out(ind);
		
		ind.line(UTF8STR("type = ") + fType->toString());
		
		for (unsigned int i = 0; i < fDecls.size(); ++i) {
			ind.line(UTF8STR("name = ") + fDecls[i]->fName);
		}
		
		fInit->myPrint(ind);
	}
	
	Class const* Field::getClass() const {
		return fClass;
	}
	
	DeclaredType const& Field::getType() const {
		return *fType;
	}
	
	bool Field::isStatic() const {
		return fStatic;
	}
	
	ClassMemberAccess Field::getAccess() const {
		return fAccess;
	}
	
	const Location* Field::getLocation() const {
		return fBeginLoc;
	}
	
	void Field::putInSymbolTable(Scope& parent) {
		LookupLocation const* ll = new LookupLocation(parent);
		fType->putInSymbolTable(ll);
		fInit->putInSymbolTable(ll);
	}
	
	CompileInfo& Field::getCompileInfo() const {
		return fCompileInfo;
	}
	
	HYPER_Common::UTF8String Field::getResolveDescription() const {
		return UTF8STR("field ") + quote(fDecls[0]->fName);
	}
	
	const Location* Field::getResolveLocation() const {
		return getLocation();
	}
	
	bool Field::doInterfaceResolve(Trace const& trace) {
		// The type must be allocatable as is, because the field is going to
		// use the physical layout of that type.
		fType->setRestrictions(true, true);
		
		bool resolveOK = fType->resolveInterface(trace);
		if (resolveOK) {
			resolveOK &=
				fType->getResolvedType()->checkAccess(
					*fContext, fType->getLocation(), getCompileInfo()
				);
		}
		
		// TODO: maybe remove this requirement ?
		if (fType->isConst() && fInit->isDefault()) {
			getCompileInfo().diag().error(
				UTF8STR("a constant field requires an explicit initializer"),
				getLocation()
			);
			resolveOK = false;
		}
		
		return resolveOK;
	}
	
	bool Field::doFullResolve(Trace const& trace) {
		return fType->resolveFull(trace) && fInit->resolveFull(trace);
	}
	
	Class const* Field::getOwnerClass() const {
		return getClass();
	}
	
	ClassMemberAccess Field::getAccessInOwner() const {
		return getAccess();
	}
	
	
	// *** FieldDecl ***
	
	FieldDecl::FieldDecl(Field* f, HYPER_Common::UTF8String name,
		Location const* loc)
	 : fF(f), fName(name), fLoc(loc)
	{
		// no action
	}
	
	Field const* FieldDecl::getField() const {
		return fF;
	}
	
	HYPER_Common::UTF8String FieldDecl::getDescription() const {
		return UTF8STR("field %1").compose(quote(fName));
	}
	
	Class const* FieldDecl::getOwnerClass() const {
		return fF->getOwnerClass();
	}
	
	ClassMemberAccess FieldDecl::getAccessInOwner() const {
		return fF->getAccessInOwner();
	}
	
	HYPER_Common::UTF8String FieldDecl::getDeclarationName() const {
		return fName;
	}
	
	const Location* FieldDecl::getLocation() const {
		return fLoc;
	}
}
