/*!
	\file functioncallexpression.cxx
	\brief Implements the FunctionCallExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2004-07-10
// Date modified:	2009-08-23

#include "functioncallexpression.h"

#include "accesschecker.h"
#include "class.h"
#include "classtype.h"
#include "compileinfo.h"
#include "constructor.h"
#include "instanceprovider.h"
#include "namespace.h"
#include "overloadedconstructor.h"
#include "procedure.h"
#include "procedureoverloadresolution.h"

#include <algorithm>

namespace HYPER_Compiler {
	
	FunctionCallExpression::FunctionCallExpression(
		Expression* function, std::vector<Expression*> arguments,
		ClassContext const* context)
	 : Expression(function->getCompileInfo(), function->getLocation()),
		fFunc(function), fArgs(arguments),
		fHasResult(TriBool::uncertain), fContext(context),
		fIsClassLValue(TriBool::uncertain),
		fIsPointerLValue(TriBool::uncertain), fProcCalled(0)
	{
		// nothing here
	}
	
	FunctionCallExpression::~FunctionCallExpression() {
		// nothing here
	}
	
	HYPER_Common::UTF8String FunctionCallExpression::toString() const {
		return fFunc->toString() + getArgumentsString();
	}
	
	void FunctionCallExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		fFunc->putInSymbolTable(loc);
		for (unsigned int i = 0; i < fArgs.size(); ++i) {
			fArgs[i]->putInSymbolTable(loc);
		}
	}
	
	bool FunctionCallExpression::doFullResolve(Trace const& trace) {
		fFunc->markAsCalled();
		bool resolveOK = fFunc->resolveFull(trace);
		
		bool illegalOrUnknownArg = false;
		for (unsigned int i = 0; i < fArgs.size(); ++i) {
			Expression* arg = fArgs[i];
			resolveOK &= arg->resolveFull(trace);
			if (!arg->checkReturnsValue() || !arg->getResult().isResolved()) {
				illegalOrUnknownArg = true;
			}
		}
		
		if (!fFunc->checkHasResult() || !fFunc->getResult().isResolved()
			|| illegalOrUnknownArg)
		{
			// no result -> cannot check call
			return false;
		}
		
		Type const* funcT = fFunc->getResult().getType();
		if (funcT != 0) {
			getCompileInfo().diag().error(
				fFunc->getLocation(),
				UTF8STR(
					"cannot use function call on expression %1, which "
					"has type %2"
				),
				quote(fFunc->toString()),
				quote(funcT->toString())
			);
			return false;
		}
		
		DeclaredItem* funcDI = fFunc->getResult().getDecl();
		if (funcDI == 0) {
			// expression called was not resolved, cannot continue
			return false;
		}
		
		if (Namespace const* ns =
			dynamic_cast<Namespace const*>(funcDI))
		{
			// cannot use function call on namespace
			getCompileInfo().diag().error(
				fFunc->getLocation(),
				UTF8STR(
					"the base operand %1 of function call expression %2 "
					"refers to %3"
				),
				quote(fFunc->toString()),
				quote(toString()),
				ns->getDescription()
			);
			resolveOK = false;
		}
		else if (Class* cl =
			dynamic_cast<Class*>(funcDI))
		{
			// Explicit constructor call
			fHasResult = true;
			fIsClassLValue = false;
			fIsPointerLValue = false;
			resolveOK &= resolveConstructorCall(cl);
		}
		else if (OverloadedProcedure const* ovp =
			dynamic_cast<OverloadedProcedure const*>(funcDI))
		{
			// procedure call
			fHasResult = true;
			resolveOK &= resolveProcedureCall(ovp);
		}
		else {
			// since funcT is 0, an error must have caused this
			hyp_assert(getCompileInfo().errorsPresent());
		}
		
		return resolveOK;
	}
	
	bool
	FunctionCallExpression::resolveConstructorCall(Class* cls)
	{
		if (cls->getASS() == kStatic) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR("%1 is static and cannot be instantiated"),
				cls->getDescription()
			);
			return false;
		}
		else if (cls->getASS() == kAbstract) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR("%1 is abstract and cannot be instantiated"),
				cls->getDescription()
			);
			return false;
		}
		else {
			// OK
		}
		
		// copy arguments list (need a vector of pointers to const Expressions)
		// FIXME: get rid of the need for this copying
		std::vector<Expression const*> args;
		args.resize(fArgs.size());
		std::copy(fArgs.begin(), fArgs.end(), args.begin());
		
		// overload resolution
		OverloadedConstructor const* ovc = cls->getConstructors();
		typedef OverloadResolution<OverloadedConstructor> ORC;
		ORC::Result call = ORC::resolveCall(ovc, args, getCompileInfo());
		
		switch (call.getResult()) {
		case ORC::Result::kNotResolved:
			return false;
			
		case ORC::Result::kNoCandidates:
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no match for constructor call %1, which is needed "
					"for %2"
				),
				quote(
					ORC::getCallPrototype(QualifiedName(UTF8STR("new")), args)
				),
				quote(toString())
			);
			
			return false;
			
		case ORC::Result::kMultipleCandidates: {
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR(
						"multiple matches for constructor call %1, which is "
						"needed for %2"
					),
					quote(
						ORC::getCallPrototype(
							QualifiedName(UTF8STR("new")), args
						)
					),
					quote(toString())
				);
			
			call.listCandidates(e);
			
			return false; }
		
		case ORC::Result::kSingleCandidate:
			break; // handled below the switch statement
		}
		
		fCstrCalled = call.getCandidates()[0]->getEntity();
		
		// check access to constructor
		bool resolveOK = true;
		resolveOK &=
			AccessChecker::check(
				*fContext, *fCstrCalled, getLocation(), getCompileInfo()
			);
		
		// check access to conversions
		OverloadResolution<OverloadedConstructor>::checkAccess(
			call.getCandidates()[0], *fContext, getLocation(), getCompileInfo()
		);
		
		setResolved();
		setResultType(Type::createClassType(false, cls));
		
		return resolveOK;
	}
	
	bool
	FunctionCallExpression::resolveProcedureCall(OverloadedProcedure const* ovp)
	{
		// copy arguments list (need a vector of pointers to const Expressions)
		// FIXME: get rid of the need for this copying
		std::vector<Expression const*> args;
		args.resize(fArgs.size());
		std::copy(fArgs.begin(), fArgs.end(), args.begin());
		
		// the context from where the procedure is called
		const InstanceProvider* instance = fFunc->getOwnerInstance();
		
		// overload resolution
		typedef ProcedureOverloadResolution POR;
		
		POR::Result call =
			POR::resolveCall(instance, ovp, args, getCompileInfo());
		
		switch (call.getResult()) {
		case POR::Result::kNotResolved:
			return false;
			
		case POR::Result::kMultipleCandidates:
		{
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR(
						"multiple matches for procedure call %1, which is "
						"needed for %2"
					),
					quote(POR::getCallPrototype(ovp->getFullName(), args)),
					quote(toString())
				);
			
			call.listCandidates(e);
			
			return false;
		}
		case POR::Result::kNoCandidates:
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no match for procedure call %1, which is needed "
					"for %2"
				),
				quote(POR::getCallPrototype(ovp->getFullName(), args)),
				quote(toString())
			);
			
			return false;
			
		case POR::Result::kSingleCandidate:
			break; // handled below the switch statement
		}
		
		// the procedure that is called
		fProcCalled = call.getCandidates()[0]->getEntity();
		
		bool resolveOK = true;
		
		InstanceType whatInstance = instance->whatInstance();
		
		//  (about the same code as in nameditemexpression)
		if (whatInstance == kConstInstance && !fProcCalled->isConst()) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"non-const procedure %1 called on const object "
					"of type %2"
				),
				quote(ovp->getFullName().toString()),
				quote(
					UTF8STR("const ")
						+ instance->getClass()->getTypeName()
				)
			);
			
			resolveOK = false;
		}
		else if (whatInstance == kNoInstance && fProcCalled->getASS() != kStatic)
		{
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"non-static procedure %1 is called without an "
					"object in expression %2:"
				),
				quote(ovp->getFullName().toString()),
				quote(toString())
			).addInfo(
				UTF8STR("declaration is: ") + fProcCalled->headerToString(),
				fProcCalled->getLocation()
			);
			
			resolveOK = false;
		}
		
		// check access to conversions
		OverloadResolution<OverloadedProcedure>::checkAccess(
			call.getCandidates()[0], *fContext, getLocation(), getCompileInfo()
		);
		
		setResolved();
		
		if (fProcCalled->hasReturnType()) {
			Type const* returnType = fProcCalled->getReturnType();
			
			if (returnType != 0) {
				// The result can be used as an LValue for classes if it
				// is a pointer.
				fIsClassLValue = returnType->isNonNullPointer();
				fIsPointerLValue = returnType->isDoublePointer();
				setResultType(returnType);
			}
			else {
				// dummy values to suppress potentially wrong errors
				fIsClassLValue = true;
				fIsPointerLValue = true;
			}
		}
		else {
			// no resulting expression
			fHasResult = false;
		}
		
		return resolveOK;
	}
	
	TriBool FunctionCallExpression::isClassLValue() const {
		return fIsClassLValue;
	}
	
	TriBool FunctionCallExpression::isPointerLValue() const {
		return fIsPointerLValue;
	}
	
	TriBool FunctionCallExpression::isCall() const {
		return true;
	}
	
	TriBool FunctionCallExpression::hasResult() const {
		return fHasResult;
	}
	
	TriBool FunctionCallExpression::resultAtCompileTime() const {
		return false;
	}
	
	HYPER_Common::UTF8String FunctionCallExpression::getArgumentsString() const {
		HYPER_Common::UTF8StringBuilder sb;
		sb += UTF8STR("(");
		
		for (unsigned int i = 0; i < fArgs.size(); ++i) {
			if (i > 0) {
				sb += UTF8STR(", ");
			}
			sb += fArgs[i]->toString();
		}
		
		sb += UTF8STR(")");
		return sb.toString();
	}
}
