/*!
	\file identifierexpression.cxx
	\brief Implements the IdentifierExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2004-07-04
// Date modified:	2008-11-08

#include "identifierexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "accesschecker.h"
#include "builtintypes.h"
#include "class.h"
#include "classrepository.h"
#include "field.h"
#include "instanceprovider.h"
#include "overloadedprocedure.h"
#include "overloadresolution.h"
#include "procedure.h"
#include "type.h"
#include "varstatement.h"

//#include "debug.h"

namespace HYPER_Compiler {
	
	IdentifierExpression::IdentifierExpression(
		HYPER_Common::UTF8String identifier, const Location* loc,
		ClassContext const* cCtxt)
	 : NamedItemExpression(loc, cCtxt), fIdent(identifier),
		fKeyword(Keywords::kNoKeyword)
	{
		//
	}
	
	IdentifierExpression::IdentifierExpression(Keywords::Keyword identifier,
		const Location* loc, ClassContext const* cCtxt)
	 : NamedItemExpression(loc, cCtxt),
		fIdent(Keywords::list().getString(identifier)), fKeyword(identifier)
	{
		//
	}
	
	IdentifierExpression::~IdentifierExpression() {
		// nothing here
	}
	
	HYPER_Common::UTF8String IdentifierExpression::toString() const {
		return fIdent;
	}
	
	bool IdentifierExpression::doFullResolve(Trace const& trace) {
		LookupLocation const* sLoc = getLookupLocation();
		hyp_assert(sLoc != 0 && "must know location in symboltable");
		
		NamedDeclaredItem* item;
		if (fKeyword != Keywords::kNoKeyword) {
			// identifier is a built-in type (keyword)
			BuiltInTypes::Primitive p = BuiltInTypes::fromKeyword(fKeyword);
			hyp_assert(p != BuiltInTypes::kNone);
			item = getCompileInfo().getClassRepository().getClassOfBuiltin(p);
		}
		else {
			// look up identifier through the symbol table
			item = sLoc->findSymbol(fIdent);
			if (item == 0) {
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR("%1 undeclared"),
					quote(fIdent)
				);
				return false;
			}
		}
		
		// name found
		setResolved();
		setResultDecl(item);
		
		bool resolveOK = true;
		
		// check if the thing this expression refers to is actually accessible
		resolveOK &=
			AccessChecker::check(
				getClassContext(), *item, getLocation(), getCompileInfo()
			);
		
		if (FieldDecl const* f = dynamic_cast<FieldDecl const*>(item)) {
			if (getClassContext().getClass() == f->getOwnerClass()) {
				// accessing field from directly inside the class it belongs to
				setOwnerInstance(InstanceProvider(getClassContext()));
			}
			else {
				// accessing field of grand parent class
				setOwnerInstance(
					InstanceProvider(f->getOwnerClass())
				);
			}
		}
		else if (OverloadedProcedure const* ovp =
			dynamic_cast<OverloadedProcedure const*>(item))
		{
			if (getClassContext().getClass() == ovp->getClass()) {
				// calling procedure from directly inside the class it belongs to
				setOwnerInstance(InstanceProvider(getClassContext()));
			}
			else {
				// accessing procedure of grand parent class
				setOwnerInstance(
					InstanceProvider(ovp->getClass())
				);
			}
			
			// TODO: move into nameditemexpression?
			resolveOK &= ovp->resolveInterfaces(trace);
		}
		
		resolveOK &= resolveNamedItem(*item);
		return resolveOK;
	}
}
