/*!
	\file ifstatement.cxx
	\brief Implementation of the IfStatement class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-08-14
// Last time modified:	2008-04-13

#include "ifstatement.h"

#include "boolconditionexpression.h"
#include "statementblock.h"

//#include "debug.h"

#include "common/utf8textflowstream.h"

namespace HYPER_Compiler {
	
	// *** IfStatement ***
	
	IfStatement::IfStatement(const Location* pos,
		StatementContext const* context, CompileInfo& info)
	 : Statement(info, pos), fElse(0), fReturnProb(SimpleProbability::always),
		fContext(context)
	{
		// nothing here
	}
	
	IfStatement::~IfStatement() {
		// nothing here
	}
	
	HYPER_Common::UTF8String IfStatement::getSyntaxName() const {
		return UTF8STR("If-Statement");
	}
	
	void IfStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.line(getSyntaxName());
		
		HYPER_Common::Indent out(ind);
		
		for (unsigned int i = 0; i < fTests.size(); ++i) {
			ind.line(UTF8STR("test: ") + fTests[i].getTest()->toString());
			fTests[i].getBlock()->myPrint(ind);
		}
		
		if (fElse) {
			ind.line(UTF8STR("else block: "));
			fElse->myPrint(ind);
		}
	}
	
	void IfStatement::addTest(BoolConditionExpression* test,
		StatementBlock* block)
	{
		fTests.push_back(TestCase(test, block, fContext));
	}
	
	void IfStatement::setElse(StatementBlock* s) {
		fElse = s;
	}
	
	HYPER_Common::UTF8String IfStatement::getResolveDescription() const {
		return UTF8STR("%1 statement").compose(QUOTED_UTF8STR("if"));
	}
	
	void IfStatement::putInSymbolTable(ScopeLocation const& loc) {
		Statement::putInSymbolTable(loc);
		
		for (unsigned int i = 0; i < fTests.size(); ++i) {
			fTests[i].putInSymbolTable(loc);
		}
		
		if (fElse) {
			fElse->putInSymbolTable(loc);
		}
	}
	
	bool IfStatement::doFullResolve(Trace const& trace) {
		//debug.line("IfStatement::doFullResolve called");
		bool resolveOK = true;
		
		for (unsigned int i = 0; i < fTests.size(); ++i) {
			resolveOK &= fTests[i].resolveFull(trace);
			fReturnProb &= fTests[i].getBlock()->getReturnProbability();
		}
		
		if (fElse) { // else block present...
			resolveOK &= fElse->resolveFull(trace);
			fReturnProb &= fElse->getReturnProbability();
		}
		else { // no else block ...
			// ... so we have an invisible else block that contains no
			// statements and never returns
			fReturnProb &= SimpleProbability::never;
		}
		
		return resolveOK;
	}
	
	SimpleProbability IfStatement::getReturnProbability() const {
		return fReturnProb;
	}
}
