/*!
	\file implementedclass.cxx
	\author Kevin André
	\brief Implements the ImplementedClass class
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-09-21
// Last time modified:	2009-08-30

#include "implementedclass.h"

#include "base/easydiagnostics.h"
#include "base/location.h"

#include "common/utf8textflowstream.h"

#include "assignmentoperator.h"
#include "autogeneratedconstructor.h"
#include "classtype.h"
#include "classrepository.h"
#include "compileinfo.h"
#include "field.h"
#include "implementedconstructor.h"
#include "implementedenum.h"
#include "implementedprocedure.h"
#include "overloadedassignmentoperator.h"
#include "overloadedconstructor.h"
#include "parameter.h"
#include "type.h"

namespace HYPER_Compiler {
	
	ImplementedClass::ImplementedClass(CompileInfo& info,
		HYPER_Common::UTF8String caption,
		ClassMemberAccess access, AbstractSealedStatic ass,
		const Location* begin, Class* parent)
	 : Class(info), SyntaxItem(), fCaption(caption), fAccess(access),
		fASS(ass), fFullName(caption), fBeginPos(begin),
		fArrayOps(0), fParent(parent)
	{
		// no action
	}
	
	ImplementedClass::~ImplementedClass() {
		// no action
	}
	
	HYPER_Common::UTF8String ImplementedClass::getSyntaxName() const {
		return UTF8STR("Class");
	}
	
	void ImplementedClass::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const
	{
		ind.print(
			UTF8STR("%1 %2 [%3]").compose(
				getSyntaxName(), fCaption, fAccess
			)
		);
		ind.line(
			(fASS == kNoASS)
				? UTF8STR(" ")
				: UTF8STR("[%1]").compose(toString(fASS))
		);
		
		HYPER_Common::Indent out(ind);
		
		for (unsigned int i = 0; i < fFields.size(); ++i) {
			fFields[i]->myPrint(ind);
		}
		
		for (unsigned int i = 0; i < fConstructors.size(); ++i) {
			fConstructors[i]->myPrint(ind);
		}
		
		for (unsigned int i = 0; i < fProcedures.size(); ++i) {
			fProcedures[i]->myPrint(ind);
		}
		
		for (unsigned int i = 0; i < fClasses.size(); ++i) {
			fClasses[i]->myPrint(ind);
		}
		
		for (unsigned int i = 0; i < fEnums.size(); ++i) {
			fEnums[i]->myPrint(ind);
		}
	}
	
	HYPER_Common::UTF8String ImplementedClass::getCaption() const {
		return fCaption;
	}
	
	HYPER_Common::UTF8String ImplementedClass::getDeclarationName() const {
		return fCaption;
	}
	
	ClassMemberAccess ImplementedClass::getAccess() const {
		return fAccess;
	}
	
	AbstractSealedStatic ImplementedClass::getASS() const {
		return fASS;
	}
	
	HYPER_Common::UTF8String ImplementedClass::getTypeName() const {
		return fFullName.toString();
	}
	
	void ImplementedClass::addMember(ImplementedEnum* e) {
		fEnums.push_back(e);
		registerClass(e);
	}
	
	void ImplementedClass::addMember(ImplementedClass* c) {
		fClasses.push_back(c);
		registerClass(c);
	}
	
	void ImplementedClass::addMember(ImplementedProcedure* p) {
		fProcedures.push_back(p);
		registerProcedure(p);
	}
	
	void ImplementedClass::addMember(ImplementedConstructor* co) {
		fConstructors.push_back(co);
		registerConstructor(co, false /* don't know yet if it's a copy constr. */);
	}
	
	void ImplementedClass::addMember(Field* f) {
		fFields.push_back(f);
		registerField(f);
	}
	
	const Location* ImplementedClass::getLocation() const {
		return fBeginPos;
	}
	
	Class const* ImplementedClass::getParentClass() const {
		return fParent;
	}
	
	Class::BinOpsMap const& ImplementedClass::getBinaryOperators() const {
		// TODO: implement real operator overloading
		return fBinaryOps;
	}
	
	Class::CremOpsMap const& ImplementedClass::getCrementOperators() const {
		// TODO: implement real operator overloading
		return fCrementOps;
	}
	
	Class::UnOpsMap const& ImplementedClass::getUnaryOperators() const {
		// TODO: implement real operator overloading
		return fUnaryOps;
	}
	
	Class::AssignOpsMap const& ImplementedClass::getAssignmentOperators() const {
		// TODO: implement real operator overloading
		return fAssignOps;
	}
	
	OverloadedArrayOperator const* ImplementedClass::getArrayOperators() const {
		// TODO: implement real operator overloading
		return fArrayOps;
	}
	
	// TODO : move this class ?
	class AutoGenAssignOp : public AssignmentOperator {
	public:
		AutoGenAssignOp(Type const* t)
		 : AssignmentOperator(
				Operators::kSimpleAssign, t, Location::getAutoGenerated()
			)
		{
			// no action
		}
		
		// TODO : change return type
		void generateCode() {
			// TODO
		}
	};
	
	bool
	ImplementedClass::doInterfaceResolve(Trace const& trace)
	{
		//debug.line("ImplementedClass::doInterfaceResolve called");
		bool resolveOK = true;
		
		for (unsigned int i = 0; i < fFields.size(); ++i) {
			resolveOK &= fFields[i]->resolveInterface(trace);
			
			if (fASS == kStatic && !fFields[i]->isStatic()) {
				getCompileInfo().diag().error(
					UTF8STR(
						"a static class cannot contain an non-static field"
					),
					fFields[i]->getLocation()
				);
				
				resolveOK = false;
			}
		}
		
		for (unsigned int i = 0; i < fConstructors.size(); ++i) {
			resolveOK &= fConstructors[i]->resolveInterface(trace);
		}
		
		if (getASS() != kStatic) {
			// create a default constructor if no user-defined constructors
			if (fConstructors.empty()) {
				registerConstructor(
					AutoGeneratedConstructor::createDefaultConstructor(this),
					false
				);
			}
			
			// look for a copy constructor & create one if needed
			
			const std::vector<Constructor*>& ovl =
				getConstructors()->getOverloadList();
			bool copyConstrFound = false;
			
			Type const* paramT = Type::createClassType(false, this);
			
			// Don't use overload resolution here, just look for an exact match.
			
			// search for exact match
			for (unsigned int i = 0; i < ovl.size(); ++i) {
				Constructor* c = ovl[i];
				
				if (c->getParameterList() == 0) {
					continue; // parameters not resolved of this one
				}
				
				if (c->getParameterList()->size() != 1) {
					continue; // only allow constructors with one parameter
				}
				
				Type const* cParamT = (*c->getParameterList())[0]->getType();
				if (cParamT == 0) { // unknown parameter type...
					continue; // don't know if this is a copy constructor
				}
				
				// copy constructor ?
				if (paramT->nonConstEquals(cParamT->getDereferenced())) {
					// yes...
					copyConstrFound = true;
					if (!c->isImplicit()) {
						// internally viewed as an implicit constructor
						c->makeImplicit();
					}
					registerCopyConstructor(c);
					break;
				}
			}
			
			// if no user-defined copy constructor then create a default one
			if (!copyConstrFound) {
				// generate a default copy constructor
				AutoGeneratedConstructor* co =
					AutoGeneratedConstructor::createCopyConstructor(this);
				
				registerConstructor(co, true);
			}
		}
		
		for (unsigned int i = 0; i < fClasses.size(); ++i) {
			resolveOK &= fClasses[i]->resolveInterface(trace);
		}
		
		for (unsigned int i = 0; i < fProcedures.size(); ++i) {
			ImplementedProcedure* p = fProcedures[i];
			AbstractSealedStatic pASS = p->getASS();
			
			switch (fASS) {
			case kNoASS:
				// All OK
				break;
			case kAbstract:
				// All OK
				break;
			case kSealed:
				if (pASS == kAbstract) {
					getCompileInfo().diag().error(
						UTF8STR(
							"a sealed class cannot contain an abstract "
								"procedure"
						),
						p->getLocation()
					);
					p->setASS(kSealed);
					resolveOK = false;
				}
				else if (pASS == kNoASS) {
					// implicit sealed
					p->setASS(kSealed);
				}
				// else OK
				break;
			case kStatic:
				switch (pASS) {
				case kNoASS:
					// implicit static
					p->setASS(kStatic);
					break;
				case kAbstract:
					getCompileInfo().diag().error(
						UTF8STR(
							"a static class cannot contain an abstract "
								"procedure"
						),
						p->getLocation()
					);
					p->setASS(kStatic);
					resolveOK = false;
					break;
				case kSealed:
					getCompileInfo().diag().error(
						UTF8STR(
							"a static class cannot contain a sealed procedure"
						),
						p->getLocation()
					);
					p->setASS(kStatic);
					resolveOK = false;
					break;
				case kStatic:
					// OK
					break;
				}
				break;
			}
			
			if (p->isConst() && p->getASS()== kStatic) {
				getCompileInfo().diag().error(
					p->getLocation(),
					UTF8STR("a static procedure cannot be %1"),
					QUOTED_UTF8STR("const")
				);
				resolveOK = false;
			}
			
			resolveOK &= p->resolveInterface(trace);
		}
		
		// create default assignment operator if needed
		if (fAssignOps.begin() == fAssignOps.end()) {
			AssignmentOperator* ass =
				new AutoGenAssignOp(Type::createClassType(false, this));
			
			// TODO: implementation of the assignment operator & check
			// copyability of fields
			
			fAssignOps[Operators::kSimpleAssign] =
				new OverloadedAssignmentOperator(ass);
		}
		
		// look for overload conflicts
		resolveOK &= checkOverloads();
		
		return resolveOK;
	}
	
	bool
	ImplementedClass::doFullResolve(Trace const& trace)
	{
		bool resolveOK = true;
		
		for (unsigned int i = 0; i < fFields.size(); ++i) {
			resolveOK &= fFields[i]->resolveFull(trace);
		}
		
		for (unsigned int i = 0; i < fConstructors.size(); ++i) {
			resolveOK &= fConstructors[i]->resolveFull(trace);
		}
		
		for (unsigned int i = 0; i < fClasses.size(); ++i) {
			resolveOK &= fClasses[i]->resolveFull(trace);
		}
		
		for (unsigned int i = 0; i < fProcedures.size(); ++i) {
			resolveOK &= fProcedures[i]->resolveFull(trace);
		}
		
		return resolveOK;
	}
}
