/*!
	\file implementedclass.h
	\author Kevin André
	\brief Contains the ImplementedClass class
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-09-21
// Last time modified:	2009-08-30

#ifndef DEF_IMPLEMENTEDCLASS_H
 #define DEF_IMPLEMENTEDCLASS_H
 
 #include "class.h"
 #include "compileinfo.h"
 #include "membertypes.h"
 #include "qualifiedname.h"
 #include "syntaxitem.h"
 
 #include <vector>

namespace HYPER_Compiler {
	
	class ImplementedProcedure;
	class ImplementedConstructor;
	class ImplementedEnum;
	class Field;
	class PartialNamespace;
	
	/// Represents a class implemented by the user
	class ImplementedClass : public Class, public SyntaxItem {
	public:
		ImplementedClass(CompileInfo& info, HYPER_Common::UTF8String caption,
			ClassMemberAccess access, AbstractSealedStatic ass,
			const Location* begin, Class* parent);
		
		~ImplementedClass();
		
		HYPER_Common::UTF8String getSyntaxName() const;
		void myPrint(HYPER_Common::UTF8TextFlowStream& ind) const;
		
		HYPER_Common::UTF8String getCaption() const;
		HYPER_Common::UTF8String getTypeName() const;
		
		ClassMemberAccess getAccess() const;
		AbstractSealedStatic getASS() const;
		
		void addMember(ImplementedEnum* e);
		void addMember(ImplementedClass* c);
		void addMember(ImplementedProcedure* p);
		void addMember(ImplementedConstructor* co);
		void addMember(Field* f);
		
		const Location* getLocation() const;
		
		Class const* getParentClass() const;
		
		BinOpsMap const& getBinaryOperators() const;
		CremOpsMap const& getCrementOperators() const;
		UnOpsMap const& getUnaryOperators() const;
		AssignOpsMap const& getAssignmentOperators() const;
		OverloadedArrayOperator const* getArrayOperators() const;
		
		HYPER_Common::UTF8String getDeclarationName() const;
		
		CompileInfo& getCompileInfo() const { return Class::getCompileInfo(); }
		
	protected:
		bool doInterfaceResolve(Trace const& trace);
		bool doFullResolve(Trace const& trace);
		
	private:
		
		HYPER_Common::UTF8String fCaption;
		ClassMemberAccess fAccess;
		AbstractSealedStatic fASS;
		std::vector<Field*> fFields;
		std::vector<ImplementedConstructor*> fConstructors;
		std::vector<ImplementedProcedure*> fProcedures;
		std::vector<ImplementedClass*> fClasses;
		std::vector<ImplementedEnum*> fEnums;
		QualifiedName fFullName;
		const Location* fBeginPos;
		BinOpsMap fBinaryOps;
		CremOpsMap fCrementOps;
		UnOpsMap fUnaryOps;
		AssignOpsMap fAssignOps;
		OverloadedArrayOperator* fArrayOps;
		Class* fParent;
	};
}
#endif
