/*!
	\file implementedconstructor.cxx
	\brief Implementation of the ImplementedConstructor class.
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-06-06
// Last time modified:	2008-11-08

#include "implementedconstructor.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

#include "class.h"
#include "compileinfo.h"
#include "declaredparameter.h"
#include "overloadedconstructor.h"
#include "overloadresolution.h"

namespace HYPER_Compiler {
	
	ImplementedConstructor::ImplementedConstructor(
		Class* c,
		ClassMemberAccess access, bool isImplicit, const Location* begin)
	 : Constructor(c->getCompileInfo()), fClass(c), fAccess(access),
		fImplicit(isImplicit), fResolvedParams(0),
		fCode(c->getCompileInfo(), begin), fBeginPos(begin), fDup(false)
	{
		// nothing here
	}
	
	HYPER_Common::UTF8String ImplementedConstructor::getSyntaxName() const {
		return UTF8STR("ImplementedConstructor");
	}
	
	void ImplementedConstructor::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.print(getSyntaxName() + UTF8STR(" [") + fAccess + UTF8STR("]"));
		ind.line(fImplicit ? UTF8STR(" [implicit]") : UTF8STR(""));
		
		HYPER_Common::Indent out(ind);
		
		for (unsigned int i = 0; i < fParams.size(); ++i) {
			DeclaredParameter* fp = fParams[i];
			ind.line(
				UTF8STR("Parameter:  %1").compose(fp->toString())
			);
		}
		
		fCode.myPrint(ind);
	}
	
	Class* ImplementedConstructor::getClass() const {
		return fClass;
	}
	
	HYPER_Common::UTF8String ImplementedConstructor::headerToString() const {
		HYPER_Common::UTF8StringBuilder sb;
		
		if (isImplicit() && fParams.size() == 1) {
			sb += UTF8STR("implicit ");
		}
		
		sb += UTF8STR("procedure new(");
		
		bool resolvedParams = (fResolvedParams != 0);
		for (unsigned int i = 0; i < fParams.size(); ++i) {
			if (i > 0) {
				sb += UTF8STR(", ");
			}
			
			if (resolvedParams) {
				sb += (*fResolvedParams)[i]->toString();
			}
			else {
				sb += fParams[i]->toString();
			}
		}
		
		sb += UTF8STR(")");
		
		return sb.toString();
	}
	
	ClassMemberAccess ImplementedConstructor::getAccess() const {
		return fAccess;
	}
	
	bool ImplementedConstructor::isImplicit() const {
		return fImplicit;
	}
	
	void ImplementedConstructor::makeImplicit() {
		fImplicit = true;
	}
	
	StatementBlock& ImplementedConstructor::getCodeBlock() {
		return fCode;
	}
	
	unsigned ImplementedConstructor::getParameterCount() const {
		return fParams.size();
	}
	
	std::vector<DeclaredParameter*>&
	ImplementedConstructor::getDeclaredParameterList()
	{
		return fParams;
	}
	
	const std::vector<Parameter const*>*
	ImplementedConstructor::getParameterList() const
	{
		return fResolvedParams;
	}
	
	const Location* ImplementedConstructor::getLocation() const {
		return fBeginPos;
	}
	
	void ImplementedConstructor::markAsDuplicate() {
		fDup = true;
	}
	
	bool ImplementedConstructor::isDuplicate() const {
		return fDup;
	}
	
	void ImplementedConstructor::putInSymbolTable(Scope& parent) {
		registerInParent(parent);
		for (unsigned i = 0; i < fParams.size(); ++i) {
			fParams[i]->putInSymbolTable(*OrderedScope::getEndLocation());
			this->registerSymbolWithDiagnostic(fParams[i], getCompileInfo());
		}
		fCode.putInSymbolTable(*OrderedScope::getEndLocation());
	}
	
	HYPER_Common::UTF8String
	ImplementedConstructor::getResolveDescription() const {
		return UTF8STR("constructor ") + headerToString();
	}
	
	const Location* ImplementedConstructor::getResolveLocation() const {
		return getLocation();
	}
	
	bool ImplementedConstructor::doInterfaceResolve(Trace const& trace) {
		bool resolveOK = true;
		
		for (unsigned int i = 0; i < fParams.size(); ++i) {
			resolveOK &= fParams[i]->resolveInterface(trace);
		}
		
		// params OK?
		if (resolveOK) {
			// create list of resolved parameters
			fResolvedParams = new std::vector<Parameter const*>();
			fResolvedParams->reserve(fParams.size());
			for (unsigned int i = 0; i < fParams.size(); ++i) {
				Parameter const* p = fParams[i]->getResolvedParameter();
				fResolvedParams->push_back(p);
			}
		}
		
		if (fImplicit && fParams.size() != 1) {
			fImplicit = false;
			getCompileInfo().diag().error(
				fBeginPos,
				UTF8STR(
					"a constructor with %1 parameters cannot be "
					"declared `implicit'"
				),
				fParams.size()
			);
			resolveOK = false;
		}
		
		return resolveOK;
	}
	
	bool ImplementedConstructor::doFullResolve(Trace const& trace) {
		// full resolve of parameters not yet needed (only interface)
		//for (unsigned int i = 0; i < fParams.size(); ++i) {
			//fParams[i]->resolveFull(trace);
		//}
		
		return fCode.resolveFull(trace);
	}
}
