/*!
	\file implementedenum.cxx
	\author Kevin André
	\brief Implements the ImplementedEnum class
	
	Copyright (C) 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2009-08-30
// Last time modified:	2009-09-06

#include "implementedenum.h"

#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

namespace HYPER_Compiler {
	
	// *** ImplementedEnum ***
	
	ImplementedEnum::ImplementedEnum(CompileInfo& info,
		HYPER_Common::UTF8String caption, ClassMemberAccess access,
		const Location* begin, Class* parent)
	 : Enum(info, access, parent), SyntaxItem(), fCaption(caption),
		fFullName(caption), fBeginPos(begin)
	{
		// no action
	}
	
	ImplementedEnum::~ImplementedEnum() {
		// no action
	}
	
	HYPER_Common::UTF8String ImplementedEnum::getSyntaxName() const {
		return UTF8STR("Enum");
	}
	
	void ImplementedEnum::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const
	{
		ind.line(
			UTF8STR("%1 %2 [%3]").compose(
				getSyntaxName(), fCaption, getAccess()
			)
		);
		
		HYPER_Common::Indent out(ind);
		
		for (unsigned int i = 0; i < fValues.size(); ++i) {
			ind.line(fValues[i]->getDeclarationName());
		}
	}
	
	HYPER_Common::UTF8String ImplementedEnum::getCaption() const {
		return fCaption;
	}
	
	HYPER_Common::UTF8String ImplementedEnum::getDeclarationName() const {
		return fCaption;
	}
	
	HYPER_Common::UTF8String ImplementedEnum::getTypeName() const {
		return fFullName.toString();
	}
	
	EnumValue*
	ImplementedEnum::addValue(HYPER_Common::UTF8String name,
		Location const* loc)
	{
		EnumValue* v = new EnumValue(this, name, loc);
		fValues.push_back(v);
		return v;
	}
	
	const Location* ImplementedEnum::getLocation() const {
		return fBeginPos;
	}
	
	bool
	ImplementedEnum::doInterfaceResolve(Trace const& trace)
	{
		//debug.line("ImplementedEnum::doInterfaceResolve called");
		bool resolveOK = true;
		
		for (unsigned i = 0; i < fValues.size(); ++i) {
			registerSymbolWithDiagnostic(fValues[i], getCompileInfo());
		}
		
		// create constructors
		createDefaultConstructor(fValues.size() == 0 ? 0 : fValues[0]);
		createCopyConstructor();
		
		// create default assignment operator
		createAssignmentOperator();
		
		// create comparison operators
		createEqualityOperator();
		createDiffersOperator();
		
		return resolveOK;
	}
	
}
