/*!
	\file implementedprocedure.cxx
	\brief Implementation of the ImplementedProcedure class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-06-20
// Last time modified:	2008-01-05

#include "implementedprocedure.h"

#include "common/utf8textflowstream.h"

#include "class.h"
#include "compileinfo.h"
#include "declaredparameter.h"
#include "declaredtype.h"
#include "overloadedprocedure.h"
#include "overloadresolution.h"
#include "type.h"

#include "base/debug.h"

namespace HYPER_Compiler {
	
	ImplementedProcedure::ImplementedProcedure(HYPER_Common::UTF8String caption,
		Class* c,
		ClassMemberAccess access,
		AbstractSealedStatic ass, bool isConst, const Location* begin)
	 : Procedure(c->getCompileInfo()), fCaption(caption), fClass(c),
		fAccess(access), fASS(ass), fConst(isConst), fResolvedParams(0),
		fReturnType(0), fResolvedReturnType(0),
		fCode(c->getCompileInfo(), begin), fFullName(caption), fBeginPos(begin),
		fEndPos(0), fDup(false)
	{
		// nothing here
	}
	
	HYPER_Common::UTF8String ImplementedProcedure::getSyntaxName() const {
		return UTF8STR("ImplementedProcedure");
	}
	
	void
	ImplementedProcedure::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const
	{
		ind.print(
			UTF8STR("%1 %2 [%3]").compose(
				getSyntaxName(), fCaption, fAccess
			)
		);
		ind.print(
			(fASS == kNoASS)
				? UTF8STR("")
				: UTF8STR("[%1]").compose(toString(fASS))
		);
		ind.line(fConst ? UTF8STR(" [const]") : UTF8STR(""));
		
		HYPER_Common::Indent out(ind);
		
		for (unsigned int i = 0; i < fParams.size(); ++i) {
			DeclaredParameter* fp = fParams[i];
			ind.line(
				UTF8STR("Parameter:  %1").compose(fp->toString())
			);
		}
		
		if (fReturnType) {
			ind.line(UTF8STR("Return type: ") + fReturnType->toString());
		}
		
		fCode.myPrint(ind);
	}
	
	Class* ImplementedProcedure::getClass() const {
		return fClass;
	}
	
	HYPER_Common::UTF8String ImplementedProcedure::getCaption() const {
		return fCaption;
	}
	
	HYPER_Common::UTF8String ImplementedProcedure::headerToString() const {
		HYPER_Common::UTF8StringBuilder sb;
		
		AbstractSealedStatic ass = getASS();
		if (ass != kNoASS) {
			sb += toString(ass);
			sb += UTF8STR(" ");
		}
		
		if (isConst()) {
			sb += UTF8STR("const ");
		}
		
		sb += UTF8STR("procedure ");
		sb += getCaption();
		sb += UTF8STR("(");
		
		bool resolvedParams = (fResolvedParams != 0);
		for (unsigned int i = 0; i < fParams.size(); ++i) {
			if (i > 0) {
				sb += UTF8STR(", ");
			}
			
			if (resolvedParams) {
				sb += (*fResolvedParams)[i]->toString();
			}
			else {
				sb += fParams[i]->toString();
			}
		}
		
		sb += UTF8STR(")");
		
		if (hasReturnType()) {
			sb += UTF8STR(" : ");
			if (fResolvedReturnType) {
				sb += getReturnType()->toString();
			}
			else {
				sb += fReturnType->toString();
			}
		}
		
		return sb.toString();
	}
	
	ClassMemberAccess ImplementedProcedure::getAccess() const {
		return fAccess;
	}
	
	void ImplementedProcedure::setASS(AbstractSealedStatic ass) {
		fASS = ass;
	}
	
	AbstractSealedStatic ImplementedProcedure::getASS() const {
		return fASS;
	}
	
	bool ImplementedProcedure::isConst() const {
		return fConst;
	}
	
	void ImplementedProcedure::setReturnType(DeclaredType* t) {
		fReturnType = t;
		//~ debug.line(UTF8STR("ImplementedProcedure: returntype %1").compose(unsigned(t)));
		hyp_assert(t != 0);
	}
	
	bool ImplementedProcedure::hasReturnType() const {
		return fReturnType != 0;
	}
	
	Type const* ImplementedProcedure::getReturnType() const {
		hyp_assert(hasReturnType());
		return fResolvedReturnType;
	}
	
	QualifiedName ImplementedProcedure::getFullName() const {
		return fFullName;
	}
	
	StatementBlock& ImplementedProcedure::getCodeBlock() {
		return fCode;
	}
	
	unsigned ImplementedProcedure::getParameterCount() const {
		return fParams.size();
	}
	
	std::vector<DeclaredParameter*>&
	ImplementedProcedure::getDeclaredParameterList()
	{
		return fParams;
	}
	
	const std::vector<Parameter const*>*
	ImplementedProcedure::getParameterList() const
	{
		return fResolvedParams;
	}
	
	const Location* ImplementedProcedure::getLocation() const {
		return fBeginPos;
	}
	
	void ImplementedProcedure::setEndLocation(const Location* loc) {
		fEndPos = loc;
	}
	
	const Location* ImplementedProcedure::getEndLocation() const {
		return fEndPos;
	}
	
	void ImplementedProcedure::markAsDuplicate() {
		fDup = true;
	}
	
	bool ImplementedProcedure::isDuplicate() const {
		return fDup;
	}
	
	void ImplementedProcedure::putInSymbolTable(Scope& parent) {
		//debug.line("ImplementedProcedure::putInSymbolTable called");
		registerInParent(parent);
		
		if (fReturnType) { // if procedure has a return type...
			fReturnType->putInSymbolTable(new LookupLocation(parent));
		}
		
		for (unsigned i = 0; i < fParams.size(); ++i) {
			fParams[i]->putInSymbolTable(*OrderedScope::getEndLocation());
			this->registerSymbolWithDiagnostic(fParams[i], getCompileInfo());
		}
		fCode.putInSymbolTable(*OrderedScope::getEndLocation());
	}
	
	HYPER_Common::UTF8String
	ImplementedProcedure::getResolveDescription() const {
		return UTF8STR("procedure %1").compose(quote(fFullName.toString()));
	}
	
	const Location* ImplementedProcedure::getResolveLocation() const {
		return getLocation();
	}
	
	bool ImplementedProcedure::doInterfaceResolve(Trace const& trace) {
		bool resolveOK = true;
		
		//~ debug.line(UTF8STR("ImplementedProcedure::doInterfaceResolve called"));
		
		for (unsigned int i = 0; i < fParams.size(); ++i) {
			resolveOK &= fParams[i]->resolveInterface(trace);
		}
		
		// params OK?
		if (resolveOK) {
			//~ debug.line(UTF8STR(" parameters resolved"));
			// create list of resolved parameters
			fResolvedParams = new std::vector<Parameter const*>();
			fResolvedParams->reserve(fParams.size());
			for (unsigned int i = 0; i < fParams.size(); ++i) {
				Parameter const* p = fParams[i]->getResolvedParameter();
				fResolvedParams->push_back(p);
			}
		}
		
		if (fReturnType) { // if procedure has a return type...
			resolveOK &= fReturnType->resolveInterface(trace);
			fResolvedReturnType = fReturnType->getResolvedType();
			//~ debug.line(UTF8STR("ImplementedProcedure: resolved returntype %1").compose(unsigned(fResolvedReturnType)));
		}
		
		return resolveOK;
	}
	
	bool ImplementedProcedure::doFullResolve(Trace const& trace) {
		bool resolveOK = true;
		
		for (unsigned int i = 0; i < fParams.size(); ++i) {
			resolveOK &= fParams[i]->resolveFull(trace);
		}
		
		resolveOK &= fCode.resolveFull(trace);
		
		if (fReturnType) { // if procedure has a return type...
			resolveOK &= fReturnType->resolveFull(trace);
			
			if (fCode.getReturnProbability().isNever()) {
				getCompileInfo().diag().error(
					fEndPos,
					UTF8STR(
						"procedure %1 never returns a result; expected is a "
							"result of type %2"
					),
					quote(headerToString()),
					quote(fReturnType->toString())
				);
				resolveOK = false;
			}
			else if (fCode.getReturnProbability().isSometimes()) {
				getCompileInfo().diag().error(
					fEndPos,
					UTF8STR(
						"end of procedure %1 can be reached without "
							"returning a result; expected is a result "
							"of type %2"
					),
					quote(headerToString()),
					quote(fReturnType->toString())
				);
				resolveOK = false;
			}
		}
		
		return resolveOK;
	}
	
	bool ImplementedProcedure::mustReturnResultInAddressByParameter() const {
		return this->hasReturnType()
			&& !this->getReturnType()->canBeInvisiblyPassedByValue();
	}
}
