/*!
	\file initializer.cxx
	\brief Implements the Initializer class
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-08-17
// Last time modified:	2008-10-15

#include "initializer.h"

#include "classcontext.h"
#include "declaredtype.h"
#include "expression.h"
#include "initializationhandler.h"
#include "type.h"

#include "common/utf8textflowstream.h"

namespace HYPER_Compiler {
	
	Initializer::Initializer(DeclaredType const* t, ClassContext const* context,
		Location const* useLoc)
	 : fType(t), fExpr(0), fContext(context), fUseLoc(useLoc),
		fInitHandler(0)
	{
		//
	}
	
	Initializer::Initializer(DeclaredType const* t, Expression* e,
		ClassContext const* context, Location const* useLoc)
	 : fType(t), fExpr(e), fContext(context), fUseLoc(useLoc),
		fInitHandler(0)
	{
		//
	}
	
	Initializer::~Initializer() {
		//
	}
	
	bool Initializer::isDefault() const {
		return fExpr == 0;
	}
	
	HYPER_Common::UTF8String Initializer::getSyntaxName() const {
		return UTF8STR("Initializer");
	}
	
	void Initializer::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.line(getSyntaxName());
		HYPER_Common::Indent i(ind);
		ind.line(UTF8STR("Type : ") + fType->toString());
		ind.print(UTF8STR("Value: "));
		if (fExpr != 0) {
			ind.line(UTF8STR("<expression> ") + fExpr->toString());
		}
		else {
			ind.line(UTF8STR("<default value for type>"));
		}
	}
	
	void Initializer::putInSymbolTable(LookupLocation const* loc) {
		if (fExpr) {
			fExpr->putInSymbolTable(loc);
		}
	}
	
	CompileInfo& Initializer::getCompileInfo() const {
		return fContext->getCompileInfo();
	}
	
	HYPER_Common::UTF8String Initializer::getResolveDescription() const {
		if (fExpr) {
			return UTF8STR("initialization expression %1").compose(
				quote(fExpr->toString())
			);
		}
		else {
			return UTF8STR("default initializer");
		}
	}
	
	bool Initializer::doFullResolve(Trace const& trace) {
		Type const* type = fType->getResolvedType();
		if (type != 0) {
			fInitHandler = type->getInitHandler();
		}
		
		if (fInitHandler == 0) {
			// type was not correctly resolved, cannot check anything
			return false;
		}
		
		bool resolveOK = true;
		
		if (fExpr != 0) { // expression initializer
			resolveOK &= fExpr->resolveFull(trace);
			if (fExpr->checkReturnsValue()) {
				TriBool ci =
					fInitHandler->checkInitByExpression(
						fExpr, *fContext, fUseLoc, fContext->getCompileInfo()
					);
				
				resolveOK &= ci.toBool(true);
				
				//if (ci.isTrue()) { // initialization (of type) OK
					// check access needed by substitution
					//fInitHandler->checkAccess(
					//	*fContext, fExpr->getLocation(), info
					//);
				//}
			}
			else {
				resolveOK = false;
			}
		}
		else { // default initializer
			TriBool ci =
				fInitHandler->checkInitByDefaultValue(
					*fContext, fUseLoc, fContext->getCompileInfo()
				);
			
			resolveOK &= ci.toBool(true);
		}
		
		return resolveOK;
	}
}
