/*!
	\file iteratestatement.cxx
	\brief Implementation of the IterateStatement class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-09-10
// Last time modified:	2009-08-23

#include "iteratestatement.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

#include "builtintypes.h"
#include "class.h"
#include "compileinfo.h"
#include "conversions.h"
#include "declaredtype.h"
#include "expression.h"
#include "operatorcalls.h"
#include "overloadedconstructor.h"
#include "overloadresolution.h"
#include "simpleparameter.h"
#include "statementblock.h"
#include "statementcontext.h"
#include "type.h"

namespace HYPER_Compiler {
	
	// *** IterateVar ***
	
	IterateVar::IterateVar(HYPER_Common::UTF8String name, const Location* loc)
	 : fName(name), fType(0), fLoc(loc)
	{
		// no action
	}
	
	void IterateVar::setIterate(IterateStatement* it) {
		fIt = it;
	}
	
	HYPER_Common::UTF8String IterateVar::getName() const {
		return fName;
	}
	
	Type const* IterateVar::getResolvedType() const {
		return fType;
	}
	
	const Location* IterateVar::getLocation() const {
		return fLoc;
	}
	
	IterateStatement* IterateVar::getIterate() const {
		return fIt;
	}
	
	HYPER_Common::UTF8String IterateVar::getDescription() const {
		return UTF8STR("iterate variable `%1'").compose(getName());
	}
	
	Class const* IterateVar::getOwnerClass() const {
		return 0;
	}
	
	ClassMemberAccess IterateVar::getAccessInOwner() const {
		hyp_assert_unreachable();
	}
	
	HYPER_Common::UTF8String IterateVar::getDeclarationName() const {
		return fName;
	}
	
	
	// *** IterateStatement ***
	
	IterateStatement::IterateStatement(IterateVar* var, Expression* begin,
		Expression* end, const Location* pos, StatementContext const* context,
		CompileInfo& info)
	 : Statement(info, pos),
		fVarDecl(var), fBeginExpr(begin), fEndExpr(end), fBlock(0),
		fContext(context)
	{
		// nothing here
	}
	
	IterateStatement::~IterateStatement() {
		// nothing here
	}
	
	HYPER_Common::UTF8String IterateStatement::getSyntaxName() const {
		return UTF8STR("Iterate-Statement");
	}
	
	void IterateStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.line(getSyntaxName());
		
		HYPER_Common::Indent out(ind);
		
		ind.line(UTF8STR("var:   ") + fVarDecl->getName());
		ind.line(UTF8STR("begin: ") + fBeginExpr->toString());
		ind.line(UTF8STR("end:   ") + fEndExpr->toString());
		
		fBlock->myPrint(ind);
	}
	
	IterateStatement*
	IterateStatement::create(HYPER_Common::UTF8String var,
		const Location* varLoc, Expression* begin, Expression* end,
		const Location* loc, StatementContext const* context, CompileInfo& info)
	{
		IterateVar* iv = new IterateVar(var, varLoc);
		
		IterateStatement* is =
			new IterateStatement(iv, begin, end, loc, context, info);
		
		iv->setIterate(is);
		
		return is;
	}
	
	void IterateStatement::setBlock(StatementBlock* s) {
		//debug.line("IterateStatement::setBlock called");
		fBlock = s;
	}
	
	HYPER_Common::UTF8String IterateStatement::getResolveDescription() const {
		return UTF8STR("iterate statement");
	}
	
	void IterateStatement::putInSymbolTable(ScopeLocation const& loc) {
		Statement::putInSymbolTable(loc);
		registerInParent(loc);
		
		LookupLocation const* ll = new LookupLocation(loc);
		
		fBeginExpr->putInSymbolTable(ll);
		fEndExpr->putInSymbolTable(ll);
		
		registerSymbol(fVarDecl);
		
		fBlock->putInSymbolTable(*getEndLocation());
	}
	
	bool IterateStatement::doFullResolve(Trace const& trace) {
		bool resolveOK = true;
		
		resolveOK &= fBeginExpr->resolveFull(trace);
		resolveOK &= fBeginExpr->checkReturnsValue();
	
		resolveOK &= fEndExpr->resolveFull(trace);
		resolveOK &= fEndExpr->checkReturnsValue();
		
		if (resolveOK) {  // begin and end must be OK
			
			
			resolveOK &= checkVarInit();
			resolveOK &= checkCrementStep();
			
			
			
		}
		
		resolveOK &= fBlock->resolveFull(trace);
		
		return resolveOK;
	}
	
	bool IterateStatement::checkVarInit() {
		Type const* beginClassT =
			fBeginExpr->getResult().getType()->getDereferenced();
		
		Type const* endClassT =
			fEndExpr->getResult().getType()->getDereferenced();
		
		Type const* varT = 0;
		if (beginClassT->constEquals(endClassT)) {
			varT = beginClassT;
		}
		else {
			SimpleParameter beginParam(beginClassT);
			SimpleParameter endParam(endClassT);
			ValuePassing beginToEnd =
				ValuePassing::check(beginClassT, endParam);
			ValuePassing endToBegin =
				ValuePassing::check(endClassT, beginParam);
			bool toEndConv = beginToEnd.isPossible();
			bool toBeginConv = endToBegin.isPossible();
			
			if (toEndConv && toBeginConv) {
				getCompileInfo().diag().error(
					UTF8STR(
						"ambiguous conversion for begin and end specification; "
						"use an explicit conversion to resolve this"
					),
					fBeginExpr->getLocation()
				).addInfo(
					fBeginExpr->getLocation(),
					UTF8STR("type of begin expr: %1"),
					quote(fBeginExpr->getResult().getType()->toString())
				).addInfo(
					fEndExpr->getLocation(),
					UTF8STR("type of end expr: %1"),
					quote(fEndExpr->getResult().getType()->toString())
				);
			}
			else if (toEndConv) {
				varT = endClassT;
			}
			else if (toBeginConv) {
				varT = beginClassT;
			}
			else {
				getCompileInfo().diag().error(
					UTF8STR(
						"incompatible begin and end specification"
					),
					fBeginExpr->getLocation()
				).addInfo(
					fBeginExpr->getLocation(),
					UTF8STR("type of begin expr: %1"),
					quote(fBeginExpr->getResult().getType()->toString())
				).addInfo(
					fEndExpr->getLocation(),
					UTF8STR("type of end expr: %1"),
					quote(fEndExpr->getResult().getType()->toString())
				);
			}
		}
		
		if (varT == 0) {
			return false; // cannot continue checking
		}
		
		varT = varT->getNonConstType(); // variable cannot be const, of course
		
		fVarDecl->fType = varT; // set type of iterate variable
		
		const Class* cls = varT->getClass();
		AbstractSealedStatic ass = cls->getASS();
		if (ass == kAbstract) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"cannot use abstract class %1 for iteration"
				),
				quote(cls->getTypeName())
			);
			
			return false;
		}
		
		OverloadedConstructor const* ovc = cls->getConstructors();
		std::vector<Expression const*> constrArgs;
		constrArgs.push_back(fBeginExpr); // constructor arg is begin expr
		
		typedef OverloadResolution<OverloadedConstructor> ORC;
		
		// check VAR.new(BEGIN)
		ORC::Result call = ORC::resolveCall(ovc, constrArgs, getCompileInfo());
		
		switch (call.getResult()) {
		case ORC::Result::kNotResolved:
			return false;
			
		case ORC::Result::kNoCandidates:
			getCompileInfo().diag().error(
				fVarDecl->getLocation(),
				UTF8STR(
					"no match for constructor call `%1.%2', "
					"which is needed for the initialization "
					"of iterate variable %3"
				),
				cls->getTypeName(),
				ORC::getCallPrototype(
					QualifiedName(UTF8STR("new")), constrArgs
				),
				quote(fVarDecl->getName())
			);
			
			return false;
			
		case ORC::Result::kMultipleCandidates: {
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(
					fVarDecl->getLocation(),
					UTF8STR(
						"multiple matches for constructor call `%1.%2', "
						"which is needed for the initialization "
						"of iterate variable %3"
					),
					cls->getTypeName(),
					ORC::getCallPrototype(
						QualifiedName(UTF8STR("new")), constrArgs
					),
					quote(fVarDecl->getName())
				);
			
			call.listCandidates(e);
			
			return false; }
		
		case ORC::Result::kSingleCandidate:
			break; // handled below the switch statement
		}
		
		const Constructor* constr = call.getCandidates()[0]->getEntity();
		
		// check for implicit use of non-implicit constructor
		if (!constr->isImplicit()) {
			getCompileInfo().diag().error(
				fVarDecl->getLocation(),
				UTF8STR(
					"cannot initialize iterate variable %1 with the following "
					"constructor because it is not implicit:"
				),
				quote(fVarDecl->getName())
			).addInfo(
				constr->headerToString(), constr->getLocation()
			);
		}
		
		return true;
	}
	
	bool IterateStatement::checkCrementStep() {
		const Type* varT = fVarDecl->getResolvedType();
		
		if (varT == 0) {
			return false;
		}
		
		const Type* endT = fEndExpr->getResult().getType();
		
		bool resolveOK = true;
		
		// check VAR < STOP
		std::vector<BinaryOperatorCandidate const*> cands =
			OperatorCalls::resolveBinaryOperatorExpression(
				varT, endT, Operators::kLess, getCompileInfo()
			);
		
		if (cands.size() > 1) { // multiple candidates
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(
					fVarDecl->getLocation(),
					UTF8STR(
						"multiple less-than operators found for the comparison "
						"of iterate variable %1 and end-expression %2"
					),
					quote(fVarDecl->getName()),
					quote(fEndExpr->toString())
				).addInfo(
					fVarDecl->getLocation(),
					UTF8STR("type of variable: %1"),
					varT->toString()
				).addInfo(
					fEndExpr->getLocation(),
					UTF8STR("type of end-expression: %1"),
					endT->toString()
				);
			
			OperatorCalls::listCandidates(cands, e);
			resolveOK = false;
		}
		else if (cands.size() == 0) { //no candidates
			getCompileInfo().diag().error(
				fVarDecl->getLocation(),
				UTF8STR(
					"no less-than operators found for the comparison "
					"of iterate variable %1 and end-expression %2"
				),
				quote(fVarDecl->getName()),
				quote(fEndExpr->toString())
			).addInfo(
				fVarDecl->getLocation(),
				UTF8STR("type of variable: %1"),
				varT->toString()
			).addInfo(
				fEndExpr->getLocation(),
				UTF8STR("type of end-expression: %1"),
				endT->toString()
			);
			
			resolveOK = false;
		}
		else { // one candidate...
			// TODO : store candidate for codegen
		}
		
		// check VAR++
		OperatorCalls::CrementCallResult result =
			OperatorCalls::resolveCrementOperator(varT, Operators::kIncrement);
		
		hyp_assert(!result.fgErrorLValue); // not set by resolveCrementOperator(expr, op)
		hyp_assert(!result.fgErrorConst); // we made sure of this
		
		// no operator++
		if (result.fgErrorNoOperator) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no operator++ for iterate variable %1, which has type %2"
				),
				quote(fVarDecl->getName()),
				quote(varT->toString())
			);
			resolveOK = false;
		}
		
		CrementOperator const* cOp = result.fgOperator;
		if (cOp != 0) {
			// TODO : store candidate for codegen
		}
		
		return resolveOK;
	}
	
	SimpleProbability IterateStatement::getReturnProbability() const {
		// FIXME: not completely accurate
		return fBlock->getReturnProbability();
	}
}
