/*!
	\file literalexpression.cxx
	\brief Implements the LiteralExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-04
// Last time modified:	2008-10-15

#include "literalexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "builtintypes.h"
#include "classcontext.h"
#include "compileinfo.h"
#include "constantvalue.h"
#include "type.h"

namespace HYPER_Compiler {
	
	LiteralExpression*
	LiteralExpression::createFromUnsignedIntNumber(unsigned long long i,
		const Location* loc, ClassContext const* cCtxt)
	{
		return
			new LiteralExpression(
				kUnsignedIntNumber, i, 0, HYPER_Common::CodePoint(' '),
				UTF8STR(""), false, loc, cCtxt
			);
	}
	
	LiteralExpression*
	LiteralExpression::createFromFloatNumber(long double f,
		const Location* loc, ClassContext const* cCtxt)
	{
		return new LiteralExpression(
			kFloatNumber, 0, f, HYPER_Common::CodePoint(' '), UTF8STR(""),
			false, loc, cCtxt
		);
	}
	
	LiteralExpression*
	LiteralExpression::createFromChar(HYPER_Common::CodePoint c,
		const Location* loc, ClassContext const* cCtxt)
	{
		return new LiteralExpression(
			kChar, 0, 0, c, UTF8STR(""), false, loc, cCtxt
		);
	}
	
	LiteralExpression*
	LiteralExpression::createFromString(HYPER_Common::UTF8String s,
		const Location* loc, ClassContext const* cCtxt)
	{
		return new LiteralExpression(
			kString, 0, 0, HYPER_Common::CodePoint(' '), s, false, loc, cCtxt
		);
	}
	
	LiteralExpression*
	LiteralExpression::createFromBool(bool isTrue,
		const Location* loc, ClassContext const* cCtxt)
	{
		return new LiteralExpression(
			kBool, 0, 0, HYPER_Common::CodePoint(' '), UTF8STR(""), isTrue,
			loc, cCtxt
		);
	}
	
	LiteralExpression*
	LiteralExpression::createNull(const Location* loc,
		ClassContext const* cCtxt)
	{
		return new LiteralExpression(
			kNull, 0, 0, HYPER_Common::CodePoint(' '), UTF8STR(""), false, loc,
			cCtxt
		);
	}
	
	LiteralExpression*
	LiteralExpression::createThis(const Location* loc,
		ClassContext const* cCtxt)
	{
		return new LiteralExpression(
			kThis, 0, 0, HYPER_Common::CodePoint(' '), UTF8STR(""), false, loc,
			cCtxt
		);
	}
	
	LiteralExpression::LiteralExpression(LiteralType t,
			unsigned long long ui,
			long double f, HYPER_Common::CodePoint c,
			HYPER_Common::UTF8String s, bool b,
			const Location* loc, ClassContext const* cCtxt)
	 : Expression(cCtxt->getCompileInfo(), loc), fCC(cCtxt),
		fLiteralType(t), fUInt(ui), fFloat(f), fChar(c),
		fString(s), fBool(b)
	{
		// nothing here
	}
	
	LiteralExpression::~LiteralExpression() {
		// nothing here
	}
	
	HYPER_Common::UTF8String LiteralExpression::toString() const {
		switch (fLiteralType) {
		case kUnsignedIntNumber:
			return fUInt;
		case kFloatNumber:
			return fFloat;
		case kChar:
			return UTF8STR("'%1'").compose(fChar);
		case kString:
			return UTF8STR("\"%1\"").compose(fString);
		case kBool:
			return (fBool) ? UTF8STR("true") : UTF8STR("false");
		case kNull:
			return UTF8STR("null");
		case kThis:
			return UTF8STR("this");
		default:
			hyp_assert_unreachable(); // shouldn't come here
		}
	}
	
	bool LiteralExpression::doFullResolve(Trace const& trace) {
		setResolved();
		bool resolveOK = true;
		
		switch (fLiteralType) {
		case kUnsignedIntNumber:
			//debug << "Unsigned number literal: " << fUInt << newl;
			
			if (fUInt < 256) { //   < 2 ^ 8
				setResultType(
					BuiltInTypes::createByte(
						getCompileInfo(), true, getLocation()
					)
				);
				//debug << " literal fits in byte" << newl;
			}
			else if (fUInt <= 65535u) { //   <= 2 ^ 16 - 1
				setResultType(
					BuiltInTypes::createNat16(
						getCompileInfo(), true, getLocation()
					)
				);
				//debug << " literal fits in nat16" << newl;
			}
			else if (fUInt <= 4294967295Lu) { //   <= 2 ^ 32 - 1
				setResultType(
					BuiltInTypes::createNat32(
						getCompileInfo(), true, getLocation()
					)
				);
				//debug << " literal fits in nat32" << newl;
			}
			else if (fUInt <= 18446744073709551615LLu) { //   <= 2 ^ 64 - 1
				//Debug::printFIXME("check for overflow of nat64 literal", info);
				setResultType(
					BuiltInTypes::createNat64(
						getCompileInfo(), true, getLocation()
					)
				);
				//debug << " literal fits in nat64" << newl;
			}
			else {
				// report overflow
				getCompileInfo().diag().error(
					UTF8STR(
						"natural number literal is too large, the maximum "
							"value allowed in nat64 is 18446744073709551615"
					),
					getLocation()
				);
				resolveOK = false;
			}
			setResultValue(new ConstantValue(fUInt));
			break;
		case kFloatNumber:
			setResultType(
				BuiltInTypes::createReal(getCompileInfo(), true, getLocation())
			);
			break;
		case kChar:
			setResultType(
				BuiltInTypes::createChar(getCompileInfo(), true, getLocation())
			);
			break;
		case kString:
			setResultType(
				BuiltInTypes::createString(
					getCompileInfo(), true, getLocation()
				)
			);
			break;
		case kBool:
			setResultType(
				BuiltInTypes::createBool(getCompileInfo(), true, getLocation())
			);
			setResultValue(new ConstantValue(fBool));
			break;
		case kNull:
			setResultType(Type::createNull(getCompileInfo()));
			break;
		case kThis:
		{
			Type const* result = fCC->getThisPointer(getLocation());
			
			if (result != 0) {
				setResultType(result);
			}
			else {
				setResolved(false);
				resolveOK = false;
			}
			break;
		}
		default:
			// shouldn't be possible
			hyp_assert_unreachable();
		}
		
		return resolveOK;
	}
	
	TriBool LiteralExpression::isClassLValue() const {
		return false;
	}
	
	TriBool LiteralExpression::isPointerLValue() const {
		return false;
	}
	
	TriBool LiteralExpression::isCall() const {
		return false;
	}
	
	TriBool LiteralExpression::hasResult() const {
		return true;
	}
	
	TriBool LiteralExpression::resultAtCompileTime() const {
		return fLiteralType == kUnsignedIntNumber; // FIXME
	}
	
	bool LiteralExpression::isNullLiteral() const {
		return fLiteralType == kNull;
	}
}
