/*!
	\file loopjumpstatement.cxx
	\brief Implementation of the LoopJumpStatement class.
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-09-15
// Last time modified:	2008-06-26

#include "loopjumpstatement.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

#include "compileinfo.h"
#include "statementcontext.h"

namespace HYPER_Compiler {
	
	LoopJumpStatement::LoopJumpStatement(LoopJumpDestination dest,
		LoopJumpType t, StatementContext const* context, const Location* loc,
		CompileInfo& info)
	 : Statement(info, loc), fDest(dest), fJumpType(t), fContext(context)
	{
		// no action
	}
	
	LoopJumpStatement::~LoopJumpStatement() {
		// no action
	}
	
	HYPER_Common::UTF8String LoopJumpStatement::getSyntaxName() const {
		return UTF8STR("loop-jumping statement");
	}
	
	void LoopJumpStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.line(getSyntaxName());
		
		HYPER_Common::Indent out(ind);
		
		ind.line(UTF8STR("Type: "));
		switch (fJumpType) {
		case kBreak:
			ind.line(UTF8STR("break"));
			break;
		case kContinue:
			ind.line(UTF8STR("continue"));
			break;
		default:
			hyp_assert_unreachable();
			break;
		}
		
		ind.line(UTF8STR("Destination: "));
		switch (fDest) {
		case kUnspecifiedJump:
			ind.line(UTF8STR("unspecified"));
			break;
		case kLoopJump:
			ind.line(UTF8STR("loop"));
			break;
		case kIterateJump:
			ind.line(UTF8STR("iterate"));
			break;
		default:
			hyp_assert_unreachable();
			break;
		}
	}
	
	LoopJumpType LoopJumpStatement::getJumpType() const {
		return fJumpType;
	}
	
	LoopJumpDestination LoopJumpStatement::getDestination() const {
		return fDest;
	}
	
	HYPER_Common::UTF8String
	LoopJumpStatement::getResolveDescription() const
	{
		switch (fJumpType) {
		case kBreak:
			return UTF8STR("`break' statement");
		case kContinue:
			return UTF8STR("`continue' statement");
		default:
			hyp_assert_unreachable();
		}
	}
	
	bool LoopJumpStatement::doFullResolve(Trace const& trace) {
		switch (fDest) { // switch on jump destination type
		case kUnspecifiedJump: // no jump target specified
		{
			// switch on most recent (nested) available destination
			switch (fContext->getIterateVersusLoop()) {
			case StatementContext::kNeitherDominates:
			{
				getCompileInfo().diag().error(
					UTF8STR("no iterate or loop statement found for this jump"),
					getLocation()
				);
				
				return false;
			}
			case StatementContext::kLoopDominates:
				// TODO: store loop for code generation
				break;
			case StatementContext::kIterateDominates:
				// TODO: store iterate for code generation
				break;
			default:
				hyp_assert_unreachable();
			}
			
			break;
		}
		case kLoopJump: // jump to loop statement
			// switch on loop type
			switch (fContext->getCurrentLoopType()) {
			case StatementContext::kNoLoop:
			{
				getCompileInfo().diag().error(
					UTF8STR("no loop statement found for this jump"),
					getLocation()
				);
				
				return false;
			}
			case StatementContext::kSimpleLoop:
			{
				// TODO: store loop for code generation
				break;
			}
			case StatementContext::kMegaLoop:
			{
				// TODO: store loop for code generation
				break;
			}
			default:
				hyp_assert_unreachable();
			}
			
			break;
		case kIterateJump: // jump to iterate statement
		{
			IterateStatement const* it = fContext->getCurrentIterate();
			if (it == 0) {
				getCompileInfo().diag().error(
					UTF8STR("no iterate statement found for this jump"),
					getLocation()
				);
				
				return false;
			}
			else {
				// TODO: store iterate for code generation
			}
			break;
		}
		default:
			hyp_assert_unreachable();
		}
		
		return true; // if we get here then there were no errors
	}
	
	SimpleProbability LoopJumpStatement::getReturnProbability() const {
		return SimpleProbability::never;
	}
}
