/*!
	\file compiler/main.cxx
	\brief Driver for the Hyper Compiler.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-02-05
// Last time modified:	2008-11-16

#include "base/compilationfailederror.h"
#include "base/errormanager.h"
#include "base/internalcompilererror.h"

#include "common/copyright.h"
#include "common/console.h"
#include "common/filepath.h"
#include "common/filesystem.h"
#include "common/utf8textflowstream.h"

#include "lexer/fileexceptions.h"

#include "compiler.h"
#include "textdiagnosticprinter.h"
#include "version.h"

#include <iostream>
#include <string>
#include <vector>

#include <boost/program_options.hpp>

using namespace HYPER_Compiler;
namespace po = boost::program_options;

using HYPER_Common::Copyright;
using HYPER_Common::CodePoint;
using HYPER_Common::Console;
using HYPER_Common::FilePath;
using HYPER_Common::FileSystem;
using HYPER_Common::InvalidPathStringException;
using HYPER_Common::UTF8String;
using HYPER_Common::UTF8PushStream;
using HYPER_Common::UTF8TextFlowStream;

void printUsage(UTF8TextFlowStream& out);
void printVersion(UTF8TextFlowStream& out);

std::pair<std::string, std::string>
bool_no_option_parser(const std::string& s)
{
	//debug.line("parsing option: " + s);
	if (s.length() > 6 && s[0] == '-' && (s[1] == 'W' || s[1] == 'f')
		&& s[2] == '-')
	{
		if (s[3] == 'n' && s[4] == 'o' && s[5] == '-') {
			// negative option
			//debug.line(" negative option");
			return std::make_pair(
				s[1] + s.substr(5), std::string("false")
			);
		}
		else {
			// positive option
			//debug.line(" positive option");
			return make_pair(s.substr(1), std::string("true"));
		}
	}
	
	// no result from this parser
	//debug.line(" could not parse it");
	return std::make_pair(std::string(), std::string());
}

void optionError(UTF8PushStream& out, char const* msg) {
	out.print(UTF8STR("hyperc: "));
	out.print(UTF8String::fromCString(msg));
	out.line(
		UTF8STR(" (use option %1 for help)").compose(
			QUOTED_UTF8STR("--help")
		)
	);
}

int main(int argc, char *argv[]) {
	// std streams
	UTF8TextFlowStream& stdOut = Console::getTextFlowStdOut();
	UTF8TextFlowStream& stdErr = Console::getTextFlowStdErr();
	
	// describe commandline options
	CompileOptions opt(stdErr);
	opt.fgDebug = true; // DEBUG enabled by default, for now
	po::options_description desc;
	desc.add_options()
		("help,h", "show usage")
		("version,v", "show program version")
		("dumpspecs", "dump compiler specs")
		("inputfile", po::value<std::vector<std::string> >())
		//("outputfile,o", po::value<std::vector<std::string> >())
		("c", po::bool_switch(&opt.fgDontLink))
		("f-syntax-only", po::value<bool>(&opt.fgSyntaxOnly))
		("f-print-ast", po::value<bool>(&opt.fgPrint_AST))
		("f-debug", po::value<bool>(&opt.fgDebug))
		("f-diagnostic-buffer", po::value<bool>(&opt.fgBufferErrors))
		("W-fixme", po::value<bool>(&opt.fgFixMeWarnings))
		("W-todo", po::value<bool>(&opt.fgToDoWarnings))
		("W-whitespace", po::value<bool>(&opt.fgWhitespaceWarnings))
		("W-hide-parameter", po::value<bool>(&opt.fgHideParameterWarnings));
	po::positional_options_description pod;
	pod.add("inputfile", 1);
	po::variables_map vm;
	int cmdStyle =
		po::command_line_style::allow_long
		| po::command_line_style::allow_short
		| po::command_line_style::allow_dash_for_short
		| po::command_line_style::long_allow_adjacent
		| po::command_line_style::long_allow_next
		| po::command_line_style::short_allow_next
		| po::command_line_style::allow_long_disguise;
	
	try {
		// parse commandline options
		po::store(
			po::command_line_parser(argc, argv).style(cmdStyle)
				.options(desc).extra_parser(bool_no_option_parser)
				.positional(pod).run(),
			vm
		);
		po::notify(vm);
	}
	catch (po::unknown_option& uo) {
		optionError(stdErr, uo.what());
		return 1;
	}
	catch (po::too_many_positional_options_error& tmpoe) {
		optionError(stdErr, tmpoe.what());
		return 1;
	}
	catch (po::invalid_command_line_syntax& icls) {
		optionError(stdErr, icls.what());
		return 1;
	}
	
	// check for option '--version'
	if (vm.count("version")) {
		printVersion(stdOut);
		return 0;
	}
	
	// check for option '--help'
	if (vm.count("help")) {
		printUsage(stdOut);
		return 0;
	}
	
	// check for option '-dumpspecs'
	if (vm.count("dumpspecs")) {
		Compiler::dumpSpecs();
		return 0;
	}
	
	// must have at least one input file
	if (vm.count("inputfile") == 0) {
		// no file args
		optionError(stdErr, "no input file");
		return 1;
	}
	
	FilePath const* workingDir = FileSystem::getWorkingDirectory();
	//std::cout << "PWD: " << workingDir->toString() << std::endl;
	
	FilePath* inputFile = 0;
	
	try {
		std::string input =
			vm["inputfile"].as<std::vector<std::string> >()[0];
		UTF8String inputU8Str = UTF8String::fromCString(input.c_str());
		inputFile = FileSystem::pathFromString(inputU8Str);
		inputFile->makeAbsolute(*workingDir);
	}
	catch (InvalidPathStringException const& ipse) {
		stdErr.line(
			UTF8STR("hyperc: invalid filename %1").compose(
				doubleQuote(ipse.getString())
			)
		);
		return 1;
	}
	
	//stdErr.line(UTF8STR("hyperc: filename is '%1'").compose(inputFile->toString()));
	
	std::list<UTF8String> hyperExts;
	hyperExts.push_back(UTF8STR("hyp"));
	hyperExts.push_back(UTF8STR("hyper"));
	
	if (!inputFile->hasExtension(hyperExts)) {
		stdErr.line(UTF8STR("hyperc: input file must be a Hyper file"));
		return 2;
	}
	
//	opt.printOptions();
	
	Compiler comp(opt);
	TextDiagnosticPrinter diagPrinter(comp.getErrorManager());
	
	try {
		// Compile
		comp.Compile(*inputFile);
		// no compile error happened
		// print warnings (if any)
		diagPrinter.print(stdErr, opt.fgBufferErrors);
	}
	catch (CompilationFailedError const& cfe) {
		// programming error in the file
		diagPrinter.print(stdErr, opt.fgBufferErrors);
		return 1;
	}
	catch (FileNotReadableException const& fnre) { // error opening file
		stdErr.line(
			UTF8STR("hyperc: could not read the file %1").compose(
				doubleQuote(inputFile->toString())
			)
		);
		return 2;
	}
	catch (InternalCompilerError const& ice) {
		stdErr.line(UTF8STR("INTERNAL COMPILER ERROR:"));
		HYPER_Common::Indent i(stdErr); // indent the text below
		stdErr.line(ice.getDescription());
		stdErr.line(
			UTF8STR(">> Please submit a full bug report.")
		);
		stdErr.line(
			UTF8STR(">> Run the compiler with option -h for more info.")
		);
		return 3;
	}
	
	return 0;
}

void printUsage(UTF8TextFlowStream& out) {
	out.line(UTF8STR("Usage:  hyperc [options] [file]"));
	out.line();
	out.line(UTF8STR("General options"));
	out.line();
	out.line(UTF8STR("  -h, --help     : display this information"));
	out.line(UTF8STR("  -v, --version  : print version information"));
	out.line(UTF8STR("  -dumpspecs     : dump compiler specs and exit"));
	out.line(UTF8STR("  -c             : compile only, don't link"));
	out.line();
	
	out.line(UTF8STR("Warning options"));
	out.line();
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"These options control the warnings that can be given. "
			"An option to enable some kind of warning type has the "
			"form -W-foo, and the corresponding option to disable the "
			"warning type is -W-no-foo. All warnings are enabled by "
			"default."
		)
	);
	out.line();
	
	out.line(UTF8STR("-W-todo"));
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"This option enables warnings for the "
			"text 'TODO' in single line comments. The TODO must be in "
			"uppercase and come right after the start of the comment. "
			"Any text after the TODO will be printed as extra information "
			"in the warning message."
		)
	);
	out.line();
	out.decreaseIndent(2);
	
	out.line(UTF8STR("-W-fixme"));
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"Similar to -W-todo but reacts to the text 'FIXME'."
		)
	);
	out.line();
	out.decreaseIndent(2);
	
	out.line(UTF8STR("-W-whitespace"));
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"Warn about whitespace in locations where "
			"it should be avoided, e.g. in front of a function operator "
			"or next to a dot operator."
		)
	);
	out.line();
	
	out.decreaseIndent(4);
	
	out.line(UTF8STR("Front end options"));
	out.line();
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"These options control various actions w.r.t. the front end. "
			"Any option -f-foo that is enabled by default can be disabled "
			"by using -f-no-foo."
		)
	);
	out.line();
	
	out.line(UTF8STR("-f-syntax-only"));
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"Stop the compiler after parsing the input file(s). This "
			"also disables semantic checkings."
		)
	);
	out.line();
	out.decreaseIndent(2);
	
	out.line(UTF8STR("-f-print-ast"));
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"Print abstract syntax tree after parsing the input file(s). "
			"This does not affect compiler operation."
		)
	);
	out.line();
	out.decreaseIndent(2);
	
	out.line(UTF8STR("-f-debug"));
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"Enable the compiler's debug mode, useful "
			"for finding bugs in the compiler. When this option is active, "
			"the compiler's version number will be printed when it is run. "
			"This option is currently enabled by default."
		)
	);
	out.line();
	out.decreaseIndent(2);
	
	out.line(UTF8STR("-f-diagnostic-buffer"));
	out.increaseIndent(2);
	out.line(
		UTF8STR(
			"This option is enabled by default. "
			"Using -f-no-diagnostic-buffer disables buffering of errors and also "
			"disables removing redundant information from the errorlisting. "
			"This is potentially useful while debugging the compiler."
		)
	);
	out.line();
	
	out.decreaseIndent(4);
	
	out.line(UTF8STR("Back end options"));
	out.line();
	out.increaseIndent(2);
	out.line(
		UTF8STR("These options control the behaviour of the back end.")
	);
	out.line();
	
	// temporary (without back end)
	out.line(
		UTF8STR(
			"(No such options available yet. This version of the compiler "
			"is still only a front end)"
		)
	);
	out.line();
	
	out.decreaseIndent(2);
	
	out.print(UTF8STR("Report bugs to: "));
	out.line(UTF8STR(Version::getBugReportLocation()));
}

void printVersion(UTF8TextFlowStream& out) {
	out.line(UTF8STR("Hyper Compiler ") + Compiler::getVersion());
	out.line(Copyright::copyrightYearsAndAuthor());
	out.line();
	out.line(Copyright::licenseAndDisclaimer());
	out.line();
	out.print(UTF8STR("Report bugs to: "));
	out.line(UTF8STR(Version::getBugReportLocation()));
}
