/*!
	\file memberselectionexpression.cxx
	\brief Implements the MemberSelectionExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2004-07-10
// Date modified:	2008-07-16

#include "memberselectionexpression.h"

#include "accesschecker.h"
#include "class.h"
#include "field.h"
#include "instanceprovider.h"
#include "namespace.h"
#include "overloadedprocedure.h"
#include "overloadresolution.h"
#include "procedure.h"
#include "type.h"
#include "varstatement.h"

namespace HYPER_Compiler {
	
	MemberSelectionExpression::MemberSelectionExpression(Expression* expr,
		HYPER_Common::UTF8String member, const Location* loc,
		ClassContext const* cCtxt)
	 : NamedItemExpression(loc, cCtxt), fExpr(expr), fMember(member)
	{
		// nothing here
	}
	
	MemberSelectionExpression::~MemberSelectionExpression() {
		// nothing here
	}
	
	HYPER_Common::UTF8String MemberSelectionExpression::toString() const {
		return UTF8STR("%1.%2").compose(fExpr->toString(), fMember);
	}
	
	void MemberSelectionExpression::putInSymbolTable(LookupLocation const* loc)
	{
		Expression::putInSymbolTable(loc);
		fExpr->putInSymbolTable(loc);
	}
	
	bool MemberSelectionExpression::doFullResolve(Trace const& trace) {
		bool resolveOK = fExpr->resolveFull(trace);
		
		if (!fExpr->checkHasResult() || !fExpr->getResult().isResolved()) {
			// no result -> cannot get member
			return false;
		}
		
		Type const* t = fExpr->getResult().getType();
		
		if (t == 0) { // no expression type
			const DeclaredItem* di = fExpr->getResult().getDecl();
			
			if (di == 0) {
				// no info present, cannot do anything
				return false;
			}
			
			if (Namespace const* ns =
				dynamic_cast<Namespace const*>(di))
			{
				NamedDeclaredItem* item = ns->findLocalSymbol(fMember);
				if (item != 0) {
					setResolved();
					setResultDecl(item/*, se->getLocation()*/);
				}
				else {
					getCompileInfo().diag().error(
						getLocation(),
						UTF8STR("%1 has no member called %2"),
						ns->getDescription(),
						quote(fMember)
					);
					resolveOK = false;
				}
			}
			else if (Class const* cl = dynamic_cast<Class const*>(di)) {
				NamedDeclaredItem* item = cl->findMember(fMember);
				
				if (item != 0) {
					setResolved();
					setResultDecl(item/*, item->getLocation()*/);
					// instance: no owner, static
					setOwnerInstance(InstanceProvider(cl));
					
					// Check if we have access to the member from here
					resolveOK &=
						AccessChecker::checkMember(
							getClassContext(), *item, fMember, getLocation(),
							getCompileInfo()
						);
				}
				else {
					getCompileInfo().diag().error(
						getLocation(),
						UTF8STR("%1 has no member called %2"),
						cl->getDescription(),
						quote(fMember)
					);
					resolveOK = false;
				}
			}
			else { // other reference to declareditem
				hyp_assert_unreachable();
			}
		}
		else { // expression has type
			t = t->getDereferenced(); // dereference if needed
			const Class* cl = t->getClass();
			setOwnerInstance(InstanceProvider(fExpr));
			
			NamedDeclaredItem* item = cl->findMember(fMember);
			if (item != 0) {
				setResolved();
				setResultDecl(item/*, se->getLocation()*/);
				
				// Check if we have access to the member from here
				resolveOK &=
					AccessChecker::checkMember(
						getClassContext(), *item, fMember, getLocation(),
						getCompileInfo()
					);
			}
			else {
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR("%1 has no member called %2"),
					cl->getDescription(),
					quote(fMember)
				);
				resolveOK = false;
			}
		}
		
		if (getResult().isResolved() && getResult().getDecl() != 0) {
			resolveOK &= resolveNamedItem(*getResult().getDecl());
		}
		
		return resolveOK;
	}
}
