/*!
	\file membertypes.cxx
	\brief Contains the types of members.
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-06-02
// Last time modified:	2008-06-26

#include "membertypes.h"

#include "base/asserts.h"

namespace HYPER_Compiler {
	
	HYPER_Common::UTF8String toString(AbstractSealedStatic ass) {
		switch (ass) {
		case kNoASS:
			return HYPER_Common::UTF8String();
		case kAbstract:
			return UTF8STR("abstract");
		case kSealed:
			return UTF8STR("sealed");
		case kStatic:
			return UTF8STR("static");
		default:
			hyp_assert_unreachable();
			return HYPER_Common::UTF8String();
		}
	}
	
	
	// *** ClassMemberAccess ***
	
	ClassMemberAccess ClassMemberAccess::createPublic() {
		return ClassMemberAccess(kPublic);
	}
	
	ClassMemberAccess ClassMemberAccess::createProtected() {
		return ClassMemberAccess(kProtected);
	}
	
	ClassMemberAccess ClassMemberAccess::createPrivate() {
		return ClassMemberAccess(kPrivate);
	}
	
	bool ClassMemberAccess::isPublic() const {
		return fA == kPublic;
	}
	
	bool ClassMemberAccess::isProtected() const {
		return fA == kProtected;
	}
	
	bool ClassMemberAccess::isPrivate() const {
		return fA == kPrivate;
	}
	
	bool ClassMemberAccess::operator == (const ClassMemberAccess& a) const {
		return fA == a.fA;
	}
	
	ClassMemberAccess::operator HYPER_Common::UTF8String() const {
		switch (fA) {
		case kPublic:	
			return UTF8STR("public");
		case kProtected:
			return UTF8STR("protected");
		case kPrivate:
			return UTF8STR("private");
		default:
			// shouldn't happen
			hyp_assert_unreachable();
			return HYPER_Common::UTF8String();
		}
	}
	
	ClassMemberAccess::ClassMemberAccess(Level a)
	 : fA(a)
	{
		// nothing here
	}
	
	
	// *** SimpleMemberAccess ***
	
	SimpleMemberAccess SimpleMemberAccess::createPublic() {
		return SimpleMemberAccess(kPublic);
	}
	
	SimpleMemberAccess SimpleMemberAccess::createPrivate() {
		return SimpleMemberAccess(kPrivate);
	}
	
	bool SimpleMemberAccess::isPublic() const {
		return fA == kPublic;
	}
	
	bool SimpleMemberAccess::isPrivate() const {
		return fA == kPrivate;
	}
	
	bool SimpleMemberAccess::operator == (const SimpleMemberAccess& a) const {
		return fA == a.fA;
	}
	
	SimpleMemberAccess::operator HYPER_Common::UTF8String() const {
		switch (fA) {
		case kPrivate:
			return UTF8STR("private");
		case kPublic:
			return UTF8STR("public");
		default:
			// shouldn't happen
			hyp_assert_unreachable();
			return HYPER_Common::UTF8String();
		}
	}
	
	SimpleMemberAccess::operator ClassMemberAccess() const {
		switch (fA) {
		case kPrivate:
			return ClassMemberAccess::createPrivate();
		case kPublic:
			return ClassMemberAccess::createPublic();
		default:
			// shouldn't happen
			hyp_assert_unreachable();
			return ClassMemberAccess::createPublic();
		}
	}
	
	SimpleMemberAccess::SimpleMemberAccess(Level a)
	 : fA(a)
	{
		// nothing here
	}
}
