/*!
	\file nameditemexpression.cxx
	\brief Implements the NamedItemExpression class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2006-07-27
// Date modified:	2009-09-05

#include "nameditemexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "class.h"
#include "declaredparameter.h"
#include "declaredtype.h"
#include "enum.h"
#include "field.h"
#include "instanceprovider.h"
#include "iteratestatement.h"
#include "namespace.h"
#include "overloadedprocedure.h"
#include "procedureoverloadresolution.h"
#include "varstatement.h"

#include "base/debug.h"

namespace HYPER_Compiler {
	
	NamedItemExpression::NamedItemExpression(const Location* loc,
		ClassContext const* cCtxt)
	 : Expression(cCtxt->getCompileInfo(), loc),
		fIsClassLValue(TriBool::uncertain),
		fIsPointerLValue(TriBool::uncertain), fIsCall(TriBool::uncertain),
		fHasResult(TriBool::uncertain), fCC(cCtxt), fProc(0)
	{
		hyp_assert(fCC != 0);
	}
	
	NamedItemExpression::~NamedItemExpression() {
		// nothing here
	}
	
	bool
	NamedItemExpression::resolveNamedItem(const DeclaredItem& decl)
	{
		bool resolveOK = true;
		
		// TODO: get rid of this dynamic_cast'ing
		
		if (dynamic_cast<Namespace const*>(&decl)
			|| dynamic_cast<Class const*>(&decl))
		{
			// Don't set result type; type must remain null pointer as the
			// expression returns no value.
			fIsClassLValue = false;
			fIsPointerLValue = false;
			fIsCall = false;
			fHasResult = true;
		}
		else if (FieldDecl const* f =
			dynamic_cast<FieldDecl const*>(&decl))
		{
			resolveOK &= resolveField(f);
		}
		else if (EnumValue const* e = dynamic_cast<EnumValue const*>(&decl)) {
			Type const* valueT = Type::createClassType(true, e->getOwnerClass());
			
			fIsClassLValue = false;
			fIsPointerLValue = false;
			fIsCall = false;
			fHasResult = true;
			
			setResultType(valueT);
		}
		else if (VarDecl const* v = dynamic_cast<VarDecl const*>(&decl)) {
			Type const* varType = v->getVar()->getResolvedType();
			
			fIsClassLValue = true;
			fIsPointerLValue = varType ? varType->isNonNullPointer() : true;
			fIsCall = false;
			fHasResult = true;
			
			if (varType) {
				setResultType(varType);
			}
			else {
				setResolved(false);
			}
		}
		else if (IterateVar const* iv =
			dynamic_cast<IterateVar const*>(&decl))
		{
			// Iterate variable cannot be modified by the user, so we pretend
			// that its type is 'const'.
			Type const* varType = iv->getResolvedType()->getConstType();
			
			fIsClassLValue = true;
			fIsPointerLValue = false;
			fIsCall = false;
			fHasResult = true;
			
			if (varType) {
				setResultType(varType);
			}
			else {
				setResolved(false);
			}
		}
		else if (DeclaredParameter const* par =
			dynamic_cast<DeclaredParameter const*>(&decl))
		{
			fIsClassLValue = true;
			fIsCall = false;
			fHasResult = true;
			
			Parameter const* p = par->getResolvedParameter();
			if (p != 0) {
				fIsPointerLValue = p->getType()->isNonNullPointer();
				setResultType(p->getType());
			}
			else {
				setResolved(false);
			}
		}
		else if (OverloadedProcedure const* ovp =
			dynamic_cast<OverloadedProcedure const*>(&decl))
		{
			//debug.line("NamedItem is an overloaded procedure");
			
			if (!isMarkedAsCalled()) { // implicit function call
				fIsCall = true;
				fHasResult = true; // we assume it returns a result for now
				resolveOK &= resolveProcedureCall(ovp);
			}
			else {
				// procedure reference is not an L-Value
				fIsClassLValue = false;
				fIsPointerLValue = false;
			}
		}
		else {
			hyp_assert_unreachable();
		}
		
		return resolveOK;
	}
	
	ClassContext const& NamedItemExpression::getClassContext() const {
		return *fCC;
	}
	
	bool NamedItemExpression::resolveField(FieldDecl const* f) {
		InstanceProvider const* instance = getOwnerInstance();
		InstanceType whatInstance = instance->whatInstance();
		
		bool resolveOK = true;
		
		if (whatInstance == kNoInstance && !f->getField()->isStatic()) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"cannot access non-static field %1 in expression %2 "
					"without an object"
				),
				quote(f->getDeclarationName()),
				quote(toString())
			);
			resolveOK = false;
		}
		
		Type const* fieldType = f->getField()->getType().getResolvedType();
		
		fIsClassLValue = true;
		fIsPointerLValue = fieldType ? fieldType->isNonNullPointer() : true;
		fIsCall = false;
		fHasResult = true;
		
		if (fieldType) {
			if (whatInstance == kConstInstance) {
				// Class instance is const, so we pretend the field has a const
				// type.
				setResultType(fieldType->getConstType());
			}
			else {
				setResultType(fieldType);
			}
		}
		else {
			setResolved(false);
		}
		
		return resolveOK;
	}
	
	bool
	NamedItemExpression::resolveProcedureCall(OverloadedProcedure const* ovp)
	{
		//debug.line("NamedItemExpression::resolveProcedureCall");
		std::vector<Expression const*> args; // empty arguments list
		
		// the context from where the procedure is called
		const InstanceProvider* instance = getOwnerInstance();
		
		typedef ProcedureOverloadResolution POR;
		
		POR::Result call =
			POR::resolveCall(instance, ovp, args, getCompileInfo());
		
		switch (call.getResult()) {
		case POR::Result::kNotResolved:
		case POR::Result::kMultipleCandidates:
			hyp_assert_unreachable();
			
		case POR::Result::kNoCandidates:
			setResolved(false);
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no match for implicit function call %1"
				),
				quote(POR::getCallPrototype(ovp->getFullName(), args))
			);
			
			return false;
			
		case POR::Result::kSingleCandidate:
			break; // handled below the switch statement
		}
		
		// the procedure that is called
		fProc = call.getCandidates()[0]->getEntity();
		
		bool resolveOK = true;
		
		InstanceType whatInstance = instance->whatInstance();
		
		//  (about the same code as in functioncallexpression)
		if (whatInstance == kConstInstance && !fProc->isConst()) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"non-const procedure %1 called on const object "
					"of type %2"
				),
				quote(ovp->getFullName().toString()),
				quote(
					UTF8STR("const ")
						+ instance->getClass()->getTypeName()
				)
			);
			
			resolveOK = false;
		}
		else if (whatInstance == kNoInstance && fProc->getASS() != kStatic)
		{
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"non-static procedure %1 is called without an "
					"object in expression %2"
				),
				quote(ovp->getFullName().toString()),
				quote(toString())
			).addInfo(
				UTF8STR("declaration is: ") + fProc->headerToString(),
				fProc->getLocation()
			);
			
			resolveOK = false;
		}
		
		if (fProc->hasReturnType()) {
			Type const* returnType = fProc->getReturnType();
			
			if (returnType != 0) {
				// The result can be used as an LValue for classes if it
				// is a pointer.
				fIsClassLValue = returnType->isNonNullPointer();
				fIsPointerLValue = returnType->isDoublePointer();
				setResultType(returnType);
			}
			else {
				// dummy values to suppress potentially wrong errors
				fIsClassLValue = true;
				fIsPointerLValue = true;
			}
		}
		else {
			// no resulting expression
			fHasResult = false;
		}
		
		return resolveOK;
	}
	
	TriBool NamedItemExpression::isClassLValue() const {
		return fIsClassLValue;
	}
	
	TriBool NamedItemExpression::isPointerLValue() const {
		return fIsPointerLValue;
	}
	
	TriBool NamedItemExpression::isCall() const {
		return fIsCall;
	}
	
	TriBool NamedItemExpression::hasResult() const {
		return fHasResult;
	}
	
	TriBool NamedItemExpression::resultAtCompileTime() const {
		return false; // FIXME
	}
}
