/*!
	\file namespace.cxx
	\brief Implements the Namespace class.
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-05-16
// Last time modified:	2008-06-26

#include "namespace.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "class.h"
#include "compileinfo.h"

namespace HYPER_Compiler {
	
	Namespace::Namespace(QualifiedName fullName, Location const* loc,
		CompileInfo& info)
	 : fShortName(fullName.getShortName()),
		fFullName(fullName), fLoc(loc), fCompileInfo(info)
	{
		// no action
	}
	
	HYPER_Common::UTF8String Namespace::getShortName() const {
		return fShortName;
	}
	
	QualifiedName Namespace::getFullName() const {
		return fFullName;
	}
	
	const Location* Namespace::getLocation() const {
		return fLoc;
	}
	
	Namespace*
	Namespace::create(QualifiedName name, Location const* loc,
		CompileInfo& info)
	{
		// create namespace for first name part
		Namespace* ns = 0;
		NamedDeclaredItem* item = info.getRootScope().findLocalSymbol(name[0]);
		if (item != 0) {
			ns = dynamic_cast<Namespace*>(item);
		}
		if (ns == 0) {
			ns = new Namespace(name[0], loc, info);
		}
		info.getRootScope().registerSymbol(ns);
		ns->putInSymbolTable(info.getRootScope());
		
		// create namespace for all remaining name parts
		for (unsigned i = 1; i < name.countNames(); ++i) {
			HYPER_Common::UTF8String shortN = name[i];
			item = ns->findLocalSymbol(shortN);
			Namespace* childNS = 0;
			if (item != 0) {
				childNS = dynamic_cast<Namespace*>(item);
			}
			if (childNS == 0) {
				childNS = new Namespace(name.subset(i + 1), loc, info);
			}
			ns->registerChild(childNS);
			ns = childNS;
		}
		
		// return most deeply nested namespace
		return ns;
	}
	
	void Namespace::putInSymbolTable(Scope& parent) {
		registerInParent(parent);
	}
	
	void Namespace::registerChild(Namespace* ns) {
		registerSymbol(ns);
		ns->putInSymbolTable(*this);
	}
	
	HYPER_Common::UTF8String Namespace::getDescription() const {
		return UTF8STR("namespace %1").compose(quote(getFullName().toString()));
	}
	
	Class const* Namespace::getOwnerClass() const {
		return 0;
	}
	
	ClassMemberAccess Namespace::getAccessInOwner() const {
		hyp_assert_unreachable();
	}
	
	HYPER_Common::UTF8String Namespace::getDeclarationName() const {
		return getShortName();
	}
	
	void Namespace::registerClass(Class* c) {
		//debug.line("Namespace::registerClass called");
		registerSymbolWithDiagnostic(c, fCompileInfo);
		c->putInSymbolTable(*this);
	}
}
