/*!
	\file newoperatorexpression.cxx
	\brief Implements the NewOperatorExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-10
// Last time modified:	2009-08-23

#include "newoperatorexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "class.h"
#include "compileinfo.h"
#include "constructor.h"
#include "declaredtype.h"
#include "overloadedconstructor.h"
#include "overloadresolution.h"
#include "type.h"

#include <algorithm>

//~ #include "base/debug.h"

namespace HYPER_Compiler {
	
	NewOperatorExpression::NewOperatorExpression(DeclaredType* type,
		std::vector<Expression*> args, const Location* loc, CompileInfo& info)
	 : Expression(info, loc),
		fType(type), fArgs(args)
	{
		// nothing here
	}
	
	NewOperatorExpression::~NewOperatorExpression() {
		// nothing here
	}
	
	HYPER_Common::UTF8String NewOperatorExpression::toString() const {
		HYPER_Common::UTF8StringBuilder sb;
		sb += UTF8STR("new ");
		sb += fType->toString();
		sb += UTF8STR("(");
		
		for (unsigned int i = 0; i < fArgs.size(); ++i) {
			if (i > 0) {
				sb += UTF8STR(", ");
			}
			sb += fArgs[i]->toString();
		}
		
		sb += UTF8STR(")");
		
		return sb.toString();
	}
	
	//~ Type const& NewOperatorExpression::getType() const {
		//~ return *fType;
	//~ }
	
	void NewOperatorExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		fType->putInSymbolTable(loc);
		for (unsigned int i = 0; i < fArgs.size(); ++i) {
			fArgs[i]->putInSymbolTable(loc);
		}
	}
	
	bool NewOperatorExpression::doFullResolve(Trace const& trace) {
		bool argsOK = true;
		bool resolveOK = true;
		
		fType->setRestrictions(true, false);
		resolveOK &= fType->resolveFull(trace);
		
		for (unsigned int i = 0; i < fArgs.size(); ++i) {
			resolveOK &= fArgs[i]->resolveFull(trace);
			argsOK &= fArgs[i]->checkReturnsValue();
		}
		
		Type const* resolvedType = fType->getResolvedType();
		
		if (resolvedType == 0 || !argsOK) {
			// can't check
			return false;
		}
		
		if (!fType->isClass()) {
			// type is not a class, so it must be some pointer type
			getCompileInfo().diag().error(
				UTF8STR(
					"creating a pointer to a pointer with a %1 expression "
						"is not allowed"
				).compose(QUOTED_UTF8STR("new")),
				getLocation()
			);
			return false;
		}
		
		setResultType(Type::createSinglePointer(false, resolvedType));
		setResolved();
		
		Class const* cls = resolvedType->getClass();
		if (cls->getASS() == kStatic) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR("class %1 is static and cannot be instantiated"),
				quote(cls->getTypeName())
			);
			return false;
		}
		else if (cls->getASS() == kAbstract) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR("class %1 is abstract and cannot be instantiated"),
				quote(cls->getTypeName())
			);
			return false;
		}
		else {
			// OK
		}
		
		OverloadedConstructor const* ovc = cls->getConstructors();
		
		// copy arguments list (need a vector of pointers to const Expressions)
		std::vector<Expression const*> args;
		args.resize(fArgs.size());
		std::copy(fArgs.begin(), fArgs.end(), args.begin());
		
		typedef OverloadResolution<OverloadedConstructor> ORC;
		
		// do overload resolution for constructor call
		ORC::Result call = ORC::resolveCall(ovc, args, getCompileInfo());
		
		switch (call.getResult()) {
		case ORC::Result::kNotResolved:
			return false;
			
		case ORC::Result::kNoCandidates:
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no match for constructor call %1, which is needed "
						"for %2"
				),
				quote(
					ORC::getCallPrototype(QualifiedName(UTF8STR("new")), args)
				),
				quote(toString())
			);
			
			return false;
			
		case ORC::Result::kMultipleCandidates: {
			EasyDiagnosticRef e =
				getCompileInfo().diag().error(
					getLocation(),
					UTF8STR(
						"multiple matches for constructor call %1, which is "
							"needed for %2"
					),
					quote(
						ORC::getCallPrototype(
							QualifiedName(UTF8STR("new")), args
						)
					),
					quote(toString())
				);
			
			call.listCandidates(e);
			
			return false; }
		
		case ORC::Result::kSingleCandidate:
			break; // handled below the switch statement
		}
		
		// TODO
		return resolveOK;
	}
	
	TriBool NewOperatorExpression::isClassLValue() const {
		return false;
	}
	
	TriBool NewOperatorExpression::isPointerLValue() const {
		return false;
	}
	
	TriBool NewOperatorExpression::isCall() const {
		return false;
	}
	
	TriBool NewOperatorExpression::hasResult() const {
		return true;
	}
	
	TriBool NewOperatorExpression::resultAtCompileTime() const {
		return false; // FIXME
	}
}
