/*!
	\file operatorcalls.cxx
	\brief Implements the OperatorCalls class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-09-06
// Last time modified:	2008-10-19

#include "operatorcalls.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "assignmentoperator.h"
#include "binaryoperator.h"
#include "class.h"
#include "classrepository.h"
#include "compileinfo.h"
#include "crementoperator.h"
#include "expression.h"
#include "overloadedassignmentoperator.h"
#include "overloadedbinaryoperator.h"
#include "simpleparameter.h"
#include "type.h"

//~ #include "base/debug.h"

using std::vector;

namespace HYPER_Compiler {
	
	// *** class BinaryOperatorCandidate ***
	
	BinaryOperatorCandidate::BinaryOperatorCandidate(BinaryOperator* o)
	 : fOperator(o), fIsBestCandidate(false), fApp(kDontKnow), fUse(kAnyUse)
	{
		// no action here
	}
	
	BinaryOperator* BinaryOperatorCandidate::getOperator() const {
		return fOperator;
	}
	
	BinaryOperatorCandidate::Applicability
	BinaryOperatorCandidate::getApplicability() const {
		return fApp;
	}
	
	void BinaryOperatorCandidate::setApplicability(
		ValuePassing left, ValuePassing right)
	{
		hyp_assert (!isCommutative());
		
		fLeftOK = std::auto_ptr<ValuePassing>(new ValuePassing(left));
		fRightOK = std::auto_ptr<ValuePassing>(new ValuePassing(right));
		
		if (left.isPossible() && right.isPossible()) {
			fApp = kOnlyNormal;
		}
		else {
			fApp = kNotApplicable;
		}
	}
	
	void BinaryOperatorCandidate::setApplicability(
		ValuePassing left, ValuePassing right,
		ValuePassing leftRev, ValuePassing rightRev)
	{
		hyp_assert (isCommutative());
		
		fLeftOK = std::auto_ptr<ValuePassing>(new ValuePassing(left));
		fRightOK = std::auto_ptr<ValuePassing>(new ValuePassing(right));
		fLeftOKReversed =
			std::auto_ptr<ValuePassing>(new ValuePassing(leftRev));
		fRightOKReversed =
			std::auto_ptr<ValuePassing>(new ValuePassing(rightRev));
		
		if (left.isPossible() && right.isPossible()) {
			if (leftRev.isPossible() && rightRev.isPossible()) {
				fApp = kBoth;
			}
			else {
				fApp = kOnlyNormal;
			}
		}
		else {
			if (leftRev.isPossible() && rightRev.isPossible()) {
				fApp = kOnlyReversed;
			}
			else {
				fApp = kNotApplicable;
			}
		}
	}
	
	bool BinaryOperatorCandidate::isApplicableInAnyWay() const {
		hyp_assert(getApplicability() != kDontKnow);
		return fApp != kNotApplicable;
	}
	
	bool BinaryOperatorCandidate::isNormalApplicable() const {
		return fApp == kOnlyNormal || fApp == kBoth;
	}
	
	bool BinaryOperatorCandidate::isReversedApplicable() const {
		return fApp == kOnlyReversed || fApp == kBoth;
	}
	
	bool BinaryOperatorCandidate::isCommutative() const {
		return fOperator->isCommutative();
	}
	
	BinaryOperatorCandidate::BetterCandidate
	BinaryOperatorCandidate::compareTo(BinaryOperatorCandidate const& c) const
	{
		// the int's in this function mean:
		//		negative int :	this candidate is better
		//		zero int :	both candidates are similar
		//		positive int :	the other candidate is better
		
		// current means the current best candidate
		int current = 0;
		
		// normal means the usability of my normal sequence
		// 		negative :	don't use it
		//		zero : 		don't care
		// 		positive : 	must use it
		int normal = 0;
		
		// reversed means the usability of my reversed sequence
		// 		negative :	don't use it
		//		zero : 		don't care
		// 		positive : 	must use it
		int reversed = 0;
		
		// like 'normal' and 'reversed' but now for the other candidate
		int normal2 = 0;
		int reversed2 = 0;
		
		if (isNormalApplicable() && c.isNormalApplicable()) {
			int left = fLeftOK->compareTo(*c.fLeftOK);
			int right = fRightOK->compareTo(*c.fRightOK);
			int bestNormalNormal = left + right;
			
			// current is now simply the first subresult
			current = bestNormalNormal;
			
			if (bestNormalNormal > 0) { // if the other is better ...
				normal = -1;
				if (normal2 >= 0) normal2 = 1;
			}
			else if (bestNormalNormal < 0) { // if this one is better ...
				if (normal >= 0) normal = 1;
				normal2 = -1;
			}
		}
		
		if (isNormalApplicable() && c.isReversedApplicable()) {
			int left = fLeftOK->compareTo(*c.fLeftOKReversed);
			int right = fRightOK->compareTo(*c.fRightOKReversed);
			int bestNormalRev = left + right;
			
			if (current == 0) { // if no better one yet ...
				// current is now this subresult
				current = bestNormalRev;
			}
			else if (current * bestNormalRev < 0) { // if signs are different ...
				return kNoBetter; // there is no best candidate
			}
			
			if (bestNormalRev > 0) { // if the other is better ...
				normal = -1;
				if (reversed2 >= 0) reversed2 = 1;
			}
			else if (bestNormalRev < 0 && normal >= 0) { // if this one is better ...
				if (normal >= 0) normal = 1;
				reversed2 = -1;
			}
		}
		
		if (isReversedApplicable() && c.isNormalApplicable()) {
			int left = fLeftOKReversed->compareTo(*c.fLeftOK);
			int right = fRightOKReversed->compareTo(*c.fRightOK);
			int bestRevNormal = left + right;
			
			if (current == 0) { // if no better one yet ...
				// current is now this subresult
				current = bestRevNormal;
			}
			else if (current * bestRevNormal < 0) { // if signs are different ...
				return kNoBetter; // there is no best candidate
			}
			
			if (bestRevNormal > 0) { // if the other is better ...
				reversed = -1;
				if (normal2 >= 0) normal2 = 1;
			}
			else if (bestRevNormal < 0) { // if this one is better ...
				if (reversed >= 0) reversed = 1;
				normal2 = -1;
			}
		}
		
		if (isReversedApplicable() && c.isReversedApplicable()) {
			int left = fLeftOKReversed->compareTo(*c.fLeftOKReversed);
			int right = fRightOKReversed->compareTo(*c.fRightOKReversed);
			int bestRevRev = left + right;
			
			if (current == 0) { // if no better one yet ...
				// current is now this subresult
				current = bestRevRev;
			}
			else if (current * bestRevRev < 0) { // if signs are different ...
				return kNoBetter; // there is no best candidate
			}
			
			if (bestRevRev > 0) { // if the other is better ...
				reversed = -1;
				if (reversed2 >= 0) reversed2 = 1;
			}
			else if (bestRevRev < 0) { // if this one is better ...
				if (reversed >= 0) reversed = 1;
				reversed2 = -1;
			}
		}
		
		if (current == 0) {
			return kNoBetter;
		}
		else if (current < 0) {
			if (normal < 0 || (normal == 0 && reversed > 0)) {
				return kFirstReversedBetter;
			}
			else {
				return kFirstNormalBetter;
			}
		}
		else {
			if (normal2 < 0 || (normal2 == 0 && reversed2 > 0)) {
				return kSecondReversedBetter;
			}
			else {
				return kSecondNormalBetter;
			}
		}
	}
	
	void BinaryOperatorCandidate::setUse(Use u) {
		fUse = u;
	}
	
	BinaryOperatorCandidate::Use
	BinaryOperatorCandidate::getUse() const {
		return fUse;
	}
	
	
	// *** class AssignmentOperatorCandidate ***
	
	AssignmentOperatorCandidate::AssignmentOperatorCandidate(
		AssignmentOperator* op)
	 : fOperator(op)
	{
		// no action
	}
	
	AssignmentOperator*
	AssignmentOperatorCandidate::getOperator() const {
		return fOperator;
	}
	
	void
	AssignmentOperatorCandidate::setApplicability(ValuePassing vp)
	{
		fVP = std::auto_ptr<ValuePassing>(new ValuePassing(vp));
	}
	
	bool AssignmentOperatorCandidate::isApplicable() const {
		return fVP->isPossible();
	}
	
	ValuePassing const&
	AssignmentOperatorCandidate::getUsedSubstitution() const
	{
		return *fVP;
	}
	
	
	// *** class OperatorCalls ***
	
	std::vector<BinaryOperatorCandidate const*>
	OperatorCalls::resolveBinaryOperatorExpression(
		const Type* left, const Type* right,
		Operators::BinaryOperator op, CompileInfo& info)
	{
		// pointer operators are not handled here
		hyp_assert(op != Operators::kPointerEquals);
		hyp_assert(op != Operators::kPointerDiffers);
		
		// get operand classes
		const Class* leftClass = left->getDereferenced()->getClass();
		const Class* rightClass = right->getDereferenced()->getClass();
		bool rightAndLeftSameClass = (leftClass == rightClass);
		
		// OVERLOADING PHASE 1 : SEARCH FOR EXACT PARAMETER MATCHES
		
		// get binary operators supported by the left and right types
		Class::BinOpsMap const& leftClassOpMap =
			leftClass->getBinaryOperators();
		Class::BinOpsMap const& rightClassOpMap =
			rightClass->getBinaryOperators();
		
		// get the operators for the operator symbol (for example: '+') we need
		Class::BinOpsMap::const_iterator leftOpIt = leftClassOpMap.find(op);
		Class::BinOpsMap::const_iterator rightOpIt = rightClassOpMap.find(op);
		OverloadedBinaryOperator* leftOps =
			(leftOpIt != leftClassOpMap.end()) ? leftOpIt->second : 0;
		OverloadedBinaryOperator* rightOps =
			(rightOpIt != rightClassOpMap.end()) ? rightOpIt->second : 0;
		
		// do we have at least one candidate (without argument type check) ?
		if (leftOps != 0 || rightOps != 0) {
			// make a list with candidates
			vector<BinaryOperatorCandidate*> candidates;
			
			if (leftOps != 0) {
				for (unsigned int i = 0; i < leftOps->getOverloads().size(); ++i)
				{
					candidates.push_back(
						new BinaryOperatorCandidate(leftOps->getOverloads()[i])
					);
				}
			}
			
			if (rightOps != 0 && !rightAndLeftSameClass) {
				for (unsigned int i = 0; i < rightOps->getOverloads().size(); ++i)
				{
					candidates.push_back(
						new BinaryOperatorCandidate(rightOps->getOverloads()[i])
					);
				}
			}
			
			refineOverloadCandidates(
				left, right, candidates, Conversions::kNoneAllowed, info
			);
			
			if (!candidates.empty()) {
				return copyConstList(candidates);
			}
			// else go on...
		}
		else { // no candidates at all ...
			// OK, then we'll go to overloading phase 2
		}
		
		// OVERLOADING PHASE 2 : ALLOW DERIVED CLASS TO BASE CLASS CONVERSIONS
		
		// get all operators
		Class::BinOpsMap const& allBinOps =
			info.getClassRepository().getBinaryOperators();
		
		// get the operator symbol we need
		Class::BinOpsMap::const_iterator allOpsIt = allBinOps.find(op);
		OverloadedBinaryOperator* ops =
			(allOpsIt != allBinOps.end()) ? allOpsIt->second : 0;
		
		// do we have at least one candidate (without argument type check) ?
		if (ops != 0) {
			vector<BinaryOperatorCandidate*> candidates;
			for (unsigned int i = 0; i < ops->getOverloads().size(); ++i) {
				candidates.push_back(
					new BinaryOperatorCandidate(ops->getOverloads()[i])
				);
			}
			
			refineOverloadCandidates(
				left, right, candidates, Conversions::kLevel1Allowed, info
			);
			
			if (!candidates.empty()) {
				// one or more candidates available
				return copyConstList(candidates);
			}
			// else go on...
		}
		else { // no candidates at all ...
			// Dont' go to next phase, because the next phase will also have no candidates.
			vector<BinaryOperatorCandidate const*> candidates;
			return candidates;
		}
		
		// OVERLOADING PHASE 3 : ALLOW USER CONVERSIONS TOO
		
		// we should have candidates, because if not phase 2 would have exited
		hyp_assert(ops != 0);
		
		vector<BinaryOperatorCandidate*> candidates;
		for (unsigned int i = 0; i < ops->getOverloads().size(); ++i) {
			candidates.push_back(
				new BinaryOperatorCandidate(ops->getOverloads()[i])
			);
		}
		
		refineOverloadCandidates(
			left, right, candidates, Conversions::kLevel2Allowed, info
		);
		
		return copyConstList(candidates);
	}
	
	OperatorCalls::CrementCallResult
	OperatorCalls::resolveCrementOperator(const Expression& operand,
		Operators::CrementOperator op)
	{
		CrementCallResult result;
		
		if (operand.isClassLValue().isFalse()) {
			result.fgErrorLValue = true;
			return result;
		}
		
		// type unknown
		if (!operand.getResult().isResolved()) {
			// cannot determine operator -> no result
			return result;
		}
		
		return resolveCrementOperator(operand.getResult().getType(), op);
	}
	
	OperatorCalls::CrementCallResult
	OperatorCalls::resolveCrementOperator(const Type* operand,
		Operators::CrementOperator op)
	{
		CrementCallResult result;
		
		// get pointer base type if needed
		const Type* t = operand->getDereferenced();
		
		if (t->isConst()) {
			result.fgErrorConst = true;
			return result;
		}
		
		hyp_assert(t->isClass());
		
		const Class* c = t->getClass();
		hyp_assert(c != 0);
		
		Class::CremOpsMap const& classOpMap = c->getCrementOperators();
		Class::CremOpsMap::const_iterator it = classOpMap.find(op);
		
		if (it == classOpMap.end()) { // no operator in class...
			result.fgErrorNoOperator = true;
			return result;
		}
		
		// class has the operator...
		
		result.fgOperator = it->second;
		return result;
	}
	
	OperatorCalls::AssignmentCallResult
	OperatorCalls::resolveAssignmentOperator(
		const Expression& left, const Type* right,
		Operators::AssignmentOperator op, CompileInfo& info)
	{
		AssignmentCallResult result;
		
		if (left.isClassLValue().isFalse()) {
			result.fgErrorLValue = true;
			return result;
		}
		
		// type unknown
		if (!left.getResult().isResolved()) {
			// cannot determine operator -> no result
			return result;
		}
		
		return resolveAssignmentOperator(
			left.getResult().getType(), right, op, info
		);
	}
	
	OperatorCalls::AssignmentCallResult
	OperatorCalls::resolveAssignmentOperator(
		const Type* left, const Type* right,
		Operators::AssignmentOperator op, CompileInfo& info)
	{
		//~ debug.line(UTF8STR("resolving assignment operator..."));
		
		AssignmentCallResult result;
		
		Type const* leftT = left->getDereferenced();
		
		if (leftT->isConst()) {
			result.fgErrorConst = true;
			return result;
		}
		
		hyp_assert(leftT->isClass());
		
		const Class* leftC = leftT->getClass();
		hyp_assert(leftC != 0);
		
		Class::AssignOpsMap const& leftClassOpMap =
			leftC->getAssignmentOperators();
		Class::AssignOpsMap::const_iterator leftOpIt =
			leftClassOpMap.find(op);
		OverloadedAssignmentOperator* leftOps =
			(leftOpIt != leftClassOpMap.end()) ? leftOpIt->second : 0;
		
		if (leftOps == 0) {
			result.fgErrorNoOperator = true;
			return result; // abort overload resolution
		}
		
		vector<AssignmentOperatorCandidate*> candidates;
		for (unsigned int i = 0; i < leftOps->getOverloads().size(); ++i) {
			//debug.line("add assignment-op. candidate");
			candidates.push_back(
				new AssignmentOperatorCandidate(leftOps->getOverloads()[i])
			);
		}
		
		// OVERLOADING PHASE 1 : SEARCH FOR EXACT PARAMETER MATCHES
		
		//~ debug.line(UTF8STR("assign-op; OVERLOAD PHASE 1"));
		
		vector<AssignmentOperatorCandidate*> cands1 = candidates;
		refineOverloadCandidates(
			right, cands1, Conversions::kNoneAllowed, info
		);
		
		if (cands1.empty()) { // no candidates left ...
			// OK, then we'll go to overloading phase 2
		}
		else if (cands1.size() == 1) { // exact one candidate ...
			// OK, we found our overload
			result.fgOperators.push_back(cands1[0]);
			return result;
		}
		else { // ambiguous overload...
			result.fgErrorMultipleCandidates = true;
			result.fgOperators = copyConstList(cands1);
			return result;
		}
		
		// OVERLOADING PHASE 2 : ALLOW DERIVED CLASS TO BASE CLASS CONVERSIONS
		
		//~ debug.line(UTF8STR("assign-op; OVERLOAD PHASE 2"));
		
		vector<AssignmentOperatorCandidate*> cands2 = candidates;
		refineOverloadCandidates(
			right, cands2, Conversions::kLevel1Allowed, info
		);
		
		if (cands2.empty()) { // no candidates left ...
			// OK, then we'll go to overloading phase 3
		}
		else if (cands2.size() == 1) { // exact one candidate ...
			// OK, we found our overload
			result.fgOperators.push_back(cands2[0]);
			return result;
		}
		else { // ambiguous overload...
			result.fgErrorMultipleCandidates = true;
			result.fgOperators = copyConstList(cands2);
			return result;
		}
		
		// OVERLOADING PHASE 3 : ALLOW USER CONVERSIONS TOO
		
		//~ debug.line(UTF8STR("assign-op; OVERLOAD PHASE 3"));
		
		vector<AssignmentOperatorCandidate*> cands3 = candidates;
		refineOverloadCandidates(
			right, cands3, Conversions::kLevel2Allowed, info
		);
		
		if (cands3.empty()) { // no candidates left ...
			result.fgErrorNoCandidates = true;
			return result;
		}
		else if (cands3.size() == 1) { // exact one candidate ...
			// OK, we found our overload
			result.fgOperators.push_back(cands3[0]);
			return result;
		}
		else { // ambiguous overload...
			result.fgErrorMultipleCandidates = true;
			result.fgOperators = copyConstList(cands3);
			return result;
		}
	}
	
	void
	OperatorCalls::listCandidates(
		const std::vector<BinaryOperatorCandidate const*>& cands,
		EasyDiagnosticRef const& d)
	{
		for (unsigned int i = 0; i < cands.size(); ++i) {
			BinaryOperator const * binOp = cands[i]->getOperator();
			d.addInfo(
				binOp->getLocation(),
				UTF8STR("candidate %1 : %2"),
				i + 1,
				binOp->toString()
			);
		}
	}
	
	void
	OperatorCalls::listCandidates(
		const Type* left,
		const std::vector<AssignmentOperatorCandidate const*>& cands,
		EasyDiagnosticRef const& d)
	{
		for (unsigned int i = 0; i < cands.size(); ++i) {
			AssignmentOperator const* assOp = cands[i]->getOperator();
			d.addInfo(
				assOp->getLocation(),
				UTF8STR("candidate %1 : %2.assignment%3(%4)"),
				i + 1,
				left->toString(),
				Operators::toString(assOp->getOperator()),
				assOp->getRValueType()->toString()
			);
		}
	}
	
	void OperatorCalls::refineOverloadCandidates(
		const Type* left, const Type* right,
		vector<BinaryOperatorCandidate*>& list,
		Conversions::AllowedLevel a,
		CompileInfo& info)
	{
		// FIRST, determine applicability and then delete all candidates that
		// cannot be used in any way.
		
		unsigned int i = 0;
		while (i < list.size()) {
			vector<BinaryOperatorCandidate*>::iterator it = list.begin() + i;
			BinaryOperatorCandidate& c = **it;
			BinaryOperator const& bo = *(c.getOperator());
			
			SimpleParameter boLeftParam(bo.getLeftOperandType());
			SimpleParameter boRightParam(bo.getRightOperandType());
			
			ValuePassing leftOk = ValuePassing::check(left, boLeftParam, a);
			ValuePassing rightOk = ValuePassing::check(right, boRightParam, a);
			
			// if the candidate is commutative ...
			if (c.isCommutative()) {
				ValuePassing leftOkRev =
					ValuePassing::check(left, boRightParam, a);
				ValuePassing rightOkRev =
					ValuePassing::check(right, boLeftParam, a);
				
				c.setApplicability(leftOk, rightOk, leftOkRev, rightOkRev);
			}
			else { // not commutative ...
				c.setApplicability(leftOk, rightOk);
			}
			
			if (c.isApplicableInAnyWay()) {
				// accepted
				++i;
			}
			else {
				// rejected
				list.erase(it);
				// ... and don't increment i here
			}
		}
		
		if (list.size() < 2) {
			// return, because we don't have multiple candidates
			return;
		}
		
		// SECOND, get the best candidates
		
		for (i = 0; i < list.size() - 1; ++i) {
			for (unsigned int j = i + 1; j < list.size(); ++j) {
				if (list[i] == 0 || list[j] == 0) {
					continue;
				}
				
				BinaryOperatorCandidate::BetterCandidate bc =
					list[i]->compareTo(*list[j]);
				
				switch (bc) {
				case BinaryOperatorCandidate::kNoBetter:
					// no action
					break;
				case BinaryOperatorCandidate::kFirstNormalBetter:
					list[i]->setUse(BinaryOperatorCandidate::kOnlyNormalUse);
					list[j] = 0;  // get rid of inferior candidate
					break;
				case BinaryOperatorCandidate::kFirstReversedBetter:
					list[i]->setUse(BinaryOperatorCandidate::kOnlyReversedUse);
					list[j] = 0;  // get rid of inferior candidate
					break;
				case BinaryOperatorCandidate::kSecondNormalBetter:
					list[j]->setUse(BinaryOperatorCandidate::kOnlyNormalUse);
					list[i] = 0;  // get rid of inferior candidate
					break;
				case BinaryOperatorCandidate::kSecondReversedBetter:
					list[j]->setUse(BinaryOperatorCandidate::kOnlyReversedUse);
					list[i] = 0;  // get rid of inferior candidate
					break;
				default:
					hyp_assert_unreachable();
				}
			}
		}
		
		// THIRD, move all candidates to the beginning of the list
		
		i = 0;
		while (i < list.size() - 1) {
			if (list[i] == 0) {
				for (unsigned int j = i + 1; j < list.size(); ++j) {
					if (list[j] != 0) {
						list[i] = list[j];
						list[j] = 0;
						i = j;
						break;
					}
				}
			}
			
			++i;
		}
		
		// FOURTH, shrink the list; pop off the zero pointers at the end
		
		while (list.back() == 0) {
			list.pop_back();
		}
	}
	
	void OperatorCalls::refineOverloadCandidates(
		const Type* rightT, vector<AssignmentOperatorCandidate*>& list,
		Conversions::AllowedLevel a, CompileInfo& info)
	{
		unsigned int i = 0;
		while (i < list.size()) {
			vector<AssignmentOperatorCandidate*>::iterator it =
				list.begin() + i;
			AssignmentOperatorCandidate* cand = *it;
			AssignmentOperator const& assOp = *cand->getOperator();
			
			SimpleParameter assOpParam(assOp.getRValueType());
			ValuePassing ok = ValuePassing::check(rightT, assOpParam, a);
			
			if (ok.isPossible()) {
				cand->setApplicability(ok);
				++i; // check next
			}
			else {
				list.erase(it);
			}
		}
	}
	
	std::vector<BinaryOperatorCandidate const*>
	OperatorCalls::copyConstList(std::vector<BinaryOperatorCandidate*> list)
	{
		std::vector<BinaryOperatorCandidate const*> cands;
		for (unsigned int i = 0; i < list.size(); ++i) {
			cands.push_back(list[i]);
		}
		return cands;
	}
	
	std::vector<AssignmentOperatorCandidate const*>
	OperatorCalls::copyConstList(
		std::vector<AssignmentOperatorCandidate*> list)
	{
		std::vector<AssignmentOperatorCandidate const*> cands;
		for (unsigned int i = 0; i < list.size(); ++i) {
			cands.push_back(list[i]);
		}
		return cands;
	}
}
