/*!
	\file operatorcalls.h
	\brief Contains the OperatorCalls class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-09-06
// Last time modified:	2008-10-11

#ifndef DEF_OPERATORCALLS_H
 #define DEF_OPERATORCALLS_H
 
 #include "conversions.h" // for Convertable etc.
 #include "important_fwdecl.h"
 #include "operators.h"
 
 #include <memory>
 #include <vector>

namespace HYPER_Compiler {
	
	class AssignmentOperator;
	class BinaryOperator;
	class CrementOperator;
	
	/// A candidate binary operator
	class BinaryOperatorCandidate {
	public:
		enum Applicability {
			kDontKnow = 0,
			kNotApplicable,
			kOnlyNormal,
			kOnlyReversed,
			kBoth
		};
		
		enum BetterCandidate {
			kNoBetter = 0,
			kFirstNormalBetter,
			kFirstReversedBetter,
			kSecondNormalBetter,
			kSecondReversedBetter
		};
		
		enum Use {
			kAnyUse = 0,
			kOnlyNormalUse,
			kOnlyReversedUse
		};
		
		BinaryOperatorCandidate(BinaryOperator* o);
		
		BinaryOperator* getOperator() const;
		
		Applicability getApplicability() const;
		void setApplicability(ValuePassing left, ValuePassing right);
		void setApplicability(ValuePassing left, ValuePassing right,
			ValuePassing leftRev, ValuePassing rightRev);
		bool isApplicableInAnyWay() const;
		bool isNormalApplicable() const;
		bool isReversedApplicable() const;
		
		bool isCommutative() const;
		
		BetterCandidate compareTo(BinaryOperatorCandidate const& c) const;
		
		void setUse(Use u);
		Use getUse() const;
		
	private:
		BinaryOperator* fOperator;
		bool fIsBestCandidate;
		Applicability fApp;
		std::auto_ptr<ValuePassing> fLeftOK;
		std::auto_ptr<ValuePassing> fRightOK;
		std::auto_ptr<ValuePassing> fLeftOKReversed;
		std::auto_ptr<ValuePassing> fRightOKReversed;
		Use fUse;
	};
	
	class AssignmentOperatorCandidate {
	public:
		AssignmentOperatorCandidate(AssignmentOperator* op);
		
		AssignmentOperator* getOperator() const;
		
		void setApplicability(ValuePassing vp);
		bool isApplicable() const;
		
		ValuePassing const& getUsedSubstitution() const;
		
	private:
		AssignmentOperator* fOperator;
		std::auto_ptr<ValuePassing> fVP;
	};
	
	/// Class with functions for resolving operator calls
	class OperatorCalls {
	public:
		
		static std::vector<BinaryOperatorCandidate const*>
		resolveBinaryOperatorExpression(const Type* left, const Type* right,
			Operators::BinaryOperator op, CompileInfo& info);
		
		/// Add the list of candidates to an error message
		static void
		listCandidates(const std::vector<BinaryOperatorCandidate const*>& cands,
			EasyDiagnosticRef const& d);
		
		
		class CrementCallResult {
		public:
			bool fgErrorLValue; // true -> expr not an LValue
			bool fgErrorConst; // true -> expr has const type
			bool fgErrorNoOperator; // true -> no such operator
			CrementOperator const* fgOperator;
			
			CrementCallResult()
			 : fgErrorLValue(false), fgErrorConst(false), fgErrorNoOperator(),
				fgOperator(0)
			{
				// empty
			}
		};
		
		static CrementCallResult
		resolveCrementOperator(const Expression& operand,
			Operators::CrementOperator op);
		
		static CrementCallResult
		resolveCrementOperator(const Type* operand,
			Operators::CrementOperator op);
		
		
		class AssignmentCallResult {
		public:
			bool fgErrorLValue; // true -> l-value not a valid l-value
			bool fgErrorConst; // true -> l-value has const type
			bool fgErrorNoOperator; // true -> no such operator in class
			bool fgErrorNoCandidates; // true -> no candidates found
			bool fgErrorMultipleCandidates; // true -> more than 1 possible candidate
			std::vector<AssignmentOperatorCandidate const*> fgOperators;
			
			AssignmentCallResult()
			 : fgErrorLValue(false), fgErrorConst(false), fgErrorNoOperator(false),
				fgErrorNoCandidates(false), fgErrorMultipleCandidates(false)
			{
				// empty
			}
		};
		
		static AssignmentCallResult
		resolveAssignmentOperator(const Expression& left, const Type* right,
			Operators::AssignmentOperator op, CompileInfo& info);
		
		static AssignmentCallResult
		resolveAssignmentOperator(const Type* left, const Type* right,
			Operators::AssignmentOperator op, CompileInfo& info);
		
		/// Add the list of candidates to an error message
		static void
		listCandidates(
			const Type* left,
			const std::vector<AssignmentOperatorCandidate const*>& cands,
			EasyDiagnosticRef const& d);
		
	private:
		
		static void
		refineOverloadCandidates(
			const Type* left, const Type* right,
			std::vector<BinaryOperatorCandidate*>& list,
			Conversions::AllowedLevel a, CompileInfo& info
		);
		
		static void
		refineOverloadCandidates(
			const Type* rValueT,
			std::vector<AssignmentOperatorCandidate*>& list,
			Conversions::AllowedLevel a, CompileInfo& info
		);
		
		static std::vector<BinaryOperatorCandidate const*>
		copyConstList(std::vector<BinaryOperatorCandidate*> list);
		
		static std::vector<AssignmentOperatorCandidate const*>
		copyConstList(std::vector<AssignmentOperatorCandidate*> list);
	};
}
#endif
