/*!
	\file outclass.cxx
	\author Kevin André
	\brief Implements the Out class
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2006-12-20
// Date modified:	2008-10-16

#include "outclass.h"

#include "base/asserts.h"
#include "base/location.h"

#include "builtinprocedure.h"
#include "builtintypes.h"
#include "parameter.h"

namespace HYPER_Compiler {
	
	enum ArgToPrint {
		kNoArgToPrint,
		kStringToPrint,
		kCharToPrint
	};
	
	class Out::ProcedureLine : public BuiltInProcedure {
	public:
		ProcedureLine(Out& o, ArgToPrint arg);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
	};
	
	class Out::ProcedurePrint : public BuiltInProcedure {
	public:
		ProcedurePrint(Out& o, ArgToPrint arg);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
	};
	
	
	// *** Out ***
	
	Out::Out(CompileInfo& info)
	 : BuiltInClass(info, kStatic)
	{
		// no action
	}
	
	Out::~Out() {
		// no action
	}
	
	HYPER_Common::UTF8String Out::getTypeName() const {
		return UTF8STR("system.Out");
	}
	
	HYPER_Common::UTF8String Out::getDeclarationName() const {
		return UTF8STR("Out");
	}
	
	void Out::buildSymbolTable() {
		createMembers();
	}
	
	void Out::createMembers() {
		// create member procedures
		
		registerProcedure(new ProcedureLine(*this, kNoArgToPrint));
		registerProcedure(new ProcedureLine(*this, kStringToPrint));
		registerProcedure(new ProcedureLine(*this, kCharToPrint));
		
		registerProcedure(new ProcedurePrint(*this, kStringToPrint));
		registerProcedure(new ProcedurePrint(*this, kCharToPrint));
	}
	
	
	// *** ProcedureLine ***
	
	Out::ProcedureLine::ProcedureLine(Out& o, ArgToPrint arg)
	 : BuiltInProcedure(
			&o, UTF8STR("line"), kStatic, ClassMemberAccess::createPublic(),
			false, 0
		)
	{
		switch (arg) {
		case kNoArgToPrint:
			break;
		case kStringToPrint:
			fParams.push_back(
				o.createParameter(UTF8STR("s"), BuiltInTypes::kString)
			);
			break;
		case kCharToPrint:
			fParams.push_back(
				o.createParameter(UTF8STR("c"), BuiltInTypes::kChar)
			);
			break;
		}
	}
	
	const std::vector<Parameter const*>*
	Out::ProcedureLine::getParameterList() const {
		return &fParams;
	}
	
	
	// *** ProcedurePrint ***
	
	Out::ProcedurePrint::ProcedurePrint(Out& o, ArgToPrint arg)
	 : BuiltInProcedure(
			&o, UTF8STR("print"), kStatic, ClassMemberAccess::createPublic(),
			false, 0
		)
	{
		switch (arg) {
		case kNoArgToPrint:
			hyp_assert_unreachable();
		case kStringToPrint:
			fParams.push_back(
				o.createParameter(UTF8STR("s"), BuiltInTypes::kString)
			);
			break;
		case kCharToPrint:
			fParams.push_back(
				o.createParameter(UTF8STR("c"), BuiltInTypes::kChar)
			);
			break;
		}
	}
	
	const std::vector<Parameter const*>*
	Out::ProcedurePrint::getParameterList() const {
		return &fParams;
	}
}
