/*!
	\file overloadresolution.cxx
	\brief Implements the OverloadResolution class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-03-26
// Last time modified:	2009-08-22

#ifndef DEF_OVERLOADRESOLUTION_CXX
 #define DEF_OVERLOADRESOLUTION_CXX
 
 #include "overloadresolution.h"
 
 #include "base/asserts.h"
 #include "base/easydiagnostics.h"
 
 #include "accesschecker.h"
 #include "constructor.h"
 #include "expression.h"
 #include "parameter.h"
 #include "type.h"
 
 #include "base/debug.h"
 
namespace HYPER_Compiler {
	
	// *** class Candidate ***
	
	template <class OE>
	OverloadResolution<OE>::Candidate::Candidate(const Entity* e)
	 : fE(e), fApp(kDontKnow), fConv(Conversions::kNoConversion)
	{
		hyp_assert(e != 0);
	}
	
	template <class OE>
	const typename OverloadResolution<OE>::Entity*
	OverloadResolution<OE>::Candidate::getEntity() const {
		return fE;
	}
	
	template <class OE>
	typename OverloadResolution<OE>::Candidate::Applicability
	OverloadResolution<OE>::Candidate::getApplicability() const {
		return fApp;
	}
	
	template <class OE>
	bool
	OverloadResolution<OE>::Candidate::isApplicable() const {
		hyp_assert(fApp != kDontKnow);
		
		return fApp == kApplicable;
	}
	
	template <class OE>
	void
	OverloadResolution<OE>::Candidate::setApplicability(
		const std::vector<ValuePassing>& list)
	{
		fApp = kApplicable;
		
		for (unsigned int i = 0; i < list.size(); ++i) {
			ValuePassing const& vp = list[i];
			
			if (fApp == kNotApplicable) {
				// total applicability is already zero, no adjustment needed
			}
			else if (vp.isPossible()) {
				// adjust maximum conversion level if needed
				Conversions::Kind cv = vp.conversionLevelNeeded();
				if (Conversions::compare(cv, fConv) < 0) {
					fConv = cv;
				}
			}
			else {
				fApp = kNotApplicable;
			}
			
			unsigned int li = fOK.size();
			if (i >= li) {
				fOK.push_back(vp);
			}
			else {
				fOK[i] = vp;
			}
		}
	}
	
	template <class OE>
	Conversions::Kind
	OverloadResolution<OE>::Candidate::getNeededMaximumConversion() const {
		return fConv;
	}
	
	template <class OE>
	std::vector<ValuePassing> const&
	OverloadResolution<OE>::Candidate::getConversions() const
	{
		return fOK;
	}
	
	/*
		Returns:
			-1 : this candidate is better
			0 : both candidates are similar
			1 : the other candidate is better
	*/
	template <class OE>
	int OverloadResolution<OE>::Candidate::compareTo(Candidate const& c) const {
		hyp_assert(fOK.size() == c.fOK.size());
		
		int current = 0;
		
		for (unsigned int i = 0; i < fOK.size(); ++i) {
			int comp =
				Conversions::compare(
					fOK[i].conversionLevelNeeded(),
					c.fOK[i].conversionLevelNeeded()
				);
			
			if (current < 0) {
				if (comp > 0) {
					return 0;
				}
			}
			else if (current > 0) {
				if (comp < 0) {
					return 0;
				}
			}
			else { // current == 0
				if (comp != 0) {
					current = comp;
				}
			}
		}
		
		return current;
	}
	
	
	// *** class Result ***
	
	template <class OE>
	void OverloadResolution<OE>::Result::listCandidates(
		EasyDiagnosticRef& diag) const
	{
		hyp_assert(fgResult == kMultipleCandidates);
		
		for (unsigned int i = 0; i < fgList.size(); ++i) {
			const Entity* ent = fgList[i]->getEntity();
			
			diag.addInfo(
				ent->getLocation(),
				UTF8STR("candidate %1 : %2"),
				i + 1, ent->headerToString()
			);
		}
	}
	
	
	// *** class OverloadResolution ***
	
	template <class OE>
	typename OverloadResolution<OE>::Result
	OverloadResolution<OE>::resolveCall(OE const* list,
		const std::vector<const Expression*>& arguments,
		CompileInfo& info)
	{
		// extract types from argument expression list
		std::vector<const Type*> argTypes;
		for (unsigned int i = 0; i < arguments.size(); ++i) {
			ExpressionResult const& argRes = arguments[i]->getResult();
			hyp_assert(argRes.isResolved());
			argTypes.push_back(argRes.getType());
		}
		
		// forward call to other function
		return resolveCall(list, argTypes, info);
	}
	
	template <class OE>
	typename OverloadResolution<OE>::Result
	OverloadResolution<OE>::resolveCall(OE const* list,
		const std::vector<const Type*>& arguments, CompileInfo& info)
	{
		const std::vector<Entity*>& overloads = list->getOverloadList();
		
		if (overloads.empty()) {
			// no candidates to start with
			return Result(std::vector<Candidate*>());
		}
		
		// generate a list of usable candidates
		
		std::vector<Candidate*> fullList;
		//~ debug.line(UTF8STR("OverloadResolution: generating full list"));
		
		for (unsigned int i = 0; i < overloads.size(); ++i) {
			Entity const* e = overloads[i];
			//~ debug.line(UTF8STR(" Entity %1").compose(i + 1));
			
			if (e->isDuplicate()) {
				continue; // don't consider duplicates as candidate
			}
			
			if (e->getParameterCount() != arguments.size()) {
				continue; // parameter count doesn't match
			}
			
			if (e->getParameterList() == 0) {
				// unresolved parameters
				//~ debug.line(UTF8STR("  unresolved parameter(s)"));
				return Result();  // cannot resolve call
			}
			
			//~ debug.line(UTF8STR("  is a possible candidate"));
			
			Candidate* c = new Candidate(e);
			
			std::vector<Parameter const*> const& params =
				*e->getParameterList();
			
			std::vector<ValuePassing> paramConversions;
			bool ok = true;
			for (unsigned int i = 0; i < arguments.size(); ++i) {
				Type const* argT = arguments[i];
				hyp_assert(argT != 0);
				
				Parameter const* param = params[i];
				
				// check for argument suitability
				ValuePassing paramOk = ValuePassing::check(argT, *param);
				
				if (paramOk.isPossible()) {
					paramConversions.push_back(paramOk);
				}
				else {
					ok = false; // candidate NOT suitable at all
					break;
				}
			}
			
			if (ok) {
				c->setApplicability(paramConversions);
				fullList.push_back(new Candidate(*c));
			}
		}
		
		//~ debug.line(
			//~ UTF8STR("OverloadResolution: full list contains %1 items").compose(
				//~ fullList.size()
			//~ )
		//~ );
		
		// candidates list is now ready
		
		if (fullList.size() < 2) {
			// No need to refine the list, as we have only one candidate or
			// none at all.
			return Result(fullList);
		}
		
		// OVERLOAD REFINING PHASE 1
		
		// copy full list to work list
		std::vector<Candidate*> candidates = fullList;
		
		refineOverloadCandidates(
			candidates, Conversions::kNoneAllowed
		);
		
		if (candidates.size() < 2) {
			return Result(candidates);
		}
		
		// OVERLOAD REFINING PHASE 2
		
		// copy full list to work list
		candidates = fullList;
		
		refineOverloadCandidates(
			candidates, Conversions::kLevel1Allowed
		);
		
		if (candidates.size() < 2) {
			return Result(candidates);
		}
		
		// OVERLOAD REFINING PHASE 3
		
		// copy full list to work list
		candidates = fullList;
		
		refineOverloadCandidates(
			candidates, Conversions::kLevel2Allowed
		);
		
		return Result(candidates);
	}
	
	
	template <class OE>
	void
	OverloadResolution<OE>::refineOverloadCandidates(
		std::vector<Candidate*>& list, Conversions::AllowedLevel aca)
	{
		hyp_assert(list.size() >= 2);
		
		// get the best candidates
		
		for (unsigned int i = 0; i < list.size() - 1; ++i) {
			if (list[i] == 0) {
				continue;
			}
			
			if (!Conversions::conversionOk(list[i]->getNeededMaximumConversion(), aca)) {
				list[i] = 0; // disable candidate
			}
			
			for (unsigned int j = i + 1; j < list.size(); ++j) {
				if (list[i] == 0 || list[j] == 0) {
					continue;
				}
				
				int better = list[i]->compareTo(*list[j]);
				
				if (better < 0) { // candidate i is better...
					// get rid of inferior candidate
					list[j] = 0;
				}
				else if (better > 0) { // candidate j is better...
					// get rid of inferior candidate
					list[i] = 0;
				}
				else {
					// no better one -> no action
				}
			}
		}
		
		// move all candidates to the beginning of the list
		
		unsigned int i = 0;
		while (i < list.size() - 1) {
			if (list[i] == 0) {
				for (unsigned int j = i + 1; j < list.size(); ++j) {
					if (list[j] != 0) {
						list[i] = list[j];
						list[j] = 0;
						i = j;
						break;
					}
				}
			}
			
			++i;
		}
		
		// shrink the list; pop off the zero pointers at the end
		
		while (list.back() == 0) {
			list.pop_back();
		}
	}
	
	template <class OE>
	HYPER_Common::UTF8String
	OverloadResolution<OE>::getCallPrototype(QualifiedName entityName,
		const std::vector<const Expression*>& args)
	{
		HYPER_Common::UTF8StringBuilder sb;
		sb += entityName.toString();
		sb += UTF8STR("(");
		
		for (unsigned int i = 0; i < args.size(); ++i) {
			if (i > 0) {
				sb += UTF8STR(", ");
			}
			ExpressionResult const& argRes = args[i]->getResult();
			hyp_assert(argRes.isResolved());
			sb += argRes.getType()->toString();
		}
		
		sb += UTF8STR(")");
		return sb.toString();
	}
	
	template <class OE>
	bool
	OverloadResolution<OE>::checkAccess(Candidate const* c,
		ClassContext const& from, Location const* loc, CompileInfo& info)
	{
		std::vector<ValuePassing> const& vps = c->getConversions();
		
		bool ok = true;
		for (unsigned int i = 0; i < vps.size(); ++i) {
			ok &= vps[i].checkAccess(from, loc, info);
		}
		
		return ok;
	}
}
#endif
