/*!
	\file overloadresolution.h
	\brief Contains the OverloadResolution class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-03-26
// Last time modified:	2009-08-23

#ifndef DEF_OVERLOADRESOLUTION_H
 #define DEF_OVERLOADRESOLUTION_H
 
 #include "common/sequenceview.h"
 
 #include "conversions.h"
 #include "qualifiedname.h"
 
 #include <vector>

namespace HYPER_Compiler {
	
	class Expression;
	
	/// Implements overload resolution in a generic way
	template <class OE>
	class OverloadResolution {
	public:
		
		typedef typename OE::OverloadEntity Entity;
		
		class Candidate {
		public:
			enum Applicability {
				kDontKnow = 0,
				kNotApplicable,
				kApplicable
			};
			
			Candidate(const Entity* e);
			
			const Entity* getEntity() const;
			
			Applicability getApplicability() const;
			bool isApplicable() const;
			void setApplicability(const std::vector<ValuePassing>& list);
			
			Conversions::Kind getNeededMaximumConversion() const;
			std::vector<ValuePassing> const& getConversions() const;
			
			/**
				\brief Comparison between two candidates
				Returns:
					-1 : this candidate is better
					0 : both candidates are similar
					1 : the other candidate is better
			*/
			int compareTo(Candidate const& c) const;
			
		private:
			const Entity* fE;
			Applicability fApp;
			std::vector<ValuePassing> fOK;
			Conversions::Kind fConv;
		};
		
		class Result {
		public:
			enum ResultType {
				kNotResolved = 0,
				kNoCandidates,
				kSingleCandidate,
				kMultipleCandidates
			};
			
		private:
			ResultType fgResult;
			std::vector<Candidate*> fgList;
			
		public:
			/// Constructor in case of procedure not resolved
			Result()
			 : fgResult(kNotResolved)
			{
				//
			}
			
			/// Constructor for a single candidate
			explicit Result(Candidate* c)
			 : fgResult(kSingleCandidate)
			{
				fgList.push_back(c);
			}
			
			/// Constructor for a list of resulting candidates
			explicit Result(std::vector<Candidate*> const& list)
			 : fgList(list)
			{
				if (list.empty()) {
					fgResult = kNoCandidates;
				}
				else if (list.size() == 1) {
					fgResult = kSingleCandidate;
				}
				else {
					fgResult = kMultipleCandidates;
				}
			}
			
			ResultType getResult() const {
				return fgResult;
			}
			
			void listCandidates(EasyDiagnosticRef& diag) const;
			
			std::vector<Candidate*> const& getCandidates() const {
				return fgList;
			}
		};
		
		
		// TODO: use SequenceView<Expression> etc. for these functions
		
		static Result
		resolveCall(OE const* list,
			const std::vector<const Expression*>& arguments,
			CompileInfo& info);
		
		static Result
		resolveCall(OE const* list, const std::vector<const Type*>& arguments,
			CompileInfo& info);
		
		
		static HYPER_Common::UTF8String
		getCallPrototype(QualifiedName entityName,
			const std::vector<const Expression*>& args);
		
		static bool checkAccess(Candidate const* c, ClassContext const& from,
			Location const* loc, CompileInfo& info);
		
	private:
		
		static void
		refineOverloadCandidates(std::vector<Candidate*>& list,
			Conversions::AllowedLevel a);
	};
}

// for templates, include the implementation
#include "overloadresolution.cxx"

#endif
