/*!
	\file parser.cxx
	\brief Implements the Parser class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-02-05
// Last time modified:	2009-09-05

#include "parser.h"

// misc. includes

#include "base/diagnostic.h"

#include "expressionparser.h"
#include "initializer.h"
#include "sourcefile.h"
#include "statementparser.h"
#include "type.h"
#include "typeparser.h"

// things that are declared
#include "declaredparameter.h"
#include "declaredtype.h"
#include "field.h"
#include "implementedclass.h"
#include "implementedconstructor.h"
#include "implementedenum.h"
#include "implementedprocedure.h"
#include "namespace.h"

// contexts
#include "constructorcontext.h"
#include "fieldcontext.h"
#include "procedurecontext.h"

using HYPER_Common::UTF8String;

namespace HYPER_Compiler {
	
	// ==========
	// Parser class
	// ==========
	
	Parser::Parser(ParserInfo const& pi)
	 : ParserBase(pi)
	{
		//
	}
	
	Parser::~Parser() {
		// nothing to do (yet)
	}
	
	SourceFile*
	Parser::doTranslationUnit(ClassRepository* repos,
		HYPER_Common::FilePath const& fileName, FileID fileID,
		bool isMainSourceFile)
	{
		skipEmptyLines();
		
		//debug << "PARSE: start translation unit" << newl;
		
		// namespace data (default)
		Namespace* ns = 0;
		
		if (lookToken().isKeyword(Keywords::kNamespace)) {
			nextToken();
			Location const* nsLoc = lookPos();
			QualifiedName nsName = doQualifiedName(true);
			ns = Namespace::create(nsName, nsLoc, getCompileInfo());
			doNewLine();
			skipEmptyLines();
		}
		
		SourceFile* sf =
			new SourceFile(
				fileID, fileName, getCompileInfo(), isMainSourceFile, ns
			);
		
		//debug << "PARSE: possible imports coming" << newl;
		
		while (lookToken().isKeyword(Keywords::kImport)) {
			nextToken(); // eat 'import' keyword
			Location const* loc = lookPos();
			QualifiedName import = doQualifiedName(true);
			sf->addImport(import, loc);
			doNewLine();
			skipEmptyLines();
		}
		
		//debug << "PARSE: possible 'begin' coming" << newl;
		
		if (lookToken().isKeyword(Keywords::kBegin)) {
			nextToken(); // eat 'begin' keyword
			const Location* p = lookPos();
			sf->setBegin(doQualifiedName(false), p);
			doNewLine();
			skipEmptyLines();
		}
		
		//debug << "PARSE: now expecting classes/enums" << newl;
		
		do {
			Keywords::Keyword k = lookToken().getKeyword();
			
			if (k == Keywords::kAbstract || k == Keywords::kClass
				|| k == Keywords::kSealed || k == Keywords::kStatic)
			{
				ImplementedClass* c =
					doClass(ClassMemberAccess::createPublic(), 0);
				sf->addMember(c);
			}
			else if (k == Keywords::kEnum) {
				ImplementedEnum* e =
					doEnum(ClassMemberAccess::createPublic(), 0);
				sf->addMember(e);
			}
			else if (lookToken().isEndOfFile()) {
				break; // stop parsing
			}
			else {
				fatal(errorExpected(
					UTF8STR("class declaration"), lookPos(), lookToken()
				));
			}
			
			skipEmptyLines();
		} while (true);
		
		return sf;
	}
	
	ImplementedEnum* Parser::doEnum(ClassMemberAccess access, Class* parent)
	{
		const Location* beginP = lookPos();
		
		doKeyword(Keywords::kEnum);
		UTF8String caption = doIdentifier();
		doNewLine();
		
		ImplementedEnum * enumAST =
			new ImplementedEnum(
				getCompileInfo(), caption, access, beginP, parent
			);
		
		do {
			skipEmptyLines();
			
			if (lookToken().isIdentifier()) {
				Location const* valueLoc = lookPos();
				enumAST->addValue(doIdentifier(), valueLoc);
				
				// possible extra identifiers on the same line
				while (lookToken().isSymbol(Token::kComma)) {
					nextToken(); // eat comma
					valueLoc = lookPos();
					enumAST->addValue(doIdentifier(), valueLoc);
				}
				
				doNewLine();
			}
			else if (lookToken().isKeyword(Keywords::kEnd)) {
				break; // end of the enum
			}
			else {
				fatal(errorExpected(
					UTF8STR(
						"enum member declaration or end of enum"
					),
					lookPos(), lookToken()
				));
			}
		} while (true);
		
		// end part
		doFlexibleEnd(Keywords::kEnum, caption);
		doNewLine();
		
		return enumAST;
	}
	
	// Parses a class
	ImplementedClass* Parser::doClass(ClassMemberAccess access,
		Class* parent)
	{
		const Location* beginP = lookPos();
		AbstractSealedStatic ass = kNoASS;
		
		Keywords::Keyword k = lookToken().getKeyword();
		
		if (k == Keywords::kAbstract) {
			ass = kAbstract;
			nextToken(); // eat 'abstract' keyword
			
			if (lookToken().isKeyword(Keywords::kConst)) {
				nonfatal(error(
					UTF8STR("\"const\" cannot be used in a class declaration"),
					lookPos()
				));
			}
			
			doKeyword(Keywords::kClass);
		}
		else if (k == Keywords::kClass) {
			nextToken(); // eat 'class' keyword
		}
		else if (k == Keywords::kSealed) {
			ass = kSealed;
			nextToken(); // eat 'sealed' keyword
			
			if (lookToken().isKeyword(Keywords::kConst)) {
				nonfatal(error(
					UTF8STR("\"const\" cannot be used in a class declaration"),
					lookPos()
				));
			}
			
			doKeyword(Keywords::kClass);
		}
		else if (k == Keywords::kStatic) {
			ass = kStatic;
			nextToken(); // eat 'static' keyword
			
			if (lookToken().isKeyword(Keywords::kConst)) {
				nonfatal(error(
					UTF8STR("\"const\" cannot be used in a class declaration"),
					lookPos()
				));
			}
			
			doKeyword(Keywords::kClass);
		}
		else {
			fatal(errorExpected(
				UTF8STR("class declaration"), lookPos(), lookToken()
			));
		}
		
		UTF8String caption = doIdentifier();
		doNewLine();
		
		ImplementedClass * curC =
			new ImplementedClass(
				getCompileInfo(), caption, access, ass, beginP, parent
			);
		
		do {
			skipEmptyLines();
			
			if (lookToken().isKeyword()) {
				Keywords::Keyword k = lookToken().getKeyword();
				
				if (k == Keywords::kEnd) {
					// OK, do nothing
				}
				else if (k == Keywords::kPrivate) {
					nextToken(); // eat 'private' keyword
					doSymbol(Token::kColon);
					doNewLine();
					doClassPart(curC, ClassMemberAccess::createPrivate());
				}
				else if (k == Keywords::kProtected) {
					ClassMemberAccess acc =
						ClassMemberAccess::createProtected();
					if (ass == kSealed) {
						acc = ClassMemberAccess::createPrivate();
						nonfatal(error(
							UTF8STR(
								"a sealed class cannot have protected sections"
							),
							lookPos()
						));
					}
					else if (ass == kStatic) {
						acc = ClassMemberAccess::createPrivate();
						nonfatal(error(
							UTF8STR(
								"a static class cannot have protected sections"
							),
							lookPos()
						));
					}
					nextToken(); // eat 'protected' keyword
					doSymbol(Token::kColon);
					doNewLine();
					doClassPart(curC, acc);
				}
				else if (k == Keywords::kPublic) {
					nextToken(); // eat 'public' keyword
					doSymbol(Token::kColon);
					doNewLine();
					doClassPart(curC, ClassMemberAccess::createPublic());
				}
				else {
					doClassPart(curC, ClassMemberAccess::createPrivate()); // Implicit private section
				}
			}
			else {
				fatal(errorExpected(
					UTF8STR(
						"class member declaration, begin of section or "
						"end of class"
					),
					lookPos(), lookToken()
				));
			}
		} while (!lookToken().isKeyword(Keywords::kEnd));
		
		// end part
		doFlexibleEnd(Keywords::kClass, caption);
		doNewLine();
		
		return curC;
	}
	
	// Parses the public, protected or private part in a class (the part header
	// not included)
	void Parser::doClassPart(ImplementedClass* cls, ClassMemberAccess access) {
		bool stop = false;
		
		do {
			skipEmptyLines();
			
			if (lookToken().isKeyword(Keywords::kEnd)) {
				// the end of this section
				break;
			}
			
			int pos = getTokenPos();
			const Location* startPos = lookPos();
			
			// jump over modifier keywords to see what we'll have to parse
			Keywords::Keyword k = lookToken().getKeyword();
			do {
				k = lookToken().getKeyword();
				switch (k) {
				case Keywords::kAbstract:
				case Keywords::kConst:
				case Keywords::kImplicit:
				case Keywords::kSealed:
				case Keywords::kStatic:
					// modifier keyword; go to next
					nextToken(); // eat the keyword
					continue; // check for another modifier keyword
				default:
					// no modifier keyword; exit loop
					k = Keywords::kNoKeyword;
				}
			} while (k != Keywords::kNoKeyword);
			
			// check for the current keyword
			k = lookToken().getKeyword();
			
			switch (k) {
			case Keywords::kClass:
				// parse a class
			{
				goBackToTokenPos(pos);
				ImplementedClass* c = doClass(access, cls);
				cls->addMember(c);
				break;
			}
			case Keywords::kEnum:
				// parse an enum
			{
				goBackToTokenPos(pos);
				ImplementedEnum* e = doEnum(access, cls);
				cls->addMember(e);
				break;
			}
			case Keywords::kPrivate:
				// the begin of the next section
				stop = true;
				break;
			case Keywords::kProcedure:
				// parse a procedure or a constructor
			{
				nextToken(); // eat 'procedure' keyword
				
				bool isConstructor = lookToken().isKeyword(Keywords::kNew);
				goBackToTokenPos(pos);
				
				if (isConstructor) { // constructor
					ImplementedConstructor* cons = doConstructor(cls, access);
					if (cls->getASS() == kStatic) {
						nonfatal(error(
							UTF8STR(
								"a static class cannot contain any "
								"constructors"
							),
							startPos
						));
					}
					else {
						cls->addMember(cons);
					}
				}
				else { // procedure
					ImplementedProcedure* p = doProcedure(cls, access);
					cls->addMember(p);
				}
				break;
			}
			case Keywords::kProtected:
				// the begin of the next section
				stop = true;
				break;
			case Keywords::kPublic:
				// the begin of the next section
				stop = true;
				break;
			case Keywords::kVar:
				// parse a field
			{
				goBackToTokenPos(pos);
				Field* f = doField(cls, access);
				cls->addMember(f);
				break;
			}
			case Keywords::kNoKeyword:
				// token is not a keyword
				if (lookToken().isIdentifier()) {
					// probably a field without the 'var' keyword
					goBackToTokenPos(pos);
					Field* f = doField(cls, access);
					cls->addMember(f);
				}
				else {
					fatal(errorExpected(
						UTF8STR(
							"class member declaration, begin of section "
							"or end of class"
						),
						lookPos(),
						lookToken()
					));
				}
				break;
			default:
				// other keywords
				fatal(error(
					UTF8STR(
						"the %1 keyword cannot be used in this context"
					).compose(quote(lookToken().getDescription())),
					lookPos()
				));
				break;
			}
		} while (!stop);
	}
	
	Field* Parser::doField(Class* c, ClassMemberAccess access) {
		const Location* p = lookPos();
		bool isConst = false;
		bool isStatic = false;
		
		if (lookToken().isKeyword(Keywords::kStatic)) {
			//debug.line("field is static");
			isStatic = true;
			nextToken(); // eat 'static' keyword
		}
		
		if (lookToken().isKeyword(Keywords::kConst)) {
			//debug.line("field is const");
			isConst = true;
			nextToken(); // eat 'const' keyword
		}
		
		// Do "var" keyword if not abbreviated version
		if ((!isConst && !isStatic) || !lookToken().isIdentifier()) {
			//debug.line("field needs var keyword");
			doKeyword(Keywords::kVar); // require 'var' keyword
		}
		
		bool stop = false;
		
		std::vector<HYPER_Common::UTF8String> names;
		std::vector<Location const*> locs;
		
		do {
			locs.push_back(lookPos());
			names.push_back(doIdentifier()); // variable name
			
			if (lookToken().isSymbol(Token::kAnd)) {
				nextToken();
			}
			else {
				stop = true;
			}
		} while (!stop);
		
		doSymbol(Token::kColon);
		FieldContext* ctxt = new FieldContext(getCompileInfo());
		TypeParser tp(this->parserInfo(), ctxt);
		DeclaredType* t = tp.doType(false);
		
		Expression* initExpr = 0;
		if (lookToken().isSymbol(Token::kEqual)) {
			nextToken(); // eat '=' symbol
			ExpressionParser ep(this->parserInfo(), ctxt);
			initExpr = ep.doExpression();
		}
		
		doNewLine();
		
		if (isConst) {
			if (t->isConst()) {
				nonfatal(error(
					UTF8STR("duplicate %1").compose(QUOTED_UTF8STR("const")),
					p
				));
			}
			else {
				t->makeConst();
			}
		}
		
		Initializer* init;
		if (initExpr) {
			init = new Initializer(t, initExpr, ctxt, p);
		}
		else {
			init = new Initializer(t, ctxt, p);
		}
		
		Field* f =
			new Field(c, isStatic, access, names, locs, t, init, p, ctxt);
		ctxt->setField(f);
		
		return f;
	}
	
	ImplementedConstructor*
	Parser::doConstructor(Class* c, ClassMemberAccess access) {
		const Location* beginLoc = lookPos();
		
		bool implicit = false;
		if (lookToken().isKeyword(Keywords::kImplicit)) {
			nextToken();
			implicit = true;
		}
		
		doKeyword(Keywords::kProcedure); // require 'procedure' keyword
		doKeyword(Keywords::kNew); // require 'new' keyword
		
		ImplementedConstructor* constr =
			new ImplementedConstructor(c, access, implicit, beginLoc);
		
		ConstructorContext* ctCtxt = new ConstructorContext(constr);
		
		// opening parentheses behind 'new' ??
		if (lookToken().isSymbol(Token::kOpenParen)) { // yes :)
			noWhitesBeforeCurrentToken();
			doParameterList(ctCtxt, constr->getDeclaredParameterList());
		}
		else { // NO !! :(
			// now we'll see if user just forgot the place the "( )"
			
			const Location* paramLoc = lookPos();
			
			// if parameter declaration by itself missing...
			if (lookToken().isNewLine()) {
				nonfatal(error(
					UTF8STR("parentheses missing after constructor header"),
					paramLoc
				));
				// continue parsing
			}
			else { // something more serious here...
				fatal(errorExpected(
					UTF8STR("parameter list"), paramLoc, lookToken()
				));
			}
		}
		
		doNewLine(); // newline behind constructor header
		StatementParser sp(this->parserInfo());
		sp.doBlockBody(constr->getCodeBlock(), ctCtxt);
		
		// end part
		doFlexibleEnd(Keywords::kProcedure, Keywords::kNew);
		doNewLine();
		
		return constr;
	}
	
	// Parses a procedure
	ImplementedProcedure* Parser::doProcedure(Class* c, ClassMemberAccess access) {
		const Location* beginP = lookPos();
		
		bool isStatic = false;
		bool isAbstract = false;
		bool isSealed = false;
		bool isConst = false;
		
		Keywords::Keyword k = lookToken().getKeyword();
		AbstractSealedStatic ass = kNoASS;
		
		if (k == Keywords::kStatic) {
			isStatic = true;
			nextToken(); // eat 'static' keyword
			ass = kStatic;
		}
		else if (k == Keywords::kSealed) {
			isSealed = true;
			nextToken(); // eat 'sealed' keyword
			ass = kSealed;
		}
		else if (k == Keywords::kAbstract) {
			isAbstract = true;
			nextToken(); // eat 'abstract' keyword
			ass = kAbstract;
		}
		
		if (lookToken().isKeyword(Keywords::kConst)) {
			isConst = true;
			nextToken(); // eat 'const' keyword
		}
		
		doKeyword(Keywords::kProcedure); // require 'procedure' keyword
		
		UTF8String caption;
		
		if (lookToken().isIdentifier()) {
			caption = doIdentifier();
		}
		else {
			fatal(errorExpected(
				UTF8STR("procedure name"), lookPos(), lookToken()
			));
		}
		
		ImplementedProcedure* curP =
			new ImplementedProcedure(caption, c, access, ass, isConst, beginP);
		
		const Location* paramPos = lookPos();
		
		ProcedureContext* pCtxt = new ProcedureContext(curP);
		
		// opening parentheses behind name ??
		if (lookToken().isSymbol(Token::kOpenParen)) { // yes :)
			noWhitesBeforeCurrentToken();
			doParameterList(pCtxt, curP->getDeclaredParameterList());
		}
		else { // NO !! :(
			// now we'll see if user just forgot the place the "( )"
			
			// if parameter declaration by itself missing...
			if (lookToken().isNewLine() || lookToken().isSymbol(Token::kColon))
			{
				nonfatal(error(
					UTF8STR("parentheses missing after procedure name"),
					paramPos
				));
				// now continue parsing
			}
			else { // something more serious here...
				fatal(errorExpected(
					UTF8STR("parameter list"), paramPos, lookToken()
				));
			}
		}
		
		if (lookToken().isSymbol(Token::kColon)) { // do optional return type
			nextToken();
			TypeParser tp(this->parserInfo(), pCtxt);
			curP->setReturnType(tp.doType(true));
		}
		
		// only do procedure body if procedure is not abstract
		if (isAbstract) {
			curP->setEndLocation(lookPos());
			doSymbol(Token::kSemicolon);
			doNewLine();
		}
		else {
			doNewLine(); // newline behind procedure header
			StatementParser sp(this->parserInfo());
			sp.doBlockBody(curP->getCodeBlock(), pCtxt);
			
			// end part
			curP->setEndLocation(lookPos());
			doFlexibleEnd(Keywords::kProcedure, caption);
			doNewLine();
		}
		
		return curP;
	}
	
	// Parses an parameter list
	void
	Parser::doParameterList(ClassContext const* cCtxt,
		std::vector<DeclaredParameter*>& params)
	{
		doSymbol(Token::kOpenParen);
		
		bool stop = false;
		
		if (lookToken().isSymbol(Token::kCloseParen)) { // empty arguments list?
			stop = true; // don't go into the while loop
		}
		
		while (!stop) {
			bool isInout = false;
			
			// optional 'inout' keyword
			if (lookToken().isKeyword(Keywords::kInout)) {
				isInout = true;
				nextToken(); // eat 'inout'
			}
			
			// local varname
			if (!lookToken().isIdentifier()) {
				fatal(errorExpected(
					UTF8STR("variable name"), lookPos(), lookToken()
				));
			}
			
			bool stop2 = false;
			
			std::vector<UTF8String> names; // list of names declared together, like "abc & defg & hello"
			std::vector<const Location*> locs; // list of locations
			
			do {
				locs.push_back(lookPos());
				names.push_back(doIdentifier()); // variable name
				
				if (lookToken().isSymbol(Token::kAnd)) {
					nextToken();
				}
				else {
					stop2 = true;
				}
			} while (!stop2);
			
			// we expect a ':' now
			doSymbol(Token::kColon);
			
			// type name
			TypeParser tp(this->parserInfo(), cCtxt);
			DeclaredType* type = tp.doType(false);
			
			// add parameters to procedure
			for (unsigned int i = 0; i < names.size(); ++i) {
				Parameter::Passing pass;
				if (isInout) {
					pass = Parameter::kInOut;
				}
				else {
					// input parameter, by reference
					//    (but small built-in types will be passed by value)
					pass = Parameter::kInByReference;
				}
				
				DeclaredParameter* par =
					new DeclaredParameter(
						getCompileInfo(), names[i], type, locs[i], pass
					);
				params.push_back(par);
			}
			
			if (lookToken().isSymbol(Token::kCloseParen)) {
				stop = true;
			}
			else if (lookToken().isSymbol(Token::kComma)) {
				nextToken(); // eat the ','
			}
			else {
				fatal(errorExpected(
					UTF8STR("',' or ')' symbol"), lookPos(), lookToken()
				));
			}
		}
		
		doSymbol(Token::kCloseParen);
	}
}
