/*!
	\file parserbase.cxx
	\brief Implements the ParserBase class.
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-09-21
// Last time modified:	2008-07-05

#include "parserbase.h"

#include "base/asserts.h"
#include "base/compileoptions.h"
#include "base/easydiagnostics.h"

#include "lexer/lexer.h"

#include "builtintypes.h"
#include "compileinfo.h"

namespace HYPER_Compiler {
	
	// *** ParserInfo ***
	
	ParserInfo::ParserInfo(Lexer& lex, CompileInfo& info)
	 : fLexer(lex), fInfo(info)
	{
		//
	}
	
	
	// *** ParserBase ***
	
	ParserBase::ParserBase(ParserInfo const& pi)
	 : fPI(pi)
	{
		//
	}
	
	ParserInfo const& ParserBase::parserInfo() const {
		return fPI;
	}
	
	// Parses a qualified name.
	// The first part is assumed to be an identifier.
	QualifiedName ParserBase::doQualifiedName(bool allowSystemNamespace) {
		if (lookToken().isKeyword(Keywords::kSystem)) {
			if (allowSystemNamespace) {
				nextToken(); // skip 'system' keyword
			}
			else {
				fatal(error(
					UTF8STR("cannot use %1 namespace name here").compose(
						QUOTED_UTF8STR("system")
					),
					lookPos()
				));
			}
			return doQualifiedName(UTF8STR("system"));
		}
		else {
			return doQualifiedName(doIdentifier());
		}
	}
	
	// Parses a qualified name, with the first part already given.
	// (so the next dot is the first to be parsed)
	QualifiedName
	ParserBase::doQualifiedName(HYPER_Common::UTF8String firstPart)
	{
		QualifiedName qn(firstPart);
		
		while (true) {
			if (lookToken().isSymbol(Token::kDot)) {
				noWhitesBeforeCurrentToken();
				nextToken(); // eat '.' symbol
				noWhitesBeforeCurrentToken();
				qn.addIdentifier(doIdentifier());
			}
			else {
				// something else than whitespace
				break; // the qualified name is done
			}
		};
		
		return qn;
	}
	
	bool ParserBase::isClassNameKeyword(const Token& t) {
		if (t.isKeyword()) {
			Keywords::Keyword k = t.getKeyword();
			return BuiltInTypes::fromKeyword(k) != BuiltInTypes::kNone;
		}
		else {
			return false;
		}
	}
	
	bool ParserBase::isEndableKeyword(Keywords::Keyword kw) {
		if (kw == Keywords::kNamespace
			|| kw == Keywords::kClass
			|| kw == Keywords::kProcedure
			|| kw == Keywords::kIterate
			|| kw == Keywords::kIf
			|| kw == Keywords::kLoop)
		{
			return true;
		}
		else {
			return false;
		}
	}
	
	void
	ParserBase::doFlexibleEnd(Keywords::Keyword entity,
		HYPER_Common::UTF8String name)
	{
		doKeyword(Keywords::kEnd);
		
		bool hadEntity = false;
		bool hadName = false;
		
		do {
			if (!hadEntity && !hadName && lookToken().isKeyword()) {
				Keywords::Keyword kw = lookToken().getKeyword();
				
				if (kw == entity) {
					nextToken(); // eat entity keyword
					hadEntity = true;
				}
				else if (isEndableKeyword(kw)) {
					fatal(error(
						UTF8STR(
							"end statement mismatch; expected %1 keyword, "
							"instead of %2 keyword"
						).compose(
							quote(toString(entity)), quote(toString(kw))
						),
						lookPos()
					));
				}
				else {
					break;
				}
			}
			else if (!hadName && lookToken().isIdentifier()) {
				HYPER_Common::UTF8String ident = lookToken().getDescription();
				
				if (name.isEmpty()) { // if no name allowed...
					nonfatal(error(
						UTF8STR(
							"name not allowed here; remove %1"
						).compose(quote(ident)),
						lookPos()
					));
					nextToken(); // eat identifier
					hadName = true; // to make this error occur maximum once for this 'end'
				}
				else { // name allowed
					if (ident != name) {
						nonfatal(error(
							UTF8STR(
								"name mismatch: identifier %1 should be %2"
							).compose(
								quote(ident), quote(name)
							),
							lookPos()
						));
					}
					
					nextToken(); // eat identifier
					hadName = true;
				}
			}
			else {
				break;
			}
		} while (true);
	}
	
	void
	ParserBase::doFlexibleEnd(Keywords::Keyword entity,
		Keywords::Keyword name)
	{
		doKeyword(Keywords::kEnd);
		
		bool hadEntity = false;
		bool hadName = false;
		
		do {
			if (!hadEntity && !hadName && lookToken().isKeyword()) {
				Keywords::Keyword kw = lookToken().getKeyword();
				
				if (kw == entity) {
					nextToken(); // eat entity keyword
					hadEntity = true;
				}
				else if (kw == name) {
					nextToken(); // eat name keyword
					hadName = true;
				}
				else if (isEndableKeyword(kw)) {
					fatal(error(
						UTF8STR(
							"end statement mismatch; expected %1 keyword, "
							"instead of %2 keyword"
						).compose(
							quote(toString(entity)), quote(toString(kw))
						),
						lookPos()
					));
				}
				else {
					break;
				}
			}
			else if (!hadName && lookToken().isIdentifier()) {
				HYPER_Common::UTF8String ident = lookToken().getDescription();
				nonfatal(error(
					UTF8STR(
						"name must be the %1 keyword; remove %2"
					).compose(
						quote(toString(name)), quote(ident)
					),
					lookPos()
				));
				nextToken(); // eat identifier
				hadName = true; // to make this error occur maximum once for this 'end'
			}
			else {
				break;
			}
		} while (true);
	}
	
	void ParserBase::skipEmptyLines() {
		while (lookToken().isNewLine()) {
			nextToken(); // eat the newline
		}
	}
	
	void ParserBase::skipRestOfLine(bool skipNewLine) {
		while (!lookToken().isNewLine()) {
			nextToken(); // eat the token
		}
		
		if (skipNewLine) {
			nextToken(); // eat the newline also
		}
	}
	
	HYPER_Common::UTF8String ParserBase::doIdentifier() {
		if (lookToken().isIdentifier()) {
			HYPER_Common::UTF8String s = lookToken().getDescription();
			nextToken(); // eat identifier
			return s;
		}
		else {
			fatal(errorExpected(UTF8STR("identifier"), lookPos(), lookToken()));
			
			// fake 'return' that will never be executed
			//return UTF8STR("");
		}
	}
	
	void ParserBase::doKeyword(Keywords::Keyword k) {
		if (lookToken().isKeyword(k)) {
			nextToken(); // eat keyword
		}
		else {
			fatal(errorExpected(
				UTF8STR("%1 keyword").compose(quote(toString(k))),
				lookPos(), lookToken()
			));
		}
	}
	
	void ParserBase::doNonfatalKeyword(Keywords::Keyword k) {
		if (lookToken().isKeyword(k)) {
			nextToken(); // eat keyword
		}
		else {
			nonfatal(errorExpected(
				UTF8STR("%1 keyword").compose(quote(toString(k))),
				lookPos(), lookToken()
			));
		}
	}
	
	// Parses a symbol
	void ParserBase::doSymbol(Token::Symbol sym) {
		if (lookToken().isSymbol(sym)) {
			nextToken(); // eat symbol
		}
		else {
			fatal(errorExpected(
				UTF8STR("%1 symbol").compose(quote(Token::toString(sym))),
				lookPos(), lookToken()
			));
		}
	}
	
	HYPER_Common::UTF8String ParserBase::doWholeNumber() {
		if (lookToken().isWholeNumber()) {
			HYPER_Common::UTF8String s = lookToken().getDescription();
			nextToken(); // eat number
			return s;
		}
		else {
			fatal(errorExpected(
				UTF8STR("whole number"), lookPos(), lookToken()
			));
		}
	}
	
	HYPER_Common::UTF8String ParserBase::doFloatingPoint() {
		if (lookToken().isFloatingPoint()) {
			HYPER_Common::UTF8String s = lookToken().getDescription();
			nextToken(); // eat number
			return s;
		}
		else {
			fatal(errorExpected(
				UTF8STR("floating point number"), lookPos(), lookToken()
			));
		}
	}
	
	// do newline with optionally some spaces in front of it
	void ParserBase::doNewLine() {
		// C/C++/Java... users might have put a semicolon at end of line
		if (lookToken().isSymbol(Token::kSemicolon)) {
			nonfatal(error(UTF8STR("semicolon not allowed here"), lookPos()));
			nextToken(); // eat the ';'
		}
		
		if (lookToken().isNewLine()) {
			// OK
			nextToken(); // eat the newline
		}
		else {
			nonfatal(errorExpected(UTF8STR("newline"), lookPos(), lookToken()));
			
			// skip the rest of the line, together with the newline
			skipRestOfLine(true);
		}
	}
	
	void ParserBase::doEndOfFile() {
		Token tok = lookToken();
		
		if (lookToken().isEndOfFile()) {
			// No getToken() call here. This would cause an error.
		}
		else {
			fatal(errorExpected(
				UTF8STR("end of file"), lookPos(), lookToken()
			));
		}
	}
	
	HYPER_Common::UTF8String ParserBase::doString() {
		if (lookToken().isString()) {
			HYPER_Common::UTF8String s = lookToken().getDescription();
			nextToken(); // eat the string
			return s;
		}
		else {
			fatal(
				errorExpected(UTF8STR("string literal"), lookPos(), lookToken())
			);
			
			// fake 'return' that will never be executed
			//return UTF8STR("");
		}
	}
	
	HYPER_Common::UTF8String ParserBase::doChar() {
		if (lookToken().isChar()) {
			HYPER_Common::UTF8String s = lookToken().getDescription();
			nextToken(); // eat the char
			return s;
		}
		else {
			fatal(errorExpected(
				UTF8STR("char literal"), lookPos(), lookToken()
			));
			
			// fake 'return' that will never be executed
			//return UTF8STR("");
		}
	}
	
	// Gets the current token
	Token ParserBase::lookToken() {
		return fPI.fLexer.lookToken();
	}
	
	// Gets the position of the current token
	const Location* ParserBase::lookPos() {
		return fPI.fLexer.lookPos();
	}
	
	// Gets the next token
	void ParserBase::nextToken() {
		//debug << "PARSE: next token" << newl;
		//debug << "   current is : " << lookToken().getDescriptiveString() << newl;
		
		if (fPI.fLexer.lookToken().isEndOfFile()) {
			fatal(error(UTF8STR("unexpected end of file!"), lookPos()));
		}
		
		fPI.fLexer.getToken();
		//debug << "   next is : " << lookToken().getDescriptiveString() << newl;
	}
	
	int ParserBase::getTokenPos() {
		return fPI.fLexer.getTokenPos();
	}
	
	void ParserBase::goBackToTokenPos(int pos) {
		fPI.fLexer.goBackToTokenPos(pos);
	}
	
	void ParserBase::goBackOneToken() {
		fPI.fLexer.goBackOneToken();
	}
	
	HYPER_Common::UTF8String ParserBase::toString(Keywords::Keyword kw) const
	{
		return fPI.fLexer.getKeywordsList().getString(kw);
	}
	
	void ParserBase::noWhitesBeforeCurrentToken() {
		if (lookToken().isPrecededByWhitespace()
			&& fPI.fInfo.getOptions().fgWhitespaceWarnings)
		{
			fPI.fInfo.diag().warning(
				UTF8STR(
					"whitespace before this token can make the "
					"sourcecode less readable"
				),
				lookPos()
			);
		}
	}
	
	void ParserBase::fatal(Diagnostic const& d) {
		fPI.fInfo.diag().add(d);
		throw FatalParseError(&d.getLocation());
	}
	
	void ParserBase::nonfatal(Diagnostic const& d) {
		fPI.fInfo.diag().add(d);
	}
	
	Diagnostic
	ParserBase::error(HYPER_Common::UTF8String msg, const Location* loc)
	{
		return Diagnostic(
			Diagnostic::kError, DiagnosticPair(msg, loc)
		);
	}
	
	Diagnostic
	ParserBase::errorExpected(HYPER_Common::UTF8String expected,
		const Location* loc, Token found)
	{
		return Diagnostic(
			Diagnostic::kError,
			DiagnosticPair(
				UTF8STR(
					"expected: %1, instead of %2"
				).compose(
					expected,
					found.getDescriptiveString()
				),
				loc
			)
		);
	}
	
	CompileInfo& ParserBase::getCompileInfo() const {
		return fPI.fInfo;
	}
}
