/*!
	\file parserbase.h
	\brief Contains the ParserBase class.
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-09-21
// Last time modified:	2008-06-17

#ifndef DEF_PARSERBASE_H
 #define DEF_PARSERBASE_H
 
 #include "common/cxx_macros.h"
 
 #include "lexer/token.h"
 
 #include "important_fwdecl.h"
 #include "qualifiedname.h"
 
namespace HYPER_Compiler {
	
	class Lexer;
	
	/// Data that parsers require to do their work
	class ParserInfo {
	public:
		ParserInfo(Lexer& lex, CompileInfo& info);
		
	private:
		friend class ParserBase;
		
		Lexer& fLexer;
		CompileInfo& fInfo;
	};
	
	/// Base class for parsers
	class ParserBase {
	protected:
		ParserBase(ParserInfo const& pi);
		
		ParserInfo const& parserInfo() const;
		
		QualifiedName doQualifiedName(bool allowSystemNamespace);
		QualifiedName doQualifiedName(HYPER_Common::UTF8String firstPart);
		
		static bool isClassNameKeyword(const Token& t);
		static bool isEndableKeyword(Keywords::Keyword kw);
		
		void doFlexibleEnd(Keywords::Keyword entity,
			HYPER_Common::UTF8String name = UTF8STR(""));
		void doFlexibleEnd(Keywords::Keyword entity, Keywords::Keyword name);
		
		// skippers
		void skipEmptyLines();
		void skipRestOfLine(bool skipNewLine);
		
		// lowlevel parsing
		HYPER_Common::UTF8String doIdentifier();
		void doKeyword(Keywords::Keyword k);
		void doNonfatalKeyword(Keywords::Keyword keyw);
		void doSymbol(Token::Symbol sym);
		HYPER_Common::UTF8String doWholeNumber();
		HYPER_Common::UTF8String doFloatingPoint();
		void doNewLine();
		void doEndOfFile();
		HYPER_Common::UTF8String doString();
		HYPER_Common::UTF8String doChar();
		void noWhitesBeforeCurrentToken();
		
		// talking to the lexer
		Token lookToken();
		const Location* lookPos();
		void nextToken();
		int getTokenPos();
		void goBackToTokenPos(int pos);
		void goBackOneToken();
		
		// keyword utilities
		HYPER_Common::UTF8String toString(Keywords::Keyword kw) const;
		
		// error reporting
		void fatal(Diagnostic const& d) ATTR_NO_RETURN;
		void nonfatal(Diagnostic const& d);
		Diagnostic error(HYPER_Common::UTF8String msg, const Location* loc);
		Diagnostic errorExpected(HYPER_Common::UTF8String expected,
			const Location* loc, Token found);
		CompileInfo& getCompileInfo() const;
		
		// data
		ParserInfo fPI;
	};
}
#endif
