/*!
	\file pointertype.cxx
	\brief Implements the PointerType class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-12-16
// Last time modified:	2008-10-21

#include "pointertype.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "compileinfo.h"
#include "conversions.h"
#include "expression.h"
#include "initializationhandler.h"
#include "simpleparameter.h"
#include "type.h"

//~ #include "base/debug.h"

namespace HYPER_Compiler {
	
	// *** PointerType ***
	
	PointerType::PointerType(bool isConst, DeclaredType* base,
		const Location* loc)
	 : DeclaredType(base->getCompileInfo(), loc), fBaseType(base),
		fConst(isConst)
	{
		//
	}
	
	PointerType::PointerType(PointerType const& pt)
	 : DeclaredType(pt), fBaseType(pt.fBaseType), fConst(pt.fConst)
	{
		// copy constructor should not be used
		hyp_assert_unreachable();
	}
	
	bool PointerType::isArray() const {
		return false;
	}
	
	bool PointerType::isPointer() const {
		return true;
	}
	
	bool PointerType::isDoublePointer() const {
		return false;
	}
	
	bool PointerType::isClass() const {
		return false;
	}
	
	bool PointerType::isConst() const {
		return fConst;
	}
	
	void PointerType::makeConst() {
		fConst = true;
	}
	
	HYPER_Common::UTF8String PointerType::toString() const {
		return UTF8STR("%1* %2").compose(
			fConst ? UTF8STR("const ") : UTF8STR(""),
			fBaseType->toString()
		);
	}
	
	void PointerType::putInSymbolTable(LookupLocation const* loc) {
		DeclaredType::putInSymbolTable(loc);
		fBaseType->putInSymbolTable(loc);
	}
	
	bool PointerType::doInterfaceResolve(Trace const& trace) {
		// disallow a runtime array size (open array is allowed though)
		fBaseType->setRestrictions(false, true);
		bool ok = fBaseType->resolveInterface(trace);
		
		//~ debug.line(UTF8STR("PointerType: ") + toString());
		
		if (ok) {
			Type const* resultType =
				Type::createSinglePointer(fConst, fBaseType->getResolvedType());
			
			//~ debug.line(UTF8STR("PointerType: resolved to ") + resultType->toString());
			setResolvedType(resultType);
		}
		
		return ok;
	}
}
