/*!
	\file procedure.h
	\brief Contains the Procedure class.
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-08-09
// Last time modified:	2008-12-20

#ifndef DEF_PROCEDURE_H
 #define DEF_PROCEDURE_H
 
 #include "declareditem.h"
 #include "qualifiedname.h"
 #include "membertypes.h"
 #include "overloadable.h"
 
 #include <vector>

namespace HYPER_Compiler {
	
	/// Represents a procedure
	class Procedure : public Overloadable, public DeclaredItem {
	public:
		Procedure(CompileInfo& info);
		virtual ~Procedure();
		
		/// Get the CompileInfo
		CompileInfo& getCompileInfo() const;
		
		/// Get the class the procedure is in
		virtual Class* getClass() const = 0;
		
		/// Gets the short name of the procedure
		virtual HYPER_Common::UTF8String getCaption() const = 0;
		
		/// Gets the full name of the procedure
		virtual QualifiedName getFullName() const = 0;
		
		HYPER_Common::UTF8String getDescription() const;
		
		Class const* getOwnerClass() const;
		ClassMemberAccess getAccessInOwner() const;
		
		/// Gets the header of the procedure as would appear in sourcecode
		virtual HYPER_Common::UTF8String headerToString() const;
		
		/// Gets the member access level of this procedure inside its class
		virtual ClassMemberAccess getAccess() const = 0;
		
		/// Gets the abstract/sealed/static/none member type of this procedure
		virtual AbstractSealedStatic getASS() const = 0;
		
		/// Check if the procedure is a static member
		bool isStatic() const {
			return getASS() == kStatic;
		}
		
		/// Check if the procedure is a const member
		virtual bool isConst() const = 0;
		
		/// Check if the procedure returns something
		virtual bool hasReturnType() const = 0;
		
		/// Get the return type for a procedure that returns something
		virtual Type const* getReturnType() const = 0;
		
		/// Get the number of parameters
		virtual unsigned getParameterCount() const = 0;
		
		/// Procedure has no parameters?
		bool hasNoParameters() const {
			return getParameterCount() == 0;
		}
		
		/// Get the parameter list of the procedure
		virtual const std::vector<Parameter const*>* getParameterList() const = 0;
		
		/// Get the location of the procedure in the source code
		virtual const Location* getLocation() const = 0;
		
		HYPER_Common::UTF8String getOverloadHash() const;
		
		/// Mark this procedure as a duplicate of another one
		virtual void markAsDuplicate() = 0;
		
		/// Check if this procedure is marked as a duplicate
		virtual bool isDuplicate() const = 0;
		
		virtual void putInSymbolTable(Scope& parent) {}
		
	private:
		CompileInfo& fCompileInfo;
	};
}
#endif
