/*!
	\file procedurecontext.cxx
	\brief Implements the ProcedureContext class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-10-28
// Last time modified:	2009-01-05

#include "procedurecontext.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "compileinfo.h"
#include "conversions.h"
#include "expression.h"
#include "implementedprocedure.h"
#include "simpleparameter.h"
#include "type.h"

namespace HYPER_Compiler {
	
	ProcedureContext::ProcedureContext(const ImplementedProcedure* p)
	 : StatementContext(p->getCompileInfo()), fProcedure(p)
	{
		// no action
	}
	
	Class const* ProcedureContext::getClass() const {
		return fProcedure->getClass();
	}
	
	bool ProcedureContext::isStatic() const {
		return fProcedure->isStatic();
	}
	
	bool ProcedureContext::isConst() const {
		return fProcedure->isConst();
	}
	
	const ImplementedProcedure* ProcedureContext::getProcedure() const {
		return fProcedure;
	}
	
	ProcedureContext* ProcedureContext::copy() const {
		return new ProcedureContext(*this);
	}
	
	Type const*
	ProcedureContext::getThisPointer(const Location* loc) const
	{
		if (isStatic()) {
			getCompileInfo().diag().error(
				loc,
				UTF8STR("%1 cannot be used in a static procedure"),
				QUOTED_UTF8STR("this")
			);
			return 0;
		}
		
		bool isConst = fProcedure->isConst();
		
		Type const* t =
			Type::createSinglePointer(
				false, Type::createClassType(isConst, fProcedure->getClass())
			);
		
		return t;
	}
	
	bool
	ProcedureContext::checkReturn(Location const* loc,
		Expression const* expr) const
	{
		if (!fProcedure->hasReturnType()) {
			if (expr == 0) {
				return true; // OK
			}
			else {
				// procedure has no return type, but a return <expr> is used
				getCompileInfo().diag().error(
					UTF8STR("no return expression needed in this procedure"),
					expr->getLocation()
				).addInfo(
					UTF8STR("expression: ") + expr->toString(),
					expr->getLocation()
				).addInfo(
					UTF8STR("procedure declaration: ")
						+ fProcedure->headerToString(),
					fProcedure->getLocation()
				);
			}
			
			return false;
		}
		
		const Type* expectedReturnType = fProcedure->getReturnType();
		
		if (expectedReturnType == 0) {
			// unresolved return type --> cannot check type compatibility
			return false;
		}
		
		if (expr == 0) { // return <nothing>
			getCompileInfo().diag().error(
				UTF8STR(
					"return expression missing in procedure with return "
						"type %1"
				).compose(quote(expectedReturnType->toString())),
				loc
			);
			
			return false;
		}
		
		if (expr->returnsValue().isNotTrue()) {
			// not sure if a result is returned from the expression
			return true;
		}
		
		Type const* exprType = expr->getResult().getType();
		hyp_assert(exprType != 0);
		
		SimpleParameter param(expectedReturnType, SimpleParameter::kCopyValue);
		ValuePassing vp = ValuePassing::check(exprType, param);
		
		if (!vp.isPossible()) {
			// no conversion possible
			getCompileInfo().diag().error(
				UTF8STR(
					"return expression has unsuitable type %1; the "
					"procedure has return type %2"
				).compose(
					quote(exprType->toString()),
					quote(expectedReturnType->toString())
				),
				expr->getLocation()
			).addInfo(
				UTF8STR("return expression: ") + expr->toString(),
				expr->getLocation()
			);
			
			return false;
		}
		
		// check access
		return vp.checkAccess(*this, expr->getLocation(), getCompileInfo());
	}
}
