/*!
	\file procedureoverloadresolution.cxx
	\brief Implements the ProcedureOverloadResolution class
	\author Kevin André
	
	Copyright (C) 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2009-08-23
// Last time modified:	2009-08-23

#include "procedureoverloadresolution.h"

#include "instanceprovider.h"
#include "procedure.h"

//~ #include "base/debug.h"

#include <vector>

namespace HYPER_Compiler {
	
	typedef OverloadResolution<OverloadedProcedure> ORP;
	
	ProcedureOverloadResolution::Result
	ProcedureOverloadResolution::resolveCall(InstanceProvider const* instance,
		OverloadedProcedure const* list,
		const std::vector<const Expression*>& arguments, CompileInfo& info)
	{
		// extract types from argument expression list
		std::vector<const Type*> argTypes;
		for (unsigned int i = 0; i < arguments.size(); ++i) {
			ExpressionResult const& argRes = arguments[i]->getResult();
			hyp_assert(argRes.isResolved());
			argTypes.push_back(argRes.getType());
		}
		
		// forward call to other function
		return resolveCall(instance, list, argTypes, info);
	}
	
	ProcedureOverloadResolution::Result
	ProcedureOverloadResolution::resolveCall(InstanceProvider const* instance,
		OverloadedProcedure const* list,
		const std::vector<const Type*>& arguments, CompileInfo& info)
	{
		// call basic overload resolution first
		ORP::Result call = ORP::resolveCall(list, arguments, info);
		std::vector<ORP::Candidate*> const& originalCands =
			call.getCandidates();
		
		// no multiple candidates?  in that case we do not touch the result
		if (call.getResult() != ORP::Result::kMultipleCandidates) {
			//~ debug.line(UTF8STR("no multiple candidates anyway"));
			return call;
		}
		
		// copy candidates to scratch list
		std::vector<ORP::Candidate*> cands;
		cands.resize(originalCands.size());
		std::copy(originalCands.begin(), originalCands.end(), cands.begin());
		
		// Find pairs of procedures that are the const and non-const versions
		// of each other, and eliminate one of them in each pair, based on the
		// type of instance we have.
		unsigned countAfterEliminations = cands.size();
		//~ debug.line(UTF8STR("Starting const/non-const elimination"));
		for (unsigned i = 0; i < cands.size(); ++i) {
			if (cands[i] == 0) {
				continue;
			}
			
			//~ IndentDebug did;
			
			Procedure const* p1 = cands[i]->getEntity();
			//~ debug.print(UTF8STR("p1: "));
			//~ debug.line(p1->headerToString());
			
			if (p1->getASS() == kStatic) {
				continue;
			}
			
			const std::vector<Parameter const*>& p1p = *p1->getParameterList();
			
			for (unsigned i2 = i + 1; i2 < cands.size(); ++i2) {
				if (cands[i2] == 0) {
					continue;
				}
				
				//~ IndentDebug did2;
				
				Procedure const* p2 = cands[i2]->getEntity();
				//~ debug.print(UTF8STR("p2: "));
				//~ debug.line(p2->headerToString());
				
				if (p2->getASS() == kStatic || p1->isConst() == p2->isConst())
				{
					continue;
				}
				
				const std::vector<Parameter const*>& p2p =
					*p2->getParameterList();
				
				if (p1p.size() != p2p.size()) {
					continue;
				}
				
				bool paramMatch = true;
				for (unsigned i3 = 0; i3 < p1p.size(); ++i3) {
					if (p1p[i3]->getType()->getOverloadHash()
						!= p2p[i3]->getType()->getOverloadHash())
					{
						paramMatch = false;
						break;
					}
				}
				if (!paramMatch) {
					continue;
				}
				
				// we found a match!
				//~ debug.line(UTF8STR("found a match!"));
				if (instance->whatInstance() != kConstInstance) {
					// eliminate the const procedure
					if (p1->isConst()) {
						cands[i] = 0; // p1
					}
					else {
						cands[i2] = 0; // p2
					}
				}
				else {
					// eliminate the non-const procedure
					if (p1->isConst()) {
						cands[i2] = 0; // p2
					}
					else {
						cands[i] = 0; // p1
					}
				}
				countAfterEliminations--;
				break;
			}
		}
		
		// multiple candidates left?
		if (countAfterEliminations > 1) {
			//~ debug.line(UTF8STR("still multiple procedure candidates left!"));
			std::vector<ORP::Candidate*> finalCands;
			finalCands.reserve(countAfterEliminations);
			for (unsigned i = 0; i < cands.size(); ++i) {
				if (cands[i] != 0) {
					finalCands.push_back(cands[i]);
				}
			}
			return Result(finalCands);
		}
		
		// only a single candidate left
		std::vector<ORP::Candidate*>::const_iterator it = cands.begin();
		while (*it == 0) {
			++it;
		}
		return Result(*it);
	}
	
	HYPER_Common::UTF8String
	ProcedureOverloadResolution::getCallPrototype(QualifiedName entityName,
		const std::vector<const Expression*>& args)
	{
		return ORP::getCallPrototype(entityName, args);
	}
	
	bool
	ProcedureOverloadResolution::checkAccess(Candidate const* c,
		ClassContext const& from, Location const* loc, CompileInfo& info)
	{
		return ORP::checkAccess(c, from, loc, info);
	}
	
}
