/*!
	\file qualifiedname.cxx
	\brief Implements the QualifiedName class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-06-19
// Last time modified:	2008-06-26

#include "qualifiedname.h"

#include "base/asserts.h"

using std::vector;

namespace HYPER_Compiler {
	
	QualifiedName::QualifiedName(HYPER_Common::UTF8String identifier)
	 : fStringRep(identifier)
	{
		HYPER_Common::CodePoint dot('.');
		
		hyp_assert(!identifier.isEmpty());
		hyp_assert(!identifier.startsWith(dot));
		hyp_assert(!identifier.endsWith(dot));
		
		// treat lengths 1 & 2 as a special case
		if (identifier.length() < 3) {
			fName = identifier;
			return;
		}
		
		// scan the identifier to see if it contains any dots  (first and last
		// chars are not dots !!)
		HYPER_Common::UTF8String::ConstIterator lastDot = identifier.end();
		{
			HYPER_Common::UTF8String::ConstIterator it = lastDot;
			--it;
			--it;
			for (; it != identifier.begin(); --it) {
				if (*it == dot) {
					lastDot = it;
					break; // exit 'for'
				}
			}
		}
		
		// if a dot was found...
		if (lastDot != identifier.end()) {
			fName = identifier.subString(lastDot.next());
			
			// scan the identifier again to create the parent qualifiers  (first
			// char is not a dot !!)
			HYPER_Common::UTF8String::ConstIterator it = identifier.begin();
			HYPER_Common::UTF8String::ConstIterator qualifierStart = it;
			for (;; ++it) {
				if (*it == dot) {
					fParent.push_back(identifier.subString(qualifierStart, it));
					
					if (it == lastDot) {
						break; // parent completed
					}
					
					// place qualifierStart behind the current dot
					qualifierStart = it.next();
				}
			}
		}
		else { // no dot found...
			fName = identifier;
		}
	}
	
	QualifiedName::QualifiedName(const QualifiedName & parent,
		HYPER_Common::UTF8String ident)
	 : fParent(parent.fParent), fName(ident),
		fStringRep(parent.fStringRep + UTF8STR(".") + ident)
	{
		fParent.push_back(parent.fName);
	}
	
	QualifiedName::QualifiedName(const QualifiedName & q)
	 : fParent(q.fParent), fName(q.fName), fStringRep(q.fStringRep)
	{
		// nothing here
	}
	
	QualifiedName::QualifiedName(
		std::vector<HYPER_Common::UTF8String> childParents,
		unsigned int childNameLength, HYPER_Common::UTF8String childStringRep)
	 : fParent(childParents), fName(childParents.back()),
		fStringRep(
			childStringRep.subString(
				childStringRep.begin(),
				childStringRep.length() - childNameLength - 1
			)
		)
	{
		fParent.pop_back();
	}
	
	QualifiedName::~QualifiedName() {
		// nothing here
	}
	
	void QualifiedName::addIdentifier(HYPER_Common::UTF8String identifier) {
		fParent.push_back(fName);
		fName = identifier;
		fStringRep = fStringRep + UTF8STR(".") + identifier;
	}
	
	HYPER_Common::UTF8String QualifiedName::getShortName() const {
		return fName;
	}
	
	HYPER_Common::UTF8String QualifiedName::getRootName() const {
		return (fParent.size() > 0 ? fParent[0] : fName);
	}
	
	HYPER_Common::UTF8String QualifiedName::toString() const {
		return fStringRep;
	}
	
	QualifiedName QualifiedName::getParent() const {
		hyp_assert(hasParent());
		return QualifiedName(fParent, fName.length(), fStringRep);
	}
	
	bool QualifiedName::hasParent() const {
		return fParent.size() > 0;
	}
	
	unsigned int QualifiedName::countNames() const {
		return (fParent.size() + 1);
	}
	
	HYPER_Common::UTF8String
	QualifiedName::operator[](unsigned int i) const
	{
		hyp_assert(i < countNames());
		
		if (i == fParent.size()) {
			return fName;
		}
		else {
			return fParent[i];
		}
	}
	
	bool QualifiedName::operator==(const QualifiedName & q) const {
		return fStringRep == q.fStringRep;
	}
	
	bool QualifiedName::operator!=(const QualifiedName & q) const {
		return fStringRep != q.fStringRep;
	}
	
	bool QualifiedName::operator<(const QualifiedName & q) const {
		return fStringRep < q.fStringRep;
	}
	
	QualifiedName& QualifiedName::operator=(const QualifiedName& q) {
		fParent = q.fParent;
		fName = q.fName;
		fStringRep = q.fStringRep;
		return *this;
	}
	
	QualifiedName QualifiedName::subset(unsigned int length) const {
		hyp_assert(length > 0);
		hyp_assert(length <= this->countNames());
		
		// copy requested?
		if (length == this->countNames()) {
			return QualifiedName(*this);
		}
		
		HYPER_Common::UTF8String subset;
		HYPER_Common::UTF8String::ConstIterator it = fStringRep.begin();
		for (; it != fStringRep.end(); ++it) {
			if ((*it).is(HYPER_Common::CodePoint::kDot)) {
				length--;
				if (length == 0) {
					return QualifiedName(
						fStringRep.subString(fStringRep.begin(), it)
					);
				}
			}
		}
		
		return QualifiedName(*this);
	}
	
	QualifiedName* QualifiedName::copy() const {
		return new QualifiedName(*this);
	}
}
