/*!
	\file realclass.cxx
	\author Kevin André
	\brief Implements the Real class
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-04-19
// Last time modified:	2008-12-26

#include "realclass.h"

#include "base/location.h"

#include "parameter.h"

namespace HYPER_Compiler {
	
	class Real::RealConstructor : public BuiltInConstructor {
	public:
		// default & copy constructor
		RealConstructor(Real& rc, bool defaultConstructor);
		
		// conversion constructors
		RealConstructor(Real& rc, BuiltInTypes::Primitive type);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParam;
	};
	
	class Real::PositiveOperator : public UnaryOperator {
	public:
		PositiveOperator(Real& rc);
		
		ConstantValue const*
			evaluateCompileTime(ConstantValue const* operand);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::NegativeOperator : public UnaryOperator {
	public:
		NegativeOperator(Real& rc);
		
		ConstantValue const*
			evaluateCompileTime(ConstantValue const* operand);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::EqualsOperator : public BinaryOperator {
	public:
		EqualsOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::DiffersOperator : public BinaryOperator {
	public:
		DiffersOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::LessOperator : public BinaryOperator {
	public:
		LessOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::LessEqOperator : public BinaryOperator {
	public:
		LessEqOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::GreaterOperator : public BinaryOperator {
	public:
		GreaterOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::GreaterEqOperator : public BinaryOperator {
	public:
		GreaterEqOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::AddOperator : public BinaryOperator {
	public:
		AddOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::SubtractOperator : public BinaryOperator {
	public:
		SubtractOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::MultiplyOperator : public BinaryOperator {
	public:
		MultiplyOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::DivideOperator : public BinaryOperator {
	public:
		DivideOperator(Real& rc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::AssignOp : public AssignmentOperator {
	public:
		AssignOp(Real& rc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::AddAssignOp : public AssignmentOperator {
	public:
		AddAssignOp(Real& rc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::SubtractAssignOp : public AssignmentOperator {
	public:
		SubtractAssignOp(Real& rc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::MultiplyAssignOp : public AssignmentOperator {
	public:
		MultiplyAssignOp(Real& rc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Real::DivideAssignOp : public AssignmentOperator {
	public:
		DivideAssignOp(Real& rc);
		
		// TODO : change return type
		void generateCode();
	};
	
	// *** Real ***
	
	Real::Real(CompileInfo& info)
	 : BuiltInClass(info, kSealed)
	{
		// no action
	}
	
	Real::~Real() {
		// no action
	}
	
	HYPER_Common::UTF8String Real::getTypeName() const {
		return UTF8STR("real");
	}
	
	HYPER_Common::UTF8String Real::getDeclarationName() const {
		return UTF8STR("real");
	}
	
	void Real::buildSymbolTable() {
		createConstructors();
		createMembers();
		createOperators();
	}
	
	void Real::createConstructors() {
		registerConstructor(new RealConstructor(*this, true), false);
		registerConstructor(new RealConstructor(*this, false), true);
		registerConstructor(
			new RealConstructor(*this, BuiltInTypes::kByte), false
		);
		registerConstructor(
			new RealConstructor(*this, BuiltInTypes::kInt16), false
		);
		registerConstructor(
			new RealConstructor(*this, BuiltInTypes::kInt32), false
		);
		registerConstructor(
			new RealConstructor(*this, BuiltInTypes::kNat16), false
		);
		registerConstructor(
			new RealConstructor(*this, BuiltInTypes::kSingle), false
		);
		registerConstructor(
			new RealConstructor(*this, BuiltInTypes::kDouble), false
		);
	}
	
	void Real::createMembers() {
		// none yet
	}
	
	void Real::createOperators() {
		registerOperator(new PositiveOperator(*this));
		registerOperator(new NegativeOperator(*this));
		
		registerOperator(new EqualsOperator(*this));
		registerOperator(new DiffersOperator(*this));
		registerOperator(new LessOperator(*this));
		registerOperator(new LessEqOperator(*this));
		registerOperator(new GreaterOperator(*this));
		registerOperator(new GreaterEqOperator(*this));
		registerOperator(new AddOperator(*this));
		registerOperator(new SubtractOperator(*this));
		registerOperator(new MultiplyOperator(*this));
		registerOperator(new DivideOperator(*this));
		
		registerOperator(new AssignOp(*this));
		registerOperator(new AddAssignOp(*this));
		registerOperator(new SubtractAssignOp(*this));
		registerOperator(new MultiplyAssignOp(*this));
		registerOperator(new DivideAssignOp(*this));
	}
	
	
	// *** RealConstructor ***
	
	Real::RealConstructor::RealConstructor(Real& rc, bool defaultConstructor)
	 : BuiltInConstructor(&rc, true)
	{
		if (!defaultConstructor) {
			fParam.push_back(
				rc.createParameter(UTF8STR("b"), BuiltInTypes::kReal)
			);
		}
	}
	
	Real::RealConstructor::RealConstructor(Real& rc,
		BuiltInTypes::Primitive type)
	 : BuiltInConstructor(&rc, true)
	{
		fParam.push_back(rc.createParameter(UTF8STR("x"), type));
	}
	
	const std::vector<Parameter const*>*
	Real::RealConstructor::getParameterList() const {
		return &fParam;
	}
	
	
	// *** PositiveOperator ***
	
	Real::PositiveOperator::PositiveOperator(Real& rc)
	 : UnaryOperator(Operators::kPositive, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::PositiveOperator::evaluateCompileTime(ConstantValue const* operand) {
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::PositiveOperator::generateCode() {
		// TODO
	}
	
	
	// *** NegativeOperator ***
	
	Real::NegativeOperator::NegativeOperator(Real& rc)
	 : UnaryOperator(Operators::kNegative, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::NegativeOperator::evaluateCompileTime(ConstantValue const* operand) {
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::NegativeOperator::generateCode() {
		// TODO
	}
	
	
	// *** EqualsOperator ***
	Real::EqualsOperator::EqualsOperator(Real& rc)
	 : BinaryOperator(Operators::kEquals, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), true, BuiltInTypes::createBool(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::EqualsOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::EqualsOperator::generateCode() {
		// TODO
	}
	
	
	// *** DiffersOperator ***
	Real::DiffersOperator::DiffersOperator(Real& rc)
	 : BinaryOperator(Operators::kDiffers, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), true, BuiltInTypes::createBool(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::DiffersOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::DiffersOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessOperator ***
	Real::LessOperator::LessOperator(Real& rc)
	 : BinaryOperator(Operators::kLess, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), false, BuiltInTypes::createBool(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::LessOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::LessOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessEqOperator ***
	Real::LessEqOperator::LessEqOperator(Real& rc)
	 : BinaryOperator(Operators::kLessOrEqual, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), false, BuiltInTypes::createBool(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::LessEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::LessEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterOperator ***
	Real::GreaterOperator::GreaterOperator(Real& rc)
	 : BinaryOperator(Operators::kGreater, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), false, BuiltInTypes::createBool(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::GreaterOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::GreaterOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterEqOperator ***
	Real::GreaterEqOperator::GreaterEqOperator(Real& rc)
	 : BinaryOperator(Operators::kGreaterOrEqual, BuiltInTypes::createReal(rc.getCompileInfo()),
		BuiltInTypes::createReal(rc.getCompileInfo()), false, BuiltInTypes::createBool(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::GreaterEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::GreaterEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** AddOperator ***
	Real::AddOperator::AddOperator(Real& rc)
	 : BinaryOperator(Operators::kAdd,
		BuiltInTypes::createReal(rc.getCompileInfo()), BuiltInTypes::createReal(rc.getCompileInfo()), true,
		BuiltInTypes::createReal(rc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::AddOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::AddOperator::generateCode() {
		// TODO
	}
	
	
	// *** SubtractOperator ***
	Real::SubtractOperator::SubtractOperator(Real& rc)
	 : BinaryOperator(Operators::kSubtract,
		BuiltInTypes::createReal(rc.getCompileInfo()), BuiltInTypes::createReal(rc.getCompileInfo()), false,
		BuiltInTypes::createReal(rc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::SubtractOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::SubtractOperator::generateCode() {
		// TODO
	}
	
	
	// *** MultiplyOperator ***
	Real::MultiplyOperator::MultiplyOperator(Real& rc)
	 : BinaryOperator(Operators::kMultiply,
		BuiltInTypes::createReal(rc.getCompileInfo()), BuiltInTypes::createReal(rc.getCompileInfo()), true,
		BuiltInTypes::createReal(rc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::MultiplyOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::MultiplyOperator::generateCode() {
		// TODO
	}
	
	
	// *** DivideOperator ***
	Real::DivideOperator::DivideOperator(Real& rc)
	 : BinaryOperator(Operators::kDivide,
		BuiltInTypes::createReal(rc.getCompileInfo()), BuiltInTypes::createReal(rc.getCompileInfo()), false,
		BuiltInTypes::createReal(rc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Real::DivideOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Real::DivideOperator::generateCode() {
		// TODO
	}
	
	
	// *** AssignOp ***
	Real::AssignOp::AssignOp(Real& rc)
	 : AssignmentOperator(Operators::kSimpleAssign, BuiltInTypes::createReal(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Real::AssignOp::generateCode() {
		// TODO
	}
	
	
	// *** AddAssignOp ***
	Real::AddAssignOp::AddAssignOp(Real& rc)
	 : AssignmentOperator(Operators::kAddAssign, BuiltInTypes::createReal(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Real::AddAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** SubtractAssignOp ***
	Real::SubtractAssignOp::SubtractAssignOp(Real& rc)
	 : AssignmentOperator(Operators::kSubtractAssign, BuiltInTypes::createReal(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Real::SubtractAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** MultiplyAssignOp ***
	Real::MultiplyAssignOp::MultiplyAssignOp(Real& rc)
	 : AssignmentOperator(Operators::kMultiplyAssign, BuiltInTypes::createReal(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Real::MultiplyAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** DivideAssignOp ***
	Real::DivideAssignOp::DivideAssignOp(Real& rc)
	 : AssignmentOperator(Operators::kDivideAssign, BuiltInTypes::createReal(rc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Real::DivideAssignOp::generateCode() {
		// TODO
	}
}
