/*!
	\file resolveable.cxx
	\brief Implements the Resolveable class
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-12-31
// Last time modified:	2008-06-26

#include "resolveable.h"

#include "base/easydiagnostics.h"

#include "compileinfo.h"

#include <deque>
#include <utility>

namespace HYPER_Compiler {
	
	/// Traces an execution of resolveable objects calling each other
	class Resolveable::Trace {
		friend class Resolveable;
		Trace const* fParent;
		Resolveable const* fR;
		bool fFull;
		
		Trace(Trace const& parent, Resolveable const* r, bool full)
		 : fParent(&parent), fR(r), fFull(full)
		{
			//
		}
		
		Trace()
		 : fParent(0), fR(0), fFull(true)
		{
			//
		}
	};
	
	Resolveable::Resolveable()
	 : fIFaceState(kNotStarted), fFullState(kNotStarted)
	{
		//
	}
	
	bool
	Resolveable::resolveInterface(Trace const& trace)
	{
		Trace t(trace, this, false);
		switch (fIFaceState) {
		case kNotStarted:
			fIFaceState = kStarted;
			{
				bool ok = doInterfaceResolve(t);
				fIFaceState = ok ? kFinished : kFailed;
			}
			break;
		case kStarted:
			circularDependencyError(t);
			return false;
		case kFinished:
		case kFailed:
			break;
		}
		
		return (fIFaceState == kFinished);
	}
	
	bool Resolveable::resolveFull(Trace const& trace) {
		Trace t(trace, this, true);
		switch (fFullState) {
		case kNotStarted:
			fFullState = kStarted;
			{
				bool ok = resolveInterface(trace);
				ok &= doFullResolve(t);
				fFullState = ok ? kFinished : kFailed;
			}
			break;
		case kStarted:
			circularDependencyError(t);
			return false;
		case kFinished:
		case kFailed:
			break;
		}
		
		return (fFullState == kFinished);
	}
	
	bool Resolveable::topLevelResolve() {
		return resolveFull(Trace());
	}
	
	bool Resolveable::resolveNotYetStarted() const {
		// Only check the state of interface-resolve, because a full resolve will
		// do the interface resolve first anyway.
		return fIFaceState == kNotStarted;
	}
	
	void
	Resolveable::circularDependencyError(Trace const& trace)
	{
		EasyDiagnosticRef ref = 
			getCompileInfo().diag().error(
				UTF8STR("circular dependency detected:"),
				trace.fParent->fR->getResolveLocation()
			);
		
		Trace const* curTrace = &trace;
		bool first = true;
		do {
			Resolveable const* curR = curTrace->fR;
			
			HYPER_Common::UTF8String descr;
			if (first) {
				descr = curR->getResolveDescription();
				first = false;
			}
			else {
				descr =
					UTF8STR("is a dependency of ")
						+ curR->getResolveDescription();
			}
			
			ref.addInfo(
				descr, curR->getResolveLocation()
			);
			
			if (curR == this && curTrace != &trace) {
				// stop when we arrive at ourself again in the dependency chain
				break;
			}
			
			curTrace = curTrace->fParent;
		} while (curTrace->fR != 0);
	}
}
