/*!
	\file returnstatement.cxx
	\brief Implements the ReturnStatement class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-12
// Last time modified:	2008-09-16

#include "returnstatement.h"

#include "expression.h"
#include "procedure.h"
#include "scopes.h"
#include "statementcontext.h"

#include "common/utf8textflowstream.h"

namespace HYPER_Compiler {
	
	ReturnStatement::ReturnStatement(Expression* returnExpression,
		StatementContext const* c, const Location* pos, CompileInfo& info)
	 : Statement(info, pos), fExpr(returnExpression), fContext(c)
	{
		// nothing here
	}
	
	ReturnStatement::~ReturnStatement() {
		// nothing here
	}
	
	HYPER_Common::UTF8String ReturnStatement::getSyntaxName() const {
		return UTF8STR("Return");
	}
	
	void ReturnStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.print(UTF8STR("Return   "));
		ind.line(fExpr ? fExpr->toString() : UTF8STR("<no result>"));
	}
	
	HYPER_Common::UTF8String ReturnStatement::getResolveDescription() const {
		return UTF8STR("%1 statement").compose(QUOTED_UTF8STR("return"));
	}
	
	void ReturnStatement::putInSymbolTable(ScopeLocation const& loc) {
		Statement::putInSymbolTable(loc);
		if (fExpr) {
			fExpr->putInSymbolTable(new LookupLocation(loc));
		}
	}
	
	bool ReturnStatement::doFullResolve(Trace const& trace) {
		bool resolveOK = true;
		
		if (fExpr) { // return <expr>
			resolveOK &= fExpr->resolveFull(trace);
			resolveOK &= fExpr->checkReturnsValue();
		}
		
		// check return statement in function of the place it is used
		resolveOK &= fContext->checkReturn(getLocation(), fExpr);
		
		return resolveOK;
	}
	
	SimpleProbability ReturnStatement::getReturnProbability() const {
		return SimpleProbability::always;
	}
}
