/**
	\file scopes.cxx
	\brief Implements the Scope class
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2008-01-04
// Date modified:	2008-06-26

#include "scopes.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "compileinfo.h"
#include "nameddeclareditem.h"

namespace HYPER_Compiler {
	
	static void
	diagDeclarationConflict(NamedDeclaredItem* original,
		NamedDeclaredItem* second, CompileInfo& info)
	{
		info.diag().error(
			second->getLocation(),
			UTF8STR("declaration of %1 conflicts with..."),
			second->getDescription()
		).addInfo(
			original->getLocation(),
			UTF8STR("...%1 here"),
			original->getDescription()
		);
	}
	
	
	// *** Scope *** //
	
	NamedDeclaredItem*
	Scope::findLocalSymbol(HYPER_Common::UTF8String const& s) const
	{
		NameMap::const_iterator it = fContent.find(s);
		if (it == fContent.end()) {
			return 0; // not found
		}
		return it->second;
	}
	
	bool Scope::registeredInParent() const {
		return fParent != 0;
	}
	
	NamedDeclaredItem*
	Scope::findSymbol(HYPER_Common::UTF8String const& s) const
	{
		if (NamedDeclaredItem* item = findLocalSymbol(s)) {
			return item;
		}
		else if (registeredInParent()) {
			return fParent->findSymbol(s);
		}
		else {
			return 0; // not found
		}
	}
	
	void Scope::registerInParent(Scope const& parent) {
		hyp_assert(!registeredInParent());
		fParent = &parent;
	}
	
	Scope::Scope()
	 : fParent(0)
	{
		//
	}
	
	bool Scope::registerSymbol(NamedDeclaredItem* symbol)
	{
		HYPER_Common::UTF8String name = symbol->getDeclarationName();
		NameMap::const_iterator it = fContent.find(name);
		if (it != fContent.end()) {
			return false;
		}
		fContent[name] = symbol;
		return true;
	}
	
	bool
	Scope::registerSymbolWithDiagnostic(NamedDeclaredItem* symbol,
		CompileInfo& info)
	{
		if (!registerSymbol(symbol)) {
			diagDeclarationConflict(
				findLocalSymbol(symbol->getDeclarationName()), symbol, info
			);
			return false;
		}
		else {
			return true;
		}
	}
	
	
	// *** OrderedScope *** //
	
	NamedDeclaredItem*
	OrderedScope::findLocalSymbol(HYPER_Common::UTF8String const& s,
		ScopeLocation const& loc) const
	{
		hyp_assert(&loc.getScope() == this && "location is of other scope");
		
		PositionType pos = loc.fPosInParent;
		while (pos > 0) {
			pos--;
			if (fContent[pos]->getDeclarationName() == s) {
				return fContent[pos];
			}
		}
		return 0; // not found
	}
	
	NamedDeclaredItem*
	OrderedScope::findSymbol(HYPER_Common::UTF8String const& s,
		ScopeLocation const& loc) const
	{
		if (NamedDeclaredItem* item = findLocalSymbol(s, loc)) {
			return item;
		}
		else if (fParentOption1) {
			return fParentOption1->findSymbol(s);
		}
		else if (fParentOption2) {
			return fParentOption2->getScope().findSymbol(s, *fParentOption2);
		}
		else {
			return 0; // not found
		}
	}
	
	OrderedScope::OrderedScope()
	 : fParentOption1(0), fParentOption2(0)
	{
		//
	}
	
	ScopeLocation const* OrderedScope::getEndLocation() {
		return new ScopeLocation(*this, fContent.size());
	}
	
	ScopeLocation const*
	OrderedScope::registerSymbol(NamedDeclaredItem* symbol)
	{
		ScopeLocation const* loc = getEndLocation();
		if (findLocalSymbol(symbol->getDeclarationName(), *loc)) {
			delete loc;
			return 0; // already such a name in this scope
		}
		fContent.push_back(symbol);
		return loc; // OK
	}
	
	bool
	OrderedScope::registerSymbolWithDiagnostic(NamedDeclaredItem* symbol,
		CompileInfo& info)
	{
		if (!registerSymbol(symbol)) {
			diagDeclarationConflict(
				findLocalSymbol(
					symbol->getDeclarationName(), *getEndLocation()
				),
				symbol, info
			);
			return false;
		}
		else {
			return true;
		}
	}
	
	bool OrderedScope::registeredInParent() const {
		return fParentOption1 || fParentOption2;
	}
	
	void OrderedScope::registerInParent(Scope const& parent) {
		hyp_assert(!registeredInParent());
		fParentOption1 = &parent;
	}
	
	void OrderedScope::registerInParent(ScopeLocation const& parent) {
		hyp_assert(!registeredInParent());
		fParentOption2 = &parent;
	}
}
