/**
	\file scopes.h
	\brief Contains the Scope class
	\author Kevin André
	
	Copyright (C) 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2008-01-04
// Date modified:	2008-02-29

#ifndef DEF_SCOPES_H
 #define DEF_SCOPES_H
 
 #include "common/utf8string.h"
 
 #include <vector>
 #include <map>

namespace HYPER_Compiler {
	
	class CompileInfo;
	class NamedDeclaredItem;
	
	/// Represents a scope containing named symbols
	class Scope {
	public:
		virtual ~Scope() {/*empty*/}
		
		virtual NamedDeclaredItem* findLocalSymbol(
			HYPER_Common::UTF8String const& s) const;
		
		virtual NamedDeclaredItem* findSymbol(
			HYPER_Common::UTF8String const& s) const;
		
	protected:
		Scope();
		
		bool registerSymbol(NamedDeclaredItem* symbol);
		bool registerSymbolWithDiagnostic(NamedDeclaredItem* symbol,
			CompileInfo& info);
		
		virtual void registerInParent(Scope const& parent);
		virtual bool registeredInParent() const;
		
	private:
		Scope const* fParent;
		typedef std::map<HYPER_Common::UTF8String, NamedDeclaredItem*> NameMap;
		NameMap fContent;
	};
	
	class OrderedScope;
	class ScopeLocation;
	
	/// Represents a scope containing named symbols in some order
	class OrderedScope {
		typedef std::vector<NamedDeclaredItem*> ContentType;
	public:
		typedef ContentType::size_type PositionType;
		
		virtual ~OrderedScope() {/*empty*/}
		
		virtual NamedDeclaredItem* findLocalSymbol(
			HYPER_Common::UTF8String const& s, ScopeLocation const& loc) const;
		
		virtual NamedDeclaredItem* findSymbol(HYPER_Common::UTF8String const& s,
			ScopeLocation const& loc) const;
		
		ScopeLocation const* getEndLocation();
		
		ScopeLocation const* registerSymbol(NamedDeclaredItem* symbol);
		bool registerSymbolWithDiagnostic(NamedDeclaredItem* symbol,
			CompileInfo& info);
		
	protected:
		OrderedScope();
		
		virtual void registerInParent(Scope const& parent);
		virtual void registerInParent(ScopeLocation const& parent);
		virtual bool registeredInParent() const;
		
	private:
		Scope const* fParentOption1;
		ScopeLocation const* fParentOption2;
		ContentType fContent;
	};
	
	/// Represents the location of something in an OrderedScope.
	class ScopeLocation {
	private:
		friend class OrderedScope;
		OrderedScope& fParent;
		OrderedScope::PositionType fPosInParent;
		
		ScopeLocation(OrderedScope& s, OrderedScope::PositionType p)
		 : fParent(s), fPosInParent(p) { /*empty*/ }
		
	public:
		OrderedScope& getScope() const { return fParent; }
		
		NamedDeclaredItem*
		findSymbol(HYPER_Common::UTF8String const& s) const
		{
			return fParent.findSymbol(s, *this);
		}
	};
	
	/// Represents a universal location, usable for the location of a Scope or
	/// an OrderedScope. Only usable for symbol lookup.
	class LookupLocation {
		Scope const* fUnordered;
		ScopeLocation const* fOrdered;
	public:
		LookupLocation(Scope const& s) : fUnordered(&s), fOrdered(0) {}
		LookupLocation(ScopeLocation const& s) : fUnordered(0), fOrdered(&s) {}
		
		NamedDeclaredItem*
		findSymbol(HYPER_Common::UTF8String const& s) const
		{
			if (fUnordered) {
				return fUnordered->findSymbol(s);
			}
			else {
				return fOrdered->findSymbol(s);
			}
		}
	};
}
#endif
