/*!
	\file simpleloopstatement.cxx
	\brief Implementation of the SimpleLoopStatement class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-09-13
// Last time modified:	2008-09-16

#include "simpleloopstatement.h"

#include "boolconditionexpression.h"
#include "statementblock.h"
#include "statementcontext.h"
#include "type.h"

#include "common/utf8textflowstream.h"

namespace HYPER_Compiler {
	
	SimpleLoopStatement::SimpleLoopStatement(const Location* pos,
		StatementContext const* context, CompileInfo& info)
	 : Statement(info, pos), fTest(0), fBlock(0), fContext(context)
	{
		// nothing here
	}
	
	SimpleLoopStatement::~SimpleLoopStatement() {
		// nothing here
	}
	
	HYPER_Common::UTF8String SimpleLoopStatement::getSyntaxName() const {
		return UTF8STR("Simple-Loop-Statement");
	}
	
	void SimpleLoopStatement::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.line(getSyntaxName());
		
		HYPER_Common::Indent out(ind);
		
		if (fCheckBeforeLoop) {
			ind.line(UTF8STR("test location: before loop"));
		}
		else {
			ind.line(UTF8STR("test location: after loop"));
		}
		
		ind.print(UTF8STR("test to exit loop: "));
		ind.print(fTest->toString());
		ind.line(fInvertTest ? UTF8STR(" [INVERTED]") : UTF8STR(""));
		
		fBlock->myPrint(ind);
	}
	
	void SimpleLoopStatement::setTest(bool checkBeforeLoop,
		BoolConditionExpression* exitLoopTest, bool mustInvertTest)
	{
		fCheckBeforeLoop = checkBeforeLoop;
		fTest = exitLoopTest;
		fInvertTest = mustInvertTest;
	}
	
	void SimpleLoopStatement::setBlock(StatementBlock* s) {
		fBlock = s;
	}
	
	StatementBlock* SimpleLoopStatement::getBlock() const {
		return fBlock;
	}
	
	HYPER_Common::UTF8String
	SimpleLoopStatement::getResolveDescription() const
	{
		return UTF8STR("%1 statement").compose(QUOTED_UTF8STR("loop"));
	}
	
	void SimpleLoopStatement::putInSymbolTable(ScopeLocation const& loc) {
		Statement::putInSymbolTable(loc);
		fTest->putInSymbolTable(new LookupLocation(loc));
		fBlock->putInSymbolTable(loc);
	}
	
	bool SimpleLoopStatement::doFullResolve(Trace const& trace) {
		bool resolveOK = fTest->resolveFull(trace);
		resolveOK &= fBlock->resolveFull(trace);
		return resolveOK;
	}
	
	SimpleProbability SimpleLoopStatement::getReturnProbability() const {
		TriBool exitTestValue = fTest->getResult().getBoolValue();
		if (fInvertTest) {
			exitTestValue = !exitTestValue;
		}
		
		SimpleProbability bodyP = fBlock->getReturnProbability();
		
		// TODO : take the chance for a break/continue into account
		if (fCheckBeforeLoop) {
			if (exitTestValue.isTrue()) {
				// loop is never entered
				return SimpleProbability::never;
			}
			
			// If the exit test is true in the first iteration, then the loop is
			// not entered.
			return SimpleProbability::never & bodyP;
		}
		else {
			return bodyP;
		}
	}
}
