/*!
	\file singleclass.cxx
	\author Kevin André
	\brief Implements the Single class
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-04-19
// Last time modified:	2008-12-26

#include "singleclass.h"

#include "base/location.h"

#include "parameter.h"

namespace HYPER_Compiler {
	
	class Single::SingleConstructor : public BuiltInConstructor {
	public:
		// default & copy constructor
		SingleConstructor(Single& sc, bool defaultConstructor);
		
		// conversion constructors
		SingleConstructor(Single& sc, BuiltInTypes::Primitive type);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParam;
	};
	
	class Single::PositiveOperator : public UnaryOperator {
	public:
		PositiveOperator(Single& sc);
		
		ConstantValue const*
			evaluateCompileTime(ConstantValue const* operand);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::NegativeOperator : public UnaryOperator {
	public:
		NegativeOperator(Single& sc);
		
		ConstantValue const*
			evaluateCompileTime(ConstantValue const* operand);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::EqualsOperator : public BinaryOperator {
	public:
		EqualsOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::DiffersOperator : public BinaryOperator {
	public:
		DiffersOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::LessOperator : public BinaryOperator {
	public:
		LessOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::LessEqOperator : public BinaryOperator {
	public:
		LessEqOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::GreaterOperator : public BinaryOperator {
	public:
		GreaterOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::GreaterEqOperator : public BinaryOperator {
	public:
		GreaterEqOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::AddOperator : public BinaryOperator {
	public:
		AddOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::SubtractOperator : public BinaryOperator {
	public:
		SubtractOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::MultiplyOperator : public BinaryOperator {
	public:
		MultiplyOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::DivideOperator : public BinaryOperator {
	public:
		DivideOperator(Single& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::AssignOp : public AssignmentOperator {
	public:
		AssignOp(Single& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::AddAssignOp : public AssignmentOperator {
	public:
		AddAssignOp(Single& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::SubtractAssignOp : public AssignmentOperator {
	public:
		SubtractAssignOp(Single& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::MultiplyAssignOp : public AssignmentOperator {
	public:
		MultiplyAssignOp(Single& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class Single::DivideAssignOp : public AssignmentOperator {
	public:
		DivideAssignOp(Single& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	
	// *** Single ***
	
	Single::Single(CompileInfo& info)
	 : BuiltInClass(info, kSealed)
	{
		// no action
	}
	
	Single::~Single() {
		// no action
	}
	
	HYPER_Common::UTF8String Single::getTypeName() const {
		return UTF8STR("single");
	}
	
	HYPER_Common::UTF8String Single::getDeclarationName() const {
		return UTF8STR("single");
	}
	
	void Single::buildSymbolTable() {
		createConstructors();
		createMembers();
		createOperators();
	}
	
	void Single::createConstructors() {
		registerConstructor(new SingleConstructor(*this, true), false);
		registerConstructor(new SingleConstructor(*this, false), true);
		registerConstructor(
			new SingleConstructor(*this, BuiltInTypes::kByte), false
		);
		registerConstructor(
			new SingleConstructor(*this, BuiltInTypes::kInt16), false
		);
		registerConstructor(
			new SingleConstructor(*this, BuiltInTypes::kInt32), false
		);
		registerConstructor(
			new SingleConstructor(*this, BuiltInTypes::kNat16), false
		);
	}
	
	void Single::createMembers() {
		// none yet
	}
	
	void Single::createOperators() {
		registerOperator(new PositiveOperator(*this));
		registerOperator(new NegativeOperator(*this));
		
		registerOperator(new EqualsOperator(*this));
		registerOperator(new DiffersOperator(*this));
		registerOperator(new LessOperator(*this));
		registerOperator(new LessEqOperator(*this));
		registerOperator(new GreaterOperator(*this));
		registerOperator(new GreaterEqOperator(*this));
		registerOperator(new AddOperator(*this));
		registerOperator(new SubtractOperator(*this));
		registerOperator(new MultiplyOperator(*this));
		registerOperator(new DivideOperator(*this));
		
		registerOperator(new AssignOp(*this));
		registerOperator(new AddAssignOp(*this));
		registerOperator(new SubtractAssignOp(*this));
		registerOperator(new MultiplyAssignOp(*this));
		registerOperator(new DivideAssignOp(*this));
	}
	
	
	// *** SingleConstructor ***
	
	Single::SingleConstructor::SingleConstructor(Single& sc,
		bool defaultConstructor)
	: BuiltInConstructor(&sc, true)
	{
		if (!defaultConstructor) {
			fParam.push_back(
				sc.createParameter(UTF8STR("b"), BuiltInTypes::kSingle)
			);
		}
	}
	
	Single::SingleConstructor::SingleConstructor(Single& sc,
		BuiltInTypes::Primitive type)
	: BuiltInConstructor(&sc, true)
	{
		fParam.push_back(sc.createParameter(UTF8STR("x"), type));
	}
	
	const std::vector<Parameter const*>*
	Single::SingleConstructor::getParameterList() const {
		return &fParam;
	}
	
	
	// *** PositiveOperator ***
	
	Single::PositiveOperator::PositiveOperator(Single& sc)
	 : UnaryOperator(Operators::kPositive, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::PositiveOperator::evaluateCompileTime(ConstantValue const* operand)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::PositiveOperator::generateCode() {
		// TODO
	}
	
	
	// *** NegativeOperator ***
	
	Single::NegativeOperator::NegativeOperator(Single& sc)
	 : UnaryOperator(Operators::kNegative, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::NegativeOperator::evaluateCompileTime(ConstantValue const* operand)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::NegativeOperator::generateCode() {
		// TODO
	}
	
	
	// *** EqualsOperator ***
	Single::EqualsOperator::EqualsOperator(Single& sc)
	 : BinaryOperator(Operators::kEquals, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), true, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::EqualsOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::EqualsOperator::generateCode() {
		// TODO
	}
	
	
	// *** DiffersOperator ***
	Single::DiffersOperator::DiffersOperator(Single& sc)
	 : BinaryOperator(Operators::kDiffers, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), true, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::DiffersOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::DiffersOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessOperator ***
	Single::LessOperator::LessOperator(Single& sc)
	 : BinaryOperator(Operators::kLess, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::LessOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::LessOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessEqOperator ***
	Single::LessEqOperator::LessEqOperator(Single& sc)
	 : BinaryOperator(Operators::kLessOrEqual, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::LessEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::LessEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterOperator ***
	Single::GreaterOperator::GreaterOperator(Single& sc)
	 : BinaryOperator(Operators::kGreater, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::GreaterOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::GreaterOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterEqOperator ***
	Single::GreaterEqOperator::GreaterEqOperator(Single& sc)
	 : BinaryOperator(Operators::kGreaterOrEqual, BuiltInTypes::createSingle(sc.getCompileInfo()),
		BuiltInTypes::createSingle(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::GreaterEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::GreaterEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** AddOperator ***
	Single::AddOperator::AddOperator(Single& sc)
	 : BinaryOperator(Operators::kAdd,
		BuiltInTypes::createSingle(sc.getCompileInfo()), BuiltInTypes::createSingle(sc.getCompileInfo()), true,
		BuiltInTypes::createSingle(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::AddOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::AddOperator::generateCode() {
		// TODO
	}
	
	
	// *** SubtractOperator ***
	Single::SubtractOperator::SubtractOperator(Single& sc)
	 : BinaryOperator(Operators::kSubtract,
		BuiltInTypes::createSingle(sc.getCompileInfo()), BuiltInTypes::createSingle(sc.getCompileInfo()), false,
		BuiltInTypes::createSingle(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::SubtractOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::SubtractOperator::generateCode() {
		// TODO
	}
	
	
	// *** MultiplyOperator ***
	Single::MultiplyOperator::MultiplyOperator(Single& sc)
	 : BinaryOperator(Operators::kMultiply,
		BuiltInTypes::createSingle(sc.getCompileInfo()), BuiltInTypes::createSingle(sc.getCompileInfo()), true,
		BuiltInTypes::createSingle(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::MultiplyOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::MultiplyOperator::generateCode() {
		// TODO
	}
	
	
	// *** DivideOperator ***
	Single::DivideOperator::DivideOperator(Single& sc)
	 : BinaryOperator(Operators::kDivide,
		BuiltInTypes::createSingle(sc.getCompileInfo()), BuiltInTypes::createSingle(sc.getCompileInfo()), false,
		BuiltInTypes::createSingle(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	Single::DivideOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void Single::DivideOperator::generateCode() {
		// TODO
	}
	
	
	// *** AssignOp ***
	Single::AssignOp::AssignOp(Single& sc)
	 : AssignmentOperator(Operators::kSimpleAssign, BuiltInTypes::createSingle(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Single::AssignOp::generateCode() {
		// TODO
	}
	
	
	// *** AddAssignOp ***
	Single::AddAssignOp::AddAssignOp(Single& sc)
	 : AssignmentOperator(Operators::kAddAssign, BuiltInTypes::createSingle(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Single::AddAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** SubtractAssignOp ***
	Single::SubtractAssignOp::SubtractAssignOp(Single& sc)
	 : AssignmentOperator(Operators::kSubtractAssign, BuiltInTypes::createSingle(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Single::SubtractAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** MultiplyAssignOp ***
	Single::MultiplyAssignOp::MultiplyAssignOp(Single& sc)
	 : AssignmentOperator(Operators::kMultiplyAssign, BuiltInTypes::createSingle(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Single::MultiplyAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** DivideAssignOp ***
	Single::DivideAssignOp::DivideAssignOp(Single& sc)
	 : AssignmentOperator(Operators::kDivideAssign, BuiltInTypes::createSingle(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void Single::DivideAssignOp::generateCode() {
		// TODO
	}
}
