/*!
	\file sourcefile.cxx
	\brief Implementation of SourceFile.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-04-05
// Last time modified:	2009-08-30

#include "sourcefile.h"

#include "base/asserts.h"
#include "base/debug.h"
#include "base/easydiagnostics.h"
#include "base/location.h"

#include "classrepository.h"
#include "compileinfo.h"
#include "implementedclass.h"
#include "implementedenum.h"
#include "namespace.h"
#include "outclass.h"
#include "overloadedprocedure.h"
#include "scopes.h"
#include "type.h"

namespace HYPER_Compiler {
	
	class SourceFile::AnonymousNamespace : public Scope {
	public:
		AnonymousNamespace(CompileInfo& info)
		 : fCompileInfo(info)
		{
			registerInParent(info.getRootScope());
		}
		
		void registerClass(Class* c) {
			registerSymbolWithDiagnostic(c, fCompileInfo);
			c->putInSymbolTable(*this);
		}
		
	private:
		CompileInfo& fCompileInfo;
	};
	
	SourceFile::SourceFile(FileID fileID,
		HYPER_Common::FilePath const& fileName, CompileInfo& info,
		bool isMainSourceFile, Namespace* ns)
	 : SyntaxItem(), fFileID(fileID), fFileName(fileName),
		fMainSource(isMainSourceFile), fBegin(0), fBeginClass(0),
		fCompileInfo(info), fContentScope(0), fNS(0), fAnonNS(0)
	{
		if (ns != 0) {
			fNS = ns;
			fContentScope = fNS;
		}
		else {
			// create anonymous 'namespace'
			fAnonNS = new AnonymousNamespace(info);
			fContentScope = fAnonNS;
		}
	}
	
	SourceFile::~SourceFile() {
		// no action
	}
	
	FileID SourceFile::getFileID() const {
		return fFileID;
	}
	
	HYPER_Common::FilePath const& SourceFile::getFileName() const {
		return fFileName;
	}
	
	HYPER_Common::UTF8String SourceFile::getSyntaxName() const {
		return UTF8STR("Source File");
	}
	
	void SourceFile::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.print(UTF8STR("begin of execution: "));
		ind.line(
			fBegin != 0 ? fBegin->fgQN.toString() : UTF8STR("<not specified>")
		);
		ind.print(UTF8STR("namespace: "));
		ind.line(
			fNS != 0 ? fNS->getFullName().toString() : UTF8STR("<anonymous>")
		);
		
		for (unsigned int i = 0; i < fEnums.size(); ++i) {
			fEnums[i]->myPrint(ind);
		}
		
		for (unsigned int i = 0; i < fClasses.size(); ++i) {
			fClasses[i]->myPrint(ind);
		}
	}
	
	void SourceFile::setBegin(QualifiedName const& name,
		const Location* loc)
	{
		fBegin = new QNLoc(name, loc);
	}
	
	void
	SourceFile::addImport(QualifiedName const& name, const Location* loc)
	{
		// look for an identical one in the list
		for (std::list<QNLoc>::iterator it2 = fImports.begin();
			it2 != fImports.end(); ++it2)
		{
			if (it2->fgQN == name) { // encountered before
				fCompileInfo.diag().error(UTF8STR("duplicate import"), loc);
				return;
			}
		}
		
		QNLoc imp(name, loc);
		fImports.push_back(imp);
		doImport(imp);
	}
	
	void
	SourceFile::addMember(ImplementedClass* c)
	{
		//debug.line("SourceFile::addMember called for class " + c->getDeclarationName());
		fClasses.push_back(c);
		if (fNS) {
			//debug.line("registering class in its namespace");
			fNS->registerClass(c);
		}
		else if (fAnonNS) {
			//debug.line("registering class in the file's anonymous namespace");
			fAnonNS->registerClass(c);
		}
	}
	
	void SourceFile::addMember(ImplementedEnum* e)
	{
		//debug.line("SourceFile::addMember called for enum " + c->getDeclarationName());
		fEnums.push_back(e);
		if (fNS) {
			//debug.line("registering class in its namespace");
			fNS->registerClass(e);
		}
		else if (fAnonNS) {
			//debug.line("registering class in the file's anonymous namespace");
			fAnonNS->registerClass(e);
		}
	}
	
	void SourceFile::doImport(QNLoc const& qni) {
		QualifiedName const& import = qni.fgQN;
		//debug.line("doImport called for " + import.toString());
		
		if (import[0] == UTF8STR("system")) { // standard library import?
			HYPER_Common::UTF8String strImp = import.toString();
			if (strImp == UTF8STR("system.stdio")) {
				fCompileInfo.getClassRepository().registerBuiltInClass(
					new Out(fCompileInfo),
					fCompileInfo.getSystemNamespace()
				);
			}
			else {
				fCompileInfo.diag().error(
					qni.fgLoc,
					UTF8STR("no such import in standard library: %1"),
					quote(strImp)
				);
			}
		}
		else {
			Debug::fixMe(
				UTF8STR("import %1 ignored").compose(quote(import.toString()))
			);
			fCompileInfo.diag().note(
				UTF8STR(
					"import %1 ignored because importing "
					"other sourcefiles is not supported yet"
				).compose(quote(import.toString())),
				qni.fgLoc
			);
		}
	}
	
	CompileInfo& SourceFile::getCompileInfo() const {
		return fCompileInfo;
	}
	
	HYPER_Common::UTF8String SourceFile::getResolveDescription() const {
		return UTF8STR("source file");
	}
	
	const Location* SourceFile::getResolveLocation() const {
		return Location::createSourceLoc(fFileID, 1, 1);
	}
	
	bool
	SourceFile::doInterfaceResolve(Trace const& trace)
	{
		//debug.line("SourceFile::doInterfaceResolve called");
		bool resolveOK = true;
		
		// build lists of operators in the repository
		fCompileInfo.getClassRepository().buildOperatorsLists();
		
		for (unsigned int i = 0; i < fClasses.size(); ++i) {
			resolveOK &= fClasses[i]->resolveInterface(trace);
		}
		
		//debug << "sourcefile interface resolve result: " << resolveOK << '\n';
		return resolveOK;
	}
	
	bool
	SourceFile::doFullResolve(Trace const& trace)
	{
		//debug.line("SourceFile::doFullResolve called");
		
		unsigned int classCount = fClasses.size();
		bool resolveOK = true;
		
		for (unsigned int i = 0; i < classCount; ++i) {
			resolveOK &= fClasses[i]->resolveFull(trace);
		}
		
		// no begin specified for a main source file
		if (fMainSource && fBegin == 0) {
			//debug.line("main sourcefile without explicit 'begin'");
			
			if (classCount == 0) {
				fCompileInfo.diag().error(
					UTF8STR(
						"this file is used as the program's start of "
							"execution, so it must contain at least one class"
					),
					Location::createSourceLoc(fFileID, 1, 1) // error at first char
				);
				resolveOK = false;
			}
			else if (classCount == 1) {
				//  if only one class in the file: take that one
				fBeginClass = fClasses[0];
			}
			else {
				// no begin is an error
				fCompileInfo.diag().error(
					UTF8STR(
						"no begin specified, but multiple classes in source "
							"file; cannot determine entry point of program"
					),
					fClasses[0]->getLocation() // location of the first class
				);
				resolveOK = false;
			}
		}
		
		if (fBegin != 0) {
			fBeginClass = findBeginClassSpecified();
			if (fBeginClass == 0) {
				resolveOK = false;
			}
		}
		
		if (fBeginClass != 0) { // we know the begin class
			// check for the procedure 'main'
			Procedure* mainP = findProcedureMain();
			if (mainP == 0) {
				resolveOK = false;
			}
			else {
				//debug.line("checking procedure 'main'");
				
				bool mainOK = true;
				
				if (mainP->getASS() != kStatic) { // not static ??
					fCompileInfo.diag().error(
						mainP->getLocation(),
						UTF8STR("procedure %1 must be static"),
						QUOTED_UTF8STR("main")
					);
					mainOK = false;
				}
				
				if (mainP->hasReturnType()) {
					Type const* returnT = mainP->getReturnType();
					Type const* byteT = BuiltInTypes::createByte(fCompileInfo);
					if (returnT != 0 && !returnT->nonConstEquals(byteT)) {
						fCompileInfo.diag().error(
							mainP->getLocation(),
							UTF8STR(
								"procedure %1 only supports %2 as its "
								"return type"
							),
							QUOTED_UTF8STR("main"),
							QUOTED_UTF8STR("byte")
						);
						mainOK = false;
					}
				}
				
				if (mainOK) {
					fBeginProc = mainP;
				}
				else {
					resolveOK = false;
				}
			}
		}
		
		//debug << "sourcefile full resolve result: " << resolveOK << '\n';
		return resolveOK;
	}
	
	Class* SourceFile::findBeginClassSpecified() {
		hyp_assert(fBegin != 0);
		
		HYPER_Common::UTF8String firstPart = fBegin->fgQN[0];
		Class* c = 0;
		// search in short names
		for (unsigned int i = 0; i < fClasses.size(); ++i) {
			ImplementedClass* cur = fClasses[i];
			
			if (firstPart == cur->getCaption()) { // match?
				c = cur;
				break;
			}
		}
		
		if (c == 0) {
			fCompileInfo.diag().error(
				fBegin->fgLoc,
				UTF8STR(
					"this source file does not contain a top-level class "
					"called %1, as is specified by %2"
				),
				quote(firstPart),
				QUOTED_UTF8STR("begin")
			);
			return 0;
		}
		
		if (fBegin->fgQN.countNames() == 1) {
			return c;
		}
		else {
			// find nested classes
			for (unsigned int i = 1; i < fBegin->fgQN.countNames(); ++i) {
				NamedDeclaredItem* item = c->findMember(fBegin->fgQN[i]);
				Class* nc = (item == 0) ? 0 : dynamic_cast<Class*>(item);
				
				if (nc == 0) { // member not found
					fCompileInfo.diag().error(
						fBegin->fgLoc,
						UTF8STR(
							"cannot find any class called %1 in %2, "
							"as specified by %3"
						),
						quote(fBegin->fgQN[i]),
						c->getDescription(),
						QUOTED_UTF8STR("begin")
					);
					return 0;
				}
				else {
					c = nc;
				}
			}
			
			return c;
		}
	}
	
	Procedure* SourceFile::findProcedureMain() {
		NamedDeclaredItem* item = fBeginClass->findMember(UTF8STR("main"));
		OverloadedProcedure* proc =
			(item == 0) ? 0 : dynamic_cast<OverloadedProcedure*>(item);
		
		if (proc == 0) { // no procedure called 'main' ?
			fCompileInfo.diag().error(
				fBeginClass->getLocation(),
				UTF8STR("no procedure %1 in %2"),
				QUOTED_UTF8STR("main"),
				fBeginClass->getDescription()
			);
			return 0;
		}
		else {
			const std::vector<Procedure*>& list = proc->getOverloadList();
			for (unsigned int i = 0; i < list.size(); ++i) {
				Procedure* p = list[i];
				if (p->hasNoParameters()) {
					// no args
					return p;
				}
			}
			
			//not found
			fCompileInfo.diag().error(
				fBeginClass->getLocation(),
				UTF8STR("no suitable overloaded procedure %1 in %2"),
				QUOTED_UTF8STR("main"),
				fBeginClass->getDescription()
			);
			return 0;
		}
	}
}
