/*!
	\file statementblock.cxx
	\brief Implementation of the StatementBlock class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-08-14
// Last time modified:	2008-06-26

#include "statementblock.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "common/utf8textflowstream.h"

#include "compileinfo.h"
#include "statement.h"

namespace HYPER_Compiler {
	
	StatementBlock::StatementBlock(CompileInfo& info, const Location* beginPos)
	 : fCompileInfo(info), fPos(beginPos), fReturnProb(SimpleProbability::never)
	{
		// nothing here
	}
	
	StatementBlock::~StatementBlock() {
		// nothing here
	}
	
	HYPER_Common::UTF8String StatementBlock::getSyntaxName() const {
		return UTF8STR("Block Of Statements");
	}
	
	void StatementBlock::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.line(getSyntaxName());
		HYPER_Common::Indent out(ind);
		
		for(unsigned int i = 0; i < fStatements.size(); ++i) {
			fStatements[i]->myPrint(ind);
		}
	}
	
	void StatementBlock::addStatement(Statement* s) {
		hyp_assert(s != 0 && "statement not null pointer");
		fStatements.push_back(s);
		s->putInSymbolTable(*this->getEndLocation());
	}
	
	bool StatementBlock::isEmpty() const {
		return fStatements.empty();
	}
	
	unsigned int StatementBlock::length() const {
		return fStatements.size();
	}
	
	const Location* StatementBlock::getBeginLocation() const {
		return fPos;
	}
	
	void StatementBlock::putInSymbolTable(Scope& parent) {
		registerInParent(parent);
	}
	
	void StatementBlock::putInSymbolTable(ScopeLocation const& loc) {
		registerInParent(loc);
	}
	
	CompileInfo& StatementBlock::getCompileInfo() const {
		return fCompileInfo;
	}
	
	HYPER_Common::UTF8String StatementBlock::getResolveDescription() const {
		return UTF8STR("statement block");
	}
	
	const Location* StatementBlock::getResolveLocation() const {
		return getBeginLocation();
	}
	
	bool StatementBlock::doInterfaceResolve(Trace const& trace) {
		// no action needed
		return true;
	}
	
	bool StatementBlock::doFullResolve(Trace const& trace) {
		bool deadCodeMsgGiven = false;
		bool resolveOK = true;
		
		for (unsigned int i = 0; i < fStatements.size(); ++i) {
			resolveOK &= fStatements[i]->resolveFull(trace);
			
			// get return probability of current statement
			SimpleProbability returnProb =
				fStatements[i]->getReturnProbability();
			
			// check for global return probability already at 'always'
			if (fReturnProb.isAlways()) {
				if (!deadCodeMsgGiven) {
					getCompileInfo().diag().error(
						UTF8STR("this statement will never be executed"),
						fStatements[i]->getLocation()
					);
					deadCodeMsgGiven = true;
					resolveOK = false;
				}
				// global return probability remains 'always'
			}
			else {
				fReturnProb |= returnProb; // 'add' to global return probability
			}
		}
		
		return resolveOK;
	}
	
	SimpleProbability StatementBlock::getReturnProbability() const {
		return fReturnProb;
	}
}
