/*!
	\file statementparser.cxx
	\brief Implements the StatementParser class.
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-09-22
// Last time modified:	2009-03-01

#include "statementparser.h"

#include "base/diagnostic.h"
#include "base/location.h"

#include "assignmentstatement.h"
#include "crementstatement.h"
#include "callstatement.h"
#include "declaredtype.h"
#include "expressionparser.h"
#include "ifstatement.h"
#include "initializer.h"
#include "iteratestatement.h"
#include "loopjumpstatement.h"
#include "megaloopstatement.h"
#include "returnstatement.h"
#include "simpleloopstatement.h"
#include "statementblock.h"
#include "statementcontext.h"
#include "typeparser.h"
#include "varstatement.h"

#include <stack>

using HYPER_Common::UTF8String;

namespace HYPER_Compiler {
	
	StatementParser::StatementParser(ParserInfo const& pi)
	 : ParserBase(pi)
	{
		//
	}
	
	void
	StatementParser::doBlockBody(StatementBlock& b,
		StatementContext const* ctxt)
	{
		while (true) {
			skipEmptyLines();
			
			Keywords::Keyword k = lookToken().getKeyword();
			
			if (k == Keywords::kEnd) {
				break;
			}
			else {
				Statement* s = doStatement(ctxt);
				if (s != 0) { // if statement parsed successfully...
					b.addStatement(s);
				}
			}
		} 
	}
	
	Statement* StatementParser::doStatement(StatementContext const* ctxt)
	{
		Statement* s = 0;
		
		Keywords::Keyword k = lookToken().getKeyword();
		
		if (k == Keywords::kBreak || k == Keywords::kContinue) {
			s = doJump(ctxt);
		}
		else if (k == Keywords::kConst || k == Keywords::kStatic || k == Keywords::kVar) {
			s = doVar(ctxt);
		}
		else if (k == Keywords::kIf) {
			s = doIf(ctxt);
		}
		else if (k == Keywords::kIterate) {
			s = doIterate(ctxt);
		}
		else if (k == Keywords::kLoop) {
			s = doLoop(ctxt);
		}
		else if (k == Keywords::kReturn) {
			s = doReturn(ctxt);
		}
		else if (k == Keywords::kAbstract) {
			fatal(error(
				UTF8STR("%1 cannot be used in a statement").compose(
					QUOTED_UTF8STR("abstract")
				),
				lookPos()
			));
		}
		else if (k == Keywords::kSealed) {
			fatal(error(
				UTF8STR("%1 cannot be used in a statement").compose(
					QUOTED_UTF8STR("sealed")
				),
				lookPos()
			));
		}
		else if (k == Keywords::kElse) {
			fatal(error(
				UTF8STR("%1 used outside %2 statement").compose(
					QUOTED_UTF8STR("else"), QUOTED_UTF8STR("if")
				),
				lookPos()
			));
		}
		else if (k == Keywords::kWhen) {
			fatal(error(
				UTF8STR("%1 used outside %2 construct").compose(
					QUOTED_UTF8STR("when"), QUOTED_UTF8STR("loop")
				),
				lookPos()
			));
		}
		else {
			// some statement not determined by initial keyword, or a syntax error
			s = doMiscStatement(ctxt);
		}
		
		return s;
	}
	
	Statement* StatementParser::doVar(StatementContext const* ctxt) {
		const Location* p = lookPos();
		
		Token curT = lookToken();
		bool isConst = false;
		bool isStatic = false;
		
		if (curT.isKeyword(Keywords::kConst)) {
			isConst = true;
			doKeyword(Keywords::kConst);
		}
		else if (curT.isKeyword(Keywords::kStatic)) {
			isStatic = true;
			doKeyword(Keywords::kStatic);
		}
		
		// Do "var" keyword if not abbreviated version
		if ((!isConst && !isStatic) || !lookToken().isIdentifier()) {
			doKeyword(Keywords::kVar); // require 'var' keyword
		}
		
		bool stop = false;
		std::vector<HYPER_Common::UTF8String> names;
		std::vector<Location const*> locs;
		
		do {
			locs.push_back(lookPos()); // variable location
			names.push_back(doIdentifier()); // variable name
			
			if (lookToken().isSymbol(Token::kAnd)) {
				nextToken();
			}
			else {
				stop = true;
			}
		} while (!stop);
		
		doSymbol(Token::kColon);
		TypeParser tp(this->parserInfo(), ctxt);
		DeclaredType* t = tp.doType(true);
		
		curT = lookToken();
		Expression* initExpr = 0;
		
		if (curT.isSymbol(Token::kEqual)) {
			nextToken(); // eat '=' symbol
			ExpressionParser ep(this->parserInfo(), ctxt);
			initExpr = ep.doExpression();
		}
		
		doNewLine();
		
		if (isConst) { // "const var" ...
			if (t->isConst()) {
				nonfatal(error(
					UTF8STR("duplicate %1").compose(QUOTED_UTF8STR("const")),
					p
				));
			}
			else {
				t->makeConst();
			}
		}
		
		Initializer* init;
		if (initExpr) {
			init = new Initializer(t, initExpr, ctxt, p);
		}
		else {
			init = new Initializer(t, ctxt, p);
		}
		
		Var* v =
			new Var(isStatic, names, locs, t, init, p, ctxt, getCompileInfo());
		
		return v;
	}
	
	Statement* StatementParser::doReturn(StatementContext const* ctxt) {
		const Location* p = lookPos();
		
		doKeyword(Keywords::kReturn); // require 'return' keyword
		
		Expression* e;
		
		if (!lookToken().isNewLine()) {
			ExpressionParser ep(this->parserInfo(), ctxt);
			e = ep.doExpression();
		}
		else {
			e = 0;
		}
		
		doNewLine();
		
		return new ReturnStatement(e, ctxt, p, getCompileInfo());
	}
	
	Statement* StatementParser::doIf(StatementContext const* ctxt) {
		IfStatement* is = new IfStatement(lookPos(), ctxt, getCompileInfo());
		BoolConditionExpression* test;
		ExpressionParser ep(this->parserInfo(), ctxt);
		
		doKeyword(Keywords::kIf); // require 'if' keyword
		test = ep.doBoolResultExpression();
		doNonfatalKeyword(Keywords::kThen);
		doNewLine();
		
		bool stop = false;
		bool hadElse = false;
		
		StatementBlock* block =
			new StatementBlock(getCompileInfo(), is->getLocation());
		is->addTest(test, block);
		
		do {
			skipEmptyLines();
			const Location* p = lookPos();
			
			if (lookToken().isKeyword(Keywords::kElse)) {  // "else" or "else if"
				nextToken(); // eat 'else' keyword
				
				// new statementblock for "else" or "else if"
				block = new StatementBlock(getCompileInfo(), p);
				
				if (lookToken().isKeyword(Keywords::kIf)) { // "else if"
					if (hadElse) {
						fatal(error(
							UTF8STR(
								"an \"else if\" is not allowed behind an "
								"\"else\""
							),
							lookPos()
						));
					}
					
					nextToken(); // eat 'if' keyword
					test = ep.doBoolResultExpression();
					doNonfatalKeyword(Keywords::kThen);
					is->addTest(test, block);
				}
				else {
					if (hadElse) {
						fatal(error(
							UTF8STR(
								"multiple \"else\" blocks are not allowed"
							),
							p
						));
					}
					
					test = 0; // no test for a simple "else"
					hadElse = true;
					
					is->setElse(block);
				}
				
				doNewLine();
			}
			else if (lookToken().isKeyword(Keywords::kEnd)) { // "end" or "end if"
				stop = true;
			}
			else { // some statement not concerning "if" and that begins with a keyword
				Statement* st = doStatement(ctxt);
				if (st != 0) {
					block->addStatement(st);
				}
			}
		} while (!stop);
		
		// end part
		doFlexibleEnd(Keywords::kIf, UTF8STR(""));
		doNewLine();
		
		return is;
	}
	
	Statement* StatementParser::doLoop(StatementContext const* ctxt) {
		const Location* beginP = lookPos();
		
		bool beginCondition = false;
		bool megaConditions = false;
		
		Statement* ls = 0;
		StatementBlock* block = new StatementBlock(getCompileInfo(), lookPos());
		
		doKeyword(Keywords::kLoop);
		
		ExpressionParser ep(this->parserInfo(), ctxt);
		StatementContext* newContext = ctxt->copy();
		
		if (lookToken().isKeyword(Keywords::kWhen)) {
			beginCondition = true;
			nextToken(); // eat 'when' keyword
			BoolConditionExpression* test = ep.doBoolResultExpression();
			SimpleLoopStatement* sls =
				new SimpleLoopStatement(beginP, ctxt, getCompileInfo());
			newContext->setCurrentLoop(sls);
			sls->setTest(true, test, true);
			sls->setBlock(block);
			ls = sls;
		}
		else if (!lookToken().isNewLine()) {
			fatal(errorExpected(
				UTF8STR("\"when\" condition or newline"), lookPos(),
				lookToken()
			));
		}
		
		doNewLine();
		
		bool stop = false;
		bool firstStatement = true;
		
		MegaLoopStatement* mls = 0;
		
		do {
			skipEmptyLines();
			
			Keywords::Keyword k = lookToken().getKeyword();
			UTF8String d = lookToken().getDescription();
			
			if (k == Keywords::kEnd) {
				stop = true;
			}
			else if (k == Keywords::kWhen) {
				const Location* p = lookPos();
				nextToken(); // eat 'when' keyword
				
				if (beginCondition) { // "when" found inside the loop, but this is a "loop when" loop!
					fatal(error(
						UTF8STR(
							"\"when\" is not allowed here, because the loop "
							"has a \"loop when <condition>\" form"
						),
						p
					));
				}
				
				if (!megaConditions) { // not (yet) in the state of a megaloop
					if (!firstStatement) { // first "when" in the loop, but not the first statement!
						fatal(error(
							UTF8STR(
								"\"when\" must be the first statement in "
								"the loop"
							),
							p
						));
					}
					else {
						mls =
							new MegaLoopStatement(
								beginP, ctxt, getCompileInfo()
							);
						newContext->setCurrentLoop(mls);
						ls = mls;
					}
				}
				
				megaConditions = true;
				BoolConditionExpression* test =
					ep.doBoolResultExpression();
				
				doNonfatalKeyword(Keywords::kThen);
				doNewLine();
				block = new StatementBlock(getCompileInfo(), p);
				mls->addTest(test, block);
			}
			else {
				Statement* st = doStatement(newContext);
				if (st != 0) {
					block->addStatement(st);
				}
			}
			
			firstStatement = false;
		} while (!stop);
		
		// end part
		doFlexibleEnd(Keywords::kLoop, UTF8STR(""));
		
		if (lookToken().isKeyword(Keywords::kWhen)) {
			if (beginCondition) {
				nonfatal(error(
					UTF8STR(
						"\"when\" is not allowed here, because the loop "
						"has a \"loop when <condition>\" form"
					),
					lookPos()
				));
				skipRestOfLine(true);
			}
			else if (megaConditions) {
				nonfatal(error(
					UTF8STR(
						"\"when\" is not allowed here, because the loop "
						"has \"when\" conditions in its body"
					),
					lookPos()
				));
				skipRestOfLine(true);
			}
			else {
				nextToken(); // eat 'when' keyword
				BoolConditionExpression* test =
					ep.doBoolResultExpression();
				SimpleLoopStatement* sls =
					new SimpleLoopStatement(beginP, ctxt, getCompileInfo());
				newContext->setCurrentLoop(sls);
				sls->setTest(false, test, false);
				sls->setBlock(block);
				ls = sls;
			}
		}
		else if (!beginCondition && !megaConditions) {
			fatal(error(
				UTF8STR("loop statement has no \"when\" condition(s)"), beginP
			));
		}
		
		doNewLine();
		
		return ls;
	}
	
	Statement* StatementParser::doIterate(StatementContext const* ctxt) {
		//debug.line("StatementParser::doIterate called");
		
		const Location* loc = lookPos();
		doKeyword(Keywords::kIterate);
		
		const Location* varLoc = lookPos();
		UTF8String name = doIdentifier();
		doKeyword(Keywords::kIn);
		
		ExpressionParser ep(this->parserInfo(), ctxt);
		
		doSymbol(Token::kOpenSquareParen);
		Expression* begin = ep.doExpression();
		
		// temporary extra check, for easier migration
		if (lookToken().isKeyword(Keywords::kTo)) {
			nonfatal(errorExpected(
				QUOTED_UTF8STR("..") + UTF8STR(" symbol"), lookPos(),
				lookToken()
			));
			nextToken(); // skip 'to'
		}
		else {
			doSymbol(Token::kDotDot);
		}
		
		Expression* end = ep.doExpression();
		doSymbol(Token::kCloseSquareParen);
		
		doNewLine();
		
		IterateStatement* iterate =
			IterateStatement::create(
				name, varLoc, begin, end, loc, ctxt, getCompileInfo()
			);
		
		StatementBlock* block = new StatementBlock(getCompileInfo(), lookPos());
		iterate->setBlock(block);
		
		StatementContext* newContext = ctxt->copy();
		newContext->setCurrentIterate(iterate);
		
		bool stop = false;
		do {
			skipEmptyLines();
			
			if (lookToken().isKeyword(Keywords::kEnd)) {
				stop = true;
			}
			else {
				Statement* st = doStatement(newContext);
				if (st != 0) {
					block->addStatement(st);
				}
			}
		} while (!stop);
		
		// end part
		doFlexibleEnd(Keywords::kIterate, UTF8STR(""));
		doNewLine();
		
		return iterate;
	}
	
	Statement* StatementParser::doJump(StatementContext const* ctxt) {
		const Location* p = lookPos();
		LoopJumpType t = kBreak;
		
		switch (lookToken().getKeyword()) {
		case Keywords::kBreak:
			// t already set
			break;
		case Keywords::kContinue:
			t = kContinue;
			break;
		default:
			fatal(errorExpected(
				UTF8STR("%1 or %2 statement").compose(
					QUOTED_UTF8STR("break"), QUOTED_UTF8STR("continue")
				),
				lookPos(), lookToken()
			));
			break;
		}
		
		nextToken(); // eat jump type keyword
		
		LoopJumpDestination d = kUnspecifiedJump;
		
		switch (lookToken().getKeyword()) {
		case Keywords::kLoop:
			d = kLoopJump;
			nextToken(); // eat 'loop' keyword
			break;
		case Keywords::kIterate:
			d = kIterateJump;
			nextToken(); // eat 'iterate' keyword
			break;
		default:
			// no destination specified
			break;
		}
		
		doNewLine();
		
		return new LoopJumpStatement(d, t, ctxt, p, getCompileInfo());
	}
	
	// Handles (function-)procedure calls, assignments and crements
	Statement*
	StatementParser::doMiscStatement(StatementContext const* ctxt)
	{
		//debug.line("trying to parse a miscellaneous statement");
		
		if (!ExpressionParser::canBeStartOfExpression(lookToken())) {
			nonfatal(error(UTF8STR("invalid statement"), lookPos()));
			skipRestOfLine(true);
			return 0;
		}
		
		Statement* si = 0;
		ExpressionParser ep(this->parserInfo(), ctxt);
		
		const Location* lValueLoc = lookPos();
		
		//debug.line("going to parse first part of statement (an expression)");
		
		// L-value or procedure call, a single expression (no direct binary or
		// unary ops)
		Expression* lv = ep.doSingleExpression();
		
		//debug.line("first part of statement parsed");
		
		const Location* operatorLoc = lookPos();
		
		// Handle assignment operators
		if (lookToken().isSymbol()) { // is this an assignment ?
			//UTF8String d = lookToken().getDescription();
			
			Operators::AssignmentOperator assOp =
				Operators::kSimpleAssign;
			Operators::CrementOperator crOp = Operators::kIncrement;
			
			bool isCrement = false;
			
			switch (lookToken().getSymbol()) {
			case Token::kEqual: // regular assignment
				// already set
				break;
			case Token::kAddAssign:
				assOp = Operators::kAddAssign;
				break;
			case Token::kSubtractAssign:
				assOp = Operators::kSubtractAssign;
				break;
			case Token::kMulAssign:
				assOp = Operators::kMultiplyAssign;
				break;
			case Token::kModuloAssign:
				assOp = Operators::kModuloAssign;
				break;
			case Token::kDivideAssign:
				assOp = Operators::kDivideAssign;
				break;
			case Token::kAndAssign:
				assOp = Operators::kAndAssign;
				break;
			case Token::kOrAssign:
				assOp = Operators::kOrAssign;
				break;
			case Token::kPointerEqual: // pointer assignment
				assOp = Operators::kPointerAssign;
				break;
			case Token::kPlusPlus:
				// already set
				isCrement = true;
				break;
			case Token::kMinusMinus:
				crOp = Operators::kDecrement;
				isCrement = true;
				break;
			default:
				nonfatal(errorExpected(
					UTF8STR(
						"assignment operator, increment/decrement operator or "
						"newline"
					),
					lookPos(), lookToken()
				));
				skipRestOfLine(true); // eat the rest of the line
				return 0;
			}
			
			// only for correct assignments the execution will continue here
			
			nextToken(); // eat assignment/crement operator symbol
			
			// don't expect an r-value for increment or decrement
			if (isCrement) {
				si =
					new CrementStatement(
						lv, crOp, operatorLoc, getCompileInfo()
					);
			}
			else {
				Expression* rv =
					ep.doExpression(); // Expression to assign (r-value)
				si =
					new AssignmentStatement(
						lv, rv, assOp, ctxt, operatorLoc, getCompileInfo()
					);
			}
			
			doNewLine(); // Statement is now done
		}
		else { // no, it is a procedure call
			doNewLine();
			
			//debug << "Creating CALL STATEMENT : call expr: " << (lv == 0 ? "NIL" : lv->toString()) << "   location: " << p->toString() << newl;
			si = new CallStatement(lv, lValueLoc, getCompileInfo());
		}
		
		return si;
	}
}
