/*!
	\file stringclass.cxx
	\author Kevin André
	\brief Implements the String class
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:	2005-10-08
// Date modified:	2008-12-26

#include "stringclass.h"

#include "base/location.h"

#include "arrayoperator.h"
#include "assignmentoperator.h"
#include "binaryoperator.h"
#include "builtinconstructor.h"
#include "builtinprocedure.h"
#include "builtintypes.h"
#include "constantvalue.h"
#include "parameter.h"
#include "type.h"

#include <boost/assign/list_of.hpp>

namespace HYPER_Compiler {
	
	class String::StringConstructor : public BuiltInConstructor {
	public:
		StringConstructor(String& sc, bool defaultConstructor);
		
		// conversion constructors
		StringConstructor(String& sc, BuiltInTypes::Primitive type);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		bool fDefaultConstr;
		std::vector<Parameter const*> fParam;
	};
	
	class String::ProcedureClear : public BuiltInProcedure {
	public:
		ProcedureClear(String& sc);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
		Type* fReturnType;
	};
	
	class String::ProcedureIsEmpty : public BuiltInProcedure {
	public:
		ProcedureIsEmpty(String& sc);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
		Type* fReturnType;
	};
	
	class String::ProcedureSize : public BuiltInProcedure {
	public:
		ProcedureSize(String& sc);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
		Type* fReturnType;
	};
	
	class String::ProcedureSubString : public BuiltInProcedure {
	public:
		ProcedureSubString(String& sc);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
		Type* fReturnType;
	};
	
	class String::ProcedureLeft : public BuiltInProcedure {
	public:
		ProcedureLeft(String& sc);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
		Type* fReturnType;
	};
	
	class String::ProcedureRight : public BuiltInProcedure {
	public:
		ProcedureRight(String& sc);
		
		const std::vector<Parameter const*>* getParameterList() const;
		
	private:
		std::vector<Parameter const*> fParams;
		Type* fReturnType;
	};
	
	class String::EqualsOperator : public BinaryOperator {
	public:
		EqualsOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::DiffersOperator : public BinaryOperator {
	public:
		DiffersOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::LessOperator : public BinaryOperator {
	public:
		LessOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::LessEqOperator : public BinaryOperator {
	public:
		LessEqOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::GreaterOperator : public BinaryOperator {
	public:
		GreaterOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::GreaterEqOperator : public BinaryOperator {
	public:
		GreaterEqOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::AddOperator : public BinaryOperator {
	public:
		AddOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* leftOp,
			ConstantValue const* rightOp);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::IndexOperator : public ArrayOperator {
	public:
		IndexOperator(String& sc);
		
		ConstantValue const*
		evaluateCompileTime(ConstantValue const* operand,
			std::vector<ConstantValue const*> params);
		
		// TODO : change return type
		virtual void generateCode();
	};
	
	class String::StringAssignOp : public AssignmentOperator {
	public:
		StringAssignOp(String& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::CharAddAssignOp : public AssignmentOperator {
	public:
		CharAddAssignOp(String& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	class String::StringAddAssignOp : public AssignmentOperator {
	public:
		StringAddAssignOp(String& sc);
		
		// TODO : change return type
		void generateCode();
	};
	
	
	// *** String class ***/
	
	String::String(CompileInfo& info)
	 : BuiltInClass(info, kSealed)
	{
		// no action
	}
	
	String::~String() {
		// no action
	}
	
	HYPER_Common::UTF8String String::getTypeName() const {
		return UTF8STR("string");
	}
	
	HYPER_Common::UTF8String String::getDeclarationName() const {
		return UTF8STR("string");
	}
	
	void String::buildSymbolTable() {
		createConstructors();
		createMembers();
		createOperators();
	}
	
	void String::createConstructors() {
		registerConstructor(new StringConstructor(*this, true), false);
		registerConstructor(new StringConstructor(*this, false), true);
		
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kByte), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kChar), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kDouble), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kInt), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kInt16), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kInt32), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kInt64), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kNat), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kNat16), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kNat32), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kNat64), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kReal), false
		);
		registerConstructor(
			new StringConstructor(*this, BuiltInTypes::kSingle), false
		);
	}
	
	void String::createMembers() {
		// create member procedures
		registerProcedure(new ProcedureClear(*this));
		registerProcedure(new ProcedureIsEmpty(*this));
		registerProcedure(new ProcedureLeft(*this));
		registerProcedure(new ProcedureRight(*this));
		registerProcedure(new ProcedureSize(*this));
		registerProcedure(new ProcedureSubString(*this));
	}
	
	void String::createOperators() {
		registerOperator(new IndexOperator(*this));
		
		registerOperator(new EqualsOperator(*this));
		registerOperator(new DiffersOperator(*this));
		registerOperator(new LessOperator(*this));
		registerOperator(new LessEqOperator(*this));
		registerOperator(new GreaterOperator(*this));
		registerOperator(new GreaterEqOperator(*this));
		registerOperator(new AddOperator(*this));
		
		registerOperator(new StringAssignOp(*this));
		registerOperator(new CharAddAssignOp(*this));
		registerOperator(new StringAddAssignOp(*this));
	}
	
	
	/* StringConstructor class */
	
	String::StringConstructor::StringConstructor(String& sc,
		bool defaultConstructor)
	 : BuiltInConstructor(&sc, true), fDefaultConstr(defaultConstructor)
	{
		if (!fDefaultConstr) {
			fParam.push_back(
				sc.createParameter(UTF8STR("s"), BuiltInTypes::kString)
			);
		}
	}
	
	String::StringConstructor::StringConstructor(String& sc, BuiltInTypes::Primitive type)
	 : BuiltInConstructor(&sc, false)
	{
		fParam.push_back(sc.createParameter(UTF8STR("x"), type));
	}
	
	const std::vector<Parameter const*>*
	String::StringConstructor::getParameterList() const {
		return &fParam;
	}
	
	
	// *** ProcedureClear ***
	
	String::ProcedureClear::ProcedureClear(String& sc)
	 : BuiltInProcedure(
			&sc, UTF8STR("clear"), kSealed, ClassMemberAccess::createPublic(),
			false, 0
		)
	{
		// no action
	}
	
	const std::vector<Parameter const*>*
	String::ProcedureClear::getParameterList() const {
		return &fParams;
	}
	
	
	// *** ProcedureIsEmpty ***
	
	String::ProcedureIsEmpty::ProcedureIsEmpty(String& sc)
	 : BuiltInProcedure(
			&sc, UTF8STR("isEmpty"), kSealed, ClassMemberAccess::createPublic(),
			true, BuiltInTypes::createBool(sc.getCompileInfo())
		)
	{
		// no action
	}
	
	const std::vector<Parameter const*>*
	String::ProcedureIsEmpty::getParameterList() const {
		return &fParams;
	}
	
	
	// *** ProcedureSize ***
	
	String::ProcedureSize::ProcedureSize(String& sc)
	 : BuiltInProcedure(
			&sc, UTF8STR("size"), kSealed, ClassMemberAccess::createPublic(),
			true, BuiltInTypes::createNat(sc.getCompileInfo())
		)
	{
		// no action
	}
	
	const std::vector<Parameter const*>*
	String::ProcedureSize::getParameterList() const {
		return &fParams;
	}
	
	
	/* ProcedureSubString class */
	
	String::ProcedureSubString::ProcedureSubString(String& sc)
	 : BuiltInProcedure(
			&sc, UTF8STR("subString"), kSealed,
			ClassMemberAccess::createPublic(),
			true, BuiltInTypes::createString(sc.getCompileInfo())
		)
	{
		fParams.push_back(
			sc.createParameter(UTF8STR("start"), BuiltInTypes::kNat)
		);
		fParams.push_back(
			sc.createParameter(UTF8STR("length"), BuiltInTypes::kNat)
		);
	}
	
	const std::vector<Parameter const*>*
	String::ProcedureSubString::getParameterList() const {
		return &fParams;
	}
	
	
	/* ProcedureLeft class */
	
	String::ProcedureLeft::ProcedureLeft(String& sc)
	 : BuiltInProcedure(
			&sc, UTF8STR("left"), kSealed, ClassMemberAccess::createPublic(),
			true, BuiltInTypes::createString(sc.getCompileInfo())
		)
	{
		fParams.push_back(
			sc.createParameter(UTF8STR("size"), BuiltInTypes::kNat)
		);
	}
	
	const std::vector<Parameter const*>*
	String::ProcedureLeft::getParameterList() const {
		return &fParams;
	}
	
	
	/* ProcedureRight class */
	
	String::ProcedureRight::ProcedureRight(String& sc)
	 : BuiltInProcedure(
			&sc, UTF8STR("right"), kSealed, ClassMemberAccess::createPublic(),
			true, BuiltInTypes::createString(sc.getCompileInfo())
		)
	{
		fParams.push_back(
			sc.createParameter(UTF8STR("size"), BuiltInTypes::kNat)
		);
	}
	
	const std::vector<Parameter const*>*
	String::ProcedureRight::getParameterList() const {
		return &fParams;
	}
	
	
	// *** EqualsOperator ***
	String::EqualsOperator::EqualsOperator(String& sc)
	 : BinaryOperator(Operators::kEquals, BuiltInTypes::createString(sc.getCompileInfo()),
		BuiltInTypes::createString(sc.getCompileInfo()), true, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	String::EqualsOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(
				leftOp->getSigned() == rightOp->getSigned()
			);
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void String::EqualsOperator::generateCode() {
		// TODO
	}
	
	
	// *** DiffersOperator ***
	String::DiffersOperator::DiffersOperator(String& sc)
	 : BinaryOperator(Operators::kDiffers, BuiltInTypes::createString(sc.getCompileInfo()),
		BuiltInTypes::createString(sc.getCompileInfo()), true, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	String::DiffersOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(
				leftOp->getSigned() != rightOp->getSigned()
			);
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void String::DiffersOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessOperator ***
	String::LessOperator::LessOperator(String& sc)
	 : BinaryOperator(Operators::kLess, BuiltInTypes::createString(sc.getCompileInfo()),
		BuiltInTypes::createString(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	String::LessOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(
				leftOp->getSigned() < rightOp->getSigned()
			);
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void String::LessOperator::generateCode() {
		// TODO
	}
	
	
	// *** LessEqOperator ***
	String::LessEqOperator::LessEqOperator(String& sc)
	 : BinaryOperator(Operators::kLessOrEqual, BuiltInTypes::createString(sc.getCompileInfo()),
		BuiltInTypes::createString(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	String::LessEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(
				leftOp->getSigned() <= rightOp->getSigned()
			);
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void String::LessEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterOperator ***
	String::GreaterOperator::GreaterOperator(String& sc)
	 : BinaryOperator(Operators::kGreater, BuiltInTypes::createString(sc.getCompileInfo()),
		BuiltInTypes::createString(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	String::GreaterOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(
				leftOp->getSigned() > rightOp->getSigned()
			);
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void String::GreaterOperator::generateCode() {
		// TODO
	}
	
	
	// *** GreaterEqOperator ***
	String::GreaterEqOperator::GreaterEqOperator(String& sc)
	 : BinaryOperator(Operators::kGreaterOrEqual, BuiltInTypes::createString(sc.getCompileInfo()),
		BuiltInTypes::createString(sc.getCompileInfo()), false, BuiltInTypes::createBool(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	String::GreaterEqOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		if (leftOp && rightOp) {
			return new ConstantValue(
				leftOp->getSigned() >= rightOp->getSigned()
			);
		}
		else {
			return 0; // don't know result
		}
	}
	
	// TODO : change return type
	void String::GreaterEqOperator::generateCode() {
		// TODO
	}
	
	
	// *** AddOperator ***
	String::AddOperator::AddOperator(String& sc)
	 : BinaryOperator(Operators::kAdd,
		BuiltInTypes::createString(sc.getCompileInfo()), BuiltInTypes::createString(sc.getCompileInfo()), true,
		BuiltInTypes::createString(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	ConstantValue const*
	String::AddOperator::evaluateCompileTime(ConstantValue const* leftOp,
		ConstantValue const* rightOp)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void String::AddOperator::generateCode() {
		// TODO
	}
	
	
	// *** IndexOperator ***
	String::IndexOperator::IndexOperator(String& sc)
	 : ArrayOperator(
			boost::assign::list_of(
				sc.createParameter(UTF8STR("index"), BuiltInTypes::kNat)
			),
			Type::createSinglePointer(
				false, BuiltInTypes::createChar(sc.getCompileInfo())
			),
			Location::getBuiltIn()
		)
	{
		// no action
	}
	
	ConstantValue const*
	String::IndexOperator::evaluateCompileTime(ConstantValue const* operand,
		std::vector<ConstantValue const*> params)
	{
		// TODO
		return 0;
	}
	
	// TODO : change return type
	void String::IndexOperator::generateCode() {
		// TODO
	}
	
	
	// *** StringAssignOp ***
	String::StringAssignOp::StringAssignOp(String& sc)
	 : AssignmentOperator(Operators::kSimpleAssign, BuiltInTypes::createString(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void String::StringAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** CharAddAssignOp ***
	String::CharAddAssignOp::CharAddAssignOp(String& sc)
	 : AssignmentOperator(Operators::kAddAssign, BuiltInTypes::createChar(sc.getCompileInfo()),
		Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void String::CharAddAssignOp::generateCode() {
		// TODO
	}
	
	
	// *** StringAddAssignOp ***
	String::StringAddAssignOp::StringAddAssignOp(String& sc)
	 : AssignmentOperator(Operators::kAddAssign,
		BuiltInTypes::createString(sc.getCompileInfo()), Location::getBuiltIn())
	{
		// no action
	}
	
	// TODO : change return type
	void String::StringAddAssignOp::generateCode() {
		// TODO
	}
}
