/*!
	\file textdiagnosticprinter.cxx
	\brief Implements the TextDiagnosticPrinter class
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 		2007-08-28
// Last time modified: 	2008-06-26

#include "textdiagnosticprinter.h"

#include "base/diagnostic.h"
#include "base/errormanager.h"
#include "base/location.h"

#include "common/filepath.h"
#include "common/utf8textflowstream.h"

#include "sourcefile.h"

using HYPER_Common::CodePoint;
using HYPER_Common::FilePath;
using HYPER_Common::Indent;
using HYPER_Common::PushStream;
using HYPER_Common::UTF8String;
using HYPER_Common::UTF8StringBuilder;
using HYPER_Common::UTF8TextFlowStream;

namespace HYPER_Compiler {
	
	TextDiagnosticPrinter::TextDiagnosticPrinter(
		ErrorManager const& errMan)
	 : fEM(errMan)
	{
		//
	}
	
	FilePath const*
	TextDiagnosticPrinter::getFile(Location const& loc) const
	{
		return fEM.getFileName(loc.getFileID());
	}
	
	void
	TextDiagnosticPrinter::printReducedLocation(Location const& loc,
		FilePath const* prefix, Location const* omitLocation,
		UTF8TextFlowStream& out) const
	{
		if (prefix != 0 && !loc.isSpecial()) {
			FilePath const* p = getFile(loc)->getRelative(*prefix);
			if (p) {
				UTF8String fileName = p->toString();
				
				// omit filename if equal to the one from omitLocation
				if (omitLocation != 0 && !omitLocation->isSpecial()
						&& loc.getFileID() == omitLocation->getFileID())
				{
					// create a placeholder that consists of all spaces
					UTF8StringBuilder sb;
					for (unsigned i = 0; i < fileName.length(); ++i) {
						sb += CodePoint(' ');
					}
					
					fileName = sb.toString();
				}
				
				out.print(
					UTF8STR("%1:%2:%3").compose(
						fileName, loc.getLine(), loc.getColumn()
					)
				);
				return;
			}
		}
		
		// not removing prefix
		fEM.print(out, loc);
	}
	
	void TextDiagnosticPrinter::printList(
		UTF8TextFlowStream& out) const
	{
		typedef ErrorManager::DiagnosticIterator EMDI;
		
		// create the list of paths
		std::list<FilePath const*> paths;
		for (EMDI it = fEM.begin(); it != fEM.end(); ++it) {
			paths.push_back(getFile(it->getLocation()));
		}
		
		//debug << "Paths: " << paths.size() << '\n';
		FilePath const* prefix = FilePath::getLargestCommonPrefix(paths);
		
		if (prefix != 0) { // print common prefix first, if present
			out.line(
				UTF8STR("In directory %1:").compose(
					doubleQuote(prefix->toString())
				)
			);
			out.line();
		}
		
		for (EMDI it = fEM.begin(); it != fEM.end(); ++it) {
			Diagnostic const& d = *it;
			Location const& loc = d.getLocation();
			
			printReducedLocation(loc, prefix, 0, out);
			out.print(UTF8STR(": %1: ").compose(d.getTypeString()));
			{
				Indent ind(out, 2); // indent 2 spaces if wrapped to next line
				out.line(d.getData().message);
			}
			
			for (Diagnostic::InfoIterator it = d.getInfoBegin();
				it != d.getInfoEnd(); ++it)
			{
				DiagnosticPair const& curInfo = *it;
				printReducedLocation(*curInfo.location, prefix, &loc, out);
				out.print(UTF8STR(": "));
				Indent ind(out, 2); // indent 2 spaces if wrapped to next line
				out.line(curInfo.message);
			}
			
			out.line();
		}
	}
	
	void TextDiagnosticPrinter::printSummary(
		UTF8TextFlowStream& out) const
	{
		if (fEM.count() == 0) { // no diagnostics present
			return; // no output
		}
		
		unsigned errors = fEM.count(Diagnostic::kError);
		unsigned warnings = fEM.count(Diagnostic::kWarning);
		unsigned notes = fEM.count(Diagnostic::kNote);
		bool printing = false;
		
		out.print(UTF8STR("*** ")); // line prefix
		
		if (errors) { // errors present
			printing = true;
			out.print(UTF8STR("Build failed *** "));
			out.print(errors);
			out.print(UTF8STR(" "));
			out.print(errors == 1 ? UTF8STR("error") : UTF8STR("errors"));
		}
		
		if (warnings) { // warnings present
			if (printing) { // separator needed?
				out.print(UTF8STR(", "));
			}
			printing = true;
			
			out.print(warnings);
			out.print(UTF8STR(" "));
			out.print(warnings == 1 ? UTF8STR("warning") : UTF8STR("warnings"));
		}
		
		if (notes) {
			if (printing) { // separator needed?
				out.print(UTF8STR(", "));
			}
			printing = true;
			
			out.print(notes);
			out.print(UTF8STR(" "));
			out.print(notes == 1 ? UTF8STR("note") : UTF8STR("notes"));
		}
		
		out.line();
	}
}
