/*!
	\file tribool.cxx
	\brief Implements the TriBool class
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-07-23
// Last time modified:	2008-06-26

#include "tribool.h"

#include "base/asserts.h"

namespace HYPER_Compiler {
	
	const TriBool TriBool::uncertain = TriBool(kUncertain);
	
	TriBool::TriBool()
	 : fValue(kFalse)
	{
		// no action
	}
	
	TriBool::TriBool(bool b)
	 : fValue(b ? kTrue : kFalse)
	{
		// no action
	}
	
	bool TriBool::toBool(bool uncertainValue) const {
		switch (fValue) {
		case kTrue:
			return true;
		case kFalse:
			return false;
		case kUncertain:
			return uncertainValue;
		default:
			hyp_assert_unreachable();
			return false; // dummy return
		}
	}
	
	TriBool TriBool::operator!() const {
		switch (fValue) {
		case kTrue:
			return TriBool(false);
		case kFalse:
			return TriBool(true);
		case kUncertain:
			return TriBool(kUncertain);
		default:
			hyp_assert_unreachable();
			return TriBool(false); // dummy return
		}
	}
	
	TriBool TriBool::operator==(TriBool b) const {
		if (isCertain() && b.isCertain()) {
			return fValue == b.fValue;
		}
		else {
			return TriBool(kUncertain);
		}
	}
	
	TriBool TriBool::operator!=(TriBool b) const {
		return !operator==(b);
	}
	
	TriBool& TriBool::operator&=(TriBool b) {
		if (isFalse() || b.isFalse()) {
			fValue = kFalse;
		}
		else if (isTrue() && b.isTrue()) {
			fValue = kTrue;
		}
		else {
			fValue = kUncertain;
		}
		
		return *this;
	}
	
	TriBool TriBool::operator&&(TriBool b) const {
		TriBool tmp = *this;
		tmp &= b;
		return tmp;
	}
	
	TriBool& TriBool::operator|=(TriBool b) {
		if (isTrue() || b.isTrue()) {
			fValue = kTrue;
		}
		else if (isFalse() && b.isFalse()) {
			fValue = kFalse;
		}
		else {
			fValue = kUncertain;
		}
		
		return *this;
	}
	
	TriBool TriBool::operator||(TriBool b) const {
		TriBool tmp = *this;
		tmp |= b;
		return tmp;
	}
}
