/*!
	\file tribool.h
	\brief Contains the TriBool class
	\author Kevin André
	
	Copyright (C) 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-04-23
// Last time modified:	2008-10-13

#ifndef DEF_TRIBOOL_H
 #define DEF_TRIBOOL_H

namespace HYPER_Compiler {
	
	/// Boolean class that supports an additional 'uncertain' value
	class TriBool {
	private:
		enum Value {
			kFalse = 1,
			kTrue = 2,
			kUncertain = 3
		};
		
	public:
		
		static const TriBool uncertain;
		
		// CONSTRUCTORS
		
		/// Create a TriBool with default value 'false'
		TriBool();
		
		/// Implicit conversion constructor for regular bools to TriBool
		TriBool(bool b);
		
		
		// INTERROGATORS
		
		bool isTrue() const {
			return fValue == kTrue;
		}
		
		bool isNotTrue() const {
			return fValue != kTrue;
		}
		
		bool isFalse() const {
			return fValue == kFalse;
		}
		
		bool isNotFalse() const {
			return fValue != kFalse;
		}
		
		bool isUncertain() const {
			return fValue == kUncertain;
		}
		
		bool isCertain() const {
			return fValue != kUncertain;
		}
		
		
		// CONVERTERS
		
		/**
			\brief Converts to bool. The argument specifies the bool value of 
			the 'uncertain' state.
		*/
		bool toBool(bool uncertainValue) const;
		
		
		// OPERATORS
		
		TriBool operator!() const;
		
		TriBool operator==(TriBool b) const;
		TriBool operator!=(TriBool b) const;
		
		TriBool& operator&=(TriBool b);
		TriBool operator&&(TriBool b) const;
		
		TriBool& operator|=(TriBool b);
		TriBool operator||(TriBool b) const;
		
	private:
		
		TriBool(Value v)
		 : fValue(v)
		{
			// no action
		}
		
		Value fValue;
	};
}
#endif
