/*!
	\file type.cxx
	\brief Implements the new Type class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2008-10-08
// Last time modified:	2009-03-01

#include "type.h"

#include "base/asserts.h"
#include "base/debug.h"  // used by dump()
#include "base/easydiagnostics.h"

#include "accesschecker.h"
#include "class.h"
#include "classrepository.h"
#include "compileinfo.h"
#include "conversions.h"
#include "expression.h"
#include "initializationhandler.h"
#include "simpleparameter.h"

namespace HYPER_Compiler {
	
	/// Initialization handler for single pointer types.
	class SinglePointerInitHandler : public InitializationHandler {
	public:
		TriBool
		checkInitByDefaultValue(ClassContext const& ctxt,
			Location const* useLoc, CompileInfo& info);
		
		TriBool
		checkInitByExpression(Expression* e, ClassContext const& ctxt,
			Location const* loc, CompileInfo& info);
		
		//void checkAccess(ClassContext const& ctxt,
		//	Location const* loc, CompileInfo& info);
		
		SinglePointerInitHandler(Type const* pt);
		
	private:
		Type const* fType;
		Expression* fExpr;
	};
	
	/// Initialization handler for double pointer types.
	class DoublePointerInitHandler : public InitializationHandler {
	public:
		TriBool
		checkInitByDefaultValue(ClassContext const& ctxt,
			Location const* useLoc, CompileInfo& info);
		
		TriBool
		checkInitByExpression(Expression* e, ClassContext const& ctxt,
			Location const* loc, CompileInfo& info);
		
		//void checkAccess(ClassContext const& ctxt,
		//	Location const* loc, CompileInfo& info);
		
		DoublePointerInitHandler(Type const* pt);
		
	private:
		Type const* fType;
		Expression* fExpr;
	};
	
	
	// *** Dimension *** //
	
	
	Dimension Dimension::create(ConstantValue::nat_t size) {
		hyp_assert(size > 0);
		return Dimension(true, size);
	}
	
	Dimension Dimension::createOpen() {
		return Dimension(true, 0);
	}
	
	Dimension Dimension::createUnknown() {
		return Dimension(false, 0);
	}
	
	bool Dimension::isOpen() const {
		return fKnown && fSize == 0;
	}
	
	bool Dimension::isUnknown() const {
		return !fKnown;
	}
	
	ConstantValue::nat_t Dimension::getSize() const {
		hyp_assert(!isOpen() && !isUnknown());
		return fSize;
	}
	
	Dimension::Dimension(bool known, ConstantValue::nat_t s)
	 : fKnown(known), fSize(s)
	{
		//
	}
	
	
	// *** Type *** //
	
	
	Type const*
	Type::createClassType(bool isConst, Class const* c)
	{
		std::vector<Dimension> dummy;
		return new Type(c->getCompileInfo(), kClass, isConst, c, dummy, 0);
	}
	
	Type const*
	Type::createArrayType(std::vector<Dimension> dims, Type const* baseType)
	{
		CompileInfo& info = baseType->getCompileInfo();
		
		Class* c =
			info.getClassRepository().getArrayClass(baseType, dims.size());
		
		return new Type(info, kArray, baseType->isConst(), c, dims,	baseType);
	}
	
	Type const*
	Type::createSinglePointer(bool isConst, Type const* baseType)
	{
		std::vector<Dimension> dummy;
		return new Type(
			baseType->getCompileInfo(), kSinglePointer, isConst, 0, dummy,
			baseType
		);
	}
	
	Type const*
	Type::createDoublePointer(Type const* baseType)
	{
		std::vector<Dimension> dummy;
		return new Type(
			baseType->getCompileInfo(), kDoublePointer, true, 0, dummy, baseType
		);
	}
	
	Type const* Type::createExtraPointer(Type const* baseType) {
		hyp_assert(!baseType->isDoublePointer());
		
		if (baseType->isNonNullPointer()) {
			return createDoublePointer(baseType);
		}
		else {
			return createSinglePointer(false, baseType);
		}
	}
	
	Type const* Type::createNull(CompileInfo& info) {
		std::vector<Dimension> dummy;
		return new Type(info, kNull, false, 0, dummy, 0);
	}
	
	// for internal use only!
	void Type::makeThisNonConst() {
		fConst = false;
	}
	
	// for internal use only!
	void Type::makeThisConst() {
		fConst = true;
		if (isArray()) {
			fClass =
				fCompileInfo.getClassRepository().getArrayClass(
					fBaseType, fDims.size()
				);
		}
	}
	
	Type const* Type::getNonConstType() const {
		if (!this->isConst()) {
			// avoid creating a new type if we're already non-const
			return this;
		}
		
		Type* t = new Type(*this);
		t->makeThisNonConst();
		return t;
	}
	
	Type const* Type::getConstType() const {
		if (this->isConst()) {
			// avoid creating a new type if we're already const
			return this;
		}
		
		Type* t = new Type(*this);
		t->makeThisConst();
		return t;
	}
	
	Type const* Type::getAllConstType() const {
		if (this->isAllConst()) {
			// avoid creating a new type if we're already completely const
			return this;
		}
		
		Type const* base = fBaseType;
		if (base != 0) { // if we have a base type
			// get a base type that is all const
			base = base->getAllConstType();
		}
		
		Type* t = new Type(*this);
		t->fBaseType = base;
		t->makeThisConst();
		return t;
	}
	
	
	// private constructor
	Type::Type(CompileInfo& info, Kind k, bool isConst, Class const* classRef,
		std::vector<Dimension> dimensions, Type const* baseType)
	 : fCompileInfo(info), fKind(k), fConst(isConst), fClass(classRef),
		fDims(dimensions), fBaseType(baseType)
	{
		// class or array type must have a valid class pointer
		hyp_assert((k != kClass && k != kArray) || classRef != 0);
		
		// check presence of basetype if applicable
		hyp_assert(k == kClass || k == kNull || baseType != 0);
		
		// check dimensions if an array type
		hyp_assert(k != kArray || !dimensions.empty());
		
		// a double pointer must have a single pointer as base type
		hyp_assert(k != kDoublePointer || baseType->fKind == kSinglePointer);
	}
	
	// copy constructor
	Type::Type(Type const& t)
	 : fCompileInfo(t.fCompileInfo), fKind(t.fKind), fConst(t.fConst),
		fClass(t.fClass), fDims(t.fDims), fBaseType(t.fBaseType)
	{
		//
	}
	
	void Type::dump() const {
		debug.line(this->toString());
	}
	
	bool Type::nonConstEquals(const Type* t) const {
		return equalTo(t, false);
	}
	
	bool Type::constEquals(const Type* t) const {
		return equalTo(t, false);
	}
	
	bool Type::equalTo(const Type* t, bool requireConstMatch) const {
		bool ignoreConst = !requireConstMatch;
		return getEqualityHash(ignoreConst) == t->getEqualityHash(ignoreConst);
	}
	
	bool Type::isOverloadableWith(Type const& t) const {
		return this->getOverloadHash() != t.getOverloadHash();
	}
	
	bool Type::isNull() const {
		return fKind == kNull;
	}
	
	bool Type::isDoublePointer() const {
		return fKind == kDoublePointer;
	}
	
	bool Type::isNonNullPointer() const {
		switch (fKind) {
		case kNull:
		case kClass:
		case kArray:
			return false;
		case kSinglePointer:
		case kDoublePointer:
			return true;
		}
		
		hyp_assert_unreachable();
	}
	
	bool Type::isPointerOrNull() const {
		switch (fKind) {
		case kClass:
		case kArray:
			return false;
		case kNull:
		case kSinglePointer:
		case kDoublePointer:
			return true;
		}
		
		hyp_assert_unreachable();
	}
	
	bool Type::isArray() const {
		return fKind == kArray;
	}
	
	bool Type::isClass() const {
		return fKind == kClass || fKind == kArray;
	}
	
	Class const* Type::getClass() const {
		return fClass;
	}
	
	std::vector<Dimension> const& Type::getDimensions() const {
		return fDims;
	}
	
	bool Type::isConst() const {
		return fConst;
	}
	
	bool Type::isAllConst() const {
		return fConst && (fBaseType == 0 || fBaseType->isAllConst());
	}
	
	Type const* Type::getBaseType() const {
		return fBaseType;
	}
	
	const std::vector<Dimension>& Type::getArrayDimensions() const {
		return fDims;
	}
	
	HYPER_Common::UTF8String Type::toString() const {
		HYPER_Common::UTF8StringBuilder sb;
		//~ debug.line(UTF8STR("Type::toString() called"));
		
		switch (fKind) {
		case kClass:
			//~ debug.line(UTF8STR("** CLASS TYPE"));
			if (fConst) {
				sb += UTF8STR("const ");
			}
			
			sb += fClass->getTypeName();
			
			return sb.toString();
		case kArray:
			//~ debug.line(UTF8STR("** ARRAY TYPE"));
			for (unsigned int i = 0; i < fDims.size(); ++i) {
				//~ debug.line(UTF8STR(">> DIMENSION"));
				sb += UTF8STR("[");
				
				Dimension const& dim = fDims[i];
				if (dim.isOpen()) {
					// no extra string piece
					//~ debug.line(UTF8STR("Open dimension"));
				}
				else if (dim.isUnknown()) {
					sb += UTF8STR("?");
					//~ debug.line(UTF8STR("Unknown dimension"));
				}
				else {
					sb += dim.getSize();
					//~ debug.line(UTF8STR("Known dimension"));
				}
				
				sb += UTF8STR("]");
			}
			
			// (array is never `const')
			sb += UTF8STR(" ");
			sb += fBaseType->toString();
			return sb.toString();
		case kNull:
			return UTF8STR("null");
		case kSinglePointer:
			//~ debug.line(UTF8STR("** POINTER TYPE"));
			if (fConst) {
				sb += UTF8STR("const ");
			}
			
			sb += UTF8STR("* ");
			sb += fBaseType->toString();
			return sb.toString();
		case kDoublePointer:
			//~ debug.line(UTF8STR("** DOUBLE POINTER TYPE"));
			sb += UTF8STR("*");
			sb += fBaseType->toString();
			return sb.toString();
		}
		
		hyp_assert_unreachable();
	}
	
	Type const* Type::getDereferenced() const {
		switch (fKind) {
		case kClass:
		case kArray:
			return this;
		case kNull:
			// should not come here -- null cannot be dereferenced
			hyp_assert_unreachable();
		case kSinglePointer:
			return fBaseType;
		case kDoublePointer:
			return fBaseType->fBaseType;
		}
		
		if (isPointerOrNull()) {
			return fBaseType;
		}
		else {
			return this;
		}
	}
	
	bool
	Type::checkAccess(ClassContext const& from, Location const* loc,
		CompileInfo& info) const
	{
		switch (fKind) {
		case kClass:
			return AccessChecker::check(from, *fClass, loc, info);
		case kNull:
			// should not come here -- null is not a type that can be declared
			hyp_assert_unreachable();
		case kArray:
		case kSinglePointer:
		case kDoublePointer:
			// check access to base
			return fBaseType->checkAccess(from, loc, info);
		}
		
		hyp_assert_unreachable();
	}
	
	HYPER_Common::UTF8String Type::getOverloadHash() const {
		HYPER_Common::UTF8StringBuilder sb;
		
		switch (fKind) {
		case kClass:
			return fClass->getTypeName();
		case kArray:
			for (unsigned int i = 0; i < fDims.size(); ++i) {
				sb += UTF8STR("[");
				
				Dimension const& dim = fDims[i];
				if (!dim.isOpen() && !dim.isUnknown()) {
					sb += dim.getSize();
				}
			}
			
			sb += UTF8STR("]");
			sb += fBaseType->getOverloadHash();
			
			return sb.toString();
		case kNull:
			// "null" should not be asked for it's overload hash
			hyp_assert_unreachable();
		case kSinglePointer:
		case kDoublePointer:
			// Pointers don't matter in overloading
			return fBaseType->getOverloadHash();
		}
		
		hyp_assert_unreachable();
	}
	
	// TODO: eliminate this function?
	HYPER_Common::UTF8String Type::getEqualityHash(bool ignoreConst) const {
		HYPER_Common::UTF8StringBuilder sb;
		
		switch (fKind) {
		case kClass:
			if (isConst() && !ignoreConst) {
				sb += UTF8STR("C~"); // mark as 'const'
			}
			
			sb += fClass->getTypeName();
			
			return sb.toString();
		case kArray:
			for (unsigned int i = 0; i < fDims.size(); ++i) {
				sb += UTF8STR("[");
				
				Dimension const& dim = fDims[i];
				if (dim.isUnknown()) {
					sb += UTF8STR("?");
				}
				else if (!dim.isOpen()) {
					sb += dim.getSize();
				}
			}
			sb += UTF8STR("]");
			sb += fBaseType->getEqualityHash(ignoreConst);
			
			return sb.toString();
		case kNull:
			// "null" should not be asked for it's equality hash
			hyp_assert_unreachable();
		case kSinglePointer:
			if (isConst() && !ignoreConst) {
				sb += UTF8STR("C~"); // mark as 'const'
			}
			// FALL-THROUGH
		case kDoublePointer:
			sb += UTF8STR("*");
			sb += fBaseType->getEqualityHash(ignoreConst);
			
			return sb.toString();
		}
		
		hyp_assert_unreachable();
	}
	
	InitializationHandler* Type::getInitHandler() const {
		switch (fKind) {
		case kClass:
		case kArray:
			return fClass->getInitHandler();
		case kNull:
			// "null" should not be asked for it's init handler
			hyp_assert_unreachable();
		case kSinglePointer:
			return new SinglePointerInitHandler(this);
		case kDoublePointer:
			return new DoublePointerInitHandler(this);
		}
		
		hyp_assert_unreachable();
	}
	
	bool Type::canBeInvisiblyPassedByValue() const {
		switch (fKind) {
		case kClass:
			return fClass->canBeInvisiblyPassedByValue();
		case kArray:
			return false;
		case kNull:
		case kSinglePointer:
		case kDoublePointer:
			return true;
		}
		
		hyp_assert_unreachable();
	}
	
	
	// *** SinglePointerInitHandler *** //
	
	SinglePointerInitHandler::SinglePointerInitHandler(Type const* pt)
	 : fType(pt), fExpr(0)
	{
		//
	}
	
	TriBool
	SinglePointerInitHandler::checkInitByDefaultValue(ClassContext const& ctxt,
		Location const* useLoc, CompileInfo& info)
	{
		// No action required. Always legal. Pointer is initialized with 'null'.
		return true;
	}
	
	TriBool
	SinglePointerInitHandler::checkInitByExpression(Expression* e,
		ClassContext const& ctxt, Location const* loc, CompileInfo& info)
	{
		if (e && e->getResult().isResolved()) {
			fExpr = e;
			
			// check suitability of initializer
			SimpleParameter param(fType, SimpleParameter::kCopyValue);
			ValuePassing vp =
				ValuePassing::check(e->getResult().getType(), param);
			
			if (vp.isPossible()) {
				//debug << "compatible initializer" << newl;
				return true;
			}
			else { // type incompatible
				info.diag().error(
					e->getLocation(),
					UTF8STR(
						"cannot initialize pointer type %1 with "
						"expression %2 which has type %3"
					),
					quote(fType->toString()),
					quote(e->toString()),
					quote(e->getResult().getType()->toString())
				);
				return false;
			}
		}
		else { // initializer unknown
			return TriBool::uncertain;
		}
	}
	
	//void
	//SinglePointerInitHandler::checkAccess(ClassContext const& ctxt,
	//	Location const* loc, CompileInfo& info)
	//{
		// TODO
	//}
	
	
	// *** DoublePointerInitHandler *** //
	
	DoublePointerInitHandler::DoublePointerInitHandler(Type const* pt)
	 : fType(pt), fExpr(0)
	{
		//
	}
	
	TriBool
	DoublePointerInitHandler::checkInitByDefaultValue(
		ClassContext const& ctxt, Location const* useLoc,
		CompileInfo& info)
	{
		// double pointers don't have a default value
		info.diag().error(
			UTF8STR(
				"double pointer cannot be assigned a default value, it must "
					"be initialized by a pointer expression"
			),
			useLoc
		);
		return false;
	}
	
	TriBool
	DoublePointerInitHandler::checkInitByExpression(Expression* e,
		ClassContext const& ctxt, Location const* loc, CompileInfo& info)
	{
		if (e && e->getResult().isResolved()) {
			fExpr = e;
			
			// check suitability of initializer
			SimpleParameter param(fType, SimpleParameter::kCopyValue);
			ValuePassing vp =
				ValuePassing::check(e->getResult().getType(), param);
			
			if (vp.isPossible()) {
				// compatible
				return true;
			}
			else { // not compatible
				info.diag().error(
					e->getLocation(),
					UTF8STR(
						"cannot initialize double pointer type %1 with "
						"expression %2 which has type %3"
					),
					quote(fType->toString()),
					quote(e->toString()),
					quote(e->getResult().getType()->toString())
				);
				
				return false;
			}
		}
		else { // initializer unknown
			return TriBool::uncertain;
		}
	}
	
	//void
	//DoublePointerInitHandler::checkAccess(ClassContext const& ctxt,
	//	Location const* loc, CompileInfo& info)
	//{
		// TODO
	//}
	
}
