/*!
	\file type.h
	\brief Contains the new Type class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2008-10-08
// Last time modified:	2009-03-01

#ifndef DEF_NEWTYPE_H
 #define DEF_NEWTYPE_H
 
 #include "constantvalue.h"
 #include "important_fwdecl.h"
 
 #include "common/cxx_macros.h"
 #include "common/utf8string.h"
 
 #include <vector>

namespace HYPER_Compiler {
	
	class InitializationHandler;
	
	/// Represents the dimension of an array
	class Dimension {
	public:
		static Dimension create(ConstantValue::nat_t size);
		static Dimension createOpen();
		static Dimension createUnknown();
		
		bool isOpen() const;
		bool isUnknown() const;
		ConstantValue::nat_t getSize() const;
		
	private:
		Dimension(bool known, ConstantValue::nat_t s);
		
		bool fKnown;
		ConstantValue::nat_t fSize;
	};
	
	/// Represents the type of a value
	class Type {
	public:
		
		// TODO: unify all Type instances. That will require (re)moving the
		// creation functions and the "getConstType" & "getAllConstType"
		// functions below.
		
		/// Create a type based on a class
		static Type const* createClassType(bool isConst, Class const* c);
		
		/// Create an array type
		static Type const* createArrayType(std::vector<Dimension> dims,
			Type const* baseType);
		
		/// Create a normal (single) pointer type
		static Type const* createSinglePointer(bool isConst,
			Type const* baseType);
		
		/// Create a double pointer type
		static Type const* createDoublePointer(Type const* baseType);
		
		/// Create an extra pointer on top of the given type
		static Type const* createExtraPointer(Type const* baseType);
		
		/// Create the type "null"
		static Type const* createNull(CompileInfo& info);
		
		/// Get a 'const' version of this type
		Type const* getConstType() const;
		
		/// Get an all 'const' version of this type
		Type const* getAllConstType() const;
		
		/// Get a non-'const' version of this type
		Type const* getNonConstType() const;
		
		
		
		/// Check if the type is the type of 'null'
		bool isNull() const;
		
		/// Check if the type is a double pointer
		bool isDoublePointer() const;
		
		/// Check if the type is a pointer or double pointer, but not 'null'
		bool isNonNullPointer() const;
		
		/// Check if the type is a pointer, double pointer or 'null'
		bool isPointerOrNull() const;
		
		/// Check if the type is an array
		bool isArray() const;
		
		/// Checks if the type is a class type, either built-in or user defined.
		/// This includes array types.
		bool isClass() const;
		
		/// Get the class of the type
		Class const* getClass() const;
		
		/// Get the list of dimensions for an array type
		std::vector<Dimension> const& getDimensions() const;
		
		/// Is this part of the type 'const'?
		bool isConst() const;
		
		/// Is the type completely 'const'?
		bool isAllConst() const;
		
		/// Get the base of this type (array element type or pointed-to type)
		Type const* getBaseType() const;
		
		/// Get a list of the dimensions for this *array* type
		const std::vector<Dimension>& getArrayDimensions() const;
		
		/// Get type syntax as a string
		HYPER_Common::UTF8String toString() const;
		
		/// Print the type to debug output
		void dump() const;
		
		/// Compare types for equality while ignoring all const keywords
		bool nonConstEquals(const Type* t) const;
		
		/// Compare types for exact equality
		bool constEquals(const Type* t) const;
		
		/// Compare types
		bool equalTo(const Type* t, bool requireConstMatch) const;
		
		Type const* getDereferenced() const;
		
		~Type() { /* empty */ }
		
		CompileInfo& getCompileInfo() const { return fCompileInfo; }
		
		/// Check if the type can be overloaded with the given type
		bool isOverloadableWith(Type const& t) const;
		
		/// Check access to this type
		bool ATTR_WARN_UNUSED_RESULT
		checkAccess(ClassContext const& from, Location const* loc,
			CompileInfo& info) const;
		
		/// Get the parts of the type that are significant for overloadable
		/// types.
		HYPER_Common::UTF8String getOverloadHash() const;
		
		/// Get the parts of the type that identify it for comparing it to
		/// another type.
		HYPER_Common::UTF8String getEqualityHash(bool ignoreConst) const;
		
		InitializationHandler* getInitHandler() const;
		
		/// Returns true if a value of this type can be passed in a register.
		/// The type must also have pass-by-value semantics.
		bool canBeInvisiblyPassedByValue() const;
		
	private:
		enum Kind {
			kNull, kClass, kSinglePointer, kDoublePointer, kArray
		};
		
		explicit Type(CompileInfo& info, Kind k, bool isConst,
			Class const* classRef, std::vector<Dimension> dimensions,
			Type const* baseType);
		
		/// Copy constructor
		Type(Type const& t);
		
		void makeThisConst();
		void makeThisNonConst();
		
		// ban the assignment operator
		void operator=(const Type& t) { /* empty */ }
		
		CompileInfo& fCompileInfo;
		Kind fKind;
		bool fConst;
		Class const* fClass;
		std::vector<Dimension> fDims;
		Type const* fBaseType;
	};
}
#endif
