/*!
	\file typeparser.cxx
	\brief Implements the TypeParser class.
	\author Kevin André
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-09-22
// Last time modified:	2008-10-15

#include "typeparser.h"

#include "base/asserts.h"
#include "base/diagnostic.h"

#include "arraytype.h"
#include "builtintypes.h"
#include "classrepository.h"
#include "classtype.h"
#include "compileinfo.h"
#include "doublepointertype.h"
#include "expressionparser.h"
#include "pointertype.h"

namespace HYPER_Compiler {
	
	TypeParser::TypeParser(ParserInfo const& pi, ClassContext const* cCtxt)
	 : ParserBase(pi), fCtxt(cCtxt)
	{
		//
	}
	
	DeclaredType* TypeParser::doType(bool allowDoublePointers) {
		int initialPos = getTokenPos();
		
		if (lookToken().isKeyword(Keywords::kConst)) {
			nextToken(); // skip 'const' (look ahead to see what's coming)
		}
		
		if (lookToken().isSymbol(Token::kAsterisk)) {
			// we have a pointer
			
			goBackToTokenPos(initialPos);
			return doPointerType(allowDoublePointers);
		}
		else if (lookToken().isSymbol(Token::kOpenSquareParen)) {
			// we have an array
			
			goBackToTokenPos(initialPos);
			return doArrayType();
		}
		else {
			// expecting a typename
			
			goBackToTokenPos(initialPos);
			return doNamedType();
		}
	}
	
	DeclaredType*
	TypeParser::doPointerType(bool allowDoublePointers)
	{
		bool isConst = false;
		const Location* constLoc = 0;
		if (lookToken().isKeyword(Keywords::kConst)) {
			// we have a const pointer
			constLoc = lookPos();
			nextToken(); // eat 'const' keyword
			isConst = true;
		}
		
		// 'typeLoc' refers to the type without 'const'
		const Location* typeLoc = lookPos();
		
		unsigned pointerCount = 0;
		do {
			nextToken(); // eat '*' symbol
			pointerCount++; // count the asterisks
			
			if (lookToken().isKeyword(Keywords::kConst)) {
				// We're going to see if this 'const' is followed by another
				// asterisk.
				Token t = lookToken(); // store the 'const' token
				nextToken();
				if (lookToken().isSymbol(Token::kAsterisk)) {
					// yup, another pointer
					
					// generate an error for the 'const' usage
					nonfatal(error(
						UTF8STR(
							"%1 cannot be used inside a double pointer"
						).compose(QUOTED_UTF8STR("const")),
						t.getPosition()
					));
					
					// proceed with next iteration
				}
				else {
					// the const applies to a non-pointer base type
					
					// put the "const" token back
					goBackOneToken();
					
					// the loop will now stop
				}
			}
		} while (lookToken().isSymbol(Token::kAsterisk));
		
		if (pointerCount == 2) { // double pointer
			if (isConst) {
				nonfatal(error(
					UTF8STR(
						"%1 cannot be used on a double pointer"
					).compose(QUOTED_UTF8STR("const")),
					constLoc
				));
			}
			
			if (allowDoublePointers) { // are double pointers allowed?
				// yes...
				return new DoublePointerType(
					new PointerType(
						false /*not const*/,
						doType(false),
						typeLoc // FIXME
					),
					typeLoc
				);
			}
			
			// double pointers not allowed in this context
			nonfatal(error(
				UTF8STR("double pointers are not allowed in this context"),
				typeLoc
			));
			
			// pretend we had only a single pointer . . .
		}
		else if (pointerCount > 2) { // more than two asterisks
			nonfatal(error(
				UTF8STR(
					"pointers cannot be nested this way; only single and "
					"double pointers are allowed"
				),
				typeLoc
			));
			
			// pretend we had only a single pointer . . .
		}
		
		// we have a single pointer, or we're pretending one because of an error
		return new PointerType(
			isConst, doType(false), typeLoc
		);
	}
	
	DeclaredType* TypeParser::doArrayType() {
		if (lookToken().isKeyword(Keywords::kConst)) {
			nonfatal(error(
				UTF8STR(
					"an array cannot be declared %1; declare the "
					"contents of the array as %1 instead"
				).compose(QUOTED_UTF8STR("const")),
				lookPos()
			));
			nextToken();
		}
		
		Location const* arrayLoc = lookPos();
		
		// process the dimensions
		std::vector<Expression*> dims;
		ExpressionParser ep(this->parserInfo(), fCtxt);
		while (lookToken().isSymbol(Token::kOpenSquareParen)) {
			nextToken(); // eat the "["
			if (lookToken().isSymbol(Token::kComma)
				|| lookToken().isSymbol(Token::kCloseSquareParen))
			{
				// no dimension size -> open array dimension
				dims.push_back(0);
			}
			else {
				dims.push_back(ep.doExpression());
			}
			
			// process optional extra dimensions
			while (lookToken().isSymbol(Token::kComma)) {
				nextToken(); // eat the ","
				if (lookToken().isSymbol(Token::kComma)
					|| lookToken().isSymbol(Token::kCloseSquareParen))
				{
					// no dimension size -> open array dimension
					dims.push_back(0);
				}
				else {
					dims.push_back(ep.doExpression());
				}
			}
			
			doSymbol(Token::kCloseSquareParen);
		}
		
		// do basetype
		DeclaredType* base = doType(false /*=disallow double pointers*/);
		
		// construct the array
		return new ArrayType(dims, base, arrayLoc);
	}
	
	DeclaredType* TypeParser::doNamedType() {
		bool isConst = false;
		if (lookToken().isKeyword(Keywords::kConst)) {
			isConst = true;
			nextToken();
		}
		
		Location const* typeLoc = lookPos();
		
		if (lookToken().isIdentifier()) {
			// Typename or typename in namespace(s)
			return new ClassType(
				getCompileInfo(), doQualifiedName(false), isConst, typeLoc
			);
		}
		
		Keywords::Keyword keyword = lookToken().getKeyword();
		BuiltInTypes::Primitive pType;
		
		switch (keyword) {
		case Keywords::kBool:
		case Keywords::kByte:
		case Keywords::kChar:
		case Keywords::kDouble:
		case Keywords::kInt:
		case Keywords::kInt16:
		case Keywords::kInt32:
		case Keywords::kInt64:
		case Keywords::kNat:
		case Keywords::kNat16:
		case Keywords::kNat32:
		case Keywords::kNat64:
		case Keywords::kObject:
		case Keywords::kReal:
		case Keywords::kSingle:
		case Keywords::kString:
			pType = BuiltInTypes::fromKeyword(keyword);
			nextToken(); // eat type keyword
			return new ClassType(
				getCompileInfo().getClassRepository().getClassOfBuiltin(pType),
				isConst, typeLoc
			);
		case Keywords::kSystem:
			// typename inside 'system' namespace
			return new ClassType(
				getCompileInfo(), doQualifiedName(true), isConst, typeLoc
			);
		default:
			// parse error
			fatal(errorExpected(UTF8STR("type"), lookPos(), lookToken()));
			return 0;
		}
	}
}
