/*!
	\file unaryoperatorexpression.cxx
	\brief Implements the UnaryOperatorExpression class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-07-04
// Last time modified:	2009-01-05

#include "unaryoperatorexpression.h"

#include "base/asserts.h"
#include "base/easydiagnostics.h"

#include "class.h"
#include "compileinfo.h"
#include "type.h"
#include "unaryoperator.h"

namespace HYPER_Compiler {
	
	UnaryOperatorExpression::UnaryOperatorExpression(Expression* expr,
		Operators::UnaryOperator op, const Location* loc)
	 : Expression(expr->getCompileInfo(), loc), fExpr(expr), fOpSym(op), fOp(0)
	{
		// nothing here
	}
	
	UnaryOperatorExpression::~UnaryOperatorExpression() {
		// nothing here
	}
	
	HYPER_Common::UTF8String UnaryOperatorExpression::toString() const {
		return UTF8STR("(%1 %2)").compose(
			Operators::toString(fOpSym), fExpr->toString()
		);
	}
	
	void UnaryOperatorExpression::putInSymbolTable(LookupLocation const* loc) {
		Expression::putInSymbolTable(loc);
		fExpr->putInSymbolTable(loc);
	}
	
	bool UnaryOperatorExpression::doFullResolve(Trace const& trace) {
		if (fExpr->isNullLiteral()) {
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"%1 cannot be used as an operand for unary "
					"operator %2"
				),
				QUOTED_UTF8STR("null"),
				Operators::toString(fOpSym)
			);
			return false;
		}
		
		bool resolveOK = fExpr->resolveFull(trace);
		resolveOK &= fExpr->checkReturnsValue();
		
		if (fExpr->isResolvedAndReturnsResult()) {
			resolveOK &= resolveOverloading();
		}
		
		return resolveOK;
	}
	
	bool UnaryOperatorExpression::resolveOverloading() {
		Type const* operandT = fExpr->getResult().getType()->getDereferenced();
		const Class* c = operandT->getClass();
		
		Class::UnOpsMap const& classOpMap = c->getUnaryOperators();
		Class::UnOpsMap::const_iterator it = classOpMap.find(fOpSym);
		
		if (it == classOpMap.end()) { // no operator in class...
			getCompileInfo().diag().error(
				getLocation(),
				UTF8STR(
					"no match for %1 in class %2, which is needed for "
					"expression %3"
				),
				UnaryOperator::getPrintPrototype(fOpSym, operandT),
				c->getDescription(),
				quote(toString())
			);
			return false;
		}
		
		// class has the operator...
		
		UnaryOperator* unOp = it->second;
		fOp = unOp;
		setResolved();
		setResultType(unOp->getResultType());
		if (ConstantValue const* opVal = fExpr->getResult().getValue()) {
			if (ConstantValue const* val = unOp->evaluateCompileTime(opVal))
			{
				setResultValue(val);
			}
		}
		
		return true;
	}
	
	TriBool UnaryOperatorExpression::isClassLValue() const {
		return false;
	}
	
	TriBool UnaryOperatorExpression::isPointerLValue() const {
		return false;
	}
	
	TriBool UnaryOperatorExpression::isCall() const {
		return false;
	}
	
	TriBool UnaryOperatorExpression::hasResult() const {
		return true;
	}
	
	TriBool UnaryOperatorExpression::resultAtCompileTime() const {
		return fExpr->resultAtCompileTime();
	}
	
	void UnaryOperatorExpression::operandTypeError() {
		getCompileInfo().diag().error(
			fExpr->getLocation(),
			UTF8STR(
				"unary operator %1 cannot be applied to operand %2, "
				"which has type %3"
			),
			quote(Operators::toString(fOpSym)),
			quote(fExpr->toString()),
			quote(fExpr->getResult().getType()->toString())
		);
	}
}
