/*!
	\file varstatement.cxx
	\brief Implementation of the Var class.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-06-30
// Last time modified:	2008-10-23

#include "varstatement.h"

#include "base/asserts.h"
#include "base/compileoptions.h"
#include "base/debug.h"
#include "base/easydiagnostics.h"

#include "accesschecker.h"
#include "class.h"
#include "compileinfo.h"
#include "declaredtype.h"
#include "initializer.h"
#include "overloadresolution.h"
#include "overloadedconstructor.h"
#include "statementcontext.h"
#include "type.h"

#include <set>

namespace HYPER_Compiler {
	
	// *** Var ***
	
	Var::Var(
		bool isStatic, std::vector<HYPER_Common::UTF8String> const& names,
		std::vector<const Location*> locs, DeclaredType* t,
		Initializer* init, const Location* loc,
		StatementContext const* context, CompileInfo& info)
	 : Statement(info, loc), fType(t), fResolvedType(0), fInit(init),
		/*fStatic(isStatic),*/ fLoc(loc), fContext(context)
	{
		hyp_assert(names.size() == locs.size());
		
		if (isStatic) {
			Debug::toDo(
				UTF8STR("static variables are not yet supported")
			);
		}
		
		for (unsigned int i = 0; i < names.size(); ++i) {
			fDecls.push_back(new VarDecl(this, names[i], locs[i]));
		}
	}
	
	Var::~Var() {
		// nothing here
	}
	
	HYPER_Common::UTF8String Var::getSyntaxName() const {
		return UTF8STR("Var");
	}
	
	void Var::myPrint(HYPER_Common::UTF8TextFlowStream& ind) const {
		ind.print(getSyntaxName());
		//~ ind.line(fStatic ? UTF8STR(" [static]") : UTF8STR(""));
		
		HYPER_Common::Indent out(ind);
		
		ind.line(UTF8STR("type = ") + fType->toString());
		
		for (unsigned int i = 0; i < fDecls.size(); ++i) {
			ind.line(UTF8STR("name = ") + fDecls[i]->fName);
		}
		
		fInit->myPrint(ind);
	}
	
	DeclaredType const* Var::getType() const {
		return fType;
	}
	
	Type const* Var::getResolvedType() const {
		return fResolvedType;
	}
	
	//bool Var::isStatic() const { //NOT YET USED
	//	return fStatic;
	//}
	
	const Location* Var::getLocation() const {
		return fLoc;
	}
	
	HYPER_Common::UTF8String Var::getResolveDescription() const {
		return UTF8STR("%1 statement").compose(QUOTED_UTF8STR("var"));
	}
	
	void Var::putInSymbolTable(ScopeLocation const& loc) {
		//debug.line("Var::putInSymbolTable called");
		Statement::putInSymbolTable(loc);
		LookupLocation const* ll = new LookupLocation(loc);
		fType->putInSymbolTable(ll);
		fInit->putInSymbolTable(ll);
		
		for (unsigned int i = 0; i < fDecls.size(); ++i) {
			loc.getScope().registerSymbolWithDiagnostic(
				fDecls[i], getCompileInfo()
			);
		}
	}
	
	bool Var::doFullResolve(Trace const& trace) {
		if (getCompileInfo().getOptions().fgHideParameterWarnings) {
			// if  -W-hide-parameter  enabled
			ScopeLocation const& loc = *getScopeLocation();
			for (unsigned int i = 0; i < fDecls.size(); ++i) {
				// lookup the name this variable uses
				NamedDeclaredItem* item =
					loc.getScope().findSymbol(fDecls[i]->fName, loc);
				
				// if the name already exists and is the name of a parameter...
				if (item != 0 && dynamic_cast<Parameter*>(item) != 0) {
					// ... give a warning
					getCompileInfo().diag().warning(
						fDecls[i]->getDescription()
							+ UTF8STR(" hides parameter with same name"),
						fDecls[i]->fLoc
					);
				}
			}
		}
		
		fType->setRestrictions(true, false);
		bool resolveOK = fType->resolveFull(trace);
		fResolvedType = fType->getResolvedType();
		
		if (fResolvedType) {
			resolveOK &=
				fResolvedType->checkAccess(
					*fContext, fType->getLocation(), getCompileInfo()
				);
		}
		
		if (fInit->isDefault() && fType->isConst()) {
			getCompileInfo().diag().error(
				UTF8STR(
					"a variable with a %1 type must be explicitly initialized"
				).compose(QUOTED_UTF8STR("const")),
				getLocation()
			);
			resolveOK = false;
		}
		
		resolveOK &= fInit->resolveFull(trace);
		return resolveOK;
	}
	
	SimpleProbability Var::getReturnProbability() const {
		return SimpleProbability::never;
	}
	
	Class const* Var::getOwnerClass() const {
		return 0;
	}
	
	ClassMemberAccess Var::getAccessInOwner() const {
		hyp_assert_unreachable();
	}
	
	
	// *** VarDecl ***
	
	VarDecl::VarDecl(Var* v, HYPER_Common::UTF8String name, Location const* loc)
	 : fV(v), fName(name), fLoc(loc)
	{
		// no action
	}
	
	Var const* VarDecl::getVar() const {
		return fV;
	}
	
	HYPER_Common::UTF8String VarDecl::getDescription() const {
		return UTF8STR("variable %1").compose(quote(fName));
	}
	
	Class const* VarDecl::getOwnerClass() const {
		return fV->getOwnerClass();
	}
	
	ClassMemberAccess VarDecl::getAccessInOwner() const {
		return fV->getAccessInOwner();
	}
	
	HYPER_Common::UTF8String VarDecl::getDeclarationName() const {
		return fName;
	}
	
	const Location* VarDecl::getLocation() const {
		return fLoc;
	}
}
