/*!
	\file highlighter.cxx
	\brief Implements the Highlighter class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 	2004-04-25
// Date modified:	2008-10-24

#include "highlighter.h"

#include "base/asserts.h"
#include "base/compilationfailederror.h"
#include "base/compileoptions.h"
#include "base/errormanager.h"

#include "common/bufferedpullstream.h"
#include "common/console.h"
#include "common/filepath.h"
#include "common/filesystem.h"
#include "common/overlaypullstream.h"
#include "common/utf8decoder.h"
#include "common/utf8encoder.h"
#include "common/utf8pushstream.h"
#include "common/utf8stdinstream.h"

#include "lexer/fileexceptions.h"
#include "lexer/lexer.h"
#include "lexer/reader.h"

#include "htmlcommentdecoder.h"
#include "htmlcommentencoder.h"

#include <fstream>
#include <map>
#include <stack>
#include <vector>

//~ #include "base/debug.h"

using HYPER_Common::BufferedPullStream;
using HYPER_Common::Console;
using HYPER_Common::FilePath;
using HYPER_Common::FileSystem;
using HYPER_Common::OverlayPullStream;
using HYPER_Common::PullStream;
using HYPER_Common::PushStream;
using HYPER_Common::TextFlowOptions;
using HYPER_Common::UTF8Decoder;
using HYPER_Common::UTF8Encoder;
using HYPER_Common::UTF8PushStream;
using HYPER_Common::UTF8StdInStream;
using HYPER_Common::UTF8StdOutStream;
using HYPER_Common::UTF8StringBuilder;
using HYPER_Common::UTF8TextFlowStream;

using namespace HYPER_Compiler;

namespace HYPER_Highlighter {
	
	Highlighter::FileRange::FileRange(unsigned long start_,
		unsigned long stop_)
	 : start(start_), stop(stop_)
	{
		hyp_assert(start <= stop);
	}
	
	Highlighter::Highlighter()
	 : fOut(Console::getStdOut()), fTxtOut(Console::getTextFlowStdOut())
 	{
		CompileOptions opt(fTxtOut);
		fInfo.setOptions(opt);
	}
	
	HYPER_Common::UTF8String Highlighter::getVersion() {
		return UTF8STR("0.2.4");
	}
	
	void Highlighter::sourceFileToHTML(FilePath* sourcefile,
		FilePath* outputfile, FilePath* stylesheet, bool createStylesheet)
	{
		//~ debug.line(UTF8STR("sourceFileToHTML called"));
		
		hyp_assert(sourcefile->isAbsolute());
		hyp_assert(outputfile->isAbsolute());
		hyp_assert(stylesheet->isAbsolute());
		
		TextFlowOptions tfOpt;
		tfOpt.setIndentChar(CodePoint::tab(), 8);
		
		// (1) Open input file
		
		PullStream<UTF8Byte>* is = FileSystem::openInputFile(*sourcefile);
		BufferedPullStream<UTF8Byte> cs(*is); // buffer
		if (!cs.streamOK()) {
			throw FileNotReadableException();
		}
		UTF8Decoder decoder;
		boost::shared_ptr<PullStream<CodePoint> > uds(decoder.puller(cs));
		Reader r(*uds, true);
		
		// (2) Read the file with the compiler's lexer and save a list of info
		
		std::vector<HLInfo> tokens;
		fInfo.registerFile(HYPER_Compiler::FileID(0), *sourcefile);
		runLexer(r, HYPER_Compiler::FileID(0), tokens);
		
		// (3) Write the stylesheet
		
		if (createStylesheet) {
			writeCSSFile(*stylesheet);
		}
		
		// (4) Now we read the input file again, manually, one character at
		// a time. At the same time we write to the output file.
		
		cs.goBackTo(0); // rewind input stream
		
		UTF8PushStream* htmlOS = FileSystem::openOutputFile(*outputfile);
		UTF8TextFlowStream htmlTOS(*htmlOS, tfOpt);
		
		UTF8String sourceShort = sourcefile->getShortName();
		UTF8String cssRelPath =
			stylesheet->getRelative(
				*outputfile->getParentDirectory()
			)->toString();
		
		//std::cout << "RELATIVE CSS FILE: " << cssRelPath << std::endl;
		
		// write prologue: html start, heading, style info, title and token count.
		htmlTOS.line(
			UTF8STR("<!DOCTYPE html PUBLIC '-//W3C//DTD XHTML 1.0 Strict//EN'")
		);
		htmlTOS.line(
			UTF8STR("   'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd'>")
		);
		htmlTOS.line(
			UTF8STR(
				"<html xmlns='http://www.w3.org/1999/xhtml' lang='en-US' "
					"xml:lang='en-US'>"
			)
		);
		htmlTOS.increaseIndent();
		htmlTOS.line(UTF8STR("<head>"));
		htmlTOS.increaseIndent();
		htmlTOS.line(
			UTF8STR(
				"<meta name='generator' content='Hyper Highlighter %1'/>"
			).compose(getVersion())
		);
		htmlTOS.line(
			UTF8STR(
				"<meta http-equiv='Content-Type' "
					"content='text/html;charset=UTF-8' />"
			)
		);
		htmlTOS.line(
			UTF8STR(
				"<link rel='StyleSheet' type='text/css' href='%1' />"
			).compose(cssRelPath)
		);
		htmlTOS.line(UTF8STR("<title>%1</title>").compose(sourceShort));
		htmlTOS.decreaseIndent();
		htmlTOS.line(UTF8STR("</head>"));
		htmlTOS.line(UTF8STR("<body>"));
		htmlTOS.increaseIndent();
		htmlTOS.line(UTF8STR("<h1>%1</h1>").compose(sourceShort));
		htmlTOS.line(
			UTF8STR(
				"<p><em>This file consists of <strong>%1</strong> "
					"tokens.</em></p>"
			).compose(tokens.size())
		);
		
		// write HTML (no indentation here)
		htmlOS->print(UTF8STR("<pre class='code'>"));
		generateHTML(cs, tokens, *htmlOS);
		htmlOS->line(UTF8STR("</pre>"));
		
		// write epilogue
		htmlTOS.line(UTF8STR("<p>&nbsp;</p>"));
		htmlTOS.line(UTF8STR("<hr/>"));
		htmlTOS.line(UTF8STR("<p>"));
		htmlTOS.increaseIndent();
		htmlTOS.line(
			UTF8STR("<a href='http://validator.w3.org/check?uri=referer'>")
		);
		htmlTOS.line(
			UTF8STR("<img src='http://www.w3.org/Icons/valid-xhtml10'")
		);
		htmlTOS.line(
			UTF8STR(
				"alt='Valid XHTML 1.0 Strict' height='31' width='88' /></a>"
			)
		);
		htmlTOS.decreaseIndent();
		htmlTOS.line(UTF8STR("</p>"));
		htmlTOS.decreaseIndent();
		htmlTOS.line(UTF8STR("</body>"));
		htmlTOS.decreaseIndent();
		htmlTOS.line(UTF8STR("</html>"));
	} // end function
	
	void Highlighter::highlightHTML(HYPER_Common::FilePath* filename) {
		hyp_assert(filename->isAbsolute());
		
		// (1) Open input
		
		PullStream<UTF8Byte>* is = FileSystem::openInputFile(*filename);
		BufferedPullStream<UTF8Byte> cs(*is); // buffer
		if (!cs.streamOK()) {
			throw FileNotReadableException();
		}
		
		// (2) Read HTML & output new highlighted code to buffer
		std::vector<FileRange> codeRanges;
		std::vector<FileRange> preRanges;
		parseHTML(cs, codeRanges, preRanges);
		
		Console::getStdOut().line(
			UTF8STR("Number of code snippets: ") + codeRanges.size()
		);
		
		if (codeRanges.empty()) { // no sourcecode in file...
			// no need to write to the file
			return;
		}
		
		// (3) Run the lexer on the code snippets
		
		// This is done before outputting the result, because errors in the
		// lexing process would cause the output file to be truncated (data
		// loss).
		
		std::vector<std::vector<HLInfo> > tokenLists;
		for (unsigned i = 0; i < codeRanges.size(); ++i) {
			// create empty list of tokens for this snippet
			tokenLists.push_back(std::vector<HLInfo>());
			
			if (codeRanges[i].start == codeRanges[i].stop) {
				Console::getStdOut().line(
					UTF8STR("Snippet %1 is empty.").compose(i + 1)
				);
				continue;
			}
			
			// overlay stream for reading snippet
			cs.goBackTo(codeRanges[i].start);
			OverlayPullStream<UTF8Byte> ops(cs,
				codeRanges[i].stop - codeRanges[i].start);
			
			// decode HTML encoding (for dash char)
			HtmlCommentDecoder hcDec;
			PullStream<UTF8Byte>* htmlCommentDecPuller = hcDec.puller(ops);
			
			// decode UTF-8
			UTF8Decoder uDec;
			PullStream<CodePoint>* utf8DecPuller =
				uDec.puller(*htmlCommentDecPuller);
			
			// register each code snippet as a virtual file
			FilePath* snipFile =
				filename->getChildPath(
					UTF8STR("code_snippet_%1").compose(i + 1)
				);
			fInfo.registerFile(HYPER_Compiler::FileID(i), *snipFile);
			
			// run the lexer
			Reader reader(*utf8DecPuller, true);
			runLexer(reader, HYPER_Compiler::FileID(i), tokenLists.back());
		}
		
		// (4) Output the new file
		
		cs.goBackTo(0); // rewind input stream
		UTF8PushStream* ocs = FileSystem::openOutputFile(*filename);
		
		UTF8Byte u8b;
		unsigned codePos = 0;
		unsigned prePos = 0;
		
		while (cs.look(u8b)) {
			cs.pullNext();
			ocs->push(UTF8Byte(u8b));
			
			unsigned long pos = cs.streamPosition();
			if (codePos < codeRanges.size()
				&& pos == codeRanges[codePos].start
				&& pos != codeRanges[codePos].stop)
			{
				// We now regenerate the sourcecode in the comment tag.
				// This is to make sure that the comments no longer contain
				// unencoded dash ("-") characters, because the sequence
				// "--" is not allowed inside a HTML comment.
				
				//~ debug.line(UTF8STR("Regenerating code in comment"));
				//~ debug.line(
					//~ UTF8STR(" range from %1 to %2").compose(
						//~ codeRanges[codePos].start, codeRanges[codePos].stop
					//~ )
				//~ );
				
				cs.goBackTo(codeRanges[codePos].start);
				OverlayPullStream<UTF8Byte> ops(cs,
					codeRanges[codePos].stop - codeRanges[codePos].start);
				
				// decode HTML encoding (for dash char)
				HtmlCommentDecoder hcDec;
				PullStream<UTF8Byte>* htmlCommentDecPuller = hcDec.puller(ops);
				
				// encode back to HTML
				HtmlCommentEncoder hcEnc;
				PushStream<UTF8Byte>* htmlCommentEncPusher = hcEnc.pusher(*ocs);
				
				// copy stream from decoder to encoder
				copyStream(*htmlCommentDecPuller, *htmlCommentEncPusher);
				
				// go to end of comment
				cs.goBackTo(codeRanges[codePos].stop);
				
				// next <pre>
				codePos++;
			}
			else if (prePos < preRanges.size()
				&& pos == preRanges[prePos].start)
			{
				// We now (re)generate the <pre></pre> construction with the
				// syntax-highlighted code in it.
				
				// generate <pre> construction with sourcecode
				ocs->print(UTF8STR("<pre class='code'>"));
				
				if (codeRanges[prePos].start != codeRanges[prePos].stop)
				{
					// the code snippet is not empty
					cs.goBackTo(codeRanges[prePos].start);
					OverlayPullStream<UTF8Byte> ops(cs,
						codeRanges[prePos].stop - codeRanges[prePos].start);
					
					// decode HTML encoding (for dash char)
					HtmlCommentDecoder hcDec;
					PullStream<UTF8Byte>* htmlCommentDecPuller =
						hcDec.puller(ops);
					
					// Generate marked-up HTML from the sourcecode in the
					// comment and a corresponding token list generated earlier.
					generateHTML(
						*htmlCommentDecPuller, tokenLists[prePos], *ocs
					);
				}
				
				ocs->print(UTF8STR("</pre>"));
				
				// go to end of old <pre>
				cs.goBackTo(preRanges[prePos].stop);
				
				// next <pre>
				prePos++;
			}
		}
		
		printErrors();
		
		// done
	}
	
	void
	Highlighter::parseHTML(HYPER_Common::PullStream<UTF8Byte>& in,
		std::vector<FileRange>& codeRanges,
		std::vector<FileRange>& preRanges)
	{
		typedef std::map<UTF8String, bool> TagMap;
		TagMap notCloseTag;
		notCloseTag[UTF8STR("br")] = true;
		notCloseTag[UTF8STR("hr")] = true;
		notCloseTag[UTF8STR("img")] = true;
		notCloseTag[UTF8STR("link")] = true;
		notCloseTag[UTF8STR("meta")] = true;
		
		UTF8Decoder dec;
		PullStream<CodePoint>* decStr = dec.puller(in);
		Reader r(*decStr, true);
		Reader::Char curC = r.lookChar();
		
		std::stack<UTF8String> tagStack;
		bool codePreStartExpected = false;
		bool codePreEndExpected = false;
		unsigned long preStart = 0;
		
		do {
			if (curC == CodePoint('<')) { // begin of tag
				unsigned long tagStart = r.lookChar().getFilePos();
				r.readNextChar();
				
				if (r.lookChar().is(CodePoint::kASCIILetter)) { // start normal tag
					//debug << "normal open tag" << newl;
					
					UTF8String tag = readTag(r);
					//debug << "tag name: " << tag << newl;
					
					if (codePreStartExpected) {
						if (tag == UTF8STR("pre")) {
							//debug << "this is a <pre> tag for source code" << newl;
							codePreStartExpected = false;
							codePreEndExpected = true;
							preStart = tagStart;
							//debug << "preStart = " << preStart << newl;
						}
						else {
							parseError(
								r,
								UTF8STR(
									"expected: <pre> tag for highlighted "
									"source code, instead of <%1 ...>"
								).compose(tag)
							);
						}
					}
					
					// rest of tag
					skipRestOfTagContent(r);
					// current char is now '>'
					
					if (r.prevChar() == CodePoint('/')) { // <... /> tag
						//debug << "this tag closes itself again (XHTML style)" << newl;
						
						if (codePreEndExpected && tag == UTF8STR("pre")) {
							//debug << "end of <pre> for sourcecode" << newl;
							codePreEndExpected = false;
							unsigned long preEnd = r.lookChar().getFilePos() + 1;
							//debug << "preEnd = " << preEnd << newl;
							preRanges.push_back(FileRange(preStart, preEnd));
							hyp_assert(codeRanges.size() == preRanges.size());
						}
					}
					else if (notCloseTag[tag]) {
						//debug << "this tag will not be closed" << newl;
					}
					else {
						// tag must be closed later
						tagStack.push(tag);
					}
					
					//debug << "end of normal tag" << newl;
				}
				else if (r.lookChar() == CodePoint('!')) { // special tag (doctype or comment ...)
					//debug << "<! ...> tag" << newl;
					
					if (codePreStartExpected) {
						parseError(
							r,
							UTF8STR(
								"expected: <pre> tag for highlighted source "
								"code, instead of <! ...>"
							)
						);
					}
					else if (codePreEndExpected) {
						parseError(
							r,
							UTF8STR("expected: </pre> tag, instead of <! ...>")
						);
					}
					
					r.readNextChar();
					
					if (r.lookChar() == CodePoint('-')) { // could be comment
						r.readNextChar();
						if (r.lookChar() == CodePoint('-')) { // it's comment
							//debug << "tag is start of comment" << newl;
							r.readNextChar();
							
							bool code = false;
							unsigned long codeStart = 0;
							if (r.lookChar() == CodePoint('C')) { // could be sourcecode
								r.readNextChar();
								if (r.lookChar() == CodePoint('O')) {
									r.readNextChar();
									if (r.lookChar() == CodePoint('D')) {
										r.readNextChar();
										if (r.lookChar() == CodePoint('E')) {
											r.readNextChar();
											//debug << "comment is marked as Hyper sourcecode" << newl;
											code = true; // it should be sourcecode
											skipWhites(r);
											doNewline(r);
											codeStart = r.lookChar().getFilePos();
											//debug << "codeStart = " << codeStart << newl;
										}
									}
								}
							}
							
							while (true) {
								if (r.lookChar() == CodePoint('-')) {
									r.readNextChar();
									if (r.lookChar() == CodePoint('-')) {
										r.readNextChar();
										if (r.lookChar() == CodePoint('>')) {
											// comment ends
											if (code) { // if this was a code-snippet comment...
												unsigned long codeEnd = r.lookChar().getFilePos() - 2;
												codeRanges.push_back(
													FileRange(codeStart, codeEnd)
												);
												codePreStartExpected = true;
											}
											break;
										}
									}
								}
								
								r.readNextChar();
							}
							
							//debug << "end of comment reached" << newl;
						}
					}
					
					skipRestOfTagContent(r);
					//debug << "end of <! ...> tag" << newl;
				}
				else if (r.lookChar() == CodePoint('/')) { // closing tag
					//debug << "closing tag" << newl;
					
					r.readNextChar();
					UTF8String tag = readTag(r);
					//debug << "tag name: " << tag << newl;
					
					if (codePreStartExpected) {
						parseError(
							r,
							UTF8STR(
								"expected: <pre> tag for highlighted source "
								"code, instead of </%1>"
							).compose(tag)
						);
					}
					
					// search for the current close tag on the "open tags" stack
					while (true) {
						if (tagStack.empty()) {
							parseError(
								r,
								UTF8STR(
									"could not close tag \"%1\" because it is "
									"not open"
								).compose(tag)
							);
						}
						
						UTF8String stackTag = tagStack.top();
						tagStack.pop();
						
						if (stackTag != tag) {
							// implicitly closing current tag
						}
						else {
							break;
						}
					}
					
					skipWhites(r);
					// tag must be closed immediately
					if (r.lookChar() != CodePoint('>')) {
						parseError(
							r,
							UTF8STR(
								"closing tag </%1 ...> cannot have parameters"
							).compose(tag)
						);
					}
					r.readNextChar();
					
					if (codePreEndExpected && tag == UTF8STR("pre")) {
						//debug << "end of <pre> for sourcecode" << newl;
						codePreEndExpected = false;
						unsigned long preEnd = r.lookChar().getFilePos();
						//debug << "preEnd = " << preEnd << newl;
						preRanges.push_back(FileRange(preStart, preEnd));
						hyp_assert(codeRanges.size() == preRanges.size());
					}
				}
				else {
					parseError(r, UTF8STR("expected: tag name, '/' or '!'"));
				}
			}
			else { // simple text outside tags
				//debug << "text char" << newl;
				r.readNextChar();
			}
			
			curC = r.lookChar();
		} while (!curC.getChar().isNull());
		
		if (preRanges.size() < codeRanges.size()) {
			parseError(
				r,
				UTF8STR("end of file reached before <pre></pre> for sourcecode")
			);
		}
		else if (!tagStack.empty()) { // still tags open
			parseError(
				r,
				UTF8STR(
					"end of file reached when there are still tags not closed; "
					"first close tag expected is </%1>"
				).compose(tagStack.top())
			);
		}
	}
	
	void Highlighter::copyStream(HYPER_Common::PullStream<UTF8Byte>& in,
		HYPER_Common::PushStream<UTF8Byte>& out)
	{
		// assuming both streams use the same character type
		UTF8Byte c;
		//~ debug.print(UTF8STR("Copystream "));
		while (in.look(c) && out.push(c)) {
			// just loop until input is exhausted or output fails
			in.pullNext();
			//~ debug.print(c.toByte());
			//~ debug.print(CodePoint(' '));
		}
		//~ debug.line();
	}
	
	void Highlighter::skipWhites(Reader& r) {
		while (r.lookChar().is(CodePoint::kIndentWhite)) {
			r.readNextChar();
		}
	}
	
	void Highlighter::doNewline(Reader& r) {
		if (r.lookChar().is(CodePoint::kLf)) {
			r.readNextChar();
		}
		else {
			parseError(r, UTF8STR("expected: newline"));
		}
	}
	
	UTF8String Highlighter::readTag(Reader& r) {
		// read tag name
		UTF8StringBuilder tagStream;
		tagStream += r.lookChar().getChar().toLower();
		r.readNextChar();
		
		while (r.lookChar().is(CodePoint::kAlphaNumeric)) {
			tagStream += r.lookChar().getChar().toLower();
			r.readNextChar();
		}
		
		return tagStream.toString();
	}
	
	void Highlighter::skipQuoted(Reader& r) {
		Reader::Char quote = r.lookChar();
		r.readNextChar();
		while (r.lookChar() != quote) {
			r.readNextChar();
		}
		r.readNextChar(); // eat terminating quote
	}
	
	void Highlighter::skipRestOfTagContent(Reader& r) {
		while (r.lookChar() != CodePoint('>')) {
			if (r.lookChar() == CodePoint('"')
					|| r.lookChar() == CodePoint('\''))
			{
				//debug << "quoted" << newl;
				skipQuoted(r);
			}
			else {
				r.readNextChar();
			}
		}
	}
	
	void Highlighter::parseError(Reader& r, UTF8String msg) {
		throw ParseError(
			r.lookChar().getLineNo(), r.lookChar().getLinePos(), msg
		);
	}
	
	void Highlighter::runLexer(Reader& r, HYPER_Compiler::FileID fileID,
		std::vector<HLInfo>& tokens)
	{
		tokens.clear();
		
		Lexer lex(r, fileID, &fInfo, true);
		bool stop = false;
		
		do {
			Token tok = lex.lookToken();
			Token::Class c = tok.getClass();
			
			//~ debug.line(
				//~ UTF8STR("Read Token: %1").compose(tok.getDescriptiveString())
			//~ );
			
			if (tok.isEndOfFile()) stop = true;
			
			HLInfo cur;
			cur.type = c;
			cur.beginPos = tok.getFileCharPos();
			cur.endPos = cur.beginPos + tok.getLength();
			//debug << "Token " << tok.getDescriptiveString() << newl;
			//debug << "pos: [" << cur.beginPos << ", " << cur.endPos << "]" << newl;
			
			if (cur.endPos < cur.beginPos) {
				cur.endPos = cur.beginPos;
			}
			
			if (!stop) {
				tokens.push_back(cur);
				lex.getToken();
			}
			
		} while (!stop);
	}
	
	void Highlighter::generateHTML(HYPER_Common::PullStream<UTF8Byte>& in,
		std::vector<HLInfo> const& tokens,
		HYPER_Common::UTF8PushStream& out)
	{
		//~ debug.line(UTF8STR("generateHTML called"));
		//~ debug.line(UTF8STR(" token info count: %1").compose(tokens.size()));
		//~ IndentDebug id;
		
		UTF8Decoder inD;
		PullStream<CodePoint>* inDStr = inD.puller(in);
		Reader* r = new Reader(*inDStr, true);
		
		hyp_assert(r->isFileOpened() && "Could not open snippet input");
		
		unsigned int curTokenNr = 0;
		bool noMarkup;
		while (!r->lookChar().getChar().isNull()) {
			//debug.line(UTF8STR("Got char: ") + r->lookChar().getChar());
			if (curTokenNr < tokens.size()) {
				HLInfo cur = tokens[curTokenNr];
				
				// start style marking for begin of token
				noMarkup = false;
				if (r->lookChar().getFilePos() == cur.beginPos) {
					switch (cur.type) {
					case Token::kComment:
						noMarkup = true;
						out.print(UTF8STR("<span class='comment'>"));
						break;
					case Token::kKeyword:
						out.print(UTF8STR("<span class='keyword'>"));
						break;
					case Token::kWholeNumber:
					case Token::kFloatingPoint:
						out.print(UTF8STR("<span class='number'>"));
						break;
					case Token::kString:
						noMarkup = true;
						out.print(UTF8STR("<span class='string'>"));
						break;
					case Token::kChar:
						out.print(UTF8STR("<span class='char'>"));
						break;
					default:
						break; // do nothing
					}
				}
			}
			
			// output char
			CodePoint c = r->lookChar().getChar();
			if (c.is(CodePoint::kSpace)) { // convert space to "&nbsp;"
				out.print(UTF8STR("&nbsp;"));
			}
			else if (c.is(CodePoint::kLessThan)) { // less-than sign
				out.print(UTF8STR("&lt;"));
			}
			else if (c.is(CodePoint::kGreaterThan)) { // greater-than sign
				out.print(UTF8STR("&gt;"));
			}
			else if (c.is(CodePoint::kAmpersand)) { // ampersand
				out.print(UTF8STR("&amp;"));
			}
			else {
				out.print(r->lookChar().getChar());
			}
			
			if (curTokenNr < tokens.size()) {
				HLInfo cur = tokens[curTokenNr];
				
				// end style marking for end of token
				if (r->lookChar().getFilePos() + 1 >= cur.endPos) {
					switch (cur.type) {
					case Token::kComment:
					case Token::kKeyword:
					case Token::kWholeNumber:
					case Token::kFloatingPoint:
					case Token::kString:
					case Token::kChar:
						// close tag
						out.print(UTF8STR("</span>"));
						break;
					case Token::kNewLine:
					case Token::kEndOfFile:
					case Token::kIdentifier:
					case Token::kSymbol:
					case Token::kEmpty:
						// no action
						break;
					}
					
					curTokenNr++;
				}
				else if (r->lookChar().getFilePos() > cur.endPos) {
					curTokenNr++;
				}
			}
			
			r->readNextChar();
			//debug.line(UTF8STR("streampos: %1").compose(inDStr->streamPosition()));
		} // end while
		
		hyp_assert(curTokenNr == tokens.size() && "not all tokens printed");
	}
	
	void Highlighter::writeCSSFile(FilePath& name) {
		UTF8PushStream* cssOS = FileSystem::openOutputFile(name);
		TextFlowOptions tfOpt;
		tfOpt.setIndentChar(CodePoint::tab(), 8);
		UTF8TextFlowStream cssTOS(*cssOS, tfOpt);
		
		cssTOS.line();
		cssTOS.line(UTF8STR("pre.code {"));
		cssTOS.increaseIndent();
		cssTOS.line(UTF8STR("font-family: monospace;"));
		cssTOS.line(UTF8STR("font-size: 9pt;"));
		cssTOS.line(UTF8STR("color: black;"));
		cssTOS.line(UTF8STR("background-color: #EEFFFF;"));
		cssTOS.line(UTF8STR("margin-left: 2em;"));
		cssTOS.line(UTF8STR("margin-right: 2em;"));
		cssTOS.line(UTF8STR("border: 1px solid #B0D0D0;"));
		cssTOS.line(UTF8STR("padding: 5px;"));
		cssTOS.decreaseIndent();
		cssTOS.line(UTF8STR("}"));
		cssTOS.line(UTF8STR(".code {"));
		cssTOS.increaseIndent();
		cssTOS.line(UTF8STR("font-family: monospace;"));
		cssTOS.line(UTF8STR("color: black;"));
		cssTOS.line(UTF8STR("background-color: #E0E0E0;"));
		cssTOS.decreaseIndent();
		cssTOS.line(UTF8STR("}"));
		cssTOS.line(UTF8STR(".comment {"));
		cssTOS.increaseIndent();
		cssTOS.line(UTF8STR("font-family: monospace;"));
		cssTOS.line(UTF8STR("color: #007F00;"));
		cssTOS.line(UTF8STR("background-color: inherit;"));
		cssTOS.decreaseIndent();
		cssTOS.line(UTF8STR("}"));
		cssTOS.line(UTF8STR(".keyword {"));
		cssTOS.increaseIndent();
		cssTOS.line(UTF8STR("font-family: monospace;"));
		cssTOS.line(UTF8STR("color: #00007F;"));
		cssTOS.line(UTF8STR("background-color: inherit;"));
		cssTOS.line(UTF8STR("font-weight: bold;"));
		cssTOS.decreaseIndent();
		cssTOS.line(UTF8STR("}"));
		cssTOS.line(UTF8STR(".number {"));
		cssTOS.increaseIndent();
		cssTOS.line(UTF8STR("font-family: monospace;"));
		cssTOS.line(UTF8STR("color: #007F7F;"));
		cssTOS.line(UTF8STR("background-color: inherit;"));
		cssTOS.decreaseIndent();
		cssTOS.line(UTF8STR("}"));
		cssTOS.line(UTF8STR(".string {"));
		cssTOS.increaseIndent();
		cssTOS.line(UTF8STR("font-family: monospace;"));
		cssTOS.line(UTF8STR("color: #7F007F;"));
		cssTOS.line(UTF8STR("background-color: inherit;"));
		cssTOS.decreaseIndent();
		cssTOS.line(UTF8STR("}"));
		cssTOS.line(UTF8STR(".char {"));
		cssTOS.increaseIndent();
		cssTOS.line(UTF8STR("font-family: monospace;"));
		cssTOS.line(UTF8STR("color: #7F007F;"));
		cssTOS.line(UTF8STR("background-color: inherit;"));
		cssTOS.decreaseIndent();
		cssTOS.line(UTF8STR("}"));
	}
	
	void Highlighter::printErrors() {
		fTxtOut.line();
		ErrorManager& em = fInfo.getErrorManager();
		typedef ErrorManager::DiagnosticIterator EMDI;
		for (EMDI it = em.begin(); it != em.end(); ++it) {
			HYPER_Compiler::Diagnostic const& d = *it;
			if (d.getType() == Diagnostic::kError) {
				em.print(fTxtOut, d);
			}
		}
	}
} // end namespace
