/*!
	\file highlighter.h
	\brief Contains the Highlighter class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created: 	2004-04-25
// Date modified:	2008-06-29

#ifndef DEF_HIGHLIGHTER_H
 #define DEF_HIGHLIGHTER_H
 
 #include "base/basiccompileinfo.h"
 #include "base/location.h"
 
 #include "common/exception.h"
 #include "common/paths_fwdecl.h"
 #include "common/streams_fwdecl.h"
 #include "common/utf8stdoutstream.h"
 #include "common/utf8textflowstream.h"
 
 #include "lexer/token.h" // for Token::Class
 
 #include <vector>

namespace HYPER_Compiler {
	class Reader;
}

/// Contains all code of the highlighter
namespace HYPER_Highlighter {
	
	using HYPER_Common::CodePoint;
	using HYPER_Common::FilePath;
	using HYPER_Common::UTF8Byte;
	using HYPER_Common::UTF8String;
	
	class ParseError : public HYPER_Common::Exception {
	public:
		unsigned long line;
		unsigned long column;
		UTF8String msg;
		
		ParseError(unsigned long line_, unsigned long column_, UTF8String msg_)
		 : line(line_), column(column_), msg(msg_)
		{
			//
		}
	};
	
	/** Provides source highlighting functionality */
	class Highlighter {
	public:
		/** Constructor */
		Highlighter();
		
		/**
			Run the highlighter on a hyper sourcefile to produce a HTML file.
			The given pathnames must be absolute.
		*/
		void sourceFileToHTML(FilePath* filename,
			FilePath* destfilename, FilePath* stylesheet,
			bool createStylesheet);
		
		/**
			Highlights code snippets in comments in a HTML file. The
			pathname must be absolute.
		*/
		void highlightHTML(FilePath* filename);
		
		/// Print errors given by the compiler
		void printErrors();
		
		/// Get the version of the highlighter
		static UTF8String getVersion();
		
	private:
		/// highlight info
		class HLInfo {
		public:
			unsigned int beginPos; // file character index of begin of token
			unsigned int endPos; // file character index of end of token
			HYPER_Compiler::Token::Class type; // token type
		};
		
		// Run the lexer and generate highlight info for all tokens
		void runLexer(HYPER_Compiler::Reader& r, HYPER_Compiler::FileID fileID,
			std::vector<HLInfo>& tokens);
		
		// Generate HTML from the highlight info
		void generateHTML(HYPER_Common::PullStream<UTF8Byte>& in,
			std::vector<HLInfo> const& tokens,
			HYPER_Common::UTF8PushStream& out);
		
		
		/// A range in the HTML source, indicated by the character start
		/// and stop positions. 'start' is included, 'stop' is one past the end
		/// of the range.
		class FileRange {
		public:
			FileRange(unsigned long start_, unsigned long stop_);
			
			unsigned long start;
			unsigned long stop;
		};
		
		void parseHTML(HYPER_Common::PullStream<UTF8Byte>& in,
			std::vector<FileRange>& codeRanges,
			std::vector<FileRange>& preRanges);
		
		void copyStream(HYPER_Common::PullStream<UTF8Byte>& in,
			HYPER_Common::PushStream<UTF8Byte>& out);
		
		void writeCSSFile(FilePath& name);
		
		void skipWhites(HYPER_Compiler::Reader& r);
		void doNewline(HYPER_Compiler::Reader& r);
		UTF8String readTag(HYPER_Compiler::Reader& r);
		void skipQuoted(HYPER_Compiler::Reader& r);
		void skipRestOfTagContent(HYPER_Compiler::Reader& r);
		void parseError(HYPER_Compiler::Reader& r, UTF8String msg);
		
		HYPER_Common::UTF8PushStream& fOut;
		HYPER_Common::UTF8TextFlowStream& fTxtOut;
		HYPER_Compiler::BasicCompileInfo fInfo;
	};
}
#endif
