/**
	\file htmlcommentdecoder.cxx
	\author Kevin André
	\brief Implements the HtmlCommentDecoder class.
	
	Copyright (C) 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2007-03-08
// Last time modified:	2008-05-04

#include "htmlcommentdecoder.h"

//~ #include "compiler/debug.h"
//~ using HYPER_Compiler::debug;

namespace HYPER_Highlighter {
	
	static const unsigned READ_AHEAD = 6;
	
	using HYPER_Common::CodePoint;
	using HYPER_Common::UTF8Byte;
	
	HtmlCommentDecoder::ProcessStatus
	HtmlCommentDecoder::processData(std::deque<HYPER_Common::UTF8Byte>& input,
		bool endOfInputReached, std::deque<HYPER_Common::UTF8Byte>& output,
		State& processState) const
	{
		/*
			We look for the sequence "&#045;" or "&#45;" (without the quotes).
			These each translate to a dash ("-").
		*/
		
		unsigned long bytesAvail = input.size();
		bool matching1 = true;
		bool matching2 = true;
		unsigned index = 0;
		
		// try to match one of the sequences
		do {
			//~ debug.line(UTF8STR("HtmlCommentDecoder: index %1").compose(index));
			
			if (matching1 && fDash1[index] == input[index]) {
				// OK, still matching
				
				if (index == kDash1Length - 1) {
					//~ debug.line(UTF8STR("HtmlCommentDecoder: match pattern 1"));
					
					// match complete
					output.push_front(UTF8Byte('-'));
					for (unsigned i = 0; i <= index; ++i) {
						input.pop_front();
					}
					return kOK;
				}
			}
			else {
				matching1 = false;
			}
			
			if (matching2 && fDash2[index] == input[index]) {
				// OK, still matching
				
				if (index == kDash2Length - 1) {
					//~ debug.line(UTF8STR("HtmlCommentDecoder: match pattern 2"));
					
					// match complete
					output.push_front(UTF8Byte('-'));
					for (unsigned i = 0; i <= index; ++i) {
						input.pop_front();
					}
					return kOK;
				}
			}
			else {
				matching2 = false;
			}
			
			++index;
			if (index >= bytesAvail && (matching1 || matching2)) {
				// need more input
				if (matching1) {
					processState.fgInputNeeded = kDash1Length;
				}
				
				if (matching2) {
					processState.fgInputNeeded = kDash2Length;
				}
				
				//~ debug.line(
					//~ UTF8STR(
						//~ "HtmlCommentDecoder: not enough input, requesting "
						//~ "at least %1"
					//~ ).compose(processState.fgInputNeeded)
				//~ );
				return kNeedMoreData;
			}
		} while (matching1 || matching2);
		
		//~ debug.line(UTF8STR("HtmlCommentDecoder: no match"));
		
		// if we get here, no match was active or found.
		output.push_front(input[0]);
		input.pop_front();
		return kOK;
	}
}
