/*!
	\file highlighter/main.cxx
	\brief Highlighter main program
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.

*/
// Date created:	2004-04-25
// Date modified:	2008-09-13

#include "highlighter.h"

#include "base/internalcompilererror.h"

#include "common/copyright.h"
#include "common/console.h"
#include "common/filepath.h"
#include "common/filesystem.h"
#include "common/logicutil.h"
#include "common/utf8stdoutstream.h"
#include "common/utf8textflowstream.h"

#include "compiler/version.h"

#include "lexer/fatalparseerror.h"
#include "lexer/fileexceptions.h"

#include <iostream>
#include <list>

#include <boost/program_options.hpp>

using namespace HYPER_Highlighter;
using namespace HYPER_Common;
using HYPER_Compiler::Version;
namespace po = boost::program_options;

void printUsage(UTF8TextFlowStream& out);
void printVersion(UTF8TextFlowStream& out);

int main(int argc, char *argv[])
{
	// describe commandline options
	po::options_description desc("Options");
	std::string cssOption;
	std::string relCssOption;
	desc.add_options()
		("help,h", "show usage")
		("version,v", "show program version")
		("inputfile", po::value<std::vector<std::string> >(), "input file")
		("outputfile", po::value<std::vector<std::string> >(), "output file")
		("css", po::value<std::string>(&cssOption))
		("rel-css", po::value<std::string>(&relCssOption))
		("css-nocreate", "don't create CSS file");
	po::positional_options_description pod;
	pod.add("inputfile", 1);
	pod.add("outputfile", 1);
	po::variables_map vm;
	int cmdStyle =
		po::command_line_style::allow_long
		| po::command_line_style::allow_short
		| po::command_line_style::allow_dash_for_short
		| po::command_line_style::long_allow_adjacent
		| po::command_line_style::long_allow_next
		| po::command_line_style::short_allow_next
		| po::command_line_style::allow_long_disguise;
	
	// std streams
	UTF8TextFlowStream& stdOut = Console::getTextFlowStdOut();
	UTF8TextFlowStream& stdErr = Console::getTextFlowStdErr();
	
	try {
		// parse commandline options
		po::store(
			po::command_line_parser(argc, argv).style(cmdStyle)
				.options(desc).positional(pod).run(), vm
		);
		po::notify(vm);
	}
	catch (po::unknown_option& uo) {
		stdErr.line(
			UTF8STR("hyperh: %1 (use option '--help' for help)").compose(
				UTF8STR(uo.what())
			)
		);
		return 1;
	}
	catch (po::too_many_positional_options_error& tmpoe) {
		stdErr.line(
			UTF8STR("hyperh: %1 (use option '--help' for help)").compose(
				UTF8STR(tmpoe.what())
			)
		);
		return 1;
	}
	catch (po::invalid_command_line_syntax& icls) {
		stdErr.line(
			UTF8STR("hyperh: %1 (use option '--help' for help)").compose(
				UTF8STR(icls.what())
			)
		);
		return 1;
	}
	
	// check for option '--version'
	if (vm.count("version")) {
		printVersion(stdOut);
		return 0;
	}
	
	// check for option 'help'
	if (vm.count("help")) {
		//std::cout << desc << "\n";
		printUsage(stdOut);
		return 0;
	}
	
	if (vm.count("inputfile") == 0) {
		// no file args
		stdErr.line(
			UTF8STR(
				"hyperh: you must specify 1 or 2 arguments "
					"(or use option '--help' for help)\n"
			)
		);
		return 1;
	}
	
	FilePath const* workingDir = FileSystem::getWorkingDirectory();
	//std::cout << "PWD: " << workingDir->toString() << std::endl;
	
	FilePath* source = 0;
	FilePath* dest = 0;
	FilePath* styleSheet = 0;
	
	std::list<UTF8String> hyperExts;
	hyperExts.push_back(UTF8STR("hyp"));
	hyperExts.push_back(UTF8STR("hyper"));
	
	std::list<UTF8String> htmlExts;
	htmlExts.push_back(UTF8STR("htm"));
	htmlExts.push_back(UTF8STR("html"));
	
	std::list<UTF8String> cssExts;
	cssExts.push_back(UTF8STR("css"));
	
	try {
		std::string input =
			vm["inputfile"].as<std::vector<std::string> >()[0];
		UTF8String inputU8Str = UTF8String::fromCString(input.c_str());
		source = FileSystem::pathFromString(inputU8Str);
		source->makeAbsolute(*workingDir);
	}
	catch (InvalidPathStringException const& ipse) {
		stdErr.line(
			UTF8STR("hyperh: invalid filename `%1'").compose(ipse.getString())
		);
		return 1;
	}
	
	if (source->hasExtension(htmlExts)) {
		// HTML highlight mode
		
		if (cssOption != "" || relCssOption != "") {
			stdErr.line(
				UTF8STR("hyperh: css options not allowed for this mode")
			);
			return 1;
		}
		
		if (vm.count("outputfile") > 0) {
			stdErr.line(
				UTF8STR("hyperh: no output file allowed for this mode")
			);
			return 1;
		}
	}
	else if (source->hasExtension(hyperExts)) {
		// Source to HTML mode
		
		if (vm.count("outputfile") != 0) {
			// user specified output file
			std::string output =
				vm["outputfile"].as<std::vector<std::string> >()[0];
			UTF8String outputU8Str = UTF8String::fromCString(output.c_str());
			dest = FileSystem::pathFromString(outputU8Str);
			dest->makeAbsolute(*workingDir);
			
			if (!dest->hasExtension(htmlExts)) {
				stdErr.line(
					UTF8STR("hyperh: output file must be a HTML file")
				);
				return 1;
			}
		}
		else {
			// output file not specified; use input file as inspiration
			dest = source->copy();
			dest->replaceExtension(UTF8STR("html"));
		}
		//std::cout << "HTML output: " << dest->toString() << std::endl;
		
		if (!LogicUtil::exclusive(cssOption != "", relCssOption != "")) {
			stdErr.line(
				UTF8STR("hyperh: different CSS options cannot be combined")
			);
			return 1;
		}
		
		if (cssOption != "") {
			// user-defined css file, relative to working directory
			styleSheet =
				FileSystem::pathFromString(
					UTF8String::fromCString(cssOption.c_str())
				);
			
			if (!styleSheet->isAbsolute()) {
				styleSheet->makeAbsolute(*workingDir);
			}
		}
		else if (relCssOption != "") {
			// user-defined css file, relative to the HTML file's directory
			styleSheet =
				FileSystem::pathFromString(
					UTF8String::fromCString(relCssOption.c_str())
				);
			
			if (!styleSheet->isAbsolute()) {
				styleSheet->makeAbsolute(*dest->getParentDirectory());
			}
		}
		else {
			// default stylesheet name
			styleSheet = FileSystem::pathFromString(UTF8STR("code.css"));
			styleSheet->makeAbsolute(*dest->getParentDirectory());
		}
		
		if (!styleSheet->hasExtension(cssExts)) {
			stdErr.line(
				UTF8STR("hyperh: css file must have a 'css' extension\n")
			);
			return 1;
		}
		//std::cout << "Stylesheet: " << styleSheet->toString()
		//	<< std::endl;
	}
	else {
		stdErr.line(
			UTF8STR(
				"hyperh: input file has inappropriate type "
					"(must be HTML or hyper source)\n"
			)
		);
		return 1;
	}
	
	try {
		Highlighter h;
		try {
			if (source->hasExtension(htmlExts)) {
				h.highlightHTML(source);
			}
			else {
				h.sourceFileToHTML(
					source, dest, styleSheet, !vm.count("css-nocreate")
				);
			}
			
			delete source;
			delete dest;
		}
		catch (HYPER_Compiler::FatalParseError const& fpe) {
			// Lexical error in the sourcefile
			h.printErrors();
			
			delete source;
			delete dest;
			return 2;
		}
		catch (HYPER_Compiler::FileNotReadableException const& fnre) {
			stdErr.line(
				UTF8STR("hyperh: could not read the file \"%1\"").compose(
					UTF8STR(argv[1])
				)
			);
			
			delete source;
			delete dest;
			return 1;
		}
		catch (ParseError const& he) {
			// HTML parse error
			stdErr.line(
				UTF8STR("HTML parse error at line %1, column %2: %3").compose(
					he.line, he.column, he.msg
				)
			);
			return 3;
		}
	}
	catch (HYPER_Compiler::InternalCompilerError const& ice) {
		// Internal compiler error in the lexical analysis, or in error printing
		stdErr.line(UTF8STR("INTERNAL ERROR:"));
		stdErr.line(ice.getDescription());
		stdErr.line(UTF8STR(">> Please submit a full bug report.\n"));
		return 4;
	}
	
	return 0;
}

void printUsage(UTF8TextFlowStream& out) {
	out.print(
		UTF8STR(
			"Usage:\n"
			"     hyperh [CSS-OPTIONS] sourcefile [html-file]\n"
			"     hyperh html-file\n"
			"     hyperh --help\n"
			"     hyperh --version\n"
		)
	);
	out.line();
	out.print(
		UTF8STR(
			"Options:\n"
			"  -h, --help      display this information\n"
			"  -v, --version   print version information\n"
			"  -css            name of CSS file\n"
			"  -rel-css        name of CSS file, relative to HTML file\n"
			"  -css-nocreate   assume that the CSS file already exist\n"
		)
	);
	out.line();
	out.print(
		UTF8STR(
			"The first form reads a Hyper sourcefile and outputs a "
				"corresponding HTML file with syntax highlighting applied "
				"to the contents of the sourcefile.\n"
			"\n"
			"The second form reads a HTML file and applies syntax "
				"highlighting to code snippets that it contains. This is "
				"useful for HTML documentation.\n"
			"\n"
			"IMPORTANT: the highlighter program uses the lexical "
				"components of the compiler in order to read Hyper "
				"sourcecode. When a lexical error is found, it is reported "
				"and no highlighter output is generated.\n"
		)
	);
	out.line();
	out.print(
		UTF8STR(
			"The first form generates an external stylesheet for code "
				"markup in the HTML file. The default name of that stylesheet "
				"is 'code.css', but another name can be specified with "
				"'-css' or '-rel-css'. The second form however "
				"assumes that you have taken care of defining the required "
				"CSS style information in the HTML file yourself. You can "
				"refer to an external style sheet or copy the style "
				"information out of a file that was generated by a command "
				"of the first form into your HTML file.\n"
		)
	);
	out.line();
	out.print(
		UTF8STR(
			"When you use the second form, the sourcecode that you want "
				"to be highlighted in the HTML file must be contained in a "
				"comment like this:\n"
			"\n"
			"  <!--CODE\n"
			"  procedure test() : bool\n"
			"    return true\n"
			"  end -->\n"
			"  <pre></pre>\n"
			"\n"
			"As shown in the example, the command is identified as a code "
				"snippet by putting 'CODE' right after the start of the "
				"comment (with no spaces in between!). A newline is then "
				"required, because the actual code must start on a separate "
				"line. Right after the comment you have to put a section of "
				"preformatted text. This can be as simple as just "
				"<pre></pre> or even <pre/>. The highlighter will then use "
				"this preformatted section to place the highlighted "
				"sourcecode in. Then when you change the sourcecode in the "
				"comment you can simply run the highlighter on the file again, "
				"and the preformatted section will be overwritten "
				"with a new one. Any <pre> section that does not directly "
				"follow a 'CODE' comment will of course be safely "
				"ignored.\n"
		)
	);
	out.line();
	out.print(
		UTF8STR(
			"The treatment of HTML comments is not according to the "
				"official standard; the program assumes that the three "
				"characters '-->' indicate the end of a comment, but "
				"that is not necessarily always the case. This means you "
				"cannot use the combination of those three characters "
				"anywhere in the sourcecode. Any sequence of two dashes "
				"'--' in the comment is considered bad practice by the "
				"standard. Because they can easily occur in sourcecode, "
				"the highlighter will convert any dash character it finds"
				" in the comment to an equivalent sequence '&#045;'. This "
				"avoids confusion for a webbrowser that complies to the "
				"standard with respect to the interpretation of '--' in "
				"comments.\n"
			"\n"
			"The current version of this program uses a very simple HTML "
				"parser. It has a few basic requirements for a HTML file "
				"to be acceptable. If one of those requirements is not met "
				"then an error message will be generated. Do not expect too "
				"much of these tests; they are only intended to guarantee "
				"the correct workings of the highlighter, and most of the "
				"content of the HTML tags is ignored because it is "
				"irrelevant for syntax highlighting. The HTML input "
				"functionality has not received much testing yet, so please "
				"report any bug(s) you might find.\n"
		)
	);
	out.line();
	
	out.print(UTF8STR("Report bugs to: "));
	out.line(UTF8STR(Version::getBugReportLocation()));
}

void printVersion(UTF8TextFlowStream& out) {
	out.line(UTF8STR("Hyper Highlighter ") + Highlighter::getVersion());
	out.line(Copyright::copyrightYearsAndAuthor());
	out.line();
	out.line(Copyright::licenseAndDisclaimer());
	out.line();
	out.print(UTF8STR("Report bugs to: "));
	out.line(UTF8STR(Version::getBugReportLocation()));
}
