/*!
	\file lexer.cxx
	\brief Implementation of the lexer.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-02-05
// Last time modified:	2008-07-11

// TODO: Add support for the exponent in a floating-point literal

#include "lexer.h"

#include "base/asserts.h"
#include "base/basiccompileinfo.h"
#include "base/compileoptions.h"
//~ #include "base/debug.h"
#include "base/easydiagnostics.h"
#include "base/location.h"

#include "fileexceptions.h"

#include <vector>

using HYPER_Common::CodePoint;
using HYPER_Common::UTF8String;
using HYPER_Common::UTF8StringBuilder;

namespace HYPER_Compiler {
	
	// **** The LEXER class ****
	
	// Constructor
	Lexer::Lexer(Reader& input, FileID sourceFileID,
		BasicCompileInfo* info, bool returnAllTokens)
	 : fFileID(sourceFileID), fCurCharLoc(0), fReader(input),
		fWhitesRead(false), fReturnAllTokens(returnAllTokens),
		fTokenPos(0), fInfo(info)
	{
		// Prepare first lookahead char
		fCurChar = fReader.lookChar();
		
		// Prepare first token
		getToken();
	}
	
	// Destructor
	Lexer::~Lexer() {
		//
	}
	
	Keywords const& Lexer::getKeywordsList() const {
		return Keywords::list();
	}
	
	// Returns the current token
	Token Lexer::lookToken() {
		return fCurToken;
	}
	
	// Gets the next token
	void Lexer::getToken() {
		//debug.line(UTF8STR("Lexer::getToken called"));
		
		if (fCurToken.getClass() == Token::kEndOfFile) {
			// At end of file throw an exception (there is no next token)
			throw EndOfFileException();
		}
		else if (!fTokensComing.empty()) {  // If Tokens on stack, use them
			Token t = fTokensComing.top();
			fTokensComing.pop();
			fTokensRead.push(t);
			fCurToken = t;
			++fTokenPos;
			return;
		}
		else { // Read characters and construct a Token to return
			HYPER_Common::CodePoint lc;
			//debug.line(UTF8STR("Lexer::getToken : read & construct token"));
			
			do {
				lc = fCurChar.getChar();
				
				//debug.line(UTF8STR("Lexer::getToken : got first char"));
				
				if (lc.is(CodePoint::kNull)) {
					// we're at end of file, return an end-of-file Token
					fCurToken =
						Token::createEndOfFile(getFileCharPos(), lookCharPos());
				}
				else if (lc.is(CodePoint::kLf)) {
					fCurToken = doNewLine();  // is a newline
				}
				else if (lc.is(CodePoint::kIndentWhite)) {
					doWhiteSpace();  // is whitespace
					fCurToken = Token(); // no token read
				}
				else if (lc.is(CodePoint::kHashMark)) {
					fCurToken = doLineComment();  // is a line comment
				}
				else if (lc == CodePoint('"')) {
					fCurToken = doString();
				}
				else if (lc == CodePoint('\'')) {
					fCurToken = doChar();
				}
				else if (lc.is(CodePoint::kASCIILetter)) {
					// could be a keyword or just an identifier
					fCurToken = doIdentOrKeyword();
				}
				else if (lc.is(CodePoint::kDigit)) {
					fCurToken = doNumber();  // must be a number
				}
				else {
					// could be symbol or invalid character.
					fCurToken = doOther();
				}
				
				if (fCurToken.getClass() == Token::kEmpty) {
					// empty token: keep running
					//debug.line(UTF8STR("Lexer::getToken : empty token, will retry"));
				}
				else {
					//debug.line(UTF8STR("Lexer::getToken : got a token"));
					if (fWhitesRead) {
						fWhitesRead = false;
						fCurToken.setPrecededByWhitespace(true);
					}
					break; // token is OK and can be returned
				}
			} while (true);
			
			fTokensRead.push(fCurToken);
			++fTokenPos;
			fLastGeneratedToken = fCurToken;
			//debug.line(UTF8STR("Lexer::getToken : token created, returning"));
			return;
		}
	}
	
	// Gets the position of the current token
	Location const* Lexer::lookPos() {
		return fCurToken.getPosition();
	}
	
	unsigned long Lexer::getFileCharPos() {
		return fCurChar.getFilePos();
	}
	
	unsigned long Lexer::getTokenPos() {
		return fTokenPos;
	}
	
	// Goes back one token
	void Lexer::goBackOneToken() {
		if (fTokensRead.empty()) {
			throw InvalidOperationException();
		}
		
		Token t = fTokensRead.top();
		fTokensRead.pop();
		fTokensComing.push(t);
		--fTokenPos;
		
		fCurToken = fTokensRead.top();
	}
	
	void Lexer::goBackToTokenPos(int tokenPos) {
		if (fTokenPos < tokenPos || tokenPos < 1)
			throw InvalidOperationException();
		
		while (fTokenPos > tokenPos) {
			goBackOneToken();
		}
	}
	
	// Processes a newline
	Token Lexer::doNewLine() {
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		
		hyp_assert(fCurChar.getChar().is(CodePoint::kLf));
		getNextChar(); // The character is processed
		
		// check for line continuation
		if (fCurChar.getChar().is(CodePoint::kUnderscore)) {
			getNextChar(); // eat underscore
			
			// eat whitespace behind the underscore
			while (fCurChar.getChar().is(CodePoint::kIndentWhite)) {
				getNextChar();
			}
			
			// return an empty token
			return Token();
		}
		else {
			// no line continuation -> return newline
			return Token::createNewLine(fcp, p);
		}
	}
	
	static bool isIdentChar(CodePoint c) {
		return c.is(CodePoint::kAlphaNumeric) || c.is(CodePoint::kUnderscore);
	}
	
	Token Lexer::doIdentOrKeyword() {
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		
		UTF8StringBuilder ident;
		CodePoint look;
		
		do {
			ident += fCurChar.getChar();
			getNextChar();
			look = fCurChar.getChar();
		} while (isIdentChar(look));
		
		Keywords::Keyword k = getKeywordsList().checkKeyword(ident.toString());
		
		if (k == Keywords::kNoKeyword) {
			return Token::createIdentifier(fcp, p, ident.toString());
		}
		else {
			return Token::createKeyword(fcp, p, k, ident.toString());
		}
	}
	
	// Processes whitespace
	void Lexer::doWhiteSpace() {
		do {
			getNextChar(); // Get next char
		} while (fCurChar.getChar().is(CodePoint::kIndentWhite));
		
		fWhitesRead = true;
	}
	
	Token
	Lexer::doCharWithOptionalAssignChar(Token::Symbol single,
		Token::Symbol withAssign)
	{
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		
		Token::Symbol sym;
		getNextChar();
		if (fCurChar.getChar() == CodePoint('=')) {
			getNextChar();
			sym = withAssign;
		}
		else {
			sym = single;
		}
		
		return Token::createSymbol(fcp, p, sym);
	}
	
	// Lexing for all other things, mostly symbols.
	Token Lexer::doOther() {
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		Location const* cLoc;
		
		Reader::Char c = fCurChar;
		Reader::Char c2;
		Token::Symbol sym = Token::kNoSymbol;
		
		// First the symbols consisting of just 1 char
		if (c.getChar().is(CodePoint::kColon)) {
			sym = Token::kColon;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kQuestionMark)) {
			sym = Token::kQuestionMark;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kSemicolon)) {
			sym = Token::kSemicolon;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kComma)) {
			sym = Token::kComma;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kLeftParenthesis)) {
			sym = Token::kOpenParen;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kRightParenthesis)) {
			sym = Token::kCloseParen;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kLeftSquareBracket)) {
			sym = Token::kOpenSquareParen;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kRightSquareBracket)) {
			sym = Token::kCloseSquareParen;
			getNextChar();
		}
		else if (c.getChar().is(CodePoint::kUnderscore)) {
			sym = Token::kUnderscore;
			getNextChar();
		}
		
		// Second, the symbols that consist of just 1 char + optional a '='
		else if (c.getChar().is(CodePoint::kAsterisk)) {
			return doCharWithOptionalAssignChar(
				Token::kAsterisk, Token::kMulAssign
			);
		}
		else if (c.getChar().is(CodePoint::kPercent)) {
			return doCharWithOptionalAssignChar(
				Token::kPercentage, Token::kModuloAssign
			);
		}
		else if (c.getChar().is(CodePoint::kAmpersand)) {
			return doCharWithOptionalAssignChar(Token::kAnd, Token::kAndAssign);
		}
		else if (c.getChar().is(CodePoint::kPipe)) {
			return doCharWithOptionalAssignChar(Token::kOr, Token::kOrAssign);
		}
		else if (c.getChar().is(CodePoint::kLessThan)) {
			return doCharWithOptionalAssignChar(
				Token::kLessThan, Token::kLessOrEqual
			);
		}
		else if (c.getChar().is(CodePoint::kGreaterThan)) {
			return doCharWithOptionalAssignChar(
				Token::kGreaterThan, Token::kGreaterOrEqual
			);
		}
		
		// Third, the special cases
		else if (c.getChar().is(CodePoint::kDot)) {
			getNextChar();
			if (fCurChar.getChar().is(CodePoint::kDot)) { // ..
				getNextChar();
				sym = Token::kDotDot;
			}
			else {
				sym = Token::kDot;
			}
		}
		else if (c.getChar().is(CodePoint::kPlus)) {
			getNextChar();
			if (fCurChar.getChar().is(CodePoint::kEqual)) { // +=
				getNextChar();
				sym = Token::kAddAssign;
			}
			else if (fCurChar.getChar().is(CodePoint::kPlus)) { // ++
				getNextChar();
				sym = Token::kPlusPlus;
			}
			else {
				sym = Token::kPlus;
			}
		}
		else if (c.getChar().is(CodePoint::kMinus)) {
			getNextChar();
			if (fCurChar.getChar().is(CodePoint::kEqual)) { // -=
				getNextChar();
				sym = Token::kSubtractAssign;
			}
			else if (fCurChar.getChar().is(CodePoint::kMinus)) { // --
				getNextChar();
				sym = Token::kMinusMinus;
			}
			else if (fCurChar.getChar().is(CodePoint::kGreaterThan)) { // ->
				getNextChar();
				sym = Token::kArrow;
			}
			else {
				sym = Token::kMinus;
			}
		}
		else if (c.getChar().is(CodePoint::kExclamationMark)) {
			getNextChar();
			if (fCurChar.getChar().is(CodePoint::kEqual)) { // !=
				getNextChar();
				sym = Token::kNotEqual;
			}
			else if (fCurChar.getChar().is(CodePoint::kDollar)) { // !$
				getNextChar();
				sym = Token::kPointerNotEqual;
			}
			else {
				sym = Token::kNot;
			}
		}
		else if (c.getChar().is(CodePoint::kEqual)) {
			getNextChar();
			if (fCurChar.getChar().is(CodePoint::kDollar)) { // =$
				getNextChar();
				sym = Token::kPointerEqual;
			}
			else {
				sym = Token::kEqual;
			}
		}
		else if (c.getChar().is(CodePoint::kSlash)) {
			cLoc = p;
			getNextChar();
			c2 = fCurChar;
			if (c2.getRelativeLinePos() == 2
					&& c2.getChar().is(CodePoint::kHashMark))
			{
				// block comment detected
				getNextChar();
				skipBlockComment();
				
				if (fReturnAllTokens) {
					return Token::createComment(
						fcp, cLoc, getFileCharPos() - fcp
					);
				}
				else {
					return Token(); // return an empty char
				}
			}
			else if (c2.getChar().is(CodePoint::kEqual)) {
				getNextChar();
				sym = Token::kDivideAssign;
			}
			else {
				sym = Token::kSlash;
			}
		}
		else {
			// unhandled character
			
			UTF8String errMsg;
			if (c.getChar().is(CodePoint::kControl)) {
				errMsg =
					UTF8STR(
						"(unprintable) illegal character encountered; "
						"it has integer value %1"
					).compose(c.getChar().toUnsigned());
			}
			else {
				errMsg =
					UTF8STR("\"%1\" is an illegal character").compose(
						c.getChar()
					);
			}
			error(errMsg, lookCharPos());
			return Token(); // return empty token
		}
		
		hyp_assert(sym != Token::kNoSymbol);
		
		// the symbol that was recognized
		return Token::createSymbol(fcp, p, sym);
	}
	
	Token Lexer::doLineComment() {
		//debug.line(UTF8STR("Lexer::doLineComment called"));
		
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		bool gotoEOL = false;
		
		getNextChar(); // skip the leading '#'
		if (fCurChar.getChar().is(CodePoint::kHashMark)) {
			// non-embeddable comment (goes to end-of-line, ignores further '#')
			getNextChar(); // skip the second '#'
			gotoEOL = true;
		}
		
		//debug.line(UTF8STR("Lexer::doLineComment : point A"));
		
		while (fCurChar.getChar().is(CodePoint::kIndentWhite)) {
			// skip spaces/tabs
			getNextChar();
		}
		
		//debug.line(UTF8STR("Lexer::doLineComment : point B"));
		
		bool todoMode = false;
		bool fixmeMode = false;
		
		Location const* pIn = lookCharPos();
		
		if (fInfo->getOptions().fgToDoWarnings
				&& fCurChar.getChar() == CodePoint('T'))
		{ // possible begin of TODO
			getNextChar();
			if (fCurChar.getChar() == CodePoint('O')) {
				getNextChar();
				if (fCurChar.getChar() == CodePoint('D')) {
					getNextChar();
					if (fCurChar.getChar() == CodePoint('O')) {
						getNextChar();
						todoMode = true;
					}
				}
			}
		}
		else if (fInfo->getOptions().fgFixMeWarnings
				&& fCurChar.getChar() == CodePoint('F'))
		{ // possible begin of FIXME
			getNextChar();
			if (fCurChar.getChar() == CodePoint('I')) {
				getNextChar();
				if (fCurChar.getChar() == CodePoint('X')) {
					getNextChar();
					if (fCurChar.getChar() == CodePoint('M')) {
						getNextChar();
						if (fCurChar.getChar() == CodePoint('E')) {
							getNextChar();
							fixmeMode = true;
						}
					}
				}
			}
		}
		
		//debug.line(UTF8STR("Lexer::doLineComment : point C"));
		
		if (todoMode || fixmeMode) { // TODO or FIXME comment
			while (fCurChar.getChar().is(CodePoint::kIndentWhite)) {
				// skip spaces/tabs
				getNextChar();
			}
			
			if (fCurChar.getChar() == CodePoint(':')) { // colon
				getNextChar();
				while (fCurChar.getChar().is(CodePoint::kIndentWhite)) {
					// skip spaces/tabs
					getNextChar();
				}
			}
			
			// use rest of comment for TODO/FIXME description
			UTF8StringBuilder warnDetail;
			while (!curCharIsOneOf(CodePoint('\n'), CodePoint('#'))) {
				// inside comment
				warnDetail += fCurChar.getChar();
				getNextChar();
			}
			
			UTF8String warn;
			if (todoMode) {
				warn = UTF8STR("TODO");
			}
			else /* if (fixmeMode) */ {
				warn = UTF8STR("FIXME");
			}
			
			UTF8String warnDetailStr = warnDetail.toString();
			if (!warnDetailStr.isEmpty()) {
				warn = warn + UTF8STR(": ") + warnDetailStr;
			}
			
			// end of comment reached
			fInfo->diag().warning(warn, pIn);
		}
		else { // regular comment
			while (!curCharIsOneOf(CodePoint('\n'), CodePoint('\0'), CodePoint('#'))) { // inside comment
				getNextChar();
			}
		}
		
		//debug.line(UTF8STR("Lexer::doLineComment : point X"));
		
		if (gotoEOL) {
			//debug.line(UTF8STR("Lexer::doLineComment : skipping until EOL"));
			skipRestOfLine();
			//debug.line(UTF8STR("Lexer::doLineComment : EOL found"));
		}
		else if (fCurChar.getChar().is(CodePoint::kHashMark)) {
			// we ended up at the artificial end of the line comment
			getNextChar(); // eat the '#'
			if (fCurChar.getChar().is(CodePoint::kHashMark)) { // another '#'
				// this was not the end of the comment, we ignore the entire line
				skipRestOfLine();
			}
		}
		
		//debug.line(UTF8STR("Lexer::doLineComment : point Y"));
		
		if (fReturnAllTokens) {
			return Token::createComment(fcp, p, getFileCharPos() - fcp);
		}
		
		// return empty token
		return Token();
	}
	
	Token Lexer::doNumber() {
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		
		UTF8StringBuilder sd;
		CodePoint look;
		bool isFloat = false;
		
		// Process digits
		while ((look = fCurChar.getChar()).is(CodePoint::kDigit)) {
			sd += look;
			getNextChar();
		}
		
		if (fCurChar.getChar().is(CodePoint::kDot)) { // a possible decimal separator
			//debug.line("possible decimal sep");
			getNextChar(); // eat the '.'
			
			if ((look = fCurChar.getChar()).is(CodePoint::kDigit)) {
				// yes, the dot indicates a floating point number
				isFloat = true;
				sd += CodePoint('.');
				do {
					sd += look;
					getNextChar();
				} while ((look = fCurChar.getChar()).is(CodePoint::kDigit));
				
				// TODO: handle exponent
			}
			else {
				//debug.line("was not a decimal sep");
				// no, the dot is not part of the number
				goBackOneChar(); // put the '.' back
				// we have a whole number
			}
		}
		
		if (fCurChar.getChar().is(CodePoint::kASCIILetter)) {
			error(UTF8STR("illegal attachment to number"), lookCharPos());
		}
		
		if (isFloat) {
			return Token::createFloatingPoint(fcp, p, sd.toString());
		}
		else {
			return Token::createWholeNumber(fcp, p, sd.toString());
		}
	}
	
	void Lexer::skipBlockComment() {
		bool endFound = false;
		HYPER_Common::CodePoint c;
		
		skipRestOfLine(); // goto end of line
		skipNewLines();
		
		do {
			c = fCurChar.getChar();
			
			if (c == CodePoint('/') || c == CodePoint('\\')) {
				getNextChar();
				if (fCurChar.getChar() == CodePoint('#')) {
					getNextChar();
					if (c == CodePoint('/')) {
						skipBlockComment();
					}
					else {
						endFound = true;
						skipRestOfLine();
					}
				}
				else {
					skipRestOfLine();
					skipNewLines();
				}
			}
			else if (c == CodePoint(0u)) {
				error(
					UTF8STR("end of file before end of block comment"),
					lookCharPos()
				);
			}
			else {
				skipRestOfLine();
				skipNewLines();
			}
		} while (! endFound);
	}
	
	Token Lexer::doString() {
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		
		getNextChar(); // eat the " (double quote)
		
		UTF8StringBuilder sd;
		HYPER_Common::CodePoint look;
		
		bool stop = false;
		
		do {
			look = fCurChar.getChar();
			if (look == CodePoint('"')) { // double quote ?
				getNextChar(); // goto next char
				
				look = fCurChar.getChar();
				
				if (look == CodePoint('"')) { // another double quote ?
					// it is a double quote within a double quoted string
					getNextChar();
					sd += look;
				}
				else {
					// don't touch this char
					stop = true; // end quote of double quoted string
				}
			}
			else if (look.is(CodePoint::kLf)) {
				error(UTF8STR("string literal must end before a newline"), p);
			}
			else if (look.is(CodePoint::kControl)) {
				error(
					UTF8STR("control characters are not allowed in a string "
						"literal"),
					p
				);
			}
			else {
				getNextChar();
				sd += look;
			}
		} while (!stop);
		
		return Token::createString(fcp, p, sd.toString());
	}
	
	Token Lexer::doChar() {
		Location const* p = lookCharPos();
		unsigned int fcp = getFileCharPos();
		
		getNextChar(); // eat the ' (quote)
		
		UTF8StringBuilder sd;
		HYPER_Common::CodePoint look = fCurChar.getChar();
		
		/*if (look == CodePoint('[')) { // hex code for character
			sd += look;
			getNextChar(); // eat the [
			
			for (unsigned int i = 0; i < 4; ++i) {
				look = fCurChar.getChar();
				
				if (look.is(CodePoint::kHexDigit)) {
					sd += look;
					getNextChar(); // eat current char
				}
				else if (look == CodePoint(']')) {
					error(
						UTF8STR("char literal code must contain 4 "
							"hexadecimal digits"),
						p
					);
				}
				else if (look.is(CodePoint::kControl)) {
					error(
						UTF8STR("control characters are not allowed in a "
							"char literal"),
						p
					);
				}
				else {
					error(
						UTF8STR(
							"character \"%1\" inside char literal is not "
							"a hexadecimal digit"
						).compose(look),
						p
					);
				}
			}
			
			if (fCurChar.getChar() == CodePoint(']')) {
				sd += ']';
				getNextChar(); // eat the ]
			}
			else {
				error(UTF8STR("char literal code does not end with \"]\""), p);
			}
		}
		else*/ if (look.is(CodePoint::kControl)) {
			error(
				UTF8STR("control characters are not allowed in a char literal"),
				p
			);
		}
		else {
			sd += look;
			getNextChar();
		}
		
		if (fCurChar.getChar() == CodePoint('\'')) {
			getNextChar(); // eat the closing ' (quote)
		}
		else {
			error(
				UTF8STR(
					"char literal does not properly end with a single quote"
				),
				p
			);
		}
		
		return Token::createChar(fcp, p, sd.toString());
	}
	
	// Eats all occuring characters before the next newline
	void Lexer::skipRestOfLine() {
		while (!fCurChar.getChar().is(CodePoint::kLf)) {
			getNextChar();
		}
	}
	
	// Eats all occuring newlines until the next non-newline char
	void Lexer::skipNewLines() {
		// Eat all Lf's
		// (The reader converted all kinds of newlines to Lf format)
		while (fCurChar.getChar().is(CodePoint::kLf)) {
			getNextChar();
		}
	}
	
	// Reads the next character
	void Lexer::getNextChar() {
		if (fCurChar.getChar() == CodePoint(0u)) {
			error(UTF8STR("unexpected end of file"), lookCharPos());
		}
		
		//debug.line("Lexer::getNextChar : asking next char...");
		fReader.readNextChar();
		fCurChar = fReader.lookChar();
		fCurCharLoc = 0; // location object is not constructed yet
		//debug.line(UTF8STR("current lexchar is now '%1'").compose(fCurChar.getChar()));
	}
	
	void Lexer::goBackOneChar() {
		//debug.line("Lexer: going back one char...");
		fReader.goBackOneChar();
		fCurChar = fReader.lookChar();
		//debug << "current lexchar is now '" << fCurChar.getChar() << "'\n";
		fCurCharLoc = 0; // location object no longer accurate
	}
	
	// Gets the location of the current char
	Location const* Lexer::lookCharPos() {
		if (fCurCharLoc == 0) {
			fCurCharLoc =
				Location::createSourceLoc(
					fFileID, fCurChar.getLineNo(), fCurChar.getLinePos()
				);
		}
		
		return fCurCharLoc;
	}
	
	bool
	Lexer::charIsOneOf(HYPER_Common::CodePoint X, HYPER_Common::CodePoint A,
		HYPER_Common::CodePoint B)
	{
		return (X == A) || (X == B);
	}
	
	bool
	Lexer::curCharIsOneOf(HYPER_Common::CodePoint A, HYPER_Common::CodePoint B)
	{
		HYPER_Common::CodePoint c = fCurChar.getChar();
		return (c == A) || (c == B);
	}
	
	bool
	Lexer::curCharIsOneOf(HYPER_Common::CodePoint A, HYPER_Common::CodePoint B,
		HYPER_Common::CodePoint C)
	{
		HYPER_Common::CodePoint c = fCurChar.getChar();
		return (c == A) || (c == B) || (c == C);
	}
	
	bool Lexer::curCharIsOneOf(HYPER_Common::CodePoint A,
		HYPER_Common::CodePoint B, HYPER_Common::CodePoint C,
		HYPER_Common::CodePoint D)
	{
		HYPER_Common::CodePoint c = fCurChar.getChar();
		return (c == A) || (c == B) || (c == C) || (c == D);
	}
	
	void Lexer::error(HYPER_Common::UTF8String msg, Location const* loc) {
		fInfo->diag().error(msg, loc);
		throw FatalParseError(loc);
	}
}
