/*!
	\file lexer.h
	\brief Lexer for the HYPER compiler.
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-02-05
// Last time modified:	2008-06-29

#ifndef DEF_LEXER_H
 #define DEF_LEXER_H
 
 #include "base/important_fwdecl.h"
 #include "base/location.h"
 
 #include "common/cxx_macros.h"
 #include "common/utf8string.h"
 
 #include "fatalparseerror.h"
 #include "reader.h"
 #include "token.h"
 
 #include <stack>

namespace HYPER_Compiler {
	
	class InvalidOperationException : public HYPER_Common::Exception {
		//
	};
	
	/// Lexical analyzer
	class Lexer {
	public:
		Lexer(Reader& input,
			FileID sourceFileID, BasicCompileInfo* info,
			bool returnAllTokens = false);
		
		void getToken();
		Token lookToken();
		Location const* lookPos();
		unsigned long getFileCharPos();
		unsigned long getTokenPos();
		
		void goBackOneToken();
		void goBackToTokenPos(int tokenPos);
		
		Keywords const& getKeywordsList() const;
		
		~Lexer();
	private:
		
		void error(HYPER_Common::UTF8String msg,
			Location const* loc) ATTR_NO_RETURN;
		
		Location const* lookCharPos();
		
		void skipRestOfLine();
		void skipNewLines();
		void doWhiteSpace();
		Token doLineComment();
		void skipBlockComment();
		Token doNewLine();
		Token doOther();
		Token doIdentOrKeyword();
		Token doNumber();
		Token doString();
		Token doChar();
		Token doCharWithOptionalAssignChar(Token::Symbol single,
			Token::Symbol withAssign);
		
		FileID fFileID;
		
		Reader::Char fCurChar;
		Location const* fCurCharLoc;
		
		// Returns true if X == A or X == B
		static bool charIsOneOf(HYPER_Common::CodePoint X,
			HYPER_Common::CodePoint A, HYPER_Common::CodePoint B);
		
		void getNextChar();
		bool curCharIsOneOf(HYPER_Common::CodePoint A,
			HYPER_Common::CodePoint B);
		bool curCharIsOneOf(HYPER_Common::CodePoint A,
			HYPER_Common::CodePoint B, HYPER_Common::CodePoint C);
		bool curCharIsOneOf(HYPER_Common::CodePoint A,
			HYPER_Common::CodePoint B, HYPER_Common::CodePoint C,
			HYPER_Common::CodePoint D);
		void goBackOneChar();
		
		Reader& fReader;
		bool fFileReadable;
		
		Token fCurToken;
		Token fLastGeneratedToken;
		bool fWhitesRead;
		
		bool fReturnAllTokens;
		
		std::stack<Token> fTokensRead;
		std::stack<Token> fTokensComing;
		int fTokenPos;
		
		BasicCompileInfo* fInfo;
	};
};
#endif
