/*!
	\file reader.cxx
	\brief Implements the Reader class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-02-20
// Last time modified:	2008-07-12

#include "reader.h"

#include "base/asserts.h"
//~ #include "base/debug.h"

#include "common/pullstream.h"

using HYPER_Common::CodePoint;

namespace HYPER_Compiler {
	
	// **** The FileChar class ****
	
	Reader::FileChar::FileChar()
	 : fC('\0'), fFilePos(0)
	{
		// no action
	}
	
	Reader::FileChar::FileChar(HYPER_Common::CodePoint c, unsigned long filePos)
	 : fC(c), fFilePos(filePos)
	{
		// no action
	}
	
	unsigned long Reader::FileChar::getFilePos() const {
		return fFilePos;
	}
	
	void Reader::FileChar::setChar(HYPER_Common::CodePoint c) {
		fC = c;
	}
	
	
	// **** The LineChar class ****
	
	Reader::LineChar::LineChar()
	 : FileChar(), fLinePos(1), fLineNo(1)
	{
		// no action
	}
	
	Reader::LineChar::LineChar(FileChar const& c, unsigned long linePos,
		unsigned long lineNo)
	 : FileChar(c), fLinePos(linePos), fLineNo(lineNo)
	{
		// no action
	}
	
	unsigned long Reader::LineChar::getLinePos() const {
		return fLinePos;
	}
	
	unsigned long Reader::LineChar::getLineNo() const {
		return fLineNo;
	}
	
	
	// **** The Char class ****
	
	Reader::Char::Char()
	 : LineChar(), fRelLinePos(1)
	{
		// no action
	}
	
	Reader::Char::Char(LineChar const& c, unsigned long relativeLinePos)
	 : LineChar(c), fRelLinePos(relativeLinePos)
	{
		// no action
	}
	
	unsigned long Reader::Char::getRelativeLinePos() const {
		return fRelLinePos;
	}
	
	void Reader::Char::setRelativeLinePos(unsigned long p) {
		fRelLinePos = p;
	}
	
	
	// **** The READER class ****
	
	// Constructor. Opens the file and reads first character.
	Reader::Reader(HYPER_Common::PullStream<HYPER_Common::CodePoint>& input,
		bool keepIndents)
	: fKeepIndents(keepIndents), fFileStream(input),
		fRelLinePos(1),
		fCurChar(), fPrevChar(),
		fLinePos(1), fLineNo(1),
		fEOFAfterLookAhead(false)
	{
		//~ debug.line(UTF8STR("Reader: initialized. Preparing first character..."));
		readNextChar(); // Read first character
		//~ debug.line(UTF8STR("Reader: first character ready."));
	}
	
	// Destructor.
	Reader::~Reader() {
		// no destruction yet
	}
	
	// Returns the current character
	Reader::Char const& Reader::lookChar() {
		return fCurChar;
	}
	
	Reader::Char const& Reader::prevChar() {
		hyp_assert(lookChar().getFilePos() >= 2);
		return fBackBuffer.top();
	}
	
	void Reader::goBackOneChar() {
		hyp_assert(lookChar().getFilePos() >= 2);
		fForwardQueue.push_front(fCurChar);
		fCurChar = fBackBuffer.top();
		fBackBuffer.pop();
	}
	
	// Reads the next character
	void Reader::readNextChar() {
		// Read next file character, but don't return leading or trailing whitespace.
		// (line-oriented)
		
		//debug.line(UTF8STR("Reader::readNextChar called"));
		
		if (!fForwardQueue.empty()) { // characters in forward buffer
			// use the next char from the buffer
			fBackBuffer.push(fCurChar);
			fCurChar = fForwardQueue.front();
			fForwardQueue.pop_front();
			//debug.line("got character from forward queue");
			return;
		}
		
		// read next char from file
		
		//debug.line("forward queue empty");
		std::deque<Char> trailingWhites;
		
		bool ignore;
		do {
			ignore = false;
			LineChar lc = fetchNextLC();
			CodePoint lcc = lc.getChar();
			
			//debug.line(UTF8STR("readNextChar : receiving char '%1'").compose(lc.getChar()));
			
			if (lcc == CodePoint(0u)) {
				// end of file reached
				//~ debug.line(UTF8STR("Reader::readNextChar : EOF reached"));
				trailingWhites.clear();
				fForwardQueue.push_back(Char(lc, fRelLinePos));
			}
			else if (lcc.is(CodePoint::kLf)) {
				trailingWhites.clear();
				fForwardQueue.push_back(Char(lc, fRelLinePos));
				fRelLinePos = 1;
			}
			else if (lcc.is(CodePoint::kIndentWhite)) {
				//debug << "white" << newl;
				// at begin of line?
				if (fRelLinePos == 1) {
					// this is indentation...
					if (!fKeepIndents) {
						// ignore this character
						ignore = true;
					}
					else {
						fForwardQueue.push_back(Char(lc, fRelLinePos));
					}
				}
				else {
					// This could be trailing whitespace, so put the whites
					// into a queue and continue reading to find out where we are.
					trailingWhites.push_back(Char(lc, fRelLinePos));
					ignore = true;
				}
			}
			else { // non-white character
				//~ debug.line(UTF8STR("Reader::readNextChar : non-white char read"));
				// trailing whites were not really trailing, add them
				while (!trailingWhites.empty()) {
					Char w = trailingWhites.front();
					w.setRelativeLinePos(fRelLinePos);
					fForwardQueue.push_back(w);
					trailingWhites.pop_front();
					++fRelLinePos;
				}
				fForwardQueue.push_back(Char(lc, fRelLinePos));
				++fRelLinePos;
			}
		} while (ignore);
		
		fBackBuffer.push(fCurChar);
		fCurChar = fForwardQueue.front();
		fForwardQueue.pop_front();
	}
	
	void Reader::didNewLine() {
		++fLineNo;
		fLinePos = 1;
	}
	
	Reader::LineChar Reader::fetchNextLC() {
		// Read next file character, but convert every CrLf and Cr to a Lf
		// and skip zero chars.
		
		//debug.line(UTF8STR("Reader::fetchNextLC called"));
		
		LineChar lc;
		
		// ensure buffer filled
		readAhead();
		
		// still no lookahead? -> end of file reached
		if (fLookAhead.empty()) {
			// set EOF character ('\0')
			lc =
				LineChar(
					FileChar(CodePoint(0u), fFileStream.streamPosition()),
					fLinePos, fLineNo
				);
		}
		else {
			lc = LineChar(fLookAhead.front(), fLinePos, fLineNo);
			fLookAhead.pop_front();
			++fLinePos;
		}
		
		CodePoint c = lc.getChar();
		if (c.is(CodePoint::kCr)) {
			// Cr and CrLf become Lf
			
			// Cr followed by Lf  (so together a CrLf) ?
			if (fLookAhead.front().getChar().is(CodePoint::kLf)) {
				// CrLf becomes Lf
				fLookAhead.pop_front();
			}
			
			lc.setChar(CodePoint::linefeed());
			didNewLine();
		}
		else if (c.is(CodePoint::kLf)) {
			didNewLine();
		}
		
		return lc;
	}
	
	bool Reader::isFileOpened() {
		return !fLookAhead.empty() || fFileStream.streamOK();
	}
	
	void Reader::readAhead() {
		static const unsigned kReadBufferMin = 5;
		static const unsigned kReadBufferMax = 250;
		
		if (fLookAhead.size() <= kReadBufferMin && fFileStream.streamOK()) {
			// buffer getting empty, so fill it again
			
			while (fLookAhead.size() < kReadBufferMax) {
				CodePoint c;
				
				unsigned long pos = fFileStream.streamPosition();
				if (fFileStream.look(c)) {
					fFileStream.pullNext();
				}
				
				bool isEOF = !fFileStream.streamOK(); // FIXME
				
				if (isEOF) {
					// the end-of-file is stored as a zero char
					fLookAhead.push_back(FileChar(CodePoint(0u), pos));
					return;
				}
				else if (c == CodePoint(0u)) { // zero char read
					// do nothing, zero chars are filtered
				}
				else {
					//debug.line(UTF8STR("readAhead : read char '%1'").compose(c));
					fLookAhead.push_back(FileChar(c, pos));
				}
			}
		}
	}
}
