/**
	\file reader.h
	\author Kevin André
	\brief Contains the Reader class
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2004-02-20
// Last time modified:	2008-07-12

#ifndef DEF_READER_H
 #define DEF_READER_H
 
 #include "common/streams_fwdecl.h"
 #include "common/unicode.h"
 
 #include <deque>
 #include <stack>

namespace HYPER_Compiler {
	
	/**
		\brief Reads a file, and delivers a file-format-independent char-by-char
		representation of its content.
		
		All kinds of newlines are translated into '\n' characters. The end of file
		is represented by a '\0' char. '\0' chars in the file are not returned. Leading
		whitespace (indentation) is stripped, and trailing whitespace as well.
	*/
	class Reader {
	public:
		/// Represents a character from a file
		class FileChar {
		public:
			FileChar();
			FileChar(HYPER_Common::CodePoint c, unsigned long filePos);
			
			HYPER_Common::CodePoint getChar() const {
				return fC;
			}
			
			unsigned long getFilePos() const;
			
			void setChar(HYPER_Common::CodePoint c);
			
			bool operator==(FileChar& c) const {
				return fC == c.fC;
			}
			
			bool operator!=(FileChar& c) const {
				return fC != c.fC;
			}
			
			bool is(HYPER_Common::CodePoint::Class c) const {
				return fC.is(c);
			}
			
		protected:
			HYPER_Common::CodePoint fC;
			unsigned long fFilePos;
		};
		
		/// Represents a character that has an line+column position
		class LineChar : public FileChar {
		public:
			LineChar();
			LineChar(FileChar const& c, unsigned long linePos,
				unsigned long lineNo);
			
			unsigned long getLinePos() const;
			unsigned long getLineNo() const;
			
		protected:
			unsigned long fLinePos;
			unsigned long fLineNo;
		};
		
		class Char : public LineChar {
		public:
			Char();
			Char(LineChar const& c, unsigned long relativeLinePos);
			unsigned long getRelativeLinePos() const;
			
			void setRelativeLinePos(unsigned long p);
			
		protected:
			unsigned long fRelLinePos;
		};
		
		/**
			\brief Creates a new Reader
			\param input  The input stream
			\param keepIndents Do not strip indentation whitespace
		*/
		Reader(HYPER_Common::PullStream<HYPER_Common::CodePoint>& input,
			bool keepIndents);
		virtual ~Reader();
		
		// FIXME: replace this by a better way to get the 'stream state'
		bool isFileOpened();
		
		void readNextChar();
		Char const& lookChar();
		Char const& prevChar();
		void goBackOneChar();
		
	private:
		
		LineChar fetchNextLC();
		void didNewLine();
		
		void readAhead();
		
		bool fKeepIndents;
		
		HYPER_Common::PullStream<HYPER_Common::CodePoint>& fFileStream;
		
		unsigned long fRelLinePos;
		Char fCurChar;
		Char fPrevChar;
		
		std::stack<Char> fBackBuffer;
		std::deque<Char> fForwardQueue;
		
		unsigned long fLinePos;
		unsigned long fLineNo;
		
		std::deque<FileChar> fLookAhead;
		bool fEOFAfterLookAhead;
	};
	
	inline bool
	operator==(Reader::FileChar const& rc, HYPER_Common::CodePoint c) {
		return rc.getChar() == c;
	}
	
	inline bool
	operator==(HYPER_Common::CodePoint c, Reader::FileChar const& rc) {
		return rc.getChar() == c;
	}
	
	inline bool
	operator!=(Reader::FileChar const& rc, HYPER_Common::CodePoint c) {
		return rc.getChar() != c;
	}
	
	inline bool
	operator!=(HYPER_Common::CodePoint c, Reader::FileChar const& rc) {
		return rc.getChar() != c;
	}
}
#endif
