/*!
	\file token.cxx
	\brief Implements the Token class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-12-04
// Last time modified:	2008-07-11

#include "token.h"

#include "base/asserts.h"
#include "base/location.h"

using HYPER_Common::CodePoint;
using HYPER_Common::UTF8String;

namespace HYPER_Compiler {
	
	// **** The TOKEN class ****
	
	HYPER_Common::UTF8String Token::toString(Symbol sym) {
		switch (sym) {
		case kNoSymbol: return UTF8STR("");
		case kDot: return UTF8STR(".");
		case kDotDot: return UTF8STR("..");
		case kComma: return UTF8STR(",");
		case kOpenParen: return UTF8STR("(");
		case kCloseParen: return UTF8STR(")");
		case kOpenSquareParen: return UTF8STR("[");
		case kCloseSquareParen: return UTF8STR("]");
		case kAsterisk: return UTF8STR("*");
		case kPlus: return UTF8STR("+");
		case kMinus: return UTF8STR("-");
		case kSlash: return UTF8STR("/");
		case kPercentage: return UTF8STR("%");
		case kAnd: return UTF8STR("&");
		case kOr: return UTF8STR("|");
		case kColon: return UTF8STR(":");
		case kSemicolon: return UTF8STR(";");
		case kEqual: return UTF8STR("=");
		case kLessThan: return UTF8STR("<");
		case kLessOrEqual: return UTF8STR("<=");
		case kGreaterThan: return UTF8STR(">");
		case kGreaterOrEqual: return UTF8STR(">=");
		case kNotEqual: return UTF8STR("!=");
		case kNot: return UTF8STR("!");
		case kQuestionMark: return UTF8STR("?");
		case kUnderscore: return UTF8STR("_");
		case kAddAssign: return UTF8STR("+=");
		case kSubtractAssign: return UTF8STR("-=");
		case kMulAssign: return UTF8STR("*=");
		case kDivideAssign: return UTF8STR("/=");
		case kModuloAssign: return UTF8STR("%=");
		case kAndAssign: return UTF8STR("&=");
		case kOrAssign: return UTF8STR("|=");
		case kPlusPlus: return UTF8STR("++");
		case kMinusMinus: return UTF8STR("--");
		case kArrow: return UTF8STR("->");
		case kPointerEqual: return UTF8STR("=$");
		case kPointerNotEqual: return UTF8STR("!$");
		default:
			hyp_assert_unreachable();
		}
	}
	
	// Constructor
	Token::Token()
	 : fClass(kEmpty), fDescr(UTF8STR("")), fKeyword(Keywords::kNoKeyword),
		fSymbol(kNoSymbol), fPos(0), fLength(0), fFileCharPos(0),
		fAfterWhites(false)
	{
		// No extra initialisation
	}
	
	Token::Token(const Token& t)
	 : fClass(t.fClass), fDescr(t.fDescr), fKeyword(t.fKeyword),
		fSymbol(t.fSymbol), fPos(t.fPos),
		fLength(t.fLength), fFileCharPos(t.fFileCharPos),
		fAfterWhites(t.fAfterWhites)
	{
		// no action here
	}
	
	Token::Token(Class tclass, HYPER_Common::UTF8String descr,
		Keywords::Keyword kw,
		Symbol sym, unsigned int fileCharPos, Location const* p,
		unsigned int length) 
	 : fClass(tclass), fDescr(descr), fKeyword(kw), fSymbol(sym),
		fPos(p), fLength(length), fFileCharPos(fileCharPos),
		fAfterWhites(false)
	{
		// No extra initialisation
	}
	
	Token::~Token() {
		// no action
	}
	
	Token::Class Token::getClass() const {
		return fClass;
	}
	
	void Token::setPrecededByWhitespace(bool p) {
		fAfterWhites = p;
	}
	
	bool Token::isPrecededByWhitespace() const {
		return fAfterWhites;
	}
	
	HYPER_Common::UTF8String Token::getDescription() const {
		return fDescr;
	}
	
	Keywords::Keyword Token::getKeyword() const {
		return fKeyword;
	}
	
	Token::Symbol Token::getSymbol() const {
		return fSymbol;
	}
	
	const Location* Token::getPosition() const {
		return fPos;
	}
	
	unsigned long Token::getLength() const {
 		return fLength;
 	}
	
	unsigned long Token::getFileCharPos() const {
		return fFileCharPos;
	}
	
	bool Token::isKeyword() const {
		return (fClass == kKeyword);
	}
	
	bool Token::isKeyword(Keywords::Keyword k) const {
		return fKeyword == k;
	}
	
	bool Token::isNewLine() const {
		return (fClass == kNewLine);
	}
	
	bool Token::isEndOfFile() const {
		return (fClass == kEndOfFile);
	}
	
	bool Token::isIdentifier() const {
		return (fClass == kIdentifier);
	}
	
	bool Token::isWholeNumber() const {
		return (fClass == kWholeNumber);
	}
	
	bool Token::isFloatingPoint() const {
		return (fClass == kFloatingPoint);
	}
	
	bool Token::isNumber() const {
		return (fClass == kWholeNumber || fClass == kFloatingPoint);
	}
	
	bool Token::isSymbol() const {
		return (fClass == kSymbol);
	}
	
	bool Token::isSymbol(Symbol sym) const {
		return (fSymbol == sym);
	}
	
	bool Token::isChar() const {
		return (fClass == kChar);
	}
	
	bool Token::isString() const {
		return (fClass == kString);
	}
	
	bool Token::isEmpty() const {
		return (fClass == kEmpty);
	}
	
	HYPER_Common::UTF8String Token::getDescriptiveString() const {
		switch (fClass) {
		case kNewLine:
			return UTF8STR("newline");
		case kEndOfFile:
			return UTF8STR("end of file");
		case kEmpty:
			return UTF8STR("nothing");
		case kKeyword:
			return UTF8STR("%1 keyword").compose(quote(fDescr));
		case kIdentifier:
			return UTF8STR("%1 identifier").compose(quote(fDescr));
		case kWholeNumber:
		case kFloatingPoint:
			return UTF8STR("number %1").compose(quote(fDescr));
		case kSymbol:
			return UTF8STR("%1 symbol").compose(quote(toString(fSymbol)));
		case kChar:
			return UTF8STR("%1 char").compose(quote(fDescr));
		case kString:
			return UTF8STR("%1 string").compose(quote(fDescr));
		case kComment:
			return UTF8STR("comment");
		default:
			hyp_assert_unreachable();
		}
	}
	
	Token Token::createNewLine(unsigned int fileCharPos, Location const* p) {
		return Token(
			kNewLine, UTF8STR(""), Keywords::kNoKeyword, kNoSymbol, fileCharPos,
			p, 0
		);
	}
	
	Token
	Token::createKeyword(unsigned int fileCharPos, Location const* p,
		Keywords::Keyword k, HYPER_Common::UTF8String keyword)
	{
		return Token(
			kKeyword, keyword, k, kNoSymbol, fileCharPos, p, keyword.length()
		);
	}
	
	Token Token::createEndOfFile(unsigned int fileCharPos, Location const* p) {
		return Token(
			kEndOfFile, UTF8STR(""), Keywords::kNoKeyword, kNoSymbol,
			fileCharPos, p, 0
		);
	}
	
	Token
	Token::createIdentifier(unsigned int fileCharPos, Location const* p,
		HYPER_Common::UTF8String ident)
	{
		return Token(
			kIdentifier, ident, Keywords::kNoKeyword, kNoSymbol,
			fileCharPos, p, ident.length()
		);
	}
	
	Token Token::createWholeNumber(unsigned int fileCharPos, Location const* p,
		HYPER_Common::UTF8String num)
	{
		return Token(
			kWholeNumber, num, Keywords::kNoKeyword, kNoSymbol,
			fileCharPos, p, num.length()
		);
	}
	
	Token
	Token::createSymbol(unsigned int fileCharPos, Location const* p, Symbol sym)
	{
		UTF8String str = toString(sym);
		return Token(
			kSymbol, str, Keywords::kNoKeyword, sym, fileCharPos, p,
			str.length()
		);
	}
	
	Token
	Token::createFloatingPoint(unsigned int fileCharPos, Location const* p,
		HYPER_Common::UTF8String floatingp)
	{
		return Token(
			kFloatingPoint, floatingp, Keywords::kNoKeyword, kNoSymbol,
			fileCharPos, p, floatingp.length()
		);
	}
	
	Token
	Token::createChar(unsigned int fileCharPos, Location const* p,
		HYPER_Common::UTF8String ch)
	{
		return Token(
			kChar, ch, Keywords::kNoKeyword, kNoSymbol, fileCharPos,
			p, ch.length() + 2
		);
	};
	
	Token
	Token::createString(unsigned int fileCharPos, Location const* p,
		HYPER_Common::UTF8String str)
	{
		unsigned int length = 2;
		for (UTF8String::ConstIterator it = str.begin(); it != str.end(); ++it)
		{
			if ((*it) == CodePoint('"')) { // double quote?
				length += 2; // double quote is escaped
			}
			else {
				length ++;
			}
		}
		
		return Token(
			kString, str, Keywords::kNoKeyword, kNoSymbol, fileCharPos,
			p, length
		);
	};
	
	Token
	Token::createComment(unsigned int fileCharPos, Location const* begin,
		unsigned int length)
	{
		return Token(
			kComment, UTF8STR(""), Keywords::kNoKeyword, kNoSymbol, fileCharPos,
			begin, length
		);
	}
}
