/*!
	\file token.h
	\brief Contains the Token class
	\author Kevin André
	
	Copyright (C) 2004, 2005, 2006, 2007, 2008  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2006-12-04
// Last time modified:	2008-07-11

#ifndef DEF_TOKEN_H
 #define DEF_TOKEN_H
 
 #include "keywords.h"
 
namespace HYPER_Compiler {
	
	class Location;
	
	/// Represents a token as read by the Lexer
	class Token {
	public:
		enum Class {
			kKeyword,
			kNewLine,
			kEndOfFile,
			kIdentifier,
			kWholeNumber,
			kFloatingPoint,
			kSymbol,
			kChar,
			kString,
			kEmpty,
			kComment
		};
		
		/// Represents the kind of a symbol token
		enum Symbol {
			kNoSymbol,
			kDot, // .
			kDotDot, // ..
			kComma, // ,
			kOpenParen, // (
			kCloseParen, // )
			kOpenSquareParen, // [
			kCloseSquareParen, // ]
			kAsterisk, // *
			kPlus, // +
			kMinus, // -
			kSlash, // /
			kPercentage, // %
			kAnd, // &
			kOr, // |
			kColon, // :
			kSemicolon, // ;
			kEqual, // =
			kLessThan, // <
			kLessOrEqual, // <=
			kGreaterOrEqual, // >=
			kGreaterThan, // >
			kNotEqual, // !=
			kNot, // !
			kQuestionMark, // ?
			kUnderscore, // _
			kAddAssign, // +=
			kSubtractAssign, // -=
			kMulAssign, // *=
			kDivideAssign, // /=
			kModuloAssign, // %=
			kAndAssign, // &=
			kOrAssign, // |=
			kPlusPlus, // ++
			kMinusMinus, // --
			kArrow, // ->
			kPointerEqual, // =$
			kPointerNotEqual, // !$
		};
		
		static HYPER_Common::UTF8String toString(Symbol sym);
		
		Token(); // Constructor
		
		Token(const Token& t);
		
		static Token
		createNewLine(unsigned int fileCharPos, Location const* p);
		
		static Token
		createKeyword(unsigned int fileCharPos, Location const* p,
			Keywords::Keyword k, HYPER_Common::UTF8String keyword);
		
		static Token
		createEndOfFile(unsigned int fileCharPos, Location const* p);
		
		static Token
		createIdentifier(unsigned int fileCharPos, Location const* p,
			HYPER_Common::UTF8String ident);
		
		static Token
		createWholeNumber(unsigned int fileCharPos, Location const* p,
			HYPER_Common::UTF8String num);
		
		static Token
		createSymbol(unsigned int fileCharPos, Location const* p, Symbol sym);
		
		static Token
		createFloatingPoint(unsigned int fileCharPos, Location const* p,
			HYPER_Common::UTF8String floatingp);
		
		static Token
		createChar(unsigned int fileCharPos, Location const* p,
			HYPER_Common::UTF8String ch);
		
		static Token
		createString(unsigned int fileCharPos, Location const* p,
			HYPER_Common::UTF8String str);
		
		static Token
		createComment(unsigned int fileCharPos, Location const* begin,
			unsigned int length);
		
		Class getClass() const;
		
		void setPrecededByWhitespace(bool p = true);
		bool isPrecededByWhitespace() const;
		
		/// Get the string version of the token
		HYPER_Common::UTF8String getDescription() const;
		
		Keywords::Keyword getKeyword() const;
		Symbol getSymbol() const;
		Location const* getPosition() const;
		
		/// Get the length of the token as number of characters in the sourcefile
		unsigned long getLength() const;
		
		unsigned long getFileCharPos() const;
		
		bool isKeyword() const;
		bool isKeyword(Keywords::Keyword k) const;
		bool isNewLine() const;
		bool isEndOfFile() const;
		bool isIdentifier() const;
		bool isWholeNumber() const;
		bool isFloatingPoint() const;
		bool isNumber() const;
		bool isSymbol() const;
		bool isSymbol(Symbol sym) const;
		bool isChar() const;
		bool isString() const;
		bool isEmpty() const;
		bool isNumericLiteral() const {
			return isWholeNumber() || isFloatingPoint();
		}
		bool isLiteral() const {
			return isNumericLiteral() || isChar() || isString();
		}
		
		HYPER_Common::UTF8String getDescriptiveString() const;
		
		~Token();
		
	private:
		Token(Class tclass, HYPER_Common::UTF8String descr,
				Keywords::Keyword kw, Symbol sym, unsigned int fileCharPos,
				Location const* p, unsigned int length);
		
		Class fClass;
		HYPER_Common::UTF8String fDescr;
		Keywords::Keyword fKeyword;
		Symbol fSymbol;
		Location const* fPos;
		unsigned int fLength;
		unsigned int fFileCharPos;
		bool fAfterWhites;
	};
}
#endif
