#! /usr/bin/env sh
#
# **************
# Tester for the compiler's command line behaviour
#    (not to be used directly, called from CTest)
# **************
#
# Copyright (C) 2008, 2009  Kevin André
#
#	This file is part of the Hyper Compiler, a compiler for the Hyper
#	programming language.
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License along
#	with this program; see the file GPLv3_license.txt.  If not, see
#	<http://www.gnu.org/licenses/>.
#


# --- script background ---

cd `dirname $0`
COMPILER="$1"
TESTNAME="$2"

if [ x"$1" = x ] ; then
	echo "ERROR: this executable should not be run directly."
	exit 3
fi

if [ ! -x "$COMPILER" ] ; then
	# not found, not a file, or not executable
	echo "COMPILER EXECUTABLE NOT FOUND -- cannot run commandline test"
	exit 4
fi


# --- set up ---

EXIT_CODE=0   # start with a successfull exit code

TESTFILE_EMPTY=./.cmdlinetest_empty.hyper
if [ ! -f $TESTFILE_EMPTY ] ; then
	echo -n "" > $TESTFILE_EMPTY
fi

TESTFILE_A=./.cmdlinetest_A.hyp
if [ ! -f $TESTFILE_A ] ; then
	echo "static class SC" > $TESTFILE_A
	echo "  procedure main()" >> $TESTFILE_A
	echo "    #TODO : implement this program" >> $TESTFILE_A
	echo "  end procedure main # FIXME: shorten this" >> $TESTFILE_A
	echo "end class SC" >> $TESTFILE_A
fi


# --- utilities ---

ok() {
	# no-op
	return
}

fail() {
	echo "FAIL: $1"
	EXIT_CODE=1
}

runcompiler() {
	"$COMPILER" "$@" >.commandoutput 2>&1
}

greptest() {
	local GREPCOUNT=`grep -c "$1" commandoutput`
	if (( $GREPCOUNT != $2 )) ; then
		fail "$3"": wrong command output"
	fi
}

# --- test ---

if [ "$TESTNAME" = plain ] ; then
	if runcompiler ; then
		fail "no arguments must generate non-zero exit code"
	else
		greptest "no input file" 1 "no arguments"
	fi
elif [ "$TESTNAME" = version ] ; then
	if runcompiler --version ; then
		greptest "Copyright" 1 "version output"
		greptest "GNU" 1 "version output"
		greptest "General" 1 "version output"
		greptest "Public" 1 "version output"
		greptest "License" 1 "version output"
		greptest "Report bugs to" 1 "version output"
	else
		fail "version output must generate zero exit code"
	fi
elif [ "$TESTNAME" = help ] ; then
	if runcompiler --help ; then
		greptest "Usage:" 1 "help output"
		greptest "Report bugs to" 1 "help output"
	else
		fail "help output must generate zero exit code"
	fi
elif [ "$TESTNAME" = noext ] ; then
	if runcompiler foobar ; then
		fail "input file without extension must generate non-zero exit code"
	else
		greptest "input file" 1 "input file without extension"
		greptest "Hyper file" 1 "input file without extension"
	fi
elif [ "$TESTNAME" = nonexist1 ] ; then
	if runcompiler nonexistingfile.hyp ; then
		fail "non-existing input file must generate non-zero exit code"
	else
		greptest "could not read" 1 "non-existing input file"
	fi
elif [ "$TESTNAME" = nonexist2 ] ; then
	if runcompiler nonexistingfile.hyper ; then
		fail "non-existing input file must generate non-zero exit code"
	else
		greptest "could not read" 1 "non-existing input file"
	fi
elif [ "$TESTNAME" = empty1 ] ; then
	if runcompiler $TESTFILE_EMPTY ; then
		fail "empty file should fail without '-c'"
	else
		greptest "1 error" 1 "empty file compile"
	fi
elif [ "$TESTNAME" = empty2 ] ; then
	if runcompiler -c $TESTFILE_EMPTY ; then
		ok
	else
		fail "empty file should be accepted with '-c'"
	fi
elif [ "$TESTNAME" = fileAplain ] ; then
	if runcompiler $TESTFILE_A ; then
		greptest "2 warnings" 1 "testfile A (plain compile)"
		greptest "warning:" 2 "testfile A (plain compile)"
		greptest "TODO:" 1 "testfile A (plain compile)"
		greptest "implement this program" 1 "testfile A (plain compile)"
		greptest "FIXME:" 1 "testfile A (plain compile)"
		greptest "shorten this" 1 "testfile A (plain compile)"
	else
		fail "testfile A should compile successfully"
	fi
elif [ "$TESTNAME" = fileAnofixme ] ; then
	if runcompiler -W-no-fixme $TESTFILE_A ; then
		greptest "1 warning" 1 "testfile A (-W-no-fixme)"
		greptest "warning:" 1 "testfile A (-W-no-fixme)"
		greptest "TODO:" 1 "testfile A (-W-no-fixme)"
		greptest "implement this program" 1 "testfile A (-W-no-fixme)"
		greptest "FIXME:" 0 "testfile A (-W-no-fixme)"
		greptest "shorten this" 0 "testfile A (-W-no-fixme)"
	else
		fail "testfile A should compile successfully (-W-no-fixme)"
	fi
elif [ "$TESTNAME" = fileAnotodo ] ; then
	if runcompiler -W-no-todo $TESTFILE_A ; then
		greptest "1 warning" 1 "testfile A (-W-no-todo)"
		greptest "warning:" 1 "testfile A (-W-no-todo)"
		greptest "TODO:" 0 "testfile A (-W-no-todo)"
		greptest "implement this program" 0 "testfile A (-W-no-todo)"
		greptest "FIXME:" 1 "testfile A (-W-no-todo)"
		greptest "shorten this" 1 "testfile A (-W-no-todo)"
	else
		fail "testfile A should compile successfully (-W-no-todo)"
	fi
else
	# if we get here, an invalid testname was specified
	fail "INVALID TESTNAME SPECIFIED: NO SUCH TEST!"
fi

# TODO : add more tests for other commandline usages

exit $EXIT_CODE
