/**
	\file testing/regtester.cxx
	\brief Regression tests the compiler with a given Hyper source file
	\author Kevin André
	
	Copyright (C) 2005, 2006, 2007, 2008, 2009  Kevin André
	
	This file is part of the Hyper Compiler, a compiler for the Hyper
	programming language.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License along
	with this program; see the file GPLv3_license.txt.  If not, see
	<http://www.gnu.org/licenses/>.
	
*/
// Date created:		2005-06-18
// Last time modified:	2008-06-26

#include "base/compilationfailederror.h"
//~ #include "base/debug.h"
#include "base/errormanager.h"
#include "base/internalcompilererror.h"
#include "base/location.h"

#include "common/console.h"
#include "common/filepath.h"
#include "common/filesystem.h"
#include "common/streams_fwdecl.h"
#include "common/utf8decoder.h"
#include "common/utf8stdoutstream.h"
#include "common/utf8textflowstream.h"

#include "compiler/compiler.h"

#include "lexer/fileexceptions.h"
#include "lexer/reader.h"

#include <map>

using namespace HYPER_Compiler;

using HYPER_Common::CodePoint;
using HYPER_Common::Console;
using HYPER_Common::FilePath;
using HYPER_Common::FileSystem;
using HYPER_Common::InvalidPathStringException;
using HYPER_Common::PullStream;
using HYPER_Common::UTF8Byte;
using HYPER_Common::UTF8Decoder;
using HYPER_Common::UTF8PushStream;
using HYPER_Common::UTF8String;
using HYPER_Common::UTF8TextFlowStream;

enum ParseMode {
	kNormal = 1,
	kSharp_Read,
	kSharpE_Read,
	kSharpER_Read,
	kSharpERR_Read,
	kSharpERRO_Read,
	kERROR_Mode,
	kERRORS_Mode,
	kSharpW_Read,
	kSharpWA_Read,
	kSharpWAR_Read,
	kSharpWARN_Read,
	kSharpWARNI_Read,
	kSharpWARNIN_Read,
	kWARNING_Mode,
	kSharpC_Read,
	kSharpCO_Read,
	kSharpCOM_Read,
	kSharpCOMP_Read,
	kSharpCOMPI_Read,
	kSharpCOMPIL_Read,
	kSharpCOMPILE_Read,
	kCompileSpecMode,
	kCompileSpecMode_O,
	kCompileSpecMode_ON,
	kCompileSpecMode_ONL,
	kNumber_Mode,
	kNumberDone_Mode,
	kLineEnd_Mode
};

unsigned test(const FilePath& filename);

int main(int argc, char *argv[]) {
	if (argc != 2) {
		return 2;
	}
	
	int t;
	
	FilePath* filename = 0;
	UTF8PushStream& stdErr = Console::getStdErr();
	try {
		filename =
			FileSystem::pathFromString(UTF8String::fromCString(argv[1]));
		
		//~ FilePath const* workingDir = FileSystem::getWorkingDirectory();
		//~ filename->makeAbsolute(*workingDir);
		
		std::list<UTF8String> hyperExts;
		hyperExts.push_back(UTF8STR("hyp"));
		hyperExts.push_back(UTF8STR("hyper"));
		
		if (!filename->hasExtension(hyperExts))
		{
			stdErr.line(UTF8STR("tester: input file must be a Hyper file"));
			return 2;
		}
		
		t = (test(*filename) > 0) ? 1 : 0;
	}
	catch (InvalidPathStringException& ipse) {
		stdErr.line(
			UTF8STR("tester: invalid filename %1").compose(
				doubleQuote(ipse.getString())
			)
		);
		t = 2;
	}
	catch (FileNotReadableException& fnre) { // error opening file
		stdErr.line(
			UTF8STR("tester: could not read the file %1").compose(
				doubleQuote(filename->toString())
			)
		);
		t = 2;
	}
	catch (InternalCompilerError& ice) { // internal compiler problem
		stdErr.line(UTF8STR("FILE : ") + filename->toString());
		stdErr.line(UTF8STR("  ! INTERNAL COMPILER ERROR !"));
		t = 1;
	}
	
	return t;
}

class TestState {
public:
	TestState(FilePath const& file, bool errorCheckingInterrupted,
		unsigned lastCompilerMessageLine)
	 : fOut(Console::getStdOut()), fFile(file), fMessagesPrinted(false),
		fProblems(0), fPrintedAbort(false), fAborted(errorCheckingInterrupted),
		fLastCompilerLine(lastCompilerMessageLine)
	{
		//
	}
	
	~TestState() {
		if (fProblems > 0 && fAborted && !fPrintedAbort) {
			printAbort();
		}
	}
	
	unsigned getProblems() {
		return fProblems;
	}
	
	void printProblem(unsigned line, UTF8String msg) {
		if (!fMessagesPrinted) {
			fOut.line(UTF8STR("FILE : ") + fFile.toString());
			fMessagesPrinted = true;
		}
		
		if (fAborted && !fPrintedAbort && line > fLastCompilerLine) {
			printAbort();
		}
		
		fOut.line(UTF8STR("  line %1 : %2").compose(line, msg));
		fProblems++;
	}
	
private:
	void printAbort() {
		fOut.line(UTF8STR("  <error checking aborted>"));
		fPrintedAbort = true;
	}
	
	UTF8PushStream& fOut;
	FilePath const& fFile;
	bool fMessagesPrinted;
	unsigned fProblems;
	bool fPrintedAbort;
	bool fAborted;
	unsigned fLastCompilerLine;
};

unsigned max(unsigned a, unsigned b) {
	return a >= b ? a : b;
}

UTF8String
getFoundVersusExpected(unsigned found, unsigned expected)
{
	CodePoint compare = (found > expected) ? CodePoint('>') : CodePoint('<');
	
	return UTF8STR("found %1 expected ::: %2 %1 %3").compose(
		compare, found, expected
	);
}

unsigned test(const FilePath& filename) {
	
	// PHASE 1 : read the specified file && parse it to look for error expectations
	
	PullStream<UTF8Byte>* input = FileSystem::openInputFile(filename);
	if (!input->streamOK()) {
		throw new FileNotReadableException();
	}
	
	UTF8Decoder decoder;
	PullStream<CodePoint>* decodeStream = decoder.puller(*input);
	Reader reader(*decodeStream, false);
	typedef std::map<unsigned int, unsigned int> ErrorMap;
	ErrorMap errors;
	ErrorMap warnings;
	Reader::Char rCur;
	ParseMode mode = kNormal;
	unsigned lineErrCount = 0;
	bool compileOnly = false;
	
	while (!(rCur = reader.lookChar()).getChar().is(CodePoint::kNull)) {
		CodePoint cur = rCur.getChar();
		
		switch (mode) {
		case kNormal:
			if (cur.is(CodePoint::kHashMark)) {
				//std::cout << "'#' read" << std::endl;
				mode = kSharp_Read;
				//debug << "entered # mode for line " << rCur.getLineNo() << newl;
			}
			break;
		case kSharp_Read:
			//std::cout << "cur = (char)" << (unsigned int)cur << std::endl;
			
			if (cur.is(CodePoint::kHashMark)) {
				// another sharp; stay in this mode
			}
			else if (cur.is(CodePoint::kIndentWhite)) {
				//std::cout << "'#' and whites read" << std::endl;
				// do nothing
			}
			else if (cur == CodePoint('E')) {
				//cout << "'# E' read" << endl;
				mode = kSharpE_Read;
				//debug << "entered #E mode for line " << rCur.getLineNo() << newl;
			}
			else if (cur == CodePoint('C')) {
				mode = kSharpC_Read;
			}
			else if (cur == CodePoint('W')) {
				//debug << "'# W' read" << newl;
				mode = kSharpW_Read;
			}
			else {
				//cout << "'#' and garbage read" << endl;
				// nothing meaningful found, return to normal mode
				mode = kNormal;
			}
			break;
		case kSharpE_Read:
			if (cur == CodePoint('R')) {
				mode = kSharpER_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpER_Read:
			if (cur == CodePoint('R')) {
				mode = kSharpERR_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpERR_Read:
			if (cur == CodePoint('O')) {
				mode = kSharpERRO_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpERRO_Read:
			if (cur == CodePoint('R')) {
				mode = kERROR_Mode;
				errors[rCur.getLineNo()] = 1;
				//debug << "entered #ERROR mode for line " << rCur.getLineNo() << newl;
			}
			else {
				mode = kNormal;
			}
			break;
		case kERROR_Mode:
			if (cur.is(CodePoint::kIndentWhite)) {
				// do nothing, stay in this mode
			}
			else if (cur == CodePoint('S')) {
				mode = kERRORS_Mode;
			}
			else if (cur == CodePoint('(')) {
				mode = kNumber_Mode;
				lineErrCount = 0;
			}
			else if (cur.is(CodePoint::kLf)) {
				mode = kNormal; // allready at line end
			}
			else {
				mode = kLineEnd_Mode;
			}
			break;
		case kERRORS_Mode:
			if (cur.is(CodePoint::kIndentWhite)) {
				// do nothing, stay in this mode
			}
			else if (cur == CodePoint('(')) {
				mode = kNumber_Mode;
				lineErrCount = 0;
			}
			else if (cur.is(CodePoint::kLf)) {
				mode = kNormal; // allready at line end
			}
			else {
				mode = kLineEnd_Mode;
			}
			break;
		case kSharpW_Read:
			if (cur == CodePoint('A')) {
				mode = kSharpWA_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpWA_Read:
			if (cur == CodePoint('R')) {
				mode = kSharpWAR_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpWAR_Read:
			if (cur == CodePoint('N')) {
				mode = kSharpWARN_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpWARN_Read:
			if (cur == CodePoint('I')) {
				mode = kSharpWARNI_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpWARNI_Read:
			if (cur == CodePoint('N')) {
				mode = kSharpWARNIN_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpWARNIN_Read:
			if (cur == CodePoint('G')) {
				mode = kWARNING_Mode;
				warnings[rCur.getLineNo()] = 1;
			}
			else {
				mode = kNormal;
			}
			break;
		case kWARNING_Mode:
			mode = kNormal;
			break;
		case kSharpC_Read:
			if (cur == CodePoint('O')) {
				mode = kSharpCO_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpCO_Read:
			if (cur == CodePoint('M')) {
				mode = kSharpCOM_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpCOM_Read:
			if (cur == CodePoint('P')) {
				mode = kSharpCOMP_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpCOMP_Read:
			if (cur == CodePoint('I')) {
				mode = kSharpCOMPI_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpCOMPI_Read:
			if (cur == CodePoint('L')) {
				mode = kSharpCOMPIL_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpCOMPIL_Read:
			if (cur == CodePoint('E')) {
				mode = kSharpCOMPILE_Read;
			}
			else {
				mode = kNormal;
			}
			break;
		case kSharpCOMPILE_Read:
			if (cur == CodePoint('-')) {
				mode = kCompileSpecMode;
			}
			else {
				mode = kNormal;
			}
			break;
		case kCompileSpecMode:
			// "# COMPILE-" read until now
			if (cur == CodePoint('O')) {
				mode = kCompileSpecMode_O;
			}
			else {
				mode = kNormal;
			}
			break;
		case kCompileSpecMode_O:
			if (cur == CodePoint('N')) {
				mode = kCompileSpecMode_ON;
			}
			else {
				mode = kNormal;
			}
			break;
		case kCompileSpecMode_ON:
			if (cur == CodePoint('L')) {
				mode = kCompileSpecMode_ONL;
			}
			else {
				mode = kNormal;
			}
			break;
		case kCompileSpecMode_ONL:
			if (cur == CodePoint('Y')) {
				compileOnly = true;
				mode = kNormal;
			}
			else {
				mode = kNormal;
			}
			break;
		case kNumber_Mode:
			if (cur.is(CodePoint::kIndentWhite)) {
				mode = kNumberDone_Mode;
			}
			else if (cur.is(CodePoint::kDigit)) {
				unsigned digit = cur.getDigitValue();
				lineErrCount *= 10;
				lineErrCount += digit;
			}
			else if (cur == CodePoint(')')) {
				mode = kLineEnd_Mode;
				errors[rCur.getLineNo()] = lineErrCount;
			}
			break;
		case kNumberDone_Mode:
			if (cur.is(CodePoint::kIndentWhite)) {
				// do nothing
			}
			else if (cur.is(CodePoint::kLf)) {
				mode = kNormal; // allready at line end
			}
			else if (cur == CodePoint(')')) {
				mode = kLineEnd_Mode;
				errors[rCur.getLineNo()] = lineErrCount;
			}
			break;
		case kLineEnd_Mode:
			if (cur.is(CodePoint::kLf)) {
				// do nothing
				mode = kNormal;
			}
			break;
		}
		
		reader.readNextChar();
	}
	
	unsigned int lastLine = rCur.getLineNo();
	
	//debug << "EXPECTED ERRORS:" << newl;
	//for (unsigned int i = 1; i <= lastLine; ++i) {
	//	debug << "[" << i << "]  " << errors[i] << " errors" << newl;
	//}
	//debug << newl;
	
	// PHASE 2 : compile the file and see what diagnostics the compiler gives
	
	UTF8TextFlowStream& stdErr = Console::getTextFlowStdErr();
	CompileOptions opt(stdErr);
	opt.fgDontLink = compileOnly;
	//opt.fgPrint_AST = true; // debug
	Compiler comp(opt);
	
	bool errorCheckingInterrupted = false;
	
	try {
		// Compile
		comp.Compile(filename);
	}
	catch (CompilationFailedError const& cfe) {
		errorCheckingInterrupted = cfe.wasFatal();
	}
	
	const ErrorManager& errMan = comp.getErrorManager();
	ErrorMap detectedErrors;
	ErrorMap detectedWarnings;
	unsigned lastCompilerMsgLine = 0;
	
	typedef ErrorManager::DiagnosticIterator EMDI;
	for (EMDI it = errMan.begin(); it != errMan.end(); ++it) {
		const Diagnostic& d = *it;
		
		//errMan.print(Console::getTextFlowStdOut(), d); // debug
		
		Location const& loc = d.getLocation();
		unsigned lineNo;
		if (loc.isSpecial()) {
			lineNo = 0;
			stdErr.line(
				UTF8STR(
					"WARNING: compiler diagnostic on non-sourcefile location "
						"(will use line number: 0)"
				)
			);
		}
		else {
			lineNo = loc.getLine();
		}
		
		if (d.getType() == Diagnostic::kError) {
			detectedErrors[lineNo] += 1;
		}
		else if (d.getType() == Diagnostic::kWarning) {
			detectedWarnings[lineNo] += 1;
		}
		
		lastCompilerMsgLine = max(lineNo, lastCompilerMsgLine);
	}
	
	// PHASE 3 : compare the two error lists
	
	TestState state(
		filename, errorCheckingInterrupted, lastCompilerMsgLine);
	
	for (unsigned int i = 0; i <= lastLine; ++i) {
		if (errors[i] == 0 && detectedErrors[i] > 0) {
			// unexpected compiler error(s)
			if (detectedErrors[i] == 1) {
				state.printProblem(i, UTF8STR("UNEXPECTED FAILURE"));
			}
			else {
				state.printProblem(
					i,
					UTF8STR("UNEXPECTED FAILURES (%1)").compose(
						detectedErrors[i]
					)
				);
			}
		}
		else if (warnings[i] == 0 && detectedWarnings[i] > 0) {
			// unexpected compiler warning(s)
			if (detectedWarnings[i] == 1) {
				state.printProblem(i, UTF8STR("UNEXPECTED WARNING"));
			}
			else {
				state.printProblem(
					i,
					UTF8STR("UNEXPECTED WARNINGS (%1)").compose(
						detectedWarnings[i]
					)
				);
			}
		}
		else if (errors[i] > 0 && detectedErrors[i] == 0) {
			// compiler error(s) missing
			state.printProblem(i, UTF8STR("UNEXPECTED SUCCESS"));
		}
		else if (errors[i] > 0 && detectedErrors[i] != errors[i]) {
			// incorrect number of compiler errors
			state.printProblem(
				i,
				UTF8STR("FAILURE NUMBER MISMATCH (%1)").compose(
					getFoundVersusExpected(detectedErrors[i], errors[i])
				)
			);
		}
		else if (warnings[i] > 0 && detectedWarnings[i] == 0) {
			state.printProblem(i, UTF8STR("MISSING WARNING(S)"));
		}
		else if (warnings[i] > 0 && detectedWarnings[i] != warnings[i]) {
			state.printProblem(
				i,
				UTF8STR("WARNING NUMBER MISMATCH (%1)").compose(
					getFoundVersusExpected(detectedWarnings[i], warnings[i])
				)
			);
		}
		else {
			// current line is OK
		}
	}
	
	return state.getProblems();
}
